/*
 * Copyright (C) 2018-2020 Cray UK
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdbool.h>
#include <pthread.h>
#include <assert.h>
#include "mamba.h"

#include "mmb_tile.h"
#include "mmb_tile_iterator.h"
#include "mmb_logging.h"
#include "mmb_memory.h"
/* Keep the full path to ease the research of the correpsonding header */
#include "../memory/allocation.h"
#include "../memory/strategy.h"
#include "../memory/memory_options.h"

#ifdef HAVE_LOOP_ANALYSIS
#include "mmb_loop.h"
#endif

/* protecting initialization */
pthread_mutex_t g_mamba_state = PTHREAD_MUTEX_INITIALIZER;
static inline void mmb_state_lock(void) {
  int stat = pthread_mutex_lock(&g_mamba_state);
  assert(stat==0);
}
static inline void mmb_state_unlock(void) {
  int stat = pthread_mutex_unlock(&g_mamba_state);
  assert(0 == stat);
}

mmbManager *g_manager;
static bool g_initialized = false;

static inline mmbError nd_to_1d_idx(mmbIndex *in_idx, mmbDimensions * in_dims, 
                                    size_t *out_idx){
  /* Validity and range checks */
  if(!in_idx || !in_dims || !out_idx)
    return MMB_INVALID_ARG;

  if(in_idx->size != in_dims->size){
    MMB_ERR("Index size %zu does not match dimensions size %zu\n", in_idx->size, in_dims->size);
    return MMB_ERROR; 
  }

  for(unsigned i = 0; i < in_idx->size; i++)
    if(in_idx->d[i] >= in_dims->d[i]) {
      MMB_ERR("Index %zu out of bounds for dimension %u size %zu\n", in_idx->d[i], i, in_dims->d[i]);
      return MMB_ERROR;
    }

  /* Convert to 1d */
  *out_idx = 0;
  for (unsigned i = 0; i < in_dims->size; i++) {
    size_t idx_i = in_idx->d[i];
    size_t sum_dims = 1;
    for(unsigned j = i+1; j < in_dims->size; j++) {
      sum_dims *= in_dims->d[i];
    }
    idx_i *= sum_dims;
    *out_idx += idx_i;
  }
  return MMB_OK;
}

mmbError mmb_init( mmbOptions *in_opts ) {
  mmbError stat = MMB_OK;

  mmb_state_lock();

  if(g_initialized) { 
    MMB_ERR("Mamba library already initialized\n");
    stat = MMB_ERROR;
    goto UNLOCK_BAILOUT;
  }

  if(!in_opts) {
    stat = mmb_options_create_default(&in_opts);
    if(stat != MMB_OK){
      MMB_ERR("Failed to create default mmb options\n");
      goto UNLOCK_BAILOUT;
    }
  }

  stat = mmb_logging_init(in_opts->logging_opts);
  if (MMB_OK != stat) {
    MMB_ERR("Failed to initialize the logging infrastructure.\n");
    goto UNLOCK_BAILOUT;
  }

  stat = mmb_init_providers(in_opts->memory_opts->providers_opts);
  if (MMB_OK != stat) {
    MMB_ERR("Failed to initialize the memory provider libraries.\n");
    goto UNLOCK_BAILOUT;
  }

  stat = mmb_init_strategies(in_opts->memory_opts->strategies_opts);
  if (MMB_OK != stat) {
    MMB_ERR("Failed to initialize the memory allocation strategies.\n");
    goto UNLOCK_BAILOUT;
  }

  stat = mmb_manager_create(&g_manager);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to create memory manager\n");
    goto UNLOCK_BAILOUT;
  }

  stat = mmb_alloc_initialize();
  if (MMB_OK != stat) {
    MMB_ERR("Failed to initialize mmbAllocationFactory.\n");
    goto UNLOCK_BAILOUT;
  }

#ifdef HAVE_LOOP_ANALYSIS
  stat = mmb_loop_initialize();
  if(stat != MMB_OK) {
    MMB_ERR("Failed to initialize mmb_loop\n");
    goto UNLOCK_BAILOUT;
  } 
#endif 

  g_initialized = true;
  MMB_DEBUG("Mamba library initialised, version: %s\n", mmb_version());

UNLOCK_BAILOUT:
  mmb_state_unlock();

  return stat;
}

mmbError mmb_is_initialized(bool *initialized) {
  if(!initialized)
    return MMB_INVALID_ARG;
  mmb_state_lock();
  *initialized = g_initialized;
  mmb_state_unlock();
  return MMB_OK;
}

mmbError mmb_register_memory(const mmbMemLayer layer,
                             const mmbExecutionContext ex_context,
                             const mmbMemSpaceConfig *config_opts,
                             mmbMemSpace **new_space)
{
  const mmbMemSpaceConfig default_config = MMB_MEMSPACE_CONFIG_DEFAULT;
  if (NULL == config_opts) {
    config_opts = &default_config;
  }
  /* /!\ mmbMemSpaceConfig->mem_size is given in bytes, not MiB. */
  mmbMemSpaceConfig space_config = *config_opts;
  if (SIZE_MAX == space_config.size_opts.mem_size // SIZE_MAX is a special value
      || SIZE_MAX / (1024*1024) < space_config.size_opts.mem_size)
    space_config.size_opts.mem_size = SIZE_MAX;
  else
    space_config.size_opts.mem_size *= 1024 * 1024;
  return mmb_manager_register_memory(g_manager, layer, ex_context, &space_config,
                                     new_space);
}

mmbError mmb_request_space(const mmbMemLayer layer,
                           const mmbExecutionContext ex_con,
                           const mmbMemSpaceConfig *config_opts,
                           mmbMemSpace **out_space)
{
  const mmbMemSpaceConfig default_config = MMB_MEMSPACE_CONFIG_DEFAULT;
  if (NULL == config_opts) {
    config_opts = &default_config;
  }
  return mmb_manager_request_space(g_manager, layer, ex_con, config_opts, out_space);
}

mmbError mmb_request_interface(mmbMemSpace *space,
                               const mmbMemInterfaceConfig *config_opts,
                               mmbMemInterface **out_interface)
{
  const mmbMemInterfaceConfig default_config = MMB_MEMINTERFACE_CONFIG_DEFAULT;
  if (NULL == config_opts) {
    config_opts = &default_config;
  }
  return mmb_memspace_request_interface(space, config_opts, out_interface);
}

mmbError mmb_memspace_configure(const mmbMemSpaceConfig *config_opts,
                                mmbMemSpace *space)
{
  if (NULL == config_opts) {
    MMB_WARN("NULL pointer given for configure option, no change to be done.\n");
    return MMB_OK;
  }
  return mmb_memspace_configure_defaults(g_manager, config_opts, space);
}

mmbError mmb_dump_memory_state(FILE *out_stream)
{
  return mmb_manager_report_state(g_manager, out_stream);
}

mmbError mmb_finalize(void) {
  mmbError stat = MMB_OK;

  mmb_state_lock();

  if(!g_initialized) {
      MMB_ERR("Mamba library not initialized\n");
      stat = MMB_ERROR;
      goto UNLOCK_BAILOUT;
  }

  stat = mmb_manager_destroy(g_manager);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to destroy memory manager\n");
    goto UNLOCK_BAILOUT;
  }

  stat = mmb_alloc_finalize();
  if (MMB_OK != stat) {
    MMB_ERR("Failed to finalize the mmbAllocationFactory.\n");
    goto UNLOCK_BAILOUT;
  }

  stat = mmb_finalize_providers();
  if (MMB_OK != stat) {
    MMB_ERR("Failed to finalize the memory provider libraries.\n");
    goto UNLOCK_BAILOUT;
  }

  stat = mmb_finalize_strategies();
  if (MMB_OK != stat) {
    MMB_ERR("Failed to finalize the memory allocation strategies.\n");
    goto UNLOCK_BAILOUT;
  }

#ifdef HAVE_LOOP_ANALYSIS
  stat = mmb_loop_finalize();
  if(stat != MMB_OK) {
    MMB_ERR("Failed to finalize mmb_loop\n");
    goto UNLOCK_BAILOUT;
  } 
#endif

  stat = mmb_logging_finalize();
  if (MMB_OK != stat) {
    MMB_ERR("Failed to finalize the logging infrastructure.\n");
    goto UNLOCK_BAILOUT;
  }

  g_initialized = false;

UNLOCK_BAILOUT:
  mmb_state_unlock();

 return stat;
}


/* protecting array state */
static inline void mmb_array_lock(mmbArray *mba) {
  int stat = pthread_mutex_lock(&mba->lock);
  assert(stat==0);
}
static inline void mmb_array_unlock(mmbArray *mba) {
  int stat = pthread_mutex_unlock(&mba->lock);
  assert(0 == stat);
}

/* prototypes for private lockfree versions of array functions */
mmbError mmb_array_tile_lockfree(mmbArray *in_mba, mmbDimensions *in_dims);
mmbError mmb_array_untile_lockfree(mmbArray *in_mba);

mmbError mmb_array_create(mmbDimensions *in_dims,
                          mmbLayout *in_layout,
                          mmbMemInterface *in_interface,
                          mmbAccessType in_access,
                          mmbArray **out_mba)
{
  mmbError stat = MMB_OK;

  if(!in_dims || !in_layout || !in_interface || !out_mba) {
    MMB_ERR("Arg can not be NULL\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }

  mmbArray *mba = (mmbArray *)calloc(1, sizeof(mmbArray));
  if(!mba) {
    MMB_ERR("Failed to allocate memory for mamba array object\n");
    stat = MMB_OUT_OF_MEMORY;
    goto BAILOUT;
  }

  if (NULL == in_interface) {
    MMB_ERR("Invalid memory interface handle (in_interface cannot be NULL).\n");
    return MMB_INVALID_ARG;
  }
  stat = mmb_layout_create_copy(in_layout, &mba->layout);
  if (stat != MMB_OK) {
    MMB_ERR("Failed to copy-construct mmbLayout\n");
    goto BAILOUT;
  }  
  mba->access = in_access;
  mba->interface = in_interface;

  size_t n_bytes;
  stat = mmb_layout_buffer_size(in_layout, in_dims, &n_bytes);
  if (MMB_OK != stat) {
    MMB_ERR("Unable to compute layout dimensions.\n");
    goto BAILOUT;
  }

  mmbAllocation *allocation;
  stat = mmb_allocate(n_bytes, in_interface, &allocation);
  if (MMB_OK != stat) {
    MMB_ERR("Failed to allocate memory for mmbArray.\n");
    goto BAILOUT;
  }

  mba->allocation = allocation;
  stat = mmb_dimensions_copy(&mba->dims, in_dims);
  if (stat != MMB_OK) {
    MMB_ERR("Failed to copy mmb_dimensions\n");
    goto BAILOUT;
  }  

  int rc = pthread_mutex_init(&mba->lock, NULL); 
  if(rc != 0) {
    MMB_ERR("Failed to initialise array mutex\n");
    goto BAILOUT;
  }   

  rc = pthread_rwlock_init(&mba->tiling_lock, NULL); 
  if(rc != 0) {
    MMB_ERR("Failed to initialise array tiling rwlock\n");
    goto BAILOUT;
  }   

  *out_mba = mba;
BAILOUT:
  return stat;
}

mmbError mmb_array_create_wrapped(void *in_data,
                                  mmbDimensions *in_dims,
                                  mmbLayout *in_layout,
                                  mmbMemInterface *in_interface,
                                  mmbAccessType in_access,
                                  mmbArray **out_mba)
{
  mmbError stat = MMB_OK;

  if(!in_data || !in_dims || !in_layout || !in_interface || !out_mba) {
    MMB_ERR("Arg can not be NULL\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }

  mmbArray *mba = (mmbArray *)calloc(1, sizeof(mmbArray));
  if(!mba) {
    MMB_ERR("Failed to allocate memory for mamba array object\n");
    stat = MMB_OUT_OF_MEMORY;
    goto BAILOUT;
  }

  if (NULL == in_interface) {
    MMB_ERR("Invalid memory interface handle (in_interface cannot be NULL).\n");
    return MMB_INVALID_ARG;
  }
  stat = mmb_layout_create_copy(in_layout, &mba->layout);
  if (stat != MMB_OK) {
    MMB_ERR("Failed to copy-construct mmbLayout\n");
    goto BAILOUT;
  }  

  mba->interface = in_interface;
  mba->access = in_access;

  if(in_data){
    size_t n_bytes;
    stat = mmb_layout_buffer_size(in_layout, in_dims, &n_bytes);
    if (MMB_OK != stat) {
      MMB_ERR("Unable to compute layout dimensions.\n");
      goto BAILOUT;
    }
    mmbAllocation *allocation;
    stat = mmb_allocation_create_wrapped(in_data, n_bytes, in_interface, &allocation);
    if (MMB_OK != stat) {
      MMB_ERR("Failed to create the appropriate wrapped allocation.\n");
      goto BAILOUT;
    }
    mba->allocation = allocation;
    stat = mmb_dimensions_copy(&mba->dims, in_dims);
    if (stat != MMB_OK) {
      MMB_ERR("Failed to copy mmb_dimensions\n");
      goto BAILOUT;
    }  
    mba->data_owned = 0;
  } else {
    MMB_ERR("Creating a wrapped mamba array required a valid data ptr\n");
    stat = MMB_ERROR;
    goto BAILOUT;
  }

  int rc = pthread_mutex_init(&mba->lock, NULL); 
  if(rc != 0) {
    MMB_ERR("Failed to initialise array mutex\n");
    goto BAILOUT;
  }   

  rc = pthread_rwlock_init(&mba->tiling_lock, NULL); 
  if(rc != 0) {
    MMB_ERR("Failed to initialise array tiling rwlock\n");
    goto BAILOUT;
  }   

  *out_mba = mba;
BAILOUT:
  return stat;
}

mmbError mmb_array_destroy(mmbArray *in_mba) {
  mmbError stat = MMB_OK;
  if (!in_mba)
    return stat;

  /* Free array data */
  stat = mmb_allocation_destroy(in_mba->allocation);
  if (stat != MMB_OK) {
    MMB_ERR("Failed to destroy mmbArray data allocation\n");
    goto BAILOUT;
  }
  
  /* Free remaining array components */
  stat = mmb_layout_destroy(in_mba->layout);
  if (stat != MMB_OK) {
    MMB_ERR("Failed to destroy mmbLayout\n");
    goto BAILOUT;
  }

  stat = mmb_dimensions_resize(&in_mba->dims, 0);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to free mmbArray dimensions\n");
    goto BAILOUT;
  }

  if(in_mba->tiling) {
    stat = mmb_array_untile(in_mba);
    if(stat != MMB_OK) {
      MMB_ERR("Failed to untile mmbArray\n");
      goto BAILOUT;
    }
  }

  int rc = pthread_mutex_destroy(&in_mba->lock); 
  if(rc != 0) {
    MMB_ERR("Failed to destroy array mutex\n");
    goto BAILOUT;
  }   

  rc = pthread_rwlock_destroy(&in_mba->tiling_lock); 
  if(rc != 0) {
    MMB_ERR("Failed to destroy array tiling rwlock\n");
    goto BAILOUT;
  }   

  free(in_mba);
BAILOUT:
  return stat;
}


mmbError mmb_array_copy(mmbArray *dst, mmbArray *src) {
  mmbError stat = MMB_OK;

  /* Validate */
  if (!dst || !src){
    MMB_ERR("Arg can not be NULL\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }
  mmb_array_lock(src);
  mmb_array_lock(dst);
  /* If src is empty, no copy necessary */
  size_t total_elements = 0;
  stat = mmb_dimensions_prod(&src->dims, &total_elements);
  if(stat != MMB_OK) {
    MMB_ERR("Unable to calculate array size\n");
    goto BAILOUT_UNLOCK;
  }
  if(total_elements == 0) {
    return MMB_OK;
  }

  /* Dst must be writeable */
  if(!(dst->access & MMB_WRITE)){
    MMB_ERR("Destination array must be writeable\n");
    stat = MMB_ERROR;
    goto BAILOUT_UNLOCK;
  }

  /* Dimensions and element size must match */
  int dims_cmp = mmb_dimensions_cmp(&dst->dims, &src->dims);
  if(dims_cmp) {
    MMB_ERR("src and dst arrays are not the same size\n");
    goto BAILOUT_UNLOCK;
  }

  int element_size_same = (dst->layout->element.size_bytes == 
                            src->layout->element.size_bytes) ? 1 : 0;

  if(!element_size_same) {
    MMB_ERR("SRC and DST arrays must have matching element size\n");
    stat = MMB_ERROR;
    goto BAILOUT_UNLOCK;
  }

  
  /* We will clear tiling if it existss */
  if(dst->tiling || src->tiling) {
    MMB_WARN("Either or both args are pre-tiled arrays, tiling will be removed during copy.\n");
  }

  /* Get a single tile for each array */
  stat = mmb_array_tile_lockfree(dst, &dst->dims);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to tile dst mmbArray for copy\n");
    goto BAILOUT_UNLOCK;
  }
  stat = mmb_array_tile_lockfree(src, &src->dims);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to tile src mmbArraymmbArray for copy\n");
    goto BAILOUT_UNLOCK;
  }

  mmbTileIterator *it_dst, *it_src;
  mmbArrayTile *tile_dst, *tile_src; 
  stat = mmb_tile_iterator_create(dst, &it_dst); 
  if(stat != MMB_OK) {
    MMB_ERR("Failed to get tile iterator from dst mmbArray for copy\n");
    goto BAILOUT_UNLOCK;
  }
  stat = mmb_tile_iterator_first(it_dst);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to move tile iterator to first for dst mmbArray for copy \n");
    goto BAILOUT_UNLOCK;
  }
  tile_dst = it_dst->tile;

  stat = mmb_tile_iterator_create(src, &it_src); 
  if(stat != MMB_OK) {
    MMB_ERR("Failed to get tile iterator from src mmbArray for copy\n");
    goto BAILOUT_UNLOCK;
  }
  stat = mmb_tile_iterator_first(it_src);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to move tile iterator to first for src mmbArray for copy \n");
    goto BAILOUT_UNLOCK;
  }
  tile_src = it_src->tile;

  /* Do the copy */
  stat = mmb_tile_copy(tile_dst, tile_src);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to copy src to dst tiles for mmbArray copy\n");
    goto BAILOUT;
  }

  /* Free the tiles we were using */
  stat = mmb_tile_iterator_destroy(it_dst);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to destroy dst tile iterator for mmbArray copy\n");
    goto BAILOUT_UNLOCK;
  }  
  stat = mmb_tile_iterator_destroy(it_src);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to destroy src tile iterator for mmbArray copy\n");
    goto BAILOUT_UNLOCK;
  }  
  
  stat = mmb_array_untile_lockfree(dst);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to untile dst mmbArray for copy\n");
    goto BAILOUT_UNLOCK;
  }
  stat = mmb_array_untile_lockfree(src);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to untile src mmbArraymmbArray for copy\n");
    goto BAILOUT_UNLOCK;
  }

BAILOUT_UNLOCK:
  mmb_array_unlock(src);
  mmb_array_unlock(dst);
BAILOUT:
  return stat;
}

/* protecting array state */
static inline void mmb_tiling_rdlock(mmbArray *mba) {
  int stat = pthread_rwlock_rdlock(&mba->tiling_lock);
  assert(stat==0);
}

static inline void mmb_tiling_wrlock(mmbArray *mba) {
  int stat = pthread_rwlock_wrlock(&mba->tiling_lock);
  assert(stat==0);
}

static inline void mmb_tiling_unlock(mmbArray *mba) {
  int stat = pthread_rwlock_unlock(&mba->tiling_lock);
  assert(stat==0);
}

mmbError mmb_array_tile(mmbArray *in_mba, mmbDimensions *in_dims) {
  mmb_array_lock(in_mba);
  mmbError stat = mmb_array_tile_lockfree(in_mba, in_dims);
  mmb_array_unlock(in_mba);
  return stat;
}

mmbError mmb_array_tile_lockfree(mmbArray *in_mba, mmbDimensions *in_dims) {
  mmbError stat = MMB_OK;
  /* Validate */
  if (!in_mba || !in_dims) {
    MMB_ERR("Arg can not be NULL\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }

  if (in_mba->tiling) {
    stat = mmb_array_untile_lockfree(in_mba);
    if(stat != MMB_OK) {
      MMB_ERR("Failed to remove previous mmbArray tiling\n");
      goto BAILOUT;
    }
  }

  if (in_mba->dims.size != in_dims->size) {
    MMB_ERR("Number of tile dimensions must match mmbArray layout dimensions\n");
    stat = MMB_ERROR;
    goto BAILOUT;
  }

  size_t elements_per_tile = 0;
  stat = mmb_dimensions_prod(in_dims, &elements_per_tile);
  if(stat != MMB_OK) {
    MMB_ERR("Unable to calculate size of tile\n");
    goto BAILOUT;
  }
  if(elements_per_tile == 0) {
    MMB_ERR("All dimensions must be non-zero, tile must have at least one element.\n");
    stat = MMB_ERROR;
    goto BAILOUT;
  }

  MMB_DEBUG("Creating tiling.\n");
  for(unsigned i = 0; i < in_dims->size; i++){
    MMB_DEBUG("Tiling dimension %d: %zu\n", i, in_dims->d[i]);
  }

  /* Store tile dimensions */
  mmbTiling *t = (mmbTiling *)calloc(1, sizeof(mmbTiling));
  stat = mmb_dimensions_copy(&t->tile_size, in_dims);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to copy tile dimensions\n");
    goto BAILOUT;
  }
  stat = mmb_dimensions_resize(&t->tile_count, in_mba->dims.size);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to resize tile count dimensions\n");
    goto BAILOUT;
  }

  /* Count tiles */
  t->total = 1;
  for (unsigned i = 0; i < in_mba->dims.size; i++) {
    t->tile_count.d[i] = (in_mba->dims.d[i] + (t->tile_size.d[i] - 1)) /
                         t->tile_size.d[i];
    t->total *= t->tile_count.d[i];
  }

  MMB_DEBUG("Total tiles: %zu\n", t->total);

  /* Init hashmap */
  t->tiles = NULL;

  in_mba->tiling = t;

BAILOUT:
  return stat;
}

mmbError mmb_array_untile(mmbArray *in_mba){
  mmb_array_lock(in_mba);
  mmbError stat = mmb_array_untile_lockfree(in_mba);
  mmb_array_unlock(in_mba);
  return stat;
}

mmbError mmb_array_untile_lockfree(mmbArray *in_mba){
  mmbError stat = MMB_OK;

  if(!in_mba) {
    MMB_ERR("Arg can not be NULL\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }
  /* Clear tile hashmap */
  if(!in_mba->tiling)
      return stat;

  mmbArrayTile *tile, *tmp;
  HASH_ITER(hh, in_mba->tiling->tiles, tile, tmp) {
    /* Free stack of tiles in LIFO order */
    while(tile->next != NULL)
      tile = tile->next;
    
    while(tile->prev != NULL){
      mmbArrayTile *prev = tile->prev;
      stat = mmb_tile_destroy(tile);
      if(stat != MMB_OK) {
        MMB_ERR("Failed to destroy tile during untiling\n");
      }
      tile = prev;
    }

    /* Free lowest tile in stack and remove from hash */
    HASH_DEL(in_mba->tiling->tiles, tile);
    mmb_tile_destroy(tile);
  }
 
  /* Cleanup */
  stat = mmb_dimensions_resize(&in_mba->tiling->tile_size, 0);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to resize tile size dimensions\n");
    goto BAILOUT;
  } 
  stat = mmb_dimensions_resize(&in_mba->tiling->tile_count, 0);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to resize tile count dimensions\n");
    goto BAILOUT;
  }
  free(in_mba->tiling);
  in_mba->tiling = NULL;
BAILOUT:
  return stat;
}

mmbError mmb_tile_count(mmbArray *in_mba, int *out_count) {
  mmbError stat = MMB_OK;
  if(!in_mba || !out_count) {
    MMB_ERR("Arg can not be NULL\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }
  /* Return counted tiles */
  *out_count = in_mba->tiling->total;
BAILOUT:
  return stat;
}

mmbError mmb_tiling_dimensions(mmbArray *in_mba, mmbDimensions **out_dims) {
  mmbError stat = MMB_OK;

  if(!in_mba || !out_dims) {
    MMB_ERR("Arg can not be NULL\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }

  if(!in_mba->tiling) {
    MMB_ERR("mmbArray is not tiled\n");
    return MMB_ERROR;
  }

  mmbDimensions* dims;
  stat =  mmb_dimensions_create(0, &dims);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to create mmb dimensions object\n");
    goto BAILOUT;
  } 

  stat = mmb_dimensions_copy(dims, &in_mba->tiling->tile_count);
  if(stat != MMB_OK) {
    MMB_ERR("Could not retrieve mmbDimensions from array tiling\n");
    goto BAILOUT;
  }

  *out_dims = dims;

BAILOUT:
  return stat;
}

mmbError mmb_tile_at(mmbArray *in_mba, mmbIndex *in_idx,
                     mmbArrayTile **out_tile) {
  mmbError stat = MMB_OK;

  if(!in_mba || !in_idx || !out_tile) {
    MMB_ERR("Arg can not be NULL\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }

  *out_tile = NULL;
  if(!in_mba->tiling) {
    stat = MMB_ERROR;
    MMB_ERR("mmbArray not tiled\n");
    goto BAILOUT;
  }

  /* Lookup in hashmap for tile at in_idx */
  size_t tile_idx = 0;
  stat = nd_to_1d_idx(in_idx, &in_mba->tiling->tile_count, &tile_idx);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to convert nd index to 1d index\n");
    goto BAILOUT;
  }

  mmbArrayTile *t = NULL;
  mmb_tiling_rdlock(in_mba);
  HASH_FIND(hh, in_mba->tiling->tiles, &tile_idx, sizeof(size_t), t);
  /* If found, move to top tile of the stack and return pointer */
  if (t != NULL) {
    while (t->next != NULL){
      t = t->next;
    }
    *out_tile = t;
    stat = MMB_OK;
    goto BAILOUT_UNLOCK;
  }
  
  mmb_tiling_unlock(in_mba);

  /* Otherwise, create tile and return using default memory space */ 

  /* Typical threaded usage would be multiple threads asking for different
    tiles, so we let each thread make its own tile before insertion
    This results in some wasted cycles if many threads ask for the same 
    not-yet-constructed tile, but minimises blocking in the more-common scenario
    where they ask for different not-yet-constructed tiles. */
  t = (mmbArrayTile *)calloc(1, sizeof(mmbArrayTile));
  if(!t) {
    stat = MMB_OUT_OF_MEMORY;
    goto BAILOUT;
  }

  size_t *dims = malloc(sizeof(size_t[6][in_mba->layout->n_dims]));
  if (NULL == dims) {
    free(t);
    stat = MMB_OUT_OF_MEMORY;
    goto BAILOUT;
  }
  t->lower   = &dims[0 * in_mba->layout->n_dims];
  t->upper   = &dims[1 * in_mba->layout->n_dims];
  t->alower  = &dims[2 * in_mba->layout->n_dims];
  t->aupper  = &dims[3 * in_mba->layout->n_dims];
  t->dim     = &dims[4 * in_mba->layout->n_dims];
  t->abs_dim = &dims[5 * in_mba->layout->n_dims];

  t->interface = in_mba->interface;
  t->access = in_mba->access;
  /* TODO: Could just use in_mba->layout pointer instead of duplicating? */
  stat = mmb_layout_create_copy(in_mba->layout, &t->layout);
  if (stat != MMB_OK) {
    MMB_ERR("Failed to copy-construct mmbLayout\n");
    free(t);
    free(dims);
    goto BAILOUT;
  }  

  for (unsigned i = 0; i < t->layout->n_dims; i++) {
    t->abs_dim[i] = in_mba->dims.d[i];
    t->lower[i] = in_idx->d[i] * in_mba->tiling->tile_size.d[i];
    t->upper[i] = (in_idx->d[i] + 1) * in_mba->tiling->tile_size.d[i];
    if (t->upper[i] > t->abs_dim[i])
      t->upper[i] = t->abs_dim[i];
    t->alower[i] = t->lower[i];
    t->aupper[i] = t->upper[i];
    t->dim[i] = t->upper[i] - t->lower[i];
  }

  /* Check contiguity */
  /* TODO: This only checks if tile is 1d, or if it covers whole array, 
   * not if subtiles are contiguous */
  int contiguous = 1;
  if(t->layout->n_dims > 1) {
    for (unsigned i = 0; i < t->layout->n_dims; i++) {
      if(t->abs_dim[i] != t->dim[i]) {
        contiguous = 0;
      }
    }
  }

  if(contiguous) {
    t->access |= MMB_CONTIGUOUS;
  }

  /* This pointer is offset using indexing macros */
  t->allocation = in_mba->allocation;

  /* Set tile index for hashmap */
  t->tidx = tile_idx;

  /* Initialise the tile lock */
  pthread_mutex_init(&t->lock, NULL); 
  
  mmb_tiling_wrlock(in_mba);
  /* Check if someone else already created this tile in the meantime
   * if so, free local data and return pre-existing tile */
  mmbArrayTile *t_temp;
  HASH_FIND(hh, in_mba->tiling->tiles, &tile_idx, sizeof(size_t), t_temp);
  if (t_temp != NULL) {
    /* We dont need the tile we constructed ourselves anymore */
    stat = mmb_layout_destroy(t->layout);
    if(stat != MMB_OK)
      MMB_WARN("Failed to destroy temporary layout object\n");
    free(dims);
    pthread_mutex_destroy(&t->lock); 
    free(t);

    t = t_temp;
    /* Move to top tile of the stack and return pointer */
    while (t->next != NULL){
      t = t->next;
    }
    *out_tile = t;
    stat = MMB_OK;
    goto BAILOUT_UNLOCK;
  }
  /* otherwise insert our tile to the map */
  HASH_ADD(hh, in_mba->tiling->tiles, tidx, sizeof(size_t), t);

  *out_tile = t;
BAILOUT_UNLOCK:
  mmb_tiling_unlock(in_mba);
BAILOUT:
  return stat;
}

mmbError mmb_tile_duplicate_at(mmbArray *in_mba, mmbIndex *in_idx,
                               mmbMemInterface *in_interface,
                               mmbArrayTile **out_tile)
{
  mmbError stat = MMB_OK;
  mmbArrayTile *t = NULL;
  stat = mmb_tile_at(in_mba, in_idx, &t);
  if (MMB_OK != stat) {
    MMB_ERR("Unable to retrive the corresponding tile.\n");
    return stat;
  }
  return mmb_tile_duplicate(t, in_interface, in_mba->access, in_mba->layout, out_tile);
}

