/*
 * Copyright (C) 2018-2020 Cray UK
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef MMB_MAMBA_H
#define MMB_MAMBA_H

#include <stdlib.h>
#include <stdbool.h>
#include <errno.h>
#include <stdio.h>
#include <pthread.h>

#include "mmb_layout.h"
#include "mmb_dimensions.h"
#include "mmb_error.h"
#include "mmb_tile.h"
#include "mmb_memory.h"
#include "mmb_options.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Globals */
extern mmbManager *g_manager;

/* Types */
typedef struct s_mmbArray {
  mmbAllocation *allocation;
  mmbLayout *layout;
  mmbMemInterface *interface;
  mmbTiling *tiling;
  pthread_rwlock_t tiling_lock;
  mmbAccessType access;
  mmbDimensions dims;
  bool data_owned;
  pthread_mutex_t lock;
} mmbArray;

/**
 * mmb_init accepts a default initialisation argument
 */
#define MMB_INIT_DEFAULT NULL

/* Functions */

/**
 * @brief      Initialize the mamba library.
 *
 * @param      in_opts  Input mamba options, may be NULL for default.
 *
 * @return     \c MMB_OK on success.
 */
mmbError mmb_init(mmbOptions *in_opts);

/**
 * @brief      Check if mamba library is initialized.
 *
 * @return     \c MMB_OK on success.
 */
mmbError mmb_is_initialized(bool *initialized);

/**
 * @brief      Register a new memory tier in which memory can be allocated.
 *
 * This function creates a new mmbMemSpace which get added to the general
 * manager and returned to the user via the \p new_space handle.
 * The new space is initialised with a maximum size of \p n_mbytes
 * mibi-bytes (2^20) of \p layer memory.
 * WARNING : if two spaces are registered to the same mmbMemLayer,
 * \p mmb_request_space will always return the first mmbMemSpace allocated.
 *
 * @param      layer The memory tier to register.
 * @param      ex_context The execution context for the space.
 * @param      config_opts Configuration parameters for the mmbMemSpace.  May
 *                         be NULL in which case \c MMB_MEMSPACE_CONFIG_DEFAULT
 *                         is assumed.  The size is given in MiB (2^20 bytes).
 * @param      new_space Newly created space.
 *
 * @return     \c MMB_OK on success.
 */
mmbError mmb_register_memory(const mmbMemLayer layer,
                             const mmbExecutionContext ex_context,
                             const mmbMemSpaceConfig *config_opts,
                             mmbMemSpace **new_space);

/**
 * @brief      Request a mmbMemSpace handle for the given layer.
 *
 * @param      layer Memory layer to which the space must correspond.
 * @param      ex_con Targeted execution layer, \c MMB_CPU for any CPU oriented
 *                    code, or \c MMB_GPU_CUDA or \c MMB_CPU_OPENCL for GPU
 *                    specific allocation.
 * @param      config_opts Configuration parameters for the mmbMemSpace.  May
 *                         be \c NULL in which case \c MMB_MEMSPACE_CONFIG_DEFAULT
 *                         is assumed.
 * @param      out_space Pointer to a mmbMemSpace handle.
 *
 * @return     \c MMB_OK on success.
 */
mmbError mmb_request_space(const mmbMemLayer layer,
                           const mmbExecutionContext ex_con,
                           const mmbMemSpaceConfig *config_opts,
                           mmbMemSpace **out_space);

/**
 * @brief      Request a mmbMemInterface handle for the given layer.
 *
 * @param      space Handle to a mmbMemSpace where the data will be allocated.
 * @param      config_opts Configuration parameters for the mmbMemInterface
 *                         May be \c NULL in which case
 *                         \c MMB_MEMINTERFACE_CONFIG_DEFAULT is assumed.
 * @param      out_interface Pointer to a mmbMemInterface handle.
 *
 * @return     \c MMB_OK on success.
 */
mmbError mmb_request_interface(mmbMemSpace *space,
                               const mmbMemInterfaceConfig *config_opts,
                               mmbMemInterface **out_interface);

/**
 * @brief       Configure the default provider and strategy for the given
 *              \p space.
 *
 * Any modification to the mem_size has no effect and is disregarded.
 * Any value set to MMB_ANY_* is kept unchanged in the target \p space.
 * Any value set to MMB_<type>_DEFAULT reset the default value with regards to
 * the manager defaults values set during initialization.
 *
 * @param       config_opts Pointer to structure containing the values to
 *                          change.
 * @param       space       Handle to space structure to modify.
 *
 * @return     \c MMB_OK on success.
 */
mmbError mmb_memspace_configure(const mmbMemSpaceConfig *config_opts,
                                mmbMemSpace *space);
/**
 * @brief      Dump the state of the memory as recorded by the mamba library.
 *
 * @return     \c MMB_OK on success.
 */
mmbError mmb_dump_memory_state(FILE *out_stream);

/**
 * @brief      Finalize the mamba library.
 *
 * @return     \c MMB_OK on success.
 */
mmbError mmb_finalize(void);

/**
 * @brief      Create new mamba array, allocating memory for specified dimensions.
 *
 * @param[in]  in_dims      Dimensions of data.
 * @param[in]  in_layout    Layout of user data.
 * @param[in]  in_interface Memory interface to be used to allocate data.
 * @param[in]  in_access    Data access rights
 * @param[out] out_mba      The output mamba array.
 *
 * @return     \c MMB_OK on success.
 */
mmbError mmb_array_create(mmbDimensions *in_dims,
                          mmbLayout *in_layout,
                          mmbMemInterface *in_interface,
                          mmbAccessType in_access,
                          mmbArray **out_mba);

/**
 * @brief      Create new mamba array wrapping an existing user array.
 *
 * @param[in]  in_data      Data from user.
 * @param[in]  in_dims      Dimensions of data.
 * @param[in]  in_layout    Layout of user data.
 * @param[in]  in_interface Memory interface to be used to allocate data.
 * @param[in]  in_access    Data access rights
 * @param[out] out_mba      The output mamba array.
 *
 * @return     \c MMB_OK on success.
 */
mmbError mmb_array_create_wrapped(void *in_data,
                                  mmbDimensions *in_dims,
                                  mmbLayout *in_layout,
                                  mmbMemInterface *in_interface,
                                  mmbAccessType in_access,
                                  mmbArray **out_mba);

/**
 * @brief      Destroy a mamba array.
 *
 * @param      in_mba   The mamba array to be destroyed.
 *
 * @return     \c MMB_OK on success.
 */
mmbError mmb_array_destroy(mmbArray *in_mba);

/**
 * @brief      Copy data from source array to destination array.
 *
 * @param      dst   The destination array to copy to.
 * @param      src   The source array to copy from.
 *
 * @return     \c MMB_OK on success.
 */
mmbError mmb_array_copy(mmbArray *dst, mmbArray *src);

/**
 * @brief      Create a tiling of mmbArray mba.
 *
 * @param[in]   in_mba   The mamba array to tile.
 * @param[in]   in_dims  The tile dimensions.
 *
 * @return     \c MMB_OK on success.
 */
mmbError mmb_array_tile(mmbArray *in_mba, mmbDimensions *in_dims);


/**
 * @brief      Remove the current tiling of mmbArray in_mba.
 *
 * @param      in_mba  The mamba array to remove tiling on.
 *
 * @return     \c MMB_OK on success.
 */
mmbError mmb_array_untile(mmbArray *in_mba);

/**
 * @brief      Get number of tiles.
 *             TODO: 1d - should have version for tile count in each dimension
 *
 * @param[in]  in_mba      The tiled mamba array to count.
 * @param[out] out_count    The output number of tiles.
 *
 * @return     \c MMB_OK on success.
 */
mmbError mmb_tile_count(mmbArray *in_mba, int *out_count);


/**
 * @brief      Query the number of tiles in each dimension of the active tiling.
 *
 * @param      in_mba    In tiled mamba array.
 * @param      out_dims  The output dimensions.
 *
 * @return     \c MMB_OK on success.
 */
mmbError mmb_tiling_dimensions(mmbArray *in_mba, mmbDimensions **out_dims);

/**
 * @brief      Get tile at location defined by in_idx.
 *
 *
 * @param[in]      in_mba   The tiled mamba array.
 * @param[in]      in_idx   The index of the tile to return.
 * @param[out]     out_tile The output tile at index in_idx.
 *
 * @return         \c MMB_OK on success.
 */
mmbError mmb_tile_at(mmbArray *in_mba, mmbIndex *in_idx,
                     mmbArrayTile **out_tile);

/**
 * @brief      Get tile at location defined by \p in_idx, into memory
 *             corresponding to the interface \p in_interface.
 *
 *
 * @param[in]      in_mba       The tiled mamba array.
 * @param[in]      in_idx       The index of the tile to return.
 * @param[in]      in_interface The memory interface to use to create the new
 *                              tiles.
 * @param[out]     out_tile     The output tile at index in_idx.
 *
 * @return         \c MMB_OK on success.
 */
mmbError mmb_tile_duplicate_at(mmbArray *in_mba, mmbIndex *in_idx,
                               mmbMemInterface *in_interface,
                               mmbArrayTile **out_tile);

/**
 * @brief Return a printable string detailing the mamba version
 * 
 * @return const char* 
 */
const char* mmb_version(void);



/**
 * @brief      2D normalised indexing into a tile object.
 *             Assumes tile indices extend from 0 to tile->dim.
 *
 * @param      T     Tile object.
 * @param      I     Outer dimension iterator.
 * @param      J     Inner dimension iterator.
 * @param      TYPE  The pointer type to cast to during indexing.
 *
 * @return     Element of tile at \p I, \p J .
 */
#define MMB_IDX_2D_NORM(T, I, J, TYPE)                                      \
  MMBALLOC2TBUF((T)->allocation, TYPE *)[((I) + (T)->lower[0]) * (T)->abs_dim[1] + (T)->lower[1] + (J)]

/**
 * @brief      2D  indexing into a tile object.
 *             Assumes tile indices extend from tile->lower to tile->upper.
 *
 * @param      T     Tile object.
 * @param      I     Outer dimension iterator.
 * @param      J     Inner dimension iterator.
 * @param      TYPE  The pointer type to cast to during indexing.
 *
 * @return     Element of tile at \p I, \p J .
 */
#define MMB_IDX_2D(T, I, J, TYPE)                                           \
  MMBALLOC2TBUF((T)->allocation, TYPE *)[(I) * (T)->abs_dim[1] + (J)]

/**
 * @brief      1D indexing into tile object.
 *
 * @param      T     Tile object.
 * @param      I     1D iterator.
 * @param      TYPE  The pointer type to cast to during indexing.
 *
 * @return     { description_of_the_return_value }
 */
#define MMB_IDX_1D(T, I, TYPE)                                           \
  MMBALLOC2TBUF((T)->allocation, TYPE *)[(I)]

/**
 * @brief      Cast a mmbAllocation into a void* buffer to directly
 *             access the data.
 *
 * @param      alloc Pointer to the mmbAllocation.
 *
 * @return a pointer to the corresponding buffer.
 */
#define MMBALLOC2BUF(alloc) MMBALLOC2TBUF(alloc, void *)

/**
 * @brief      Cast a mmbAllocation into a buffer of type \p type
 *             to directly access the data.
 *
 * @param      alloc Pointer to the mmbAllocation.
 * @param      type Ty.
 *
 * @return a pointer to the corresponding buffer.
 */
#ifdef __cplusplus
#define MMBALLOC2TBUF(alloc, type) (*reinterpret_cast<type *>(alloc))
#else
#define MMBALLOC2TBUF(alloc, type) (*(type *)(alloc))
#endif

/**
 * @brief      Cast a buffer as a *FAKE* mmbAllocation (no information
 *             set). This macro is only to be used in conjuction with
 *             \p MMBALLOC2BUF or \p MMBALLOC2TBUF .
 *
 * @param      buffer Buffer to be wrapped.
 *
 * @return a properly casted fake mmbAllocation handle.
 */
#ifdef __cplusplus
#define BUF2MMBALLOC(buf) reinterpret_cast<mmbAllocation *>(&(buf))
#else
#define BUF2MMBALLOC(buf) ((mmbAllocation *)(&(buf)))
#endif

#ifdef __cplusplus
}
#endif

#endif /* MMB_MAMBA_H */
