/*
 * Copyright (C) 2018-2020 Cray UK
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stddef.h>
#include <stdbool.h>
#include <string.h>
#include <pthread.h>
#include <assert.h>

#include "mmb_error.h"
#include "mmb_logging.h"
#include "mmb_memory.h"

#include "i_memory.h"
#include "allocation.h"

#define N_ELT_FACTORY 1

typedef struct mmbAllocationFactory {
  mmbAllocation ptrs[N_ELT_FACTORY][8 * sizeof(unsigned int)];
  unsigned int avail[N_ELT_FACTORY];
  struct mmbAllocationFactory *next;
} mmbAllocationFactory;

static mmbAllocationFactory *global_factory = NULL;
/* protecting global_factory */
pthread_mutex_t global_factory_mtx = PTHREAD_MUTEX_INITIALIZER;
#define LOCK_GLOBAL_FACTORY do {                                  \
    MMB_DEBUG("Locking global factory\n");                            \
    int _gflstat = pthread_mutex_lock(&global_factory_mtx);       \
    assert(_gflstat==0);                                          \
  } while(0)

#define UNLOCK_GLOBAL_FACTORY do {                                \
    MMB_DEBUG("Unlocking global factory\n");                          \
    int _gflstat = pthread_mutex_unlock(&global_factory_mtx);     \
    assert(_gflstat==0);                                          \
  } while(0)

/* must be called under lock */
static mmbError mmb_alloc_factory_new(mmbAllocationFactory **out_factory)
{
  /* Check parameter */
  if (NULL == out_factory) {
    MMB_ERR("Out parameter \"out_factory\" cannot be NULL.\n");
    return MMB_INVALID_ARG;
  }
  /* Allocate 0 initialized structure */
  mmbAllocationFactory *new_fact = calloc(1, sizeof *new_fact);
  if (NULL == new_fact) {
    MMB_ERR("Unable to allocate mmbAllocation factory.\n");
    return MMB_OUT_OF_MEMORY;
  }
  memset(new_fact->avail, ~0, sizeof(unsigned int [N_ELT_FACTORY]));
  *out_factory = new_fact;
  return MMB_OK;
}

/* must be called under lock */
static mmbError mmb_alloc_factory_delete(mmbAllocationFactory *in_factory)
{
  /* Check parameter */
  if (NULL == in_factory) {
    MMB_WARN("Invalid factory handle.\n");
    return MMB_INVALID_ARG;
  }
  /* Check if unfree'd mmbAllocation left */
  /* block is (8 * sizeof(unsigned int)), i.e., one bit-field. */
  for (size_t block = 0; N_ELT_FACTORY > block; ++block) {
    if (0 != (~in_factory->avail[block])) {
      MMB_WARN("block %zu: avail=%x\n", block, in_factory->avail[block]);
      for (size_t it = 0; 8 * sizeof(unsigned int) > it; ++it) {
        if ((~in_factory->avail[block]) & (1<<it)) {
          MMB_WARN("Allocation pending of size %zu at address %p.\n",
               in_factory->ptrs[block][it].n_bytes,
               in_factory->ptrs[block][it].ptr);
        }
      }
    }
  }
  free(in_factory);
  return MMB_OK;
}

mmbError mmb_alloc_check_params(const mmbAllocation *alloc)
{
  if (NULL == alloc) {
    MMB_WARN("Invalid allocation handle.\n");
    return MMB_INVALID_ALLOCATION;
  }
  /* Do we allow a "valid" allocation that would point on NULL? */
  if (NULL == alloc->ptr) {
    MMB_WARN("Invalid allocation pointing on NULL.\n");
    return MMB_INVALID_ALLOCATION;
  }
  if (MMB_OK != mmb_meminterface_check_params(alloc->interface)) {
    MMB_WARN("Invalid interface handle in allocation handle.\n");
    return MMB_INVALID_ALLOCATION;
  }
  return MMB_OK;
}

mmbError mmb_alloc_initialize(void)
{
  mmbError stat = MMB_OK;
  LOCK_GLOBAL_FACTORY;
  if (NULL == global_factory) {
    stat = mmb_alloc_factory_new(&global_factory);
  }
  UNLOCK_GLOBAL_FACTORY;
  return stat;
}

mmbError mmb_alloc_finalize(void)
{
  mmbError stat;
  mmbAllocationFactory *fact = NULL;

  LOCK_GLOBAL_FACTORY;

  for (; NULL != global_factory; global_factory = fact) {
    fact = global_factory->next;
    stat = mmb_alloc_factory_delete(global_factory);
    if (MMB_OK != stat) {
      MMB_WARN("Error when deallocating the mmbAllocation factory.\n");
      goto BAILOUT_UNLOCK;
    }
  }
BAILOUT_UNLOCK:
  UNLOCK_GLOBAL_FACTORY;

  return stat;
}

mmbError mmb_alloc_allocate(void *ptr, const size_t n_bytes,
                            mmbMemInterface *interface,
                            const bool owned, mmbAllocation **alloc)
{
  /* Check parameter */
  mmbError stat;

  if (NULL == alloc) {
    MMB_ERR("Out parameter \"alloc\" cannot be NULL.\n");
    return MMB_INVALID_ARG;
  }

  LOCK_GLOBAL_FACTORY;

  if (NULL == global_factory) {
    MMB_ERR("Factory not initialized.\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT_UNLOCK;
  }
  /* Look for an available slot */
  mmbAllocationFactory *last = NULL, *factory = global_factory;
  mmbAllocation *new_alloc = NULL;
  while (NULL != factory) {
    for (size_t block = 0; N_ELT_FACTORY > block; ++block) {
      if (0 != factory->avail[block]) {
        for (size_t it = 0; 8 * sizeof(unsigned int) > it; ++it) {
          if (factory->avail[block] & (1<<it)) {
            new_alloc = &factory->ptrs[block][it];
            factory->avail[block] &= ~(1<<it);
            goto SET_AND_OUT;
          }
        }
      }
    }
    last = factory;
    factory = factory->next;
  }
  /* if this point is reached no suitable available entry has been found.
   * Allocate a new set of mmbAllocation's. */
  stat = mmb_alloc_factory_new(&last->next);
  if (MMB_OK != stat) {
    MMB_WARN("Unable to allocate new set of mmbAllocation. Allocation fails.\n");
    goto BAILOUT_UNLOCK;
  }
  new_alloc = &last->next->ptrs[0][0];
  last->next->avail[0] &= ~1;

SET_AND_OUT:
  new_alloc->ptr = ptr;
  new_alloc->n_bytes = n_bytes;
  new_alloc->interface = interface;
  new_alloc->owned = owned;
  *alloc = new_alloc;
  stat = MMB_OK;

BAILOUT_UNLOCK:
  UNLOCK_GLOBAL_FACTORY;

  return stat;
}

mmbError mmb_alloc_free(mmbAllocation *alloc)
{
  mmbError stat;
  /* Check parameter */
  stat = mmb_alloc_check_params(alloc);
  if (MMB_OK != stat) {
    MMB_WARN("Invalid \"alloc\" handle.\n");
    return MMB_INVALID_ALLOCATION;
  }
  LOCK_GLOBAL_FACTORY;

  if (NULL == global_factory) {
    MMB_ERR("Factory not initialized.\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT_UNLOCK;
  }
  /* Find proper set that contains the allocated alloc. */
  for (mmbAllocationFactory *factory = global_factory;
       NULL != factory;
       factory = factory->next) {
    if (alloc >= *factory->ptrs && alloc < &factory->ptrs[N_ELT_FACTORY][0]) {
      const ptrdiff_t offset = alloc - &factory->ptrs[0][0];
      const size_t block = offset / (8 * sizeof(unsigned int));
      const size_t index = offset % (8 * sizeof(unsigned int));
      factory->avail[block] |= 1<<index;
      stat = MMB_OK;
      goto BAILOUT_UNLOCK;
    }
  }
  /* Could not find the proper set : incoherent state that should have happen */
  MMB_ERR("Impossible to find the corresponding mmbAllocation record.\n");
  stat = MMB_ERROR;

BAILOUT_UNLOCK:
  UNLOCK_GLOBAL_FACTORY;
  return stat;
}
