/*
 * Copyright (C) 2018-2020 Cray UK
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <inttypes.h>
#include <pthread.h>
#include <assert.h>

#include "mmb_logging.h"
#include "mmb_memory.h"

#include "i_memory.h"
#include "memory_config.h"
#include "mmb_memory_options.h"

#if ENABLE_DISCOVERY
#include "topology.h"
#endif /* ENABLE_DISCOVERY */

#include "interface.h"
#include "allocation.h"

#include "strategy.h"
#include "strategies/basic.h"
#include "strategies/pooled.h"
#include "strategies/thread_safe.h"

static inline void mmb_manager_lock(mmbManager *mngr)
{
  int stat = pthread_mutex_lock(&mngr->spaces_mtx);
  assert(0 == stat);
}
static inline void mmb_manager_unlock(mmbManager *mngr)
{
  int stat = pthread_mutex_unlock(&mngr->spaces_mtx);
  assert(0 == stat);
}

static const struct mmb_dispatch_strategy_entry
mmb_strategies_table[MMB_STRATEGY__MAX] = {
  [MMB_STRATEGY_NONE]
    = { .name = "NONE",
        .init               = mmb_strat_init_interface__basic,
        .finalize           = mmb_strat_finalize_interface__basic,
        .allocate           = mmb_strat_allocate__basic,
        .free               = mmb_strat_free__basic,
        .copy               = mmb_strat_copy__basic,
    },
  [MMB_POOLED]
    = { .name = "POOLED",
        .init               = mmb_strat_init_interface__pooled,
        .finalize           = mmb_strat_finalize_interface__pooled,
        .allocate           = mmb_strat_allocate__pooled,
        .free               = mmb_strat_free__pooled,
        .copy               = mmb_strat_copy__pooled,
    },
  [MMB_THREAD_SAFE]
    = { .name = "THREAD SAFE",
        .init               = mmb_strat_init__thread_safe,
        .finalize           = mmb_strat_finalize__thread_safe,
        .allocate           = mmb_strat_allocate__thread_safe,
        .free               = mmb_strat_free__thread_safe,
        .copy               = mmb_strat_copy__thread_safe,
    },
};

/* Possibly modifiable at configure/compile time and at runtime with
 * environment variables. */
static const mmbDefaultContext GLOBAL_DEFAULT_CONTEXT = {
  .provider          = MMB_CONFIG_PROVIDER_DEFAULT,
  .strategy          = MMB_CONFIG_STRATEGY_DEFAULT,
  .execution_context = { MMB_CPU, MMB_CONFIG_EXECUTION_CONTEXT_GPU_DEFAULT },
};

/**
 * @brief Translate the \p in_strat to a value that can be used by the strategy
 *        building blocks.
 *
 * @param [in] in_strat Requested strategy (can be \c MMB_STRATEGY_DEFAULT).
 * @param [in] default_strategy The value to assign \p out_strat if \p in_strat
 *                              equals \c MMB_STRATEGY_DEFAULT).
 * @param [out] out_strat The proper strategy value (cannot be
 *                        \c MMB_STRATEGY_DEFAULT).
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
static mmbError
check_and_set_mmbstrategy(const mmbStrategy in_strat,
                          const mmbStrategy default_strategy,
                          mmbStrategy *out_strat)
{
  mmbError status = MMB_OK;

  /* Check parameters */
  if (NULL == out_strat) {
    MMB_ERR("Out parameter cannot be NULL.\n");
    return MMB_INVALID_ARG;
  }

  switch (in_strat) {
    case MMB_STRATEGY_ANY:
    case MMB_STRATEGY_DEFAULT:
      *out_strat = default_strategy;
      break;
    case MMB_POOLED:
    case MMB_THREAD_SAFE:
    case MMB_STRATEGY_NONE:
      *out_strat = in_strat;
      break;
    case MMB_LIMITED_GREEDY:
      MMB_WARN("This strategy (%s) is not yet available.\n",
           mmb_strategy_get_string(in_strat));
      status = MMB_UNIMPLEMENTED;
      break;
    default:
      MMB_ERR("Invalid strategy requested (%d).\n", in_strat);
      status = MMB_INVALID_STRATEGY;
  }
  return status;
}

/**
 * @brief Translate the \p in_provider to a value that can be used by the
 *        strategy building blocks.
 *
 * @param [in] in_provider Requested provider (can be \c MMB_PROVIDER_DEFAULT).
 * @param [in] default_provider The value to assign \p out_provider if
 *                              \p in_provider equals \c MMB_PROVIDER_DEFAULT).
 * @param [out] out_provider The proper provider value (cannot be
 *                           \c MMB_PROVIDER_DEFAULT).
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
static mmbError
check_and_set_mmbprovider(const mmbProvider in_provider,
                          const mmbProvider default_provider,
                          mmbProvider *out_provider)
{
  mmbError status = MMB_OK;

  /* Check parameters */
  if (NULL == out_provider) {
    MMB_ERR("Out parameter cannot be NULL.\n");
    return MMB_INVALID_ARG;
  }

  switch (in_provider) {
    case MMB_PROVIDER_ANY:
    case MMB_PROVIDER_DEFAULT:
      *out_provider = default_provider;
      break;
    case MMB_NATIVE:
    case MMB_SICM:
      *out_provider = in_provider;
      break;
    default:
      MMB_ERR("Invalid memory provider requested (%d).\n", in_provider);
      status = MMB_INVALID_PROVIDER;
  }
  return status;
}

/**
 * @brief Translate the \p in_ex_con to a value that can be used by the space.
 *
 * @param [in] in_ex_con Requested strategy (can be
 *                       \c MMB_EXECUTION_CONTEXT_DEFAULT).
 * @param [in] default_ex_con The value to assign \p out_ex_con if \p in_ex_con
 *                            equals \c MMB_EXECUTION_CONTEXT_DEFAULT).
 * @param [out] out_ex_con The proper strategy value (cannot be
 *                         \c MMB_EXCUTION_CONTEXT_DEFAULT).
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
static mmbError
check_and_set_mmbexecutioncontext(const mmbExecutionContext in_ex_con,
                                  const mmbExecutionContext default_ex_con,
                                  mmbExecutionContext *out_ex_con)
{
  mmbError status = MMB_OK;

  /* Check parameters */
  if (NULL == out_ex_con) {
    MMB_ERR("Out parameter cannot be NULL.\n");
    return MMB_INVALID_ARG;
  }

  switch (in_ex_con) {
    case MMB_EXECUTION_CONTEXT_DEFAULT:
      if (MMB_EXECUTION_CONTEXT_NONE == default_ex_con) {
        MMB_ERR("No default execution context available for the requested layer.\n");
        status = MMB_INVALID_EXECUTION_CONTEXT;
      } else {
        *out_ex_con = default_ex_con;
      }
      break;
    case MMB_CPU:
    case MMB_GPU_CUDA:
      *out_ex_con = in_ex_con;
      break;
    case MMB_GPU_OPENCL:
      MMB_WARN("This execution context \"%s\" is not yet available.\n",
           mmb_execution_context_get_string(in_ex_con));
      status = MMB_UNIMPLEMENTED;
      break;
    default:
      MMB_ERR("Invalid execution context requested (%d).\n", in_ex_con);
      status = MMB_INVALID_EXECUTION_CONTEXT;
  }
  return status;
}

/**
 * @name mmbMemSpace
 * @{ */

/**
 * @name Internal API
 * @{ */

static inline void mmb_memspace_lock_interfaces(mmbMemSpace *space)
{
  int stat = pthread_mutex_lock(&space->interfaces_mtx);
  assert(0 == stat);
}
static inline void mmb_memspace_unlock_interfaces(mmbMemSpace *space)
{
  int stat = pthread_mutex_unlock(&space->interfaces_mtx);
  assert(0 == stat);
}
static inline void mmb_memspace_lock_defaults(mmbMemSpace *space)
{
  int stat = pthread_mutex_lock(&space->defaults_mtx);
  assert(0 == stat);
}
static inline void mmb_memspace_unlock_defaults(mmbMemSpace *space)
{
  int stat = pthread_mutex_unlock(&space->defaults_mtx);
  assert(0 == stat);
}
static inline void mmb_memspace_lock(mmbMemSpace *space)
{
  mmb_memspace_lock_interfaces(space);
  mmb_memspace_lock_defaults(space);
  mmb_memspace_lock_size(space);
  mmb_memspace_rdlock_max_bytes(space);
}
static inline void mmb_memspace_unlock(mmbMemSpace *space)
{
  mmb_memspace_unlock_size(space);
  mmb_memspace_unlock_defaults(space);
  mmb_memspace_unlock_interfaces(space);
  mmb_memspace_unlock_max_bytes(space);
}

static mmbError
mmb_memspace_create(const mmbMemLayer layer, const size_t max_bytes,
                    const mmbExecutionContext ex_context,
                    const mmbProvider default_provider,
                    const mmbStrategy default_strat,
                    const mmbDefaultContext *global_default_ctx,
                    mmbMemSpace **out_space)
{
  mmbError stat = MMB_OK;

  /* Check parameters */
  if (NULL == out_space) {
    MMB_ERR("Out parameter cannot be NULL.\n");
    return MMB_INVALID_ARG;
  }
  if (MMB_MEMLAYER__MAX <= layer) {
    MMB_ERR("Invalid layer requested (%d).\n", layer);
    stat = MMB_INVALID_LAYER;
    goto BAILOUT;
  }
  mmbStrategy strategy = global_default_ctx->strategy;
  stat = check_and_set_mmbstrategy(default_strat, strategy, &strategy);
  if (MMB_OK != stat) {
    goto BAILOUT;
  }
  mmbProvider provider = global_default_ctx->provider;
  stat = check_and_set_mmbprovider(default_provider, provider, &provider);
  if (MMB_OK != stat) {
    goto BAILOUT;
  }
  mmbExecutionContext ex_con;
  ex_con = global_default_ctx->execution_context[MMB_GDRAM == layer];
  stat = check_and_set_mmbexecutioncontext(ex_context, ex_con, &ex_con);
  if (MMB_OK != stat) {
    goto BAILOUT;
  }

  /* Check whether the requested provider can provide support
   * for the requested layer */
  bool check_provider;
  stat = mmb_provider_compatibility_check(provider, layer, ex_con, &check_provider);
  if (MMB_OK != stat) {
    goto BAILOUT;
  } else if (false == check_provider) {
    stat = MMB_INVALID_PROVIDER;
    MMB_WARN("The requested provider (%s) cannot provide support for the requested "
         "memory layer (%s) with the required execution context (%s).\n",
         mmb_provider_get_string(provider), mmb_layer_get_string(layer),
         mmb_execution_context_get_string(ex_con));
    goto BAILOUT;
  }

  /* This should be a space specific initialisation
   * e.g. to set default interface (at compile time) */
  mmbMemSpace *s = calloc (1, sizeof(mmbMemSpace));
  if (s == NULL) {
    MMB_ERR("Failed to allocate memory space.\n");
    stat = MMB_OUT_OF_MEMORY;
    goto BAILOUT;
  }
  mmbMemInterface **interfaces = calloc(MMB_MAX_INTERFACES, sizeof *interfaces);
  if (NULL == interfaces) {
    MMB_ERR("Failed to allocate the array of interfaces.\n");
    stat = MMB_OUT_OF_MEMORY;
    free(s);
    goto BAILOUT;
  }
  s->interfaces = interfaces;
  s->max_interfaces = MMB_MAX_INTERFACES;
  s->layer = layer;
  s->default_strategy = strategy;
  s->default_provider = provider;
  s->max_bytes = max_bytes;
  s->ex_context = ex_con;

  /* Initialize the mutexes */
  {
    int status;
    status = pthread_mutex_init(&s->defaults_mtx, NULL);
    if (0 != status) {
      MMB_ERR("Unable to initialize defaults' mutex (err=%d).\n", status);
      stat = MMB_ERROR;
      free(s);
      goto BAILOUT;
    }
    status = pthread_mutex_init(&s->interfaces_mtx, NULL);
    if (0 != status) {
      MMB_ERR("Unable to initialize space's mutex (err=%d).\n", status);
      stat = MMB_ERROR;
      free(s);
      goto BAILOUT;
    }
    status = pthread_mutex_init(&s->alloc_size_mtx, NULL);
    if (0 != status) {
      MMB_ERR("Unable to initialize allocated size's mutex (err=%d).\n", status);
      stat = MMB_ERROR;
      free(s);
      goto BAILOUT;
    }
    status = pthread_rwlock_init(&s->max_bytes_mtx, NULL);
    if (0 != status) {
      MMB_ERR("Unable to initialize read-write lock on space's maximum allocated "
          "size (err=%d).\n", status);
      stat = MMB_ERROR;
      free(s);
      goto BAILOUT;
    }
  }

  /* Create default interface */
  mmbMemInterface *interface;
  stat = mmb_meminterface_create(provider, s->default_strategy, &interface);
  if (MMB_OK != stat) {
    MMB_ERR("Failed to allocate memory interface.\n");
    free(s);
    goto BAILOUT;
  }
  interface->space = s;
  stat = mmb_meminterface_init(interface);
  if (stat != MMB_OK) {
    MMB_ERR("Failed to initialise memory interface.\n");
    mmb_meminterface_destroy(interface);
    free(s);
    goto BAILOUT;
  }

  /* TODO: Replace this with hashmap on memory provider */
  s->interfaces[s->num_interfaces++] = interface;

  *out_space = s;
BAILOUT:
  return stat;
}

static mmbError mmb_memspace_destroy(mmbMemSpace *space)
{
  mmbError stat = MMB_OK;
  /* Check parameter */
  if (NULL == space) {
    MMB_WARN("Invalid space (cannot be NULL).\n");
    return MMB_INVALID_ARG;
  }

  /* Free all interfaces */
  for (unsigned i = 0; i < space->num_interfaces; i++) {
    stat = mmb_meminterface_finalize(space->interfaces[i]);
    if (MMB_OK != stat) {
      MMB_WARN("Failed to finalize memory interface.\n");
      goto BAILOUT;
    }
    stat = mmb_meminterface_destroy(space->interfaces[i]);
    if (stat != MMB_OK) {
      MMB_WARN("Failed to destroy memory interface.\n");
      goto BAILOUT;
    }
  }
  free(space->interfaces);

  /* Destroy space mutexes */
  {
    int status;
    status = pthread_mutex_destroy(&space->defaults_mtx);
    if (0 != status) {
      MMB_ERR("Unable to destroy defaults' mutex (err=%d).\n", status);
      stat = MMB_ERROR;
      goto BAILOUT;
    }
    status = pthread_mutex_destroy(&space->interfaces_mtx);
    if (0 != status) {
      MMB_ERR("Unable to destroy space's mutex (err=%d).\n", status);
      stat = MMB_ERROR;
      goto BAILOUT;
    }
    status = pthread_mutex_destroy(&space->alloc_size_mtx);
    if (0 != status) {
      MMB_ERR("Unable to destroy allocated size's mutex (err=%d).\n", status);
      stat = MMB_ERROR;
      goto BAILOUT;
    }
  }

  free(space);
BAILOUT:
  return stat;
}

/**  @} */

mmbError mmb_memspace_configure_defaults(const struct mmbManager *mngr,
                                         const struct mmbMemSpaceConfig *config_opts,
                                         mmbMemSpace *space)
{
  mmbError stat = MMB_OK;
  if (NULL == space) {
    MMB_ERR("Invalid space (cannot be NULL).\n");
    return MMB_INVALID_ARG;
  }
  if (NULL == config_opts) {
    MMB_WARN("NULL pointer given for configure option, no change to be done.\n");
    return MMB_OK;
  }

  mmbProvider provider = config_opts->interface_opts.provider;
  mmbStrategy strategy = config_opts->interface_opts.strategy;

  mmb_memspace_lock_defaults(space);
  if (MMB_PROVIDER_ANY != config_opts->interface_opts.provider) {
    stat = check_and_set_mmbprovider(provider, mngr->default_ctx.provider,
                                     &provider);
    if (MMB_OK != stat) {
      goto UNLOCK_BAILOUT;
    }
  }
  if (MMB_STRATEGY_ANY != config_opts->interface_opts.strategy) {
    stat = check_and_set_mmbstrategy(strategy, mngr->default_ctx.strategy,
                                     &strategy);
    if (MMB_OK != stat) {
      goto UNLOCK_BAILOUT;
    }
  }
  stat = mmb_memspace_resize(&config_opts->size_opts, space);
  if (MMB_OK != stat) {
    goto UNLOCK_BAILOUT;
  }

  /* Update only if no error occured */
  space->default_provider = provider;
  space->default_strategy = strategy;

UNLOCK_BAILOUT:
  mmb_memspace_unlock_defaults(space);

  return stat;
}

mmbError mmb_memspace_resize(const mmbSizeConfig *size_opts, mmbMemSpace *space)
{
  mmbError stat = MMB_OK;
  /* check parameters */
  if (NULL == size_opts) { // Should never happen
    stat = MMB_INVALID_ARG;
    MMB_ERR("Invalid size argument.\n");
    goto BAILOUT;
  }
  /* Set size if not given MMB_SET_ANY */
  size_t new_size = 0;
  switch (size_opts->action) {
    case MMB_SIZE_ANY:
      goto BAILOUT;
    case MMB_SIZE_SET:
      new_size = size_opts->mem_size;
      /* fallthrough */
    case MMB_SIZE_NONE:
      mmb_memspace_wrlock_max_bytes(space);
      mmb_memspace_lock_size(space);
      if (new_size < space->allocated_bytes) {
        MMB_ERR("The size requested (%zu) cannot be smaller than the already "
            "allocated size (%zu).\n", new_size, space->allocated_bytes);
        stat = MMB_INVALID_SIZE;
      } else {
        space->max_bytes = new_size;
      }
      mmb_memspace_unlock_size(space);
      mmb_memspace_unlock_max_bytes(space);
      break;
    default:
      MMB_ERR("Invalid size action parameter.\n");
      stat = MMB_INVALID_ARG;
  }
BAILOUT:
  return stat;
}

mmbError mmb_memspace_request_interface(mmbMemSpace *space,
                                        const struct mmbMemInterfaceConfig *config_opts,
                                        mmbMemInterface **out_interface)
{
  mmbError stat = MMB_OK;
  mmbMemInterface *interface = NULL;

  /* check parameters */
  if (NULL == space) {
    MMB_ERR("Invalid space (cannot be NULL).\n");
    return MMB_INVALID_ARG;
  }
  const mmbMemInterfaceConfig default_config = MMB_MEMINTERFACE_CONFIG_DEFAULT;
  if (NULL == config_opts) {
    config_opts = &default_config;
  }
  if (NULL == out_interface) {
    MMB_ERR("Out parameter cannot be NULL.\n");
    return MMB_INVALID_ARG;
  }
  mmbStrategy strategy = space->default_strategy;
  stat = check_and_set_mmbstrategy(config_opts->strategy, strategy, &strategy);
  if (MMB_OK != stat) {
    return stat;
  }
  mmbProvider provider = space->default_provider;
  stat = check_and_set_mmbprovider(config_opts->provider, provider, &provider);
  if (MMB_OK != stat) {
    return stat;
  }

  mmb_memspace_lock_interfaces(space);

  /* Lookup to be replaced with a hashmap research */
  for (size_t i = 0; i < space->num_interfaces; ++i) {
    interface = space->interfaces[i];
    if (NULL != interface) {
      if (   (MMB_STRATEGY_ANY == config_opts->strategy
              || strategy == interface->strategy)
          && (MMB_PROVIDER_ANY == config_opts->provider
              || provider == interface->provider)
         ) {
        *out_interface = interface;
        goto unlock_end;
      }
    }
  }

  /* No suitable provider found. Try to create a new one */
  if (space->num_interfaces == space->max_interfaces) {
    size_t max_interfaces = space->max_interfaces + MMB_MAX_INTERFACES;
    size_t realloc_size = sizeof(mmbMemInterface *[max_interfaces]);
    mmbMemInterface **interfaces = realloc(space->interfaces, realloc_size);
    if (NULL == interfaces) {
      MMB_ERR("No slot available for any extra interface and "
          "the reallocation failed.\n");
      stat = MMB_OUT_OF_MEMORY;
      goto unlock_end;
    }
    space->max_interfaces = max_interfaces;
    space->interfaces = interfaces;
  }
  stat = mmb_meminterface_create(provider, strategy, &interface);
  if (MMB_OK != stat) {
    goto unlock_end;
  }
  interface->space = space;
  stat = mmb_meminterface_init(interface);
  if (MMB_OK != stat) {
    goto unlock_end;
  }
  space->interfaces[space->num_interfaces++] = interface;

  *out_interface = interface;

unlock_end:
  mmb_memspace_unlock_interfaces(space);
  return stat;
}

/**  @} */

mmbError mmb_meminterface_create(const mmbProvider provider,
                                 const mmbStrategy strategy,
                                 mmbMemInterface **out_interface)
{
  mmbError stat = MMB_OK;

  /* Check parameters */
  if (NULL == out_interface) {
    MMB_ERR("Out parameter cannot be NULL.\n");
    return MMB_INVALID_ARG;
  }

  size_t alloc_size_data;
  stat = mmb_strat_get_data_size(provider, strategy, &alloc_size_data);
  if (MMB_OK != stat) {
    return stat;
  }

  mmbMemInterface *interface;
  interface = malloc(alloc_size_data + sizeof(mmbMemInterface));
  if (interface == NULL) {
    MMB_ERR("Failed to allocate new memory space (out of memory).\n");
    return MMB_OUT_OF_MEMORY;
  }

  interface->provider = provider;
  interface->strategy = strategy;
  *out_interface = interface;

  return stat;
}

mmbError mmb_meminterface_destroy(mmbMemInterface *interface)
{
  if (MMB_OK == mmb_meminterface_check_params(interface)) {
    free(interface);
    return MMB_OK;
  } else {
    MMB_WARN("Memory interface already freed or invalid interface handle.\n");
    return MMB_INVALID_ARG;
  }
}

mmbError mmb_meminterface_init(mmbMemInterface *interface)
{
  mmbError stat = MMB_OK;
  /* Check parameters */
  stat = mmb_meminterface_check_params(interface);
  if (MMB_OK != stat) {
    return stat;
  }
  stat = mmb_strategies_table[interface->strategy]
          .init(interface);
  return stat;
}

mmbError mmb_meminterface_finalize(mmbMemInterface *interface)
{
  mmbError stat = MMB_OK;
  /* Check parameters */
  stat = mmb_meminterface_check_params(interface);
  if (MMB_OK != stat) {
    return stat;
  }
  stat = mmb_strategies_table[interface->strategy]
          .finalize(interface);
  return stat;
}

mmbError mmb_meminterface_check_params(const mmbMemInterface *interface)
{
  if (NULL == interface) {
    MMB_WARN("Invalid interface (cannot be NULL).\n");
    return MMB_INVALID_ARG;
  } else if (NULL == interface->space) {
    MMB_WARN("Invalid space handle in given interface (cannot be NULL).\n");
    return MMB_INVALID_ARG;
  }
  return MMB_OK;
}

mmbError mmb_allocation_create_wrapped(void* in_ptr, const size_t n_bytes,
                                       mmbMemInterface *interface,
                                       mmbAllocation **out_allocation)
{
  mmbError stat = MMB_OK;

  /* Check_parameters */
  stat = mmb_meminterface_check_params(interface);
  if (MMB_OK != stat) {
    return stat;
  }

  stat = mmb_alloc_allocate(in_ptr, n_bytes, interface, false, out_allocation);
  if (MMB_OK != stat) {
    MMB_WARN("Failed to allocate structure for mmbAllocation object.\n");
  }
  return stat;
}

mmbError mmb_allocation_destroy(mmbAllocation *in_allocation)
{
  /* Check parameters */
  mmbError stat = mmb_alloc_check_params(in_allocation);
  if (MMB_OK != stat) {
    return stat;
  }
  if (in_allocation->owned) {
    return mmb_free(in_allocation);
  } else {
    return mmb_alloc_free(in_allocation);
  }
}

mmbError mmb_allocate(const size_t n_bytes, mmbMemInterface *interface,
                      mmbAllocation **out_allocation)
{
  mmbError stat = MMB_OK;

  /* Check parameters */
  stat = mmb_meminterface_check_params(interface);
  if (MMB_OK != stat) {
    return stat;
  }

  mmbAllocation *a;
  stat = mmb_alloc_allocate(NULL, n_bytes, interface, true, &a);
  if (stat != MMB_OK) {
    MMB_WARN("Could not allocate mmbAllocation object.\n");
    return stat;
  }

  stat = mmb_strategies_table[interface->strategy]
          .allocate(n_bytes, interface, a);

  /* set structure if everything went smoothly. Otherwise, free allocation and
   * propagate error. */
  if (MMB_OK != stat) {
    mmb_alloc_free(a);
    return stat;
  }

  *out_allocation = a;

  return stat;
}

mmbError mmb_free(mmbAllocation *allocation)
{
  mmbError stat = MMB_OK;

  /* Check parameters */
  stat = mmb_alloc_check_params(allocation);
  if (MMB_OK != stat) {
    MMB_ERR("Failed mmb_alloc_check_params.\n");
    return stat;
  }

  if (false == allocation->owned) {
    MMB_ERR("Wrong allocation handle. Allocation should be owned.\n");
    return MMB_INVALID_ARG;
  }

  stat = mmb_strategies_table[allocation->interface->strategy]
          .free(allocation);
  if (MMB_OK != stat) {
    MMB_ERR("Failed strategy free on allocation.\n");
    return stat;
  }

  stat = mmb_alloc_free(allocation);
  if (MMB_OK != stat) {
    MMB_ERR("Failed mmb_alloc_free on allocation.\n");
    return stat;
  }
  return stat;
}

mmbError mmb_copy(mmbAllocation *dst, mmbAllocation *src)
{
  mmbError stat = MMB_OK;

  /* Check parameters */
  stat = mmb_alloc_check_params(dst);
  if (MMB_OK != stat) {
    return stat;
  }
  stat = mmb_alloc_check_params(src);
  if (MMB_OK != stat) {
    return stat;
  }

  stat = mmb_strategies_table[dst->interface->strategy]
          .copy(dst, src);

  return stat;
}

mmbError mmb_allocation_get_slice(mmbAllocation *in_alloc,
                                  const ptrdiff_t offset, const size_t n_bytes,
                                  mmbAllocation **out_allocation)
{
  mmbError stat = MMB_OK;

  /* Check parameters */
  stat = mmb_alloc_check_params(in_alloc);
  if (MMB_OK != stat) {
    return stat;
  }

  void *in_ptr = (char *)in_alloc->ptr + offset;
  mmbMemInterface *interface = in_alloc->interface;
  stat = mmb_allocation_create_wrapped(in_ptr, n_bytes, interface, out_allocation);
  if (MMB_OK != stat) {
    MMB_WARN("Unable to wrap the sub slice of the given allocation.\n");
  }
  return stat;
}

mmbError mmb_allocation_get_ptr(const mmbAllocation *in_alloc, void **out_ptr)
{
  mmbError stat = MMB_OK;
  /* Check parameters */
  stat = mmb_alloc_check_params(in_alloc);
  if (MMB_OK != stat) {
    return stat;
  }
  if (NULL == out_ptr) {
    MMB_ERR("Invalid \"out_ptr\" handle. Parameter cannot be NULL.\n");
    return MMB_INVALID_ARG;
  }
  *out_ptr = in_alloc->ptr;
  return stat;
}

mmbError mmb_allocation_get_layer(const mmbAllocation *alloc, mmbMemLayer *layer)
{
  mmbError stat = MMB_OK;
  /* Check parameters */
  stat = mmb_alloc_check_params(alloc);
  if (MMB_OK != stat) {
    return stat;
  }
  *layer = alloc->interface->space->layer;
  return stat;
}

mmbError mmb_allocation_get_execution_context(const mmbAllocation *alloc,
                                              mmbExecutionContext *ex_context)
{
  mmbError stat = MMB_OK;
  /* Check parameters */
  stat = mmb_alloc_check_params(alloc);
  if (MMB_OK != stat) {
    return stat;
  }
  *ex_context = alloc->interface->space->ex_context;
  return stat;
}

mmbError mmb_allocation_get_provider(const mmbAllocation *alloc,
                                           mmbProvider *provider)
{
  mmbError stat = MMB_OK;
  /* Check parameters */
  stat = mmb_alloc_check_params(alloc);
  if (MMB_OK != stat) {
    return stat;
  }
  *provider = alloc->interface->provider;
  return stat;
}

mmbError mmb_allocation_get_strategy(const mmbAllocation *alloc,
                                     mmbStrategy *strat)
{
  mmbError stat = MMB_OK;
  /* Check parameters */
  stat = mmb_alloc_check_params(alloc);
  if (MMB_OK != stat) {
    return stat;
  }
  *strat = alloc->interface->strategy;
  return stat;
}

static mmbError mmb_manager_initialize(mmbManager *mngr)
{
  mmbError stat = MMB_OK;

  /* Initialize mutex */
  {
    int status = pthread_mutex_init(&mngr->spaces_mtx, NULL);
    if (0 != status) {
      MMB_ERR("Unable to initialize the manager's mutex (err=%d).\n", status);
      stat = MMB_ERROR;
      goto BAILOUT;
    }
  }

#if ENABLE_DISCOVERY
  uint64_t size;
  uint64_t layers[MMB_MEMLAYER__MAX] = { 0 };
  uint64_t ex_cons[MMB_EXECUTION_CONTEXT__MAX] = { 0 };

  stat = mmb_topology_init(&mngr->topology, layers, ex_cons);
  if (MMB_OK != stat) {
    goto BAILOUT;
  }

  // TODO find a way to pass the numa id back from topology ?
  mmbMemSpaceConfig space_conf = {
    .interface_opts = {
      .provider = MMB_PROVIDER_ANY,
      .strategy = MMB_STRATEGY_ANY,
    },
  };

  for (int layer = MMB_DRAM; MMB_MEMLAYER__MAX > layer; ++layer) {
    size = layers[layer];
    if (MMB_GDRAM != layer && 0 < size) {
      /* register memory */
      space_conf.size_opts.mem_size = size;
      stat = mmb_manager_register_memory(mngr, layer, MMB_CPU, &space_conf, NULL);
      if (MMB_OK == stat) {
        MMB_DEBUG("layer: %s, context: MMB_CPU, size: %" PRIu64 "\n",
              mmb_layer_get_string(layer), size);
      } else {
        MMB_WARN("Unable to register MMB_CPU for layer #%d.\n", layer);
        stat = MMB_OK;
      }
    }
  }
  for (int ec = MMB_CPU; MMB_EXECUTION_CONTEXT__MAX > ec; ++ec) {
    /* Check which contexts are valid */
    size = ex_cons[ec];
    if (0 < size) {
      /* register memory */
      space_conf.size_opts.mem_size = size;
      stat = mmb_manager_register_memory(mngr, MMB_GDRAM, ec, &space_conf, NULL);
      if (MMB_OK == stat) {
        MMB_DEBUG("layer: MMB_GDRAM, context: %s, size: %" PRIu64 "\n",
              mmb_execution_context_get_string(ec), size);
      } else {
        MMB_WARN("Unable to register execution context %s for GDRAM.\n",
             mmb_execution_context_get_string(ec));
        stat = MMB_OK;
      }
    }
  }
#endif /* ENABLE_DISCOVERY */

  /* Check the validity of compile defined GLOBAL_DEFAULT_CONTEXT variable */
  mmbDefaultContext default_ctx = GLOBAL_DEFAULT_CONTEXT;
  const char *global_default_provider_env = MMB_CONFIG_PROVIDER_DEFAULT_ENV_NAME;
  const char *global_default_provider_str = getenv(global_default_provider_env);
  if (NULL != global_default_provider_str) {
    /* provider overwritten at run-time in environment */
    default_ctx.provider = mmb_provider_get_symbol(global_default_provider_str);
  } else {
    global_default_provider_str = mmb_provider_get_string(default_ctx.provider);
  }
  switch (default_ctx.provider) {
    case MMB_NATIVE:
    case MMB_SICM:
      break;
    default:
      MMB_ERR("Default memory provider misconfigured (%s). Please rebuild the "
          "library with another value or set environment variable %s.\n",
          global_default_provider_str, global_default_provider_env);
      stat = MMB_ERROR;
      goto BAILOUT;
  }
  const char *global_default_strategy_env = MMB_CONFIG_STRATEGY_DEFAULT_ENV_NAME;
  const char *global_default_strategy_str = getenv(global_default_strategy_env);
  if (NULL != global_default_strategy_str) {
    /* strategy overwritten at run-time in environment */
    default_ctx.strategy = mmb_strategy_get_symbol(global_default_strategy_str);
  } else {
    global_default_strategy_str = mmb_strategy_get_string(default_ctx.strategy);
  }
  switch (default_ctx.strategy) {
    case MMB_STRATEGY_NONE:
    case MMB_POOLED:
    case MMB_THREAD_SAFE:
      break;
    case MMB_LIMITED_GREEDY:
      MMB_WARN("This strategy (%s) is not yet available.\n",
           mmb_strategy_get_string(default_ctx.strategy));
      /* FALLTHROUGH */
    default:
      MMB_ERR("Default strategy misconfigured (%s). Please rebuild the "
          "library with another value or set environment variable %s.\n",
          global_default_strategy_str, global_default_strategy_env);
      stat = MMB_ERROR;
      goto BAILOUT;
  }
  const char *global_default_ex_con_env = MMB_CONFIG_EXECUTION_CONTEXT_GPU_DEFAULT_ENV_NAME;
  const char *global_default_ex_con_str = getenv(global_default_ex_con_env);
  if (NULL != global_default_ex_con_str) {
    /* execution context overwritten at run-time in environment */
    default_ctx.execution_context[1] =
      mmb_execution_context_get_symbol(global_default_ex_con_str);
  } else {
    global_default_ex_con_str =
      mmb_execution_context_get_string(default_ctx.execution_context[1]);
  }
  switch (default_ctx.execution_context[1]) {
    case MMB_CPU:
    case MMB_GPU_CUDA:
    case MMB_EXECUTION_CONTEXT_NONE:
      break;
    case MMB_GPU_OPENCL:
      MMB_WARN("This execution (%s) context is not yet available.\n",
           mmb_execution_context_get_string(default_ctx.execution_context[1]));
      /* FALLTHROUGH */
    default:
      MMB_ERR("Invalid default execution context requested for GPUs (%s). Please "
          "rebuild the library with another value or set environment variable "
          "%s.\n", global_default_ex_con_str, global_default_ex_con_env);
      stat = MMB_ERROR;
      goto BAILOUT;
  }
  mngr->default_ctx = default_ctx;

BAILOUT:
  return stat;
}

mmbError mmb_manager_create(mmbManager **out_mngr)
{
  /* Creates a new manager instance */
  mmbError stat = MMB_OK;

  /* Check parameters */
  if (NULL == out_mngr) {
    MMB_ERR("Out parameter cannot be NULL.\n");
    return MMB_INVALID_ARG;
  }

  mmbManager *mngr = calloc(1, sizeof(mmbManager));
  if (NULL == mngr) {
    MMB_ERR("Error, malloc of manager failed.\n");
    stat = MMB_OUT_OF_MEMORY;
    goto BAILOUT;
  }

  stat = mmb_manager_initialize(mngr);
  if (MMB_OK != stat) {
    MMB_WARN("Failed to initializing manager (%s).\n",
         mmb_error_description(stat));
    goto BAILOUT;
  }

  *out_mngr = mngr;
BAILOUT:
  return stat;
}

mmbError mmb_manager_destroy(mmbManager *mngr)
{
  mmbError stat = MMB_OK;

  /* Check parameters */
  if (NULL == mngr) {
    MMB_WARN("Invalid manager handle (cannot be NULL).\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }

  mmb_manager_lock(mngr);

  /* free all spaces manager handles */
  if (mngr->num_spaces != 0) {
    for (size_t i = 0; i < mngr->num_spaces ; i++) {
      stat =  mmb_memspace_destroy(mngr->spaces[i]);
      if (stat != MMB_OK) {
        MMB_WARN("Unable to free memory space.\n");
        goto BAILOUT;
      }
    }
  }
  free(mngr->spaces);

  mmb_manager_unlock(mngr);

  /* Finalize mutex */
  {
    int status = pthread_mutex_destroy(&mngr->spaces_mtx);
    if (0 != status) {
      MMB_ERR("Unable to destroy the manager's mutex (err=%d).\n", status);
      stat = MMB_ERROR;
      goto BAILOUT;
    }
  }

#if ENABLE_DISCOVERY
  mmb_topology_finalize(mngr->topology);
#endif /* ENABLE_DISCOVERY */

  /* free the manager */
  free(mngr);

BAILOUT:
  return stat;
}

mmbError mmb_manager_register_memory(mmbManager *mngr, const mmbMemLayer layer,
                                     const mmbExecutionContext ex_context,
                                     const mmbMemSpaceConfig *config_opts,
                                     mmbMemSpace **new_space)
{
  mmbError stat = MMB_OK;
  mmbMemSpace *space = NULL;

  /* Check parameters */
  if (NULL == mngr) {
    MMB_ERR("Invalid manager provided (cannot be NULL).\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }
  if (NULL == config_opts) {
    MMB_ERR("Invalid configuration handle (cannot be NULL).\n");
    return MMB_INVALID_ARG;
  }
  if (0 > layer || MMB_MEMLAYER__MAX <= layer) {
    MMB_ERR("Invalid memory layer provided.\n");
    return MMB_INVALID_LAYER;
  }
  if (-2 > ex_context || MMB_EXECUTION_CONTEXT__MAX <= ex_context) {
    MMB_ERR("Invalid execution context provided.\n");
    return MMB_INVALID_EXECUTION_CONTEXT;
  }

  /* creates a new manager instance and keeps track of spaces */

  switch (layer) {
    case MMB_DRAM:
    case MMB_GDRAM:
    case MMB_HBM:
    case MMB_NVDIMM:
      break;
    default:
      MMB_ERR("Only currently supported layers are %s, %s, %s and %s.\n",
           mmb_layer_get_string(MMB_DRAM),
           mmb_layer_get_string(MMB_GDRAM),
           mmb_layer_get_string(MMB_HBM),
           mmb_layer_get_string(MMB_NVDIMM));
      stat = MMB_UNIMPLEMENTED;
      goto BAILOUT;
  }

  mmb_manager_lock(mngr);

  /* Extend the number of slots for the spaces if required */
  if (mngr->num_spaces == mngr->max_spaces) {
    size_t max_spaces = mngr->max_spaces + MMB_MAX_SPACES;
    size_t realloc_size = sizeof(mmbMemSpace[max_spaces]);
    mmbMemSpace **spaces = realloc(mngr->spaces, realloc_size);
    if (NULL == spaces) {
      /* Same as missing space for adding a new interface to a space. See
       * function mmb_memspace_request_interface. */
      MMB_ERR("Error, MAX_SPACES exceeded (%d) and unable to reallocate!\n",
          mngr->max_spaces);
      stat = MMB_OUT_OF_MEMORY;
      goto UNLOCK_BAILOUT;
    }
    mngr->spaces = spaces;
    mngr->max_spaces = max_spaces;
  }

  stat = mmb_memspace_create(layer,
                             config_opts->size_opts.mem_size,
                             ex_context,
                             config_opts->interface_opts.provider,
                             config_opts->interface_opts.strategy,
                             &mngr->default_ctx,
                             &space);
  if (stat != MMB_OK) {
    MMB_WARN("Unable to create the requested new memspace (%s)\n",
         mmb_error_description(stat));
    goto UNLOCK_BAILOUT;
  }
  mngr->spaces[mngr->num_spaces++] = space;

UNLOCK_BAILOUT:
  mmb_manager_unlock(mngr);

BAILOUT:
  /* Returns newly created handle (if any space was created) */
  if (NULL != space && NULL != new_space) {
    *new_space = space;
  }

  return stat;
}

mmbError mmb_manager_request_space(mmbManager *mngr, const mmbMemLayer layer,
                                   const mmbExecutionContext ex_context,
                                   const mmbMemSpaceConfig *config_opts,
                                   mmbMemSpace **out_space)
{
  mmbError stat = MMB_OK;

  /* Check parameters */
  if (NULL == mngr) {
    MMB_ERR("Invalid manager handle (cannot be NULL).\n");
    stat = MMB_ERROR;
    goto BAILOUT;
  }
  if (NULL == out_space) {
    MMB_ERR("Out parameter cannot be NULL.\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }
  if (NULL == config_opts) {
    MMB_ERR("Invalid configuration handle (cannot be NULL).\n");
    return MMB_INVALID_ARG;
  }
  if (MMB_MEMLAYER__MAX <= layer) {
    MMB_ERR("Invalid layer requested (%d).\n", layer);
    stat = MMB_INVALID_LAYER;
    goto BAILOUT;
  }
  mmbExecutionContext ex_con;
  ex_con = mngr->default_ctx.execution_context[MMB_GDRAM == layer];
  stat = check_and_set_mmbexecutioncontext(ex_context, ex_con, &ex_con);
  if (MMB_OK != stat) {
    goto BAILOUT;
  }
  size_t request_size;
  switch (config_opts->size_opts.action) {
    case MMB_SIZE_ANY:
      request_size = 0;
      break;
    case MMB_SIZE_SET:
      request_size = config_opts->size_opts.mem_size;
      break;
    default:
      MMB_ERR("Invalid size option action value. Valid values are %s or %s.\n",
          mmb_size_config_action_get_string(MMB_SIZE_ANY),
          mmb_size_config_action_get_string(MMB_SIZE_SET));
      goto BAILOUT;
  }

  /* TODO: Replace with hash lookup */
  for (size_t i = 0; i < mngr->num_spaces; i++) {
    mmbMemSpace *sp = mngr->spaces[i];
    /* Protected reading of values that may be changed while we read */
    mmb_memspace_lock_defaults(sp);
    mmb_memspace_lock_size(sp);
    mmb_memspace_rdlock_max_bytes(sp);
    const mmbProvider space_default_provider = sp->default_provider;
    const mmbStrategy space_default_strategy = sp->default_strategy;
    const size_t available_size = sp->max_bytes - sp->allocated_bytes;
    mmb_memspace_unlock_defaults(sp);
    mmb_memspace_unlock_size(sp);
    mmb_memspace_unlock_max_bytes(sp);
    /* Test value */
    if (layer == sp->layer && ex_con == sp->ex_context
        && (MMB_PROVIDER_ANY == config_opts->interface_opts.provider
            || config_opts->interface_opts.provider == space_default_provider)
        && (MMB_STRATEGY_ANY == config_opts->interface_opts.strategy
            || config_opts->interface_opts.strategy == space_default_strategy)
        && request_size <= available_size
       ) {
      *out_space = mngr->spaces[i];
      goto BAILOUT;
    }
  }

  MMB_WARN("Space not found for layer \"%s\" with execution context \"%s\"\n",
       mmb_layer_get_string(layer), mmb_execution_context_get_string(ex_con));
  stat = MMB_NOT_FOUND;

BAILOUT:
  return stat;
}

mmbError mmb_manager_report_state(mmbManager *mngr, FILE *out)
{
  mmbError stat = MMB_OK;
  /* Check parameters */
  if (NULL == mngr) {
    MMB_ERR("Invalid manager handle parameter (cannot be NULL).\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }
  if (NULL == out) {
    MMB_ERR("Invalid output stream (cannot be NULL).\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }
  const unsigned LABEL_LEN = 15;
  const unsigned TAB_INDENT = 6;
  mmb_manager_lock(mngr);
  fprintf(out, "[");
  for (size_t i = 0; mngr->num_spaces > i; ++i) {
    fprintf(out, "\n%*s%2zu: {\n", 2, "", i);
    const unsigned INDENT = TAB_INDENT + 2;
    mmbMemSpace *sp = mngr->spaces[i];
    /* Locking info */
    mmb_memspace_lock(sp);
    /* Print info */
    fprintf(out, "%*s%*s: %s,\n", INDENT, "", LABEL_LEN,
        "layer", mmb_layer_get_string(sp->layer));
    fprintf(out, "%*s%*s: %s,\n", INDENT, "", LABEL_LEN,
        "exe context", mmb_execution_context_get_string(sp->ex_context));
    fprintf(out, "%*s%*s: %s,\n", INDENT, "", LABEL_LEN,
        "def. provider", mmb_provider_get_string(sp->default_provider));
    fprintf(out, "%*s%*s: %s,\n", INDENT, "", LABEL_LEN,
        "def. strategy", mmb_strategy_get_string(sp->default_strategy));
    fprintf(out, "%*s%*s: %zu,\n", INDENT, "", LABEL_LEN,
        "max_bytes", sp->max_bytes);
    fprintf(out, "%*s%*s: %zu,\n", INDENT, "", LABEL_LEN,
        "allocated bytes", sp->allocated_bytes);
    fprintf(out, "%*s%*s: %zu,\n", INDENT, "", LABEL_LEN,
        "available bytes", sp->max_bytes - sp->allocated_bytes);
    fprintf(out, "%*s%*s: %zu/%zu\n", INDENT, "", LABEL_LEN,
        "#interfaces", sp->num_interfaces, sp->max_interfaces);
    fprintf(out, "%*s    }%c", 2, "", mngr->num_spaces - 1 > i ? ',' : '\n');
    /* Unlocking info */
    mmb_memspace_unlock(sp);
  }
  fprintf(out, "]\n");
  fflush(stdout);
  mmb_manager_unlock(mngr);
BAILOUT:
  return stat;
}

