/*
 * Copyright (C) 2018-2020 Cray UK
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef MMB_MEMORY_H
#define MMB_MEMORY_H

#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#include <stdbool.h>

#include "mmb_error.h"
#include "memory_config.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @name mmb_memory
 * @{ */

/**
 * @name mmbStrategy
 * @{ */

/**
 * @name Public API
 * @{ */

typedef enum mmbMemLayer {
  MMB_DRAM = 0,
  MMB_GDRAM,
  MMB_HBM,
  MMB_NVDIMM, // Non-volatile
  MMB_NADRAM, // Network attached
  MMB_SSD,
  MMB_MEMLAYER__MAX
} mmbMemLayer;

/**  @} */

/**  @} */

/**
 * @name mmbStrategy
 * @{ */

/**
 * @name Public API
 * @{ */

typedef enum mmbStrategy {
  MMB_STRATEGY_ANY = -2,
  MMB_STRATEGY_DEFAULT = -1,
  MMB_STRATEGY_NONE,
  MMB_THREAD_SAFE,
  MMB_LIMITED_GREEDY,
  MMB_POOLED,
  MMB_STRATEGY__MAX
} mmbStrategy;

/**  @} */

/**  @} */

/**
 * @name mmbMemInterface
 * @{ */

/**
 * @name Public API
 * @{ */

typedef enum mmbProvider {
  MMB_PROVIDER_ANY = -2,
  MMB_PROVIDER_DEFAULT = -1,
  MMB_NATIVE,
  MMB_SICM,
  MMB_PROVIDER__MAX
} mmbProvider;

typedef struct mmbMemInterface mmbMemInterface;

/**  @} */

/**  @} */

/**
 * @name mmbExecutionContext
 * @{ */

/**
 * @name Public API
 * @{ */

typedef enum mmbExecutionContext {
  MMB_EXECUTION_CONTEXT_NONE = -2,
  MMB_EXECUTION_CONTEXT_DEFAULT = -1,
  MMB_CPU,
  MMB_GPU_CUDA,
  MMB_GPU_OPENCL,
  MMB_EXECUTION_CONTEXT__MAX
} mmbExecutionContext;

/**  @} */

/**  @} */

/**
 * @name mmbMemSpace
 * @{ */

/**
 * @name Public API
 * @{ */

/* Early declaration, with definition lower, but symbols have to be reserved
 * here */
struct mmbManager;
struct mmbMemInterfaceConfig;
struct mmbMemSpaceConfig;

typedef struct mmbMemSpace mmbMemSpace;

mmbError mmb_memspace_request_interface(mmbMemSpace *space,
                                        const struct mmbMemInterfaceConfig *config_opts,
                                        mmbMemInterface **out_interface);

mmbError mmb_memspace_configure_defaults(const struct mmbManager *mngr,
                                         const struct mmbMemSpaceConfig *config_opts,
                                         mmbMemSpace *space);

/**  @} */

/**  @} */

/**
 * @name mmbManager
 * @{ */

/**
 * @name Public API
 * @{ */

typedef struct mmbManager mmbManager;

mmbError mmb_manager_create(mmbManager **out_mngr);
mmbError mmb_manager_destroy(mmbManager *mngr);
/* /!\ config_opts->mem_size is given in bytes, not MiB. */
mmbError mmb_manager_register_memory(mmbManager *mngr, const mmbMemLayer layer,
                                     const mmbExecutionContext ex_context,
                                     const struct mmbMemSpaceConfig *config_opts,
                                     mmbMemSpace **out_space);
mmbError mmb_manager_request_space(mmbManager *mngr, const mmbMemLayer in_layer,
                                   const mmbExecutionContext ex_context,
                                   const struct mmbMemSpaceConfig *config_opts,
                                   mmbMemSpace **out_space);
mmbError mmb_manager_report_state(mmbManager *mngr, FILE *out);

/**  @} */

/**  @} */

/**
 * @name mmbAllocation
 * @{ */

/**
 * @name Public API
 * @{ */

typedef struct mmbAllocation {
  void         *ptr;
  size_t        n_bytes;
  struct mmbMemInterface *interface;
  bool          owned;
} mmbAllocation;

mmbError mmb_allocation_create_wrapped(void* in_ptr, const size_t n_bytes,
                                       mmbMemInterface *interface,
                                       mmbAllocation **out_allocation);
mmbError mmb_allocation_destroy(mmbAllocation *in_allocation);

mmbError mmb_allocate(const size_t n_bytes, mmbMemInterface *interface,
                      mmbAllocation **out_allocation);
mmbError mmb_free(mmbAllocation *allocation);
/**
 * The amount of data to copy is equal to
 * MIN( \p dest->n_bytes, \p src->n_bytes ).
 */
mmbError mmb_copy(mmbAllocation *dest, mmbAllocation *src);
mmbError mmb_allocation_get_slice(mmbAllocation *in_alloc,
                                  const ptrdiff_t offset, const size_t n_bytes,
                                  mmbAllocation **out_allocation);
mmbError mmb_allocation_get_ptr(const mmbAllocation *allocation, void **out_ptr);
/**
 * @brief Returns the layer the provided allocation is located to.
 *
 * @param [in]  alloc Handle to a valid mmbAllocation
 * @param [out] layer Pointer to a allocated mmbMemLayer
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_allocation_get_layer(const mmbAllocation *alloc, mmbMemLayer *layer);
/**
 * @brief Returns the execution context requested for the provided allocation.
 *
 * @param [in]  alloc Handle to a valid mmbAllocation
 * @param [out] ex_context Pointer to a allocated mmbExecutionContext
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_allocation_get_execution_context(const mmbAllocation *alloc,
                                              mmbExecutionContext *ex_context);
/**
 * @brief Returns the memory provider used to allocate the memory for the given
 *        \c mmbAllocation.
 *
 * @param [in]  alloc Handle to a valid mmbAllocation
 * @param [out] provider Pointer to a allocated mmbProvider
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_allocation_get_provider(const mmbAllocation *alloc,
                                     mmbProvider *provider);
/**
 * @brief Returns the strategy used to allocate the memory for the given
 *        \c mmbAllocation.
 *
 * @param [in]  alloc Handle to a valid mmbAllocation
 * @param [out] strat Pointer to a allocated mmbStrategy
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_allocation_get_strategy(const mmbAllocation *alloc,
                                     mmbStrategy *strat);

/**  @} */

/**  @} */

/**  @} */

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* MMB_MEMORY_H */
