/*
 * Copyright (C) 2018-2020 Cray UK
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdbool.h>
#include <string.h>

#include "mmb_error.h"
#include "mmb_logging.h"
#include "mmb_memory.h"

#include "../i_memory.h"
#include "../interface.h"
#include "linux_native.h"

/**
 * @name mmb_memory
 * @{ */

/**
 * @name mmbMemInterface
 * @{ */

/**
 * @name GNU/Linux native memory provider
 * @{ */

mmbError mmb_init__linux(mmbMemInterface *interface)
{
  (void) interface;
  MMB_DEBUG("Calling linux initializer()\n");
  return MMB_OK;
}

mmbError mmb_finalize__linux(mmbMemInterface *interface)
{
  (void) interface;
  MMB_DEBUG("Calling linux finalizer()\n");
  return MMB_OK;
}

mmbError mmb_supports__linux(const mmbMemLayer layer, bool *is_supported)
{
  if (NULL == is_supported) {
    return MMB_INVALID_ARG;
  }
  switch (layer) {
    case MMB_DRAM:
    case MMB_GDRAM:
      *is_supported = true;
      break;
    case MMB_HBM:
    case MMB_NVDIMM:
    case MMB_NADRAM:
    case MMB_SSD:
    default:
      *is_supported = false;
      break;
  }
  return MMB_OK;
}

/**
 * @name GNU/Linux native DRAM memory provider
 * @{ */

mmbError mmb_allocate__linux_dram(const size_t n_bytes,
                                  mmbMemInterface *interface,
                                  void **out_allocation)
{
  (void) interface;
  mmbError stat = MMB_OK;

  MMB_DEBUG("Calling linux_dram allocator()\n");

  void *ptr = malloc(n_bytes);
  if(NULL == ptr) {
    MMB_ERR("Could not allocate %zu bytes for mmbAllocation\n", n_bytes);
    stat = MMB_OUT_OF_MEMORY;
    goto BAILOUT;
  }
  *out_allocation = ptr;

BAILOUT:
  return stat;
}

mmbError mmb_free__linux_dram(void *allocation, mmbMemInterface *inter,
                              const size_t n_bytes)
{
  (void) n_bytes;
  mmbError stat = MMB_OK;

  MMB_DEBUG("Calling linux_dram deallocator(%p)\n", allocation);

  if (MMB_DRAM != inter->space->layer) {
    stat = MMB_INVALID_LAYER;
    goto BAILOUT;
  }

  free(allocation);

BAILOUT:
  return stat;
}

mmbError mmb_copy__linux_dram(void *dst, mmbMemInterface *dst_inter,
                              const void *src, const mmbMemInterface *src_inter,
                              const size_t n_bytes)
{
  mmbError stat = MMB_OK;

  MMB_DEBUG("Calling linux_dram copy()\n");

  if (dst_inter->space->ex_context != src_inter->space->ex_context) {
    stat = MMB_INVALID_EXECUTION_CONTEXT;
    goto BAILOUT;
  }

  MMB_DEBUG("copying %zu bytes from %p to %p.\n", n_bytes, src, dst);
  memcpy(dst, src, n_bytes);

BAILOUT:
  return stat;
}

/**  @} */

/**
 * @name GNU/Linux CUDA memory interface
 * @{ */

#if HAVE_CUDA

#include <cuda_runtime_api.h>

mmbError mmb_allocate__linux_cuda(const size_t n_bytes,
                                  mmbMemInterface *interface,
                                  void **out_allocation)
{
  (void) interface;
  mmbError stat = MMB_OK;

  MMB_DEBUG("Calling linux_cuda allocator()\n");

  void *ptr = NULL;
  cudaError_t cuda_err = cudaMalloc(&ptr, n_bytes);
  if(cudaSuccess != cuda_err) {
    switch (cuda_err) {
      case cudaErrorInvalidValue:
        MMB_ERR("Could not allocate %zu bytes for mmbAllocation.\n", n_bytes);
        stat = MMB_INVALID_ARG;
        break;
      case cudaErrorMemoryAllocation:
        MMB_ERR("Could not allocate %zu bytes for mmbAllocation.\n", n_bytes);
        stat = MMB_OUT_OF_MEMORY;
        break;
      default:
        MMB_ERR("Could not allocate %zu bytes for mmbAllocation: %s\n",
            n_bytes, cudaGetErrorString(cuda_err));
        stat = MMB_ERROR;
    }
    return stat;
  }

  *out_allocation = ptr;

  return stat;
}

mmbError mmb_free__linux_cuda(void *allocation, mmbMemInterface *inter,
                              const size_t n_bytes)
{
  (void) inter;
  (void) n_bytes;
  mmbError stat = MMB_OK;

  MMB_DEBUG("Calling linux_cuda deallocator(%p)\n", allocation);

  cudaError_t cuda_err = cudaFree(allocation);
  switch (cuda_err) {
    case cudaSuccess:
      break;
    case cudaErrorInvalidValue:
      MMB_ERR("Error while freeing allocation: %s\n", cudaGetErrorString(cuda_err));
      stat = MMB_INVALID_ARG;
      break;
    default:
      MMB_ERR("Error while freeing allocation: %s\n", cudaGetErrorString(cuda_err));
      stat = MMB_ERROR;
  }

  return stat;
}

mmbError mmb_copy__linux_cuda(void *dst, mmbMemInterface *dst_inter,
                              const void *src, const mmbMemInterface *src_inter,
                              const size_t n_bytes)
{
  mmbError stat = MMB_OK;

  MMB_DEBUG("Calling linux_cuda copy()\n");

  cudaError_t c_err;
  const mmbExecutionContext dst_ex_con = dst_inter->space->ex_context;
  const mmbExecutionContext src_ex_con = src_inter->space->ex_context;
  enum cudaMemcpyKind cpy_kind;
  char const *cpy_kind_str;
  if (MMB_CPU == dst_ex_con && MMB_CPU == src_ex_con) {
    cpy_kind = cudaMemcpyHostToHost;
    cpy_kind_str = "H2H";
  } else if (MMB_GPU_CUDA == dst_ex_con && MMB_CPU == src_ex_con) {
    cpy_kind = cudaMemcpyHostToDevice;
    cpy_kind_str = "H2D";
  } else if (MMB_CPU == dst_ex_con && MMB_GPU_CUDA == src_ex_con) {
    cpy_kind = cudaMemcpyDeviceToHost;
    cpy_kind_str = "D2H";
  } else if (MMB_GPU_CUDA == dst_ex_con && MMB_GPU_CUDA == src_ex_con) {
    cpy_kind = cudaMemcpyDeviceToDevice;
    cpy_kind_str = "D2D";
  } else {
    MMB_ERR("Invalid execution context for source or destination of the copy.\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }

  MMB_DEBUG("copying %zu bytes from %p to %p (%s)\n", n_bytes, src, dst,
        cpy_kind_str);

  c_err = cudaMemcpy(dst, src, n_bytes, cpy_kind);

  switch (c_err) {
    case cudaSuccess:
      stat = MMB_OK;
      break;
    case cudaErrorInvalidValue:
    case cudaErrorInvalidMemcpyDirection:
      stat = MMB_INVALID_ARG;
      MMB_ERR("Error while copying data: %s\n", cudaGetErrorString(c_err));
      break;
    default:
      stat = MMB_ERROR;
      MMB_ERR("Error while copying data: %s\n", cudaGetErrorString(c_err));
      break;
  }

BAILOUT:
  return stat;
}

#endif /* HAVE_CUDA */

/**  @} */

/**  @} */

/**  @} */

/**  @} */
