/*
 * Copyright (C) 2018-2020 Cray UK
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef MMB_LINUX_NATIVE_H
#define MMB_LINUX_NATIVE_H

#include <stdlib.h>

#include "../../common/mmb_error.h"
#include "../mmb_memory.h"

#include "../i_memory.h"

/**
 * @name mmb_memory
 * @{ */

/**
 * @name mmbMemInterface
 * @{ */

/**
 * @name GNU/Linux native memory provider
 * @{ */

/**
 * @brief Place holder. This provider doesn't really need initialization.
 *
 * @param [in] interface Handle to the structure to initialize.
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_init__linux(mmbMemInterface *interface);

/**
 * @brief Place holder. This provider doesn't really need finalization.
 *
 * @param [in] interface Handle to the structure to finalize.
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_finalize__linux(mmbMemInterface *interface);

/**
 * @brief Check whether the requested layer is supported by the native
 *        linux provider.
 *
 * @param [in] layer Requested layer.
 * @param [out] is_supported Handle for returned value.
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_supports__linux(const mmbMemLayer layer, bool *is_supported);

/**
 * @name GNU/Linux native DRAM memory provider
 * @{ */

/**
 * @brief Data structure used to retain allocation NUMA node for \c MMB_NATIVE
 *        provider.
 */
typedef struct mmbLinux {
  int numa_domain;
} mmbLinux;

/**
 * @brief size of the \p mmbNative structure.
 *
 * This parameter is required to initialize the static table in memory.c.
 * This value is used to compute the required size for an
 * \c struct mmbMemInterface.
 */
#define mmb_data_size__linux sizeof(struct mmbLinux)

/**
 * @brief Allocate \p n_bytes of data in DRAM in a Unix environment.
 *
 * @param [in] n_bytes Size of the requested allocation.
 * @param [in] interface Handle to the interface data structure.
 * @param [out] allocation Pointer to the allocated data.
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_allocate__linux_dram(const size_t n_bytes,
                                  mmbMemInterface *interface,
                                  void **allocation);

/**
 * @brief Free the memory and structures allocated by
 *        \p mmb_allocate__linux_dram.
 *
 * @param [in] allocation Data to be free'd.
 * @param [in] inter Interface managing the given data.
 * @param [in] n_bytes Size of the allocation to free.
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_free__linux_dram(void *allocation, mmbMemInterface *inter,
                              const size_t n_bytes);

/**
 * @brief Copy the data described by \p src in the memory space described by
 *        \p dst.
 *
 * @param [out] dst Destination data handle.
 * @param [in] dst_inter Interface managing the given destination data.
 * @param [in] src Source data handle.
 * @param [in] src_inter Interface managing the given source data.
 * @param [in] n_bytes The amount of data to move from \p src to \p dst.
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_copy__linux_dram(void *dst, mmbMemInterface *dst_inter,
                              const void *src, const mmbMemInterface *src_inter,
                              const size_t n_bytes);

/**  @} */

/**
 * @name GNU/Linux CUDA memory provider
 * @{ */

#if HAVE_CUDA

/**
 * @brief Allocate \p n_bytes of data in GDRAM in a Unix environment using CUDA
 *        API.
 *
 * @param [in] n_bytes Size of the requested allocation.
 * @param [in] interface Handle to the interface data structure.
 * @param [in] data Allocation's interface data.
 * @param [out] allocation Pointer to the allocated data.
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_allocate__linux_cuda(const size_t n_bytes,
                                  mmbMemInterface *interface,
                                  void **allocation);

/**
 * @brief Free the memory and structures allocated by
 *        \p mmb_allocate__linux_cuda.
 *
 * @param [in] allocation Allocated data to be free'd.
 * @param [in] inter Interface managing the given data.
 * @param [in] n_bytes Size of the allocation to free.
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_free__linux_cuda(void *allocation, mmbMemInterface *inter,
                              const size_t n_bytes);

/**
 * @brief Copy the data described by \p src in the memory space described by
 *        \p dst.
 *
 * @param [out] dst Destination data handle.
 * @param [in] dst_inter Interface managing the given destination data.
 * @param [in] src Source data handle.
 * @param [in] src_inter Interface managing the given source data.
 * @param [in] n_bytes The amount of data to move from \p src to \p dst.
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_copy__linux_cuda(void *dst, mmbMemInterface *dst_inter,
                              const void *src, const mmbMemInterface *src_inter,
                              const size_t n_bytes);

#endif /* HAVE_CUDA */

/**  @} */

/**  @} */

/**  @} */

/**  @} */

#endif /* MMB_LINUX_NATIVE_H */
