/*
 * Copyright (C) 2020      Cray UK
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef MMB_STRATEGY_THREAD_SAFE_H
#define MMB_STRATEGY_THREAD_SAFE_H

#include <stddef.h>

#include "../../common/mmb_error.h"
#include "../mmb_memory.h"

#include "../i_memory.h"

/**
 * @name mmb_memory
 * @{ */

/**
 * @name mmbStrategy
 * @{ */

/**
 * @name Thread Safe Strategy
 * @{ */

/**
 * @name Public API
 *
 * This strategy provides a mutual exclusion mechanism based on the pthread
 * library to create a critical section around other the provider, or for other
 * other strategy when thread safety is necesary.
 *
 * @{ */

/**
 * @brief Chain the functions for the initialization of the parameters used to
 *        apply the thread safe strategy.
 *
 * @param [in] interface Handle to the mmbMemInterface structure to initialize
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_strat_init__thread_safe(mmbMemInterface *interface);

/**
 * @brief Chain the functions for the finalization of the parameters used to
 *        apply the thread safe strategy.
 *
 * @param [in] interface Handle to the mmbMemInterface structure to finalize
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_strat_finalize__thread_safe(mmbMemInterface *interface);

/**
 * @brief Chain the function in order to apply the thread safe strategy for the
 *        allocate operation.
 *
 * @param [in] n_bytes Size of the requested allocation
 * @param [in] interface Handle to the mmbMemInterface structure used to
 *                       allocate
 * @param [out] allocation Handle to a pre-allocated structure as provided by the
 *                         memory interface
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_strat_allocate__thread_safe(const size_t n_bytes,
                                         mmbMemInterface *interface,
                                         mmbAllocation *allocation);

/**
 * @brief Chain the functions in order to apply the thread safe strategy free
 *        operation.
 *
 * @param [in] allocation Allocation handle pointing to the data to be free'd
 *                        by the memory interface
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_strat_free__thread_safe(mmbAllocation *allocation);

/**
 * @brief Chain the function in order to apply the thread safe strategy for the
 *        copy operation.
 *
 * @param [inout] dest Handle to the allocation structre of the written data
 * @param [in] src Handle to the allocation structure of the read data
 * @param [in] args Pointer to the followings parameters as set in the strategy
 *                  builder.
 *
 * @return \c MMB_OK on success, an error type otherwise.
 */
mmbError mmb_strat_copy__thread_safe(mmbAllocation *dest,
                                     const mmbAllocation *src);

/**  @} */

/**  @} */

/**  @} */

/**  @} */

#endif /* MMB_STRATEGY_THREAD_SAFE_H */
