#include <stdlib.h>
#include <stdarg.h>
#include <stdbool.h>

#include "mmb_memory.h"
#include "mmb_logging.h"

#include "i_memory.h"
#include "interface.h"
#include "providers/linux_native.h"
#if HAVE_SICM
#include "providers/sicm.h"
#endif /* HAVE_SICM */

#include "strategies/apply.h"
#include "strategies/i_strategy.h"
#include "strategies/i_basic.h"
#include "strategies/i_pooled.h"
#include "strategies/i_thread_safe.h"

/**
 * @name mmb_memory
 * @{ */

/**
 * @name mmbStrategy
 * @{ */

/**
 * @brief Provider dispatch table.
 *
 * We use C99 named initializers, and rely on C static data being 0-initialized
 * so that the 'enabled' fields of the entries are 0 if we actually stored an
 * entry for the respective substrate.
 */
static const struct mmb_dispatch_provider_entry
mmb_providers_table[MMB_PROVIDER__MAX] = {
  [MMB_NATIVE] = {
    .name     = "LINUX_POSIX",
    .supports = mmb_supports__linux,
    .ops = {
      [MMB_CPU] = {
        .enabled  = true,
        .init     = mmb_init__linux,
        .finalize = mmb_finalize__linux,
        .allocate = mmb_allocate__linux_dram,
        .free     = mmb_free__linux_dram,
        .copy     = mmb_copy__linux_dram,
      },
#if HAVE_CUDA
      [MMB_GPU_CUDA] = {
        .enabled  = true,
        .init     = mmb_init__linux,
        .finalize = mmb_finalize__linux,
        .allocate = mmb_allocate__linux_cuda,
        .free     = mmb_free__linux_cuda,
        .copy     = mmb_copy__linux_cuda,
      },
#endif /* HAVE_CUDA */
    }
  },
#if HAVE_SICM
  [MMB_SICM] = {
    .name     = "SICM_LOW",
    .supports = mmb_supports__sicm,
    .init     = mmb_init__sicm,
    .finalize = mmb_finalize__sicm,
    .ops = {
      [MMB_CPU]
        = {
          .enabled  = true,
          .init     = mmb_init_interface__sicm,
          .finalize = mmb_finalize_interface__sicm,
          .allocate = mmb_allocate__sicm,
          .free     = mmb_free__sicm,
          .copy     = mmb_copy__sicm,
        },
    }
  },
#endif /* HAVE_SICM */
};

/**
 * @brief Size of provider specific data.
 *
 * This table is used to retrieve the addresses of the beginning of strategy
 * specific data in the interface structure.
 */
const size_t mmb_provider_data_sizes[MMB_PROVIDER__MAX] = {
  [MMB_NATIVE]  = mmb_data_size__linux,
#if HAVE_SICM
  [MMB_SICM]    = mmb_data_size__sicm,
#endif /* HAVE_SICM */
};

/**
 * @brief Strategies dispatch table.
 *
 * This table is used to retrieve the addresses of the functions used as
 * building blocks when defining a new mmbStrategy in a strategy builder.
 */
const struct mmb_dispatch_strategy_va_entry
mmb_strategies_va_table[MMB_STRATEGY__MAX] = {
  [MMB_STRATEGY_NONE]
    = {
        .init_interface     = mmb_strat_init_interface_va__basic,
        .finalize_interface = mmb_strat_finalize_interface_va__basic,
        .allocate           = mmb_strat_allocate_va__basic,
        .free               = mmb_strat_free_va__basic,
        .copy               = mmb_strat_copy_va__basic,
        .data_size          = mmb_strat_data_size__basic,
    },
  [MMB_POOLED]
    = {
        .init               = mmb_strat_init__pooled,
        .init_interface     = mmb_strat_init_interface_va__pooled,
        .finalize_interface = mmb_strat_finalize_interface_va__pooled,
        .allocate           = mmb_strat_allocate_va__pooled,
        .free               = mmb_strat_free_va__pooled,
        .copy               = mmb_strat_copy_va__pooled,
        .data_size          = mmb_strat_data_size__pooled,
    },
  [MMB_THREAD_SAFE]
    = {
        .init_interface     = mmb_strat_init_interface_va__thread_safe,
        .finalize_interface = mmb_strat_finalize_interface_va__thread_safe,
        .allocate           = mmb_strat_allocate_va__thread_safe,
        .free               = mmb_strat_free_va__thread_safe,
        .copy               = mmb_strat_copy_va__thread_safe,
        .data_size          = mmb_strat_data_size__thread_safe,
    },
};

mmbError mmb_strat_get_data_size(const mmbProvider provider,
                                 const mmbStrategy strategy,
                                 size_t *out_data_size)
{
  mmbError stat = MMB_INVALID_ARG;
  size_t data_size = 0;
  if (NULL == out_data_size) {
    MMB_ERR("Invalid pointer \"out_data_size\".\n");
    goto BAILOUT;
  }
  switch (provider) {
    case MMB_NATIVE:
#if HAVE_SICM
    case MMB_SICM:
#endif /* HAVE_SICM */
      data_size += mmb_provider_data_sizes[provider];
      break;
    default:
      MMB_ERR("Invalid provider data size requested.\n");
      goto BAILOUT;
  }
  switch (strategy) {
    case MMB_STRATEGY_NONE:
    case MMB_POOLED:
    case MMB_THREAD_SAFE:
      data_size += mmb_strategies_va_table[strategy].data_size;
      break;
    default:
      MMB_ERR("Invalid strategy data size requested.\n");
      goto BAILOUT;
  }
  *out_data_size = data_size;
  stat = MMB_OK;
BAILOUT:
  return stat;
}

mmbError mmb_provider_compatibility_check(const mmbProvider provider,
                                          const mmbMemLayer layer,
                                          const mmbExecutionContext ex_con,
                                          bool *is_available)
{
  mmbError stat = MMB_OK;
  if (NULL == is_available) {
    return MMB_INVALID_ARG;
  } else if (MMB_MEMLAYER__MAX <= layer) {
    return MMB_INVALID_LAYER;
  } else if (MMB_PROVIDER__MAX <= provider) {
    return MMB_INVALID_PROVIDER;
  } else if (MMB_EXECUTION_CONTEXT__MAX <= ex_con) {
    return MMB_INVALID_EXECUTION_CONTEXT;
  }
  const bool avail = mmb_providers_table[provider].ops[ex_con].enabled;
  if (avail) {
    stat = mmb_providers_table[provider].supports(layer, is_available);
  } else {
    *is_available = avail;
  }
  return stat;
}

mmbError mmb_init_providers(const mmbProviderInitParamSet params)
{
  mmbError stat = MMB_OK;
  if (NULL == params) {
    MMB_WARN("Param list empty, use default parameters.\n");
  }
  mmbProvider provider;
  const mmbProviderInitParam *param = params;
  for (provider = MMB_NATIVE; provider < MMB_PROVIDER__MAX; ++provider) {
    if (NULL != mmb_providers_table[provider].init)
      stat = mmb_providers_table[provider].init(param);
    if (MMB_OK != stat) {
      MMB_WARN("Failed to initialize provider %s (err=%d)\n",
          mmb_provider_get_string(provider), stat);
      goto BAILOUT;
    }
    param = (NULL == params) ? params : param + 1;
  }
BAILOUT:
  return stat;
}

mmbError mmb_finalize_providers(void)
{
  mmbError stat = MMB_OK;
  mmbProvider provider;
  for (provider = MMB_NATIVE; provider < MMB_PROVIDER__MAX; ++provider) {
    if (NULL != mmb_providers_table[provider].finalize)
      stat = mmb_providers_table[provider].finalize();
    if (MMB_OK != stat)
      MMB_WARN("Failed to finalize provider %s (err=%d)\n",
          mmb_provider_get_string(provider), stat);
  }
  return stat;
}

mmbError mmb_init_strategies(const mmbStrategyInitParamSet params)
{
  mmbError stat = MMB_OK;
  if (NULL == params) {
    MMB_WARN("Param list empty, use default parameters.\n");
  }
  mmbStrategy strategy;
  const mmbStrategyInitParam *param = params;
  for (strategy = MMB_STRATEGY_NONE; strategy < MMB_STRATEGY__MAX; ++strategy) {
    if (NULL != mmb_strategies_va_table[strategy].init)
      stat = mmb_strategies_va_table[strategy].init(param);
    if (MMB_OK != stat) {
      MMB_WARN("Failed to initialize strategy %s (err=%d)\n",
          mmb_strategy_get_string(strategy), stat);
      goto BAILOUT;
    }
    param = (NULL == params) ? params : param + 1;
  }
BAILOUT:
  return stat;
}

mmbError mmb_finalize_strategies(void)
{
  mmbError stat = MMB_OK;
  mmbStrategy strategy;
  for (strategy = MMB_STRATEGY_NONE; strategy < MMB_STRATEGY__MAX; ++strategy) {
    if (NULL != mmb_strategies_va_table[strategy].finalize)
      stat = mmb_strategies_va_table[strategy].finalize();
    if (MMB_OK != stat)
      MMB_WARN("Failed to finalize strategy %s (err=%d)\n",
          mmb_strategy_get_string(strategy), stat);
  }
  return stat;
}

/** va_args versions */

/**
 * @name Private API
 *
 * Because of the need to access mmb_providers_table for the MMB_STRATEGY_NONE
 * strategy functions, these functions are to be found in this file (the
 * provider dispatch table is static, so not exposed outside this file).
 *
 * @{ */

mmbError mmb_strat_init_interface_va__basic(mmbMemInterface *interface,
                                            void *data,
                                            va_list args)
{
  (void) data;
  (void) args;
  const mmbProvider provider = interface->provider;
  const mmbExecutionContext ex_con = interface->space->ex_context;
  return mmb_providers_table[provider].ops[ex_con].init(interface);
}

mmbError mmb_strat_finalize_interface_va__basic(mmbMemInterface *interface,
                                                void *data,
                                                va_list args)
{
  (void) data;
  (void) args;
  const mmbProvider provider = interface->provider;
  const mmbExecutionContext ex_con = interface->space->ex_context;
  return mmb_providers_table[provider].ops[ex_con].finalize(interface);
}

mmbError mmb_strat_allocate_va__basic(const size_t n_bytes,
                                      mmbMemInterface *interface,
                                      void *data,
                                      void **allocation,
                                      va_list args)
{
  (void) data;
  (void) args;
  mmbError stat = MMB_OK;
  const mmbProvider provider = interface->provider;
  const mmbExecutionContext ex_con = interface->space->ex_context;

  /* Secure the access to the max_bytes field in the space */
  mmb_memspace_rdlock_max_bytes(interface->space);

  /* Check we arent over the limit */
  mmb_memspace_lock_size(interface->space);
  {
    const size_t total_bytes = interface->space->allocated_bytes + n_bytes;
    if (total_bytes > interface->space->max_bytes) {
      MMB_WARN("Allocation failed - memory space is full\n");
      stat = MMB_ERROR;
      mmb_memspace_unlock_size(interface->space);
      goto BAILOUT;
    }
    /* Pre-reserve the allocated bytes.  In case of error when allocating, we
     * will adjust the value after the call to allocate. */
    interface->space->allocated_bytes += n_bytes;
  }
  mmb_memspace_unlock_size(interface->space);

  stat = mmb_providers_table[provider].ops[ex_con].allocate(n_bytes, interface,
                                                            allocation);
  if (MMB_OK != stat) {
    /* Adjust value on error */
    mmb_memspace_lock_size(interface->space);
    interface->space->allocated_bytes -= n_bytes;
    mmb_memspace_unlock_size(interface->space);
  }

BAILOUT:
  mmb_memspace_unlock_max_bytes(interface->space);

  return stat;
}

mmbError mmb_strat_free_va__basic(void *alloc,
                                  void *data,
                                  mmbMemInterface *interface,
                                  const size_t n_bytes,
                                  va_list args)
{
  (void) data;
  (void) args;
  mmbError stat = MMB_OK;
  const mmbProvider provider = interface->provider;
  const mmbExecutionContext ex_con = interface->space->ex_context;
  stat = mmb_providers_table[provider].ops[ex_con].free(alloc, interface, n_bytes);
  if (MMB_OK == stat) {
    mmb_memspace_lock_size(interface->space);
    interface->space->allocated_bytes -= n_bytes;
    mmb_memspace_unlock_size(interface->space);
  }
  return stat;
}

mmbError mmb_strat_copy_va__basic(void *dst, void *dst_data,
                                  mmbMemInterface *dst_interface,
                                  const void * src,
                                  const mmbMemInterface *src_interface,
                                  const size_t n_bytes,
                                  va_list args)
{
  (void) dst_data;
  (void) args;
  mmbError stat;
  const mmbProvider dst_provider = dst_interface->provider;
  const mmbProvider src_provider = src_interface->provider;
  const mmbExecutionContext dst_ex_con = dst_interface->space->ex_context;
  const mmbExecutionContext src_ex_con = src_interface->space->ex_context;
  stat = mmb_providers_table[dst_provider].ops[dst_ex_con]
            .copy(dst, dst_interface, src, src_interface, n_bytes);
  if (MMB_INVALID_PROVIDER == stat || MMB_INVALID_LAYER == stat
      || MMB_INVALID_EXECUTION_CONTEXT == stat) {
    stat = mmb_providers_table[src_provider].ops[src_ex_con]
              .copy(dst, dst_interface, src, src_interface, n_bytes);
  }
  return stat;
}

/**  @} */

/**  @} */

/**  @} */

/**  @} */
