/*
 * Copyright (C) 2020 Cray UK
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* needed before inclusion of cheat.h: */
#ifndef __BASE_FILE__
#define __BASE_FILE__ __FILE__
#endif

#include "cheat.h"
#include "cheats.h"

#include <mamba.h>

#define ONE_KiB (1U << 10)
#define ONE_MiB (ONE_KiB << 10)
#define ONE_GiB (ONE_MiB << 10)
#define ONE_TiB (ONE_GiB << 10)

/* Set-up / Tear-down */

#ifndef MAX_LOG_LEVEL
#  ifdef MMB_LOG_DEBUG
#    define MAX_LOG_LEVEL MMB_LOG_DEBUG
#  else
#    define MAX_LOG_LEVEL 3
#  endif
#endif

CHEAT_DECLARE( mmbOptions *MMB_TESTS_INIT_DEFAULT; )

CHEAT_SET_UP(
    cheat_assert_int(mmb_options_create_default(&MMB_TESTS_INIT_DEFAULT), MMB_OK);
    cheat_assert_int(mmb_options_set_debug_level(MMB_TESTS_INIT_DEFAULT,
                                                 MAX_LOG_LEVEL),          MMB_OK);
    cheat_assert_int(mmb_init(MMB_TESTS_INIT_DEFAULT), MMB_OK);
)

CHEAT_TEAR_DOWN(
    cheat_assert_int(mmb_finalize(), MMB_OK);
    cheat_assert_int(mmb_options_destroy(MMB_TESTS_INIT_DEFAULT), MMB_OK);
)

/* Tests */

CHEAT_DECLARE(
  /* Space configuration handle is similar for all spaces as it does not
   * depends on the execution context nor on the layer. */
  const mmbMemSpaceConfig g_config = {
    .size_opts = { .action = MMB_SIZE_SET, .mem_size = 8000, },
    .interface_opts = MMB_MEMINTERFACE_CONFIG_DEFAULT,
  };

  void *default_pointer_addr = (void *) 0xABBA;
)

/* Check the error checking of the layer parameter */
CHEAT_TEST(register_invalid_def_null_null,
    mmbError stat = MMB_OK;
    mmbMemLayer              layer  = -1;
    mmbExecutionContext      ex_con = MMB_EXECUTION_CONTEXT_DEFAULT;
    const mmbMemSpaceConfig *conf   = NULL;
    stat = mmb_register_memory(layer, ex_con, conf, NULL);
    cheat_assert_not_int(stat, MMB_OK);
    cheat_assert_int(stat, MMB_INVALID_LAYER);
    /* max */
    layer = MMB_MEMLAYER__MAX;
    stat = mmb_register_memory(layer, ex_con, conf, NULL);
    cheat_assert_not_int(stat, MMB_OK);
    cheat_assert_int(stat, MMB_INVALID_LAYER);
    /* above */
    layer = 2 * MMB_MEMLAYER__MAX;
    stat = mmb_register_memory(layer, ex_con, conf, NULL);
    cheat_assert_not_int(stat, MMB_OK);
    cheat_assert_int(stat, MMB_INVALID_LAYER);
)

/* Check the error checking of the execution context parameter */
CHEAT_TEST(register_dram_invalid_def_null,
    mmbError stat = MMB_OK;
    mmbMemLayer              layer  = MMB_DRAM;
    mmbExecutionContext      ex_con = -3;
    const mmbMemSpaceConfig *conf   = NULL;
    stat = mmb_register_memory(layer, ex_con, conf, NULL);
    cheat_assert_not_int(stat, MMB_OK);
    cheat_assert_int(stat, MMB_INVALID_EXECUTION_CONTEXT);
    /* NONE */
    ex_con = MMB_EXECUTION_CONTEXT_NONE;
    stat = mmb_register_memory(layer, ex_con, conf, NULL);
    cheat_assert_not_int(stat, MMB_OK);
    cheat_assert_int(stat, MMB_INVALID_EXECUTION_CONTEXT);
    /* max */
    ex_con = MMB_EXECUTION_CONTEXT__MAX;
    stat = mmb_register_memory(layer, ex_con, conf, NULL);
    cheat_assert_not_int(stat, MMB_OK);
    cheat_assert_int(stat, MMB_INVALID_EXECUTION_CONTEXT);
    /* above */
    ex_con = 2 * MMB_EXECUTION_CONTEXT__MAX;
    stat = mmb_register_memory(layer, ex_con, conf, NULL);
    cheat_assert_not_int(stat, MMB_OK);
    cheat_assert_int(stat, MMB_INVALID_EXECUTION_CONTEXT);
)

/* Check the space parameter is not changed when calling mmb_register_memory
 * with bad parameters. */
CHEAT_TEST(register_error_ptr_unchanged,
    mmbError stat = MMB_OK;
    /* Invalid layer, valid execution context */
    {
      mmbMemLayer              layer  = -1;
      mmbExecutionContext      ex_con = MMB_EXECUTION_CONTEXT_DEFAULT;
      const mmbMemSpaceConfig *conf   = NULL;
      mmbMemSpace             *space  = default_pointer_addr;
      stat = mmb_register_memory(layer, ex_con, conf, &space);
      cheat_assert_not_int(stat, MMB_OK);
      cheat_assert_int(stat, MMB_INVALID_LAYER);
      cheat_assert_pointer(space, default_pointer_addr);
      /* max */
      layer = MMB_MEMLAYER__MAX;
      stat = mmb_register_memory(layer, ex_con, conf, &space);
      cheat_assert_not_int(stat, MMB_OK);
      cheat_assert_int(stat, MMB_INVALID_LAYER);
      cheat_assert_pointer(space, default_pointer_addr);
      /* above */
      layer = 2 * MMB_MEMLAYER__MAX;
      stat = mmb_register_memory(layer, ex_con, conf, &space);
      cheat_assert_not_int(stat, MMB_OK);
      cheat_assert_int(stat, MMB_INVALID_LAYER);
      cheat_assert_pointer(space, default_pointer_addr);
    }
    /* Valid layer, invalid execution context */
    {
      mmbMemLayer              layer  = MMB_DRAM;
      mmbExecutionContext      ex_con = -3;
      const mmbMemSpaceConfig *conf   = NULL;
      mmbMemSpace             *space  = default_pointer_addr;
      stat = mmb_register_memory(layer, ex_con, conf, &space);
      cheat_assert_not_int(stat, MMB_OK);
      cheat_assert_int(stat, MMB_INVALID_EXECUTION_CONTEXT);
      cheat_assert_pointer(space, default_pointer_addr);
      /* NONE */
      ex_con = MMB_EXECUTION_CONTEXT_NONE;
      stat = mmb_register_memory(layer, ex_con, conf, &space);
      cheat_assert_not_int(stat, MMB_OK);
      cheat_assert_int(stat, MMB_INVALID_EXECUTION_CONTEXT);
      cheat_assert_pointer(space, default_pointer_addr);
      /* max */
      ex_con = MMB_EXECUTION_CONTEXT__MAX;
      stat = mmb_register_memory(layer, ex_con, conf, &space);
      cheat_assert_not_int(stat, MMB_OK);
      cheat_assert_int(stat, MMB_INVALID_EXECUTION_CONTEXT);
      cheat_assert_pointer(space, default_pointer_addr);
      /* above */
      ex_con = 2 * MMB_EXECUTION_CONTEXT__MAX;
      stat = mmb_register_memory(layer, ex_con, conf, &space);
      cheat_assert_not_int(stat, MMB_OK);
      cheat_assert_int(stat, MMB_INVALID_EXECUTION_CONTEXT);
      cheat_assert_pointer(space, default_pointer_addr);
    }
)

/* Test correctness of not providing any configuration and no return pointer */
CHEAT_TEST(register_dram_def_null_null,
    mmbError stat = MMB_OK;
    mmbMemLayer              layer  = MMB_DRAM;
    mmbExecutionContext      ex_con = MMB_EXECUTION_CONTEXT_DEFAULT;
    const mmbMemSpaceConfig *conf   = NULL;
    stat = mmb_register_memory(layer, ex_con, conf, NULL);
    cheat_assert_int(stat, MMB_OK);
)

/* Test correctness of not providing default configuration and no return pointer */
CHEAT_TEST(register_dram_def_def_null,
    mmbError stat = MMB_OK;
    mmbMemLayer              layer  = MMB_DRAM;
    mmbExecutionContext      ex_con = MMB_EXECUTION_CONTEXT_DEFAULT;
    const mmbMemSpaceConfig *conf   = &g_config;
    stat = mmb_register_memory(layer, ex_con, conf, NULL);
    cheat_assert_int(stat, MMB_OK);
)

/* mmb_register_memory(DRAM, DEFAULT_EX_CONTEXT, NULL, valid_ptr) */
CHEAT_TEST(register_dram_def_null_ptr,
    mmbError stat = MMB_OK;
    mmbMemLayer              layer  = MMB_DRAM;
    mmbExecutionContext      ex_con = MMB_EXECUTION_CONTEXT_DEFAULT;
    const mmbMemSpaceConfig *conf   = NULL;
    mmbMemSpace             *space  = NULL;
    stat = mmb_register_memory(layer, ex_con, conf, &space);
    cheat_assert_int(stat, MMB_OK);
    cheat_assert_not_pointer(space, NULL);
    /* check with default_pointer_addr to check NULL is not a special value. */
    space = default_pointer_addr;
    stat = mmb_register_memory(layer, ex_con, conf, &space);
    cheat_assert_int(stat, MMB_OK);
    cheat_assert_not_pointer(space, NULL);
)

/* mmb_register_memory(DRAM, DEFAULT_EX_CON, &g_config, valid_ptr) */
CHEAT_TEST(register_dram_def_def_ptr,
    mmbError stat = MMB_OK;
    mmbMemLayer              layer  = MMB_DRAM;
    mmbExecutionContext      ex_con = MMB_EXECUTION_CONTEXT_DEFAULT;
    const mmbMemSpaceConfig  *conf  = &g_config;
    mmbMemSpace             *space  = NULL;
    stat = mmb_register_memory(layer, ex_con, conf, &space);
    cheat_assert_int(stat, MMB_OK);
    cheat_assert_not_pointer(space, NULL);
)

/* mmb_register_memory(DRAM, CPU, NULL, valid_ptr) */
CHEAT_TEST(register_dram_cpu_null_ptr,
    mmbError stat = MMB_OK;
    mmbMemLayer              layer  = MMB_DRAM;
    mmbExecutionContext      ex_con = MMB_CPU;
    const mmbMemSpaceConfig *conf   = NULL;
    mmbMemSpace             *space  = NULL;
    stat = mmb_register_memory(layer, ex_con, conf, &space);
    cheat_assert_int(stat, MMB_OK);
    cheat_assert_not_pointer(space, NULL);
)

/* mmb_register_memory(DRAM, CPU, DEFAULT, valid_ptr) */
CHEAT_TEST(register_dram_cpu_def_ptr,
    mmbError stat = MMB_OK;
    mmbMemLayer              layer  = MMB_DRAM;
    mmbExecutionContext      ex_con = MMB_CPU;
    const mmbMemSpaceConfig *conf   = &g_config;
    mmbMemSpace             *space  = NULL;
    stat = mmb_register_memory(layer, ex_con, conf, &space);
    cheat_assert_int(stat, MMB_OK);
    cheat_assert_not_pointer(space, NULL);
)

/* mmb_register_memory(GDRAM, CPU, NULL, valid_ptr) */
CHEAT_TEST(register_gdram_gpu_null_ptr,
    mmbError stat = MMB_OK;
    mmbMemLayer              layer  = MMB_DRAM;
    mmbExecutionContext      ex_con = MMB_CPU;
    const mmbMemSpaceConfig *conf   = NULL;
    mmbMemSpace             *space  = NULL;
    stat = mmb_register_memory(layer, ex_con, conf, &space);
    cheat_assert_int(stat, MMB_OK);
    cheat_assert_not_pointer(space, NULL);
)

/* test mmb_register_memory() with mem_size = 0 */
CHEAT_TEST(register_zero,
    mmbError stat = MMB_OK;
    mmbMemLayer         layer  = MMB_DRAM;
    mmbExecutionContext ex_con = MMB_CPU;
    mmbMemSpaceConfig   config = g_config;
    mmbMemSpace        *space  = NULL;
    config.size_opts.mem_size  = 0;
    stat = mmb_register_memory(layer, ex_con, &config, &space);
    cheat_assert_int(stat, MMB_OK);
    cheat_assert_not_pointer(space, NULL);
)

/* test mmb_register_memory() with mem_size = 0 */
CHEAT_TEST(register_big,
    mmbError stat = MMB_OK;
    mmbMemLayer         layer  = MMB_DRAM;
    mmbExecutionContext ex_con = MMB_CPU;
    mmbMemSpaceConfig   config = g_config;
    mmbMemSpace        *space  = NULL;
    config.size_opts.mem_size  = 1 * ONE_MiB;   // 1 TibiByte as the unit is
                                                // MiB in mmb_register_memory
    stat = mmb_register_memory(layer, ex_con, &config, &space);
    cheat_assert_int(stat, MMB_OK);
    cheat_assert_not_pointer(space, NULL);
)
