/* -*- mode:c -*- */
/** @file
 ** @brief Maestro OpenFabric connections header file
 **/
/*
 * Copyright (C) 2019 Cray Computer GmbH
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef MAESTRO_I_OFI_H_
#define MAESTRO_I_OFI_H_ 1

/**@addtogroup MSTRO_Internal
 **@{
 **/

/**@defgroup MSTRO_I_OFI Network layer
 **@{
 ** This is the Internal Network layer interface
 **/


#include "maestro/status.h"

#include <stdbool.h>
#include <stdatomic.h>

/* we assume everyone has IP */
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

/* we checked for libfabric */
#include <rdma/fabric.h>

#ifdef HAVE_IB
#include <infiniband/ib.h>
#endif

/* for message envelopes: */
#include "maestro/i_pool_manager_protocol.h" 


enum mstro_endpoint_type {
  MSTRO_EP_INVALID = 0, /**< an illegal value */
  /** libfabric native endpoints. FIXME: not all supported yet. */
  MSTRO_EP_OFI_IN4 = 1, /**< open fabric FI_SOCKADDR_IN */
  MSTRO_EP_OFI_IN6,     /**< open fabric FI_SOCKADDR_IN6 */
  MSTRO_EP_OFI_IB,      /**< open fabric FI_SOCKADDR_IB */
  MSTRO_EP_OFI_PSMX,    /**< open fabric FI_ADDR_PSMX */
  MSTRO_EP_OFI_PSMX2,   /**< open fabric FI_ADDR_PSMX2 */
  MSTRO_EP_OFI_GNI,     /**< open fabric FI_ADDR_GNI */
  MSTRO_EP_OFI_BGQ,     /**< open fabric FI_ADDR_BGQ */
  MSTRO_EP_OFI_MLX,     /**< open fabric FI_ADDR_MLX */
  MSTRO_EP_OFI_STR,     /**< open fabric FI_ADDR_STR */
  /** other transport endpoints */
  MSTRO_EP__MAX
};

static const
char *mstro_ep_descriptor_names[MSTRO_EP__MAX] = {
  [MSTRO_EP_INVALID]  = "MSTRO_EP_INVALID",
  /** libfabric native endpoints. FIXME: not all supported yet. */
  [MSTRO_EP_OFI_IN4]  = "MSTRO_EP_OFI_IN4",
  [MSTRO_EP_OFI_IN6]  = "MSTRO_EP_OFI_IN6",
  [MSTRO_EP_OFI_IB]   = "MSTRO_EP_OFI_IB",
  [MSTRO_EP_OFI_PSMX] = "MSTRO_EP_OFI_PSMX",
  [MSTRO_EP_OFI_PSMX2]= "MSTRO_EP_OFI_PSMX2",
  [MSTRO_EP_OFI_GNI]  = "MSTRO_EP_OFI_GNI",
  [MSTRO_EP_OFI_BGQ]  = "MSTRO_EP_OFI_BGQ",
  [MSTRO_EP_OFI_MLX]  = "MSTRO_EP_OFI_MLX",
  [MSTRO_EP_OFI_STR]  = "MSTRO_EP_OFI_STR",
    /** other transport endpoints */
};

/** maximum length of an OFI key. FIXME: this should not exist, and we
 * should support arbitrary length keys. Will be fixed in #81 when
 * chaning the serialization of endpoints to protobuf */
#define MSTRO_OFI_KEY_LEN_MAX 20

/** for each endpoint type there needs to be a serialized descriptor */
struct mstro_endpoint_descriptor_ {
  enum mstro_endpoint_type type; /**< type of endpoint described */
  union {
    struct sockaddr_in  in4;      /**< for MSTRO_EP_OFI_IN4 */
    struct sockaddr_in6 in6;      /**< for MSTRO_EP_OFI_IN6 */
#ifdef HAVE_IB
    struct sockaddr_ib  ib;       /**< for MSTRO_EP_OFI_IB */
#endif
    uint64_t            psmx;     /**< for MSTRO_EP_OFI_PSMX */
    uint64_t            psmx2[2]; /**< for MSTRO_EP_OFI_PSMX2 */
    uint64_t            gni[6];   /**< for MSTRO_EP_OFI_GNI */
    uint64_t            bgq;      /**< for MSTRO_EP_OFI_BGQ */
    uint64_t            mlx;      /**< for MSTRO_EP_OFI_MLX */
    char *              str;      /**< for MSTRO_EP_OFI_STR */
    void *              addr;     /**< generic 'whatever it is' reader-slot */
  };
  char *oob_cookie;               /**< a cookie (currently used for DRC, and only on GNI only) */
  char *name;                     /**< a printable name */
  /* temporary (until protobuf-based endpoint handling comes along) */
  uint64_t info_addr;    /** well-known app descriptor: block memory address */
  uint64_t info_keysize; /** well-known app descriptor: block memory key size */
  char     info_key[MSTRO_OFI_KEY_LEN_MAX];     /** well-known app descriptor: block memory key */

  /**< permit chaining descriptors in a singly-linked list */
  struct mstro_endpoint_descriptor_ *next; 
    
};

typedef struct mstro_endpoint_descriptor_ *mstro_endpoint_descriptor;

/** structure holding information about an endpoint */
struct mstro_endpoint {
  struct fid_cq     *cq;
  struct fid_av     *av;
  struct fid_ep     *ep;
  struct fid_domain *domain;
  struct fid_eq     *eq;
  struct fid_fabric *fabric;
  struct fi_info    *fi;
  struct fid_mr     *component_info_mr;
  uint64_t           component_info_addr;
  size_t             component_info_keysize;
  uint8_t           *component_info_raw_key;
  struct fid_mr     *peer_info_mr; /**< memory registration for reading config block from a peer */
  /** the address to be passed to clients */
  /* void *addr;            /\**< address buffer *\/ */
  /* size_t addrlen;        /\**< address length *\/ */
  char *addr_serialized; /**< string-form of an endpoint
                          * for passing OOB */
  mstro_endpoint_descriptor descr; /**< descriptor */
  struct mstro_endpoint *next; /**< make structure sinly-linked list linkable */
};


/** intermediate datastructure used in serialization */
struct serialized_endpoint_element {
  int64_t  type;         /**< enum mstro_endpoint converted to int */
  char    *strval;       /**< base64-encoded version of tpl_map encoded
                          * endpoint address; exact tpl_map format string
                          * depends on TYPE */
  char    *oob_cookie;   /**< DRC or other cookie for endpoint */
  uint64_t info_addr;    /** well-known app descriptor: block memory address */
  uint64_t info_keysize; /** well-known app descriptor: block memory key size */
  char     info_key[MSTRO_OFI_KEY_LEN_MAX];     /** well-known app descriptor: block memory key */
};
/** we serialize the linked list as a variable-length array of tag
 * (32-bit unsigned integer) and string (result of serializing the
 * individual entry, without the tag, and base64-encoding the result)
 * and the memory registration info necessary to do RDMA reads of the
 * @ref g_mstro_component_info block.
 *
 *Each descriptor type needs to implement a suitable
 * serialization for the descriptor slot it uses; many can share the
 * same one though.*/
mstro_status
mstro_ep_desc_serialize(char **result_p,
                        const struct mstro_endpoint * ep);

mstro_status
mstro_ep_desc_deserialize(mstro_endpoint_descriptor *result_p,
                          const char *serialized_eps);

const char *
mstro_ep_desc_describe(mstro_endpoint_descriptor desc);

const char *
mstro_endpoint_describe(struct mstro_endpoint *ep);

mstro_status
mstro_ep_desc_to_ofi_addr(uint32_t *addr_format,
                          void **addr,
                          size_t *addrlen,
                          const mstro_endpoint_descriptor epd);
mstro_status
mstro_ofi__check_compatibility(bool *suitable_p,
                               struct mstro_endpoint_descriptor_ *remote,
                               struct mstro_endpoint *local);

/**@brief Initialize OFI layer.
 **
 **
 ** Populates @ref g_endpoints with an enabled endpoint for each useful
 ** OFI endpoint discovered
 **
 **/
mstro_status
mstro_ofi_init(void);

/**@brief De-initialize OFI layer.
 **
 ** Clears @ref g_endoiunts and any other OFI-related infrastructure (memory registrations etc.)
 **/
mstro_status
mstro_ofi_finalize(void);

/**@brief Run pool manager loop on OFI endpoints 
 **
 **/
mstro_status
mstro_ofi_pm_loop(_Atomic bool *terminate);

/**@brief obtain the serialized OFI endopint info required to reach the PM */
mstro_status
mstro_ofi_pm_info(char **result_p);

mstro_status
mstro_ofi__submit_message_wait(struct mstro_endpoint *ep, fi_addr_t dst,
                               struct mstro_msg_envelope *msg);

mstro_status
mstro_ofi__submit_message_nowait(struct mstro_endpoint *ep, fi_addr_t dst,
                                 struct mstro_msg_envelope *msg);

/** select the best match between remote and local endpoints */
mstro_status
mstro_ofi__select_endpoint(struct mstro_endpoint_descriptor_ *remote,
                           struct mstro_endpoint **local_p,
                           fi_addr_t *remote_addr_p);

/**@} (end of group MSTRO_I_OFI) */
/**@} (end of group MSTRO_Internal) */

#endif /* MAESTRO_I_OFI_H_ */
