/* -*- mode:c -*- */
/** @file
 ** @brief Maestro Pool functions, local and global
 **/
/*
 * Copyright (C) 2019 Cray Computer GmbH
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef MAESTRO_I_POOL_H_
#define MAESTRO_I_POOL_H_ 1

#include "maestro/i_state.h"
#include "maestro/i_cdo.h"
#include "maestro/i_uthash.h"
#include <pthread.h>

/**@addtogroup MSTRO_Internal
 **@{
 **/

/**@defgroup MSTRO_I_Pool Internal Maestro Pool Management
 **@{
 ** This is the Internal Pool Management API, as developed for D3.2
 **/

/**@brief Add CDO to pool, either in a fresh entry (for the CDO-ID) or add
 ** to existing entry for CDO-ID.
 **
 ** The *cdo* state will be set to the *new_state* upon successful addition.
 ** 
 ** @param[in]	cdo		A CDO handle
 ** @param[out]	new_state	cdo new state	
 **
 ** @returns A status code, ::MSTRO_OK on success.
 **/
mstro_status
mstro_pool__add(mstro_cdo cdo, mstro_cdo_state new_state);

/**@brief Remove CDO from the pool
 **
 ** The *cdo* state will be set to the *new_state* upon success.
 ** 
 ** @param[in]	cdo		A CDO handle
 ** @param[out]	new_state	cdo new state	
 **
 ** @returns A status code, ::MSTRO_OK on success.
 **/
mstro_status
mstro_pool__remove(mstro_cdo cdo, mstro_cdo_state new_state);

/**@brief Demand CDO from pool 
 **
 ** The *cdo* state will be set to the *new_state* upon success.
 ** 
 ** @param[in]	cdo		A CDO handle
 ** @param[out]	new_state	cdo new state	
 **
 ** @returns A status code, ::MSTRO_OK on success.
 **/
mstro_status
mstro_pool__demand(mstro_cdo cdo, mstro_cdo_state new_state);

/**@brief Retract CDO  op from local pool 
 **
 ** The *cdo* state will be set to the *new_state* upon success.
 ** 
 ** @param[in]	cdo		A CDO handle
 ** @param[out]	new_state	cdo new state	
 **
 ** @returns A status code, ::MSTRO_OK on success.
 **/
mstro_status
mstro_pool__retract(mstro_cdo cdo, mstro_cdo_state new_state);


/**@brief Initialize pool
 **
 ** Initializes the local part of the pool, and notifies pool manager
 ** of it (if there is one)
 **
 ** @returns A status code, ::MSTRO_OK on success.
 **
 **/
mstro_status
mstro_pool_init(void);

/**@brief Finalize pool
 **
 ** Shuts down local pool functionality.
 **
 ** Notifies pool manager to possibly pull CDOs that are still needed
 ** off the local resources, then shuts down local pool operations.
 **
 ** @returns A status code, ::MSTRO_OK on success.
 **/
mstro_status
mstro_pool_finalize(void);


/**@brief Find a CDO in the pool by CDOID that has data
 **
 ** Locates a CDO by its CDOID that can provide the data of the CDO.
 **
 ** *result is set to the CDO handle if found, or NULL if not.
 **
 **@returns A status code, ::MSTRO_OK on success, ::MSTRO_FAIL if not found
 **/
mstro_status
mstro_pool__find_source_cdo(const struct mstro_cdo_id *cdoid,
                            const Mstro__Pool__Attributes *desired_attributes,
                            mstro_cdo *result);

/**@brief Find a CDO in the pool by CDOID that needs data
 **
 ** Locates a CDO by its CDOID that can is ready to accept data.
 **
 ** Typically this is a CDO that appeared as argument to REQUIRE or
 ** DEMAND, but could also be a 'shallow' CDO whose resources are
 ** currently in use for other purposes.
 **
 ** *result is set to the CDO handle if found, or NULL if not.
 **
 **@returns A status code, ::MSTRO_OK on success, ::MSTRO_FAIL if not found
 **/
mstro_status
mstro_pool__find_sink_cdo(const struct mstro_cdo_id *cdoid,
                          const Mstro__Pool__Attributes *available_attributes,
                          mstro_cdo *result);


/**@brief Notify the pool of a change of the state in the CDO
 **/
mstro_status
mstro_pool__notify(mstro_cdo cdo);
/**@} (end of group MSTRO_I_Pool) */
/**@} (end of group MSTRO_Internal) */

#endif /* MAESTRO_I_POOL_H_ */
