/* -*- mode:c -*- */
/** @file
 ** @brief Maestro attributes
 **/

/*
 * Copyright (C) 2019 Cray Computer GmbH
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the copyright holder nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "maestro/attributes.h"
#include "maestro/i_attributes.h"
#include "maestro/logging.h"
#include "attributes/maestro-schema.h"

#include "c-timestamp/timestamp.h"

#include <inttypes.h>


/* simplify logging */
#define DEBUG(...) LOG_DEBUG(MSTRO_LOG_MODULE_ATTR,__VA_ARGS__)
#define INFO(...)  LOG_INFO(MSTRO_LOG_MODULE_ATTR,__VA_ARGS__)
#define WARN(...)  LOG_WARN(MSTRO_LOG_MODULE_ATTR,__VA_ARGS__)
#define ERR(...)   LOG_ERR(MSTRO_LOG_MODULE_ATTR,__VA_ARGS__)


/* FIXME: we should have a table of name/type/... and make the const
 * char* refer to the entries in that */
const char *MSTRO_ATTR_CORE_CDO_NAME =
    ".maestro.core.cdo.name";

const char *MSTRO_ATTR_CORE_CDO_ALLOCATE_NOW =
    ".maestro.core.cdo.allocate-now";

const char *MSTRO_ATTR_CORE_CDO_MAESTRO_PROVIDED_STORAGE =
    ".maestro.core.cdo.maestro-provided-storage";

const char *MSTRO_ATTR_CORE_CDO_RAW_PTR = 
    ".maestro.core.cdo.raw-ptr";

const char *MSTRO_ATTR_CORE_CDO_MAMBA_ARRAY = 
    ".maestro.core.cdo.mamba-array";

const char *MSTRO_ATTR_CORE_CDO_SCOPE_LOCAL_SIZE =
    ".maestro.core.cdo.scope.local-size";

const char* MSTRO_ATTR_CORE_CDO_LAYOUT_ELEMENT_SIZE =
	".maestro.core.cdo.layout.element-size";

const char* MSTRO_ATTR_CORE_CDO_LAYOUT_NDIMS =
	".maestro.core.cdo.layout.ndims";

const char* MSTRO_ATTR_CORE_CDO_LAYOUT_ORDER =
	".maestro.core.cdo.layout.order";

const char* MSTRO_ATTR_CORE_CDO_LAYOUT_DIMS_SIZE =
	".maestro.core.cdo.layout.dims-size";

const char *MSTRO_ATTR_CORE_CDO_ISGROUP =
    ".maestro.core.cdo.isgroup";

const char *MSTRO_ATTR_CORE_CDO_GROUP_MEMBERS =
    ".maestro.core.cdo.group-members";



/* .... */


mstro_status
mstro_cdo_attr_table__alloc(mstro_cdo_attr_table *result)
{
  *result = calloc(1,sizeof(struct mstro_cdo_attr_table_));
  if (*result == NULL)
    return MSTRO_NOMEM;

  // defaults are set elsewhere in mstro_cdo_attribute_set_default()

  return MSTRO_OK;
}

mstro_status
mstro_cdo_attr_table__destroy(mstro_cdo_attr_table tab)
{
  if (tab == NULL)
    return MSTRO_INVARG;
  if(tab->name)
    free(tab->name);
  if (tab->id)
    free(tab->id);
  if(tab->lifetime)
    free(tab->lifetime);
  if(tab->redundancy)
    free(tab->redundancy);
//  if (tab->scope.layout)
//    free(tab->scope.layout);
  if(tab->scope.distribution)
    free(tab->scope.distribution);

  free(tab);

  return MSTRO_OK;
}

mstro_status
mstro_cdo_attribute_set(mstro_cdo cdo, const char* key, void* val)
{
  if(cdo==NULL || key==NULL)
    return MSTRO_INVARG;
  if (key[strlen(key)] != '\0') {
    ERR("Invalid key: not a cstring\n");
    return MSTRO_INVARG;
  }

  if(mstro_cdo_state_check(cdo, MSTRO_CDO_STATE_SEALED)) {
    ERR("cannot set attributes on sealed CDO\n");
    return MSTRO_FAIL;
  }

  char *tmpfqkey=NULL;
  const char *fqkey=NULL;
  if(key[0]!='.') {
    /* qualify it */
    tmpfqkey = malloc(strlen(cdo->current_namespace)+1+strlen(key)+1);
    if(tmpfqkey==NULL) {
      ERR("Cannot allocate for fully-qualified key\n");
      return MSTRO_NOMEM;
    }
    strcpy(tmpfqkey,cdo->current_namespace);
    size_t l = strlen(tmpfqkey);
    if(tmpfqkey[l-1]!='.')  /* add '.' if needed ; we could require current_namespace to always have one ...*/
      tmpfqkey[l++]='.';
    strcpy(tmpfqkey+l,key);
    fqkey = tmpfqkey;
  } else {
    fqkey=key;
  }
  
  mstro_status status
      = mstro_attribute_dict_set(cdo->attributes, fqkey,
                                 MSTRO_CDO_ATTR_VALUE_INVALID, /* we dont help in checking */
                                 val, false);
  if(tmpfqkey)
    free(tmpfqkey);
  return status;
}

mstro_status
mstro_cdo_attribute_get(mstro_cdo cdo, const char* key,
                        enum mstro_cdo_attr_value_type *type,
                        const void ** val_p)
{
  if (key == NULL || cdo == NULL) {
    ERR("Invalid CDO or key\n");
    return MSTRO_INVARG;
  }
  if(val_p==NULL) {
    ERR("NULL value destination location\n");
    return MSTRO_INVOUT;
  }

  const char *fqkey=key;
  char *tmpkey = NULL;
  if(key[0]!='.') {
    tmpkey = malloc(strlen(cdo->current_namespace)+strlen(key)+1);
    if(tmpkey==NULL) {
      ERR("Cannot allocate fully qualified key name\n");
      return MSTRO_NOMEM;
    }
    strcpy(tmpkey, cdo->current_namespace);
    strcat(tmpkey, key);
    fqkey=tmpkey;
  }

  mstro_status status
      = mstro_attribute_dict_get(cdo->attributes, fqkey,
                                 type, val_p, NULL, true);
  if(tmpkey)
    free(tmpkey);
  return status;
}

mstro_status
mstro_cdo_attribute_set_default(mstro_cdo cdo)
{
  if (cdo == NULL)
    return MSTRO_INVARG;
  /* we can rely on the dict being non-NULL (at least the core schema is loaded) */

  mstro_schema schema;
  mstro_status s= mstro_attribute_dict_get_schema(cdo->attributes, &schema);
  if(s!=MSTRO_OK)
    return s;
  
  return mstro_attribute_dict_set_defaults(schema,
                                           true,
                                           &(cdo->attributes));
}

mstro_status
mstro_cdo_attribute_set_yaml(mstro_cdo cdo, const char* keyval_in_yaml) 
{
  if(cdo==NULL || keyval_in_yaml==NULL)
    return MSTRO_INVARG;

  if(mstro_cdo_state_check(cdo, MSTRO_CDO_STATE_SEALED)) {
    ERR("Cannot add yaml attributes on SEALED CDO\n");
    return MSTRO_FAIL;
  }

  mstro_schema schema;
  mstro_status s= mstro_attribute_dict_get_schema(cdo->attributes, &schema);
  if(s!=MSTRO_OK)
    return s;
  
  return mstro_attributes_parse(schema,
                                keyval_in_yaml,
                                &cdo->attributes,
                                cdo->current_namespace);
}

mstro_status
mstro_timestamp_parse(const char *str, size_t len, mstro_timestamp *tsp)
{
  if(str==NULL) {
    ERR("NULL string, cannot parse as timestamp\n");
    return MSTRO_INVARG;
  }
  timestamp_t aux;
  timestamp_t *ts;
  if(tsp==NULL) {
    ts=&aux;
  } else {
    ts = (timestamp_t*)tsp;
  }
  
  int err = timestamp_parse(str, len, ts);
  if(err==0)
    return MSTRO_OK;
  else
    return MSTRO_FAIL;
}

mstro_status
mstro_timestamp_valid(const mstro_timestamp *tsp, bool *valid)
{
  *valid = timestamp_valid((timestamp_t*)tsp);
  return MSTRO_OK;
}

size_t
mstro_timestamp_format(char *dst, size_t len, const mstro_timestamp *tsp)
{
  return timestamp_format(dst, len, (const timestamp_t*)tsp);
}

size_t
mstro_timestamp_format_precision(char *dst, size_t len, const mstro_timestamp *tsp, int precision)
{
  return timestamp_format_precision(dst, len, (const timestamp_t*)tsp, precision);
}

/** Compare timestamps, returning <0, 0, >0 */
int
mstro_timestamp_compare(const mstro_timestamp *tsp1, const mstro_timestamp *tsp2)
{
  /* timstamp_compare does not respect offset ... FIXME there */
  uint64_t lshift = tsp1->offset * 60;
  uint64_t rshift = tsp2->offset * 60;
  if((tsp1->sec +lshift  < tsp2->sec + rshift))
    return -1;
  else if((tsp1->sec +lshift  == tsp2->sec + rshift)
          && tsp1->nsec == tsp2->nsec)
    return 0;
  else
    return +1;
}

/** Convert timestamp to C style (caller-allocated) struct tm, UTC */
mstro_status
mstro_timestamp_to_tm_utc(const mstro_timestamp *tsp, struct tm *tmp)
{
  if(tsp==NULL) 
    return MSTRO_INVARG;
  if(tmp==NULL)
    return MSTRO_INVOUT;

  struct tm *res = timestamp_to_tm_utc((const timestamp_t*)tsp, tmp);
  if(res==NULL)
    return MSTRO_FAIL;
  
  return MSTRO_OK;
}

/** Convert timestamp to C style (caller-allocated) struct tm, local time zone */
mstro_status
mstro_timestamp_to_tm_local(const mstro_timestamp *tsp, struct tm *tmp)
{
  if(tsp==NULL) 
    return MSTRO_INVARG;
  if(tmp==NULL)
    return MSTRO_INVOUT;

  struct tm *res = timestamp_to_tm_local((const timestamp_t*)tsp, tmp);
  if(res==NULL)
    return MSTRO_FAIL;
  
  return MSTRO_OK;
}




mstro_status
mstro_blob_create(size_t len, void *data, mstro_blob **result_p)
{
  if(result_p==NULL)
    return MSTRO_INVOUT;
  *result_p = malloc(sizeof(mstro_blob));
  if(!*result_p)
    return MSTRO_NOMEM;
  (*result_p)->len = len;
  (*result_p)->data = data;
  return MSTRO_OK;
}

/** convenience function to dispose a blob allocated with mstro_blob_create()
 */
mstro_status
mstro_blob_dispose(mstro_blob *b)
{
  if(b==NULL)
    return MSTRO_INVARG;
  free(b);
  return MSTRO_OK;
}
