/* -*- mode:c -*- */
/** @file
 ** @brief Maestro interface with OFI layer
 **/

/*
 * Copyright (C) 2019 Cray Computer GmbH
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the copyright holder nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "maestro/i_drc.h"
#include "maestro/logging.h"
#include "maestro/env.h"
#include "maestro/i_utlist.h"
#include "maestro/i_tpl.h"
#include "maestro/i_base64.h"
#include "maestro/i_pool_manager_protocol.h"
#include "maestro/i_ofi.h"
#include "maestro/i_globals.h"
#include "maestro/i_pool.h"

#include "i_pool_client.h"
#include "i_pool_manager.h"
#include "maestro/i_memlock.h"

#include "../transport/transport.h"
#include <rdma/fabric.h>
#include <rdma/fi_cm.h>
#include <rdma/fi_domain.h>
#include <rdma/fi_errno.h>
#include <rdma/fi_tagged.h>
#include <rdma/fi_rma.h>

#include <limits.h>

#include "maestro/i_uthash.h"
#include <stdatomic.h>

#include <sys/mman.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <string.h>
#include <stdbool.h>
#include <sched.h>
#include <unistd.h>

#ifdef HAVE_MIO
// #define MSTRO_MIO_SEM_STR_MAXLEN 	128
 #include "transport/transport_mio.h"
 #include "mio.h"
#endif


/* simplify logging */
#define NOISE(...) LOG_NOISE(MSTRO_LOG_MODULE_COMM,__VA_ARGS__)
#define DEBUG(...) LOG_DEBUG(MSTRO_LOG_MODULE_COMM,__VA_ARGS__)
#define INFO(...)  LOG_INFO(MSTRO_LOG_MODULE_COMM,__VA_ARGS__)
#define WARN(...)  LOG_WARN(MSTRO_LOG_MODULE_COMM,__VA_ARGS__)
#define ERR(...)   LOG_ERR(MSTRO_LOG_MODULE_COMM,__VA_ARGS__)



#define NFREE(x) do { if(x) free(x); } while(0)








struct mstro_endpoint_set {
  size_t size; /**< number of endpoints */
  size_t num_free; /**< number of free EPS entries beyond SIZE */
  char *serialized_eps; /**< if non-NULL, a serialized version of the set */
  struct mstro_endpoint eps[1]; /**< endpoints; auto-extended structure */
};

#define MSTRO_OFI_CAPS    (FI_MSG|FI_RMA|FI_TAGGED)
#define MSTRO_OFI_MODE    (FI_CONTEXT|FI_CONTEXT2|FI_RX_CQ_DATA)
#define MSTRO_OFI_MRMODE  (FI_MR_VIRT_ADDR | FI_MR_ALLOCATED | FI_MR_PROV_KEY | FI_MR_LOCAL | FI_MR_MMU_NOTIFY)
#define MSTRO_OFI_EP_TYPE FI_EP_RDM
#define MSTRO_OFI_VERSION FI_VERSION(1,10)

static mstro_drc_info g_drc_info = NULL;

/* /\* The pool manager identifier *\/ */
/* static mstro_msg_source g_pool_srcid  = 1; */




/** globals */
/** global: all endpoints. In manager typically all OFI endpoints; in
 * worker may include other endpoint types */
static struct mstro_endpoint_set *g_endpoints = NULL;




static inline
mstro_status
mstro_ep_desc_free(mstro_endpoint_descriptor desc)
{
  if(desc) {
    mstro_endpoint_descriptor el,tmp;
    LL_FOREACH_SAFE(desc,el,tmp) {
      if(el->name)
        free(el->name);
      free(el);
    }
  }
  return MSTRO_OK;
}

mstro_status
mstro_ep_desc_create_ofi(mstro_endpoint_descriptor *result_p,
                         const struct fi_info *fi, struct fid_ep *ep)
{
  if(result_p==NULL)
    return MSTRO_INVOUT;
  if(fi==NULL || ep==NULL)
    return MSTRO_INVARG;

  enum mstro_endpoint_type ept = MSTRO_EP_INVALID;
  
  switch(fi->addr_format) {
    /** This is the list documented for libfabric 1.5 */ 
    case FI_SOCKADDR_IN:
      ept = MSTRO_EP_OFI_IN4;        break;
    case FI_SOCKADDR_IN6:
      ept = MSTRO_EP_OFI_IN6;        break;
    case FI_SOCKADDR_IB:
      ept = MSTRO_EP_OFI_IB;         break;
    case FI_ADDR_PSMX:
      ept = MSTRO_EP_OFI_PSMX;       break;
    case FI_ADDR_GNI:
      ept = MSTRO_EP_OFI_GNI;        break;
#if FI_VERSION_GE(FI_VERSION(FI_MAJOR_VERSION,FI_MINOR_VERSION), FI_VERSION(1,5))
    case FI_ADDR_PSMX2:
      ept = MSTRO_EP_OFI_PSMX2;      break;
    case FI_ADDR_BGQ:
      ept = MSTRO_EP_OFI_BGQ;        break;
    case FI_ADDR_MLX:
      ept = MSTRO_EP_OFI_MLX;        break;
    case FI_ADDR_STR:
      ept = MSTRO_EP_OFI_STR;        break;
#endif
      /* unsupported ones: */
    case FI_FORMAT_UNSPEC:
    case FI_SOCKADDR:
      ERR("fi_info entry with unsupported format, ignoring it\n");
      break;
      /* unexpected ones: */
    default:
      ERR("fi_info entry with unexpected format %d, ignoring it\n");
      break;
  }
  if(ept!=MSTRO_EP_INVALID) {
    mstro_endpoint_descriptor target
        = calloc(1, sizeof(struct mstro_endpoint_descriptor_));
    if(target==NULL) {
      ERR("Failed to allocate ep descriptor entry\n");
      /* FIXME: avoid leaking */
      return MSTRO_FAIL;
    }
    target->type = ept;
    target->name = NULL;
    /** all targets will fit 2x uint64, which is what psmx2 allocates
     * in the union, except for a string provider. For that we need to
     * query here */
    size_t addr_len;
    char *buf;
#if FI_VERSION_GE(FI_VERSION(FI_MAJOR_VERSION,FI_MINOR_VERSION), FI_VERSION(1,5))
    if(fi->addr_format==FI_ADDR_STR) {
      int s = fi_getname(&ep->fid, NULL, &addr_len);
      if(s<0) {
        if(s!=-FI_ETOOSMALL) {
          ERR("Failed to obtain endpoint name: %d (%s)\n",
              s, fi_strerror(-s));
          return MSTRO_FAIL;
        }
      }
      buf = malloc(addr_len+1);
      if(buf==NULL) {
        ERR("Failed to allocate endpoint name buffer\n");
        free(target);
        return MSTRO_NOMEM;
      }
      buf[addr_len]='\0';
    } else
#endif
    {
      /* gni has largest buffer */
      buf = (char*)&target->gni[0];
      addr_len = sizeof(target->gni);
    }
    size_t assumed_addr_len = addr_len;
    int s = fi_getname(&ep->fid, buf, &addr_len);
    if(s<0) {
      if(s==-FI_ETOOSMALL) {
        ERR("Unexpectedly large endpoint name requirement: Tried %zu, need %zu\n",
            assumed_addr_len, addr_len);
      } else {
        ERR("Failed to obtain endpoint name: %d, (%s)\n",
            s, fi_strerror(-s));
      }
      free(target);
      return MSTRO_FAIL;
    }
#if FI_VERSION_GE(FI_VERSION(FI_MAJOR_VERSION,FI_MINOR_VERSION), FI_VERSION(1,5))
    if(fi->addr_format==FI_ADDR_STR)
      target->str = (char*)buf;
#endif
    assert(g_drc_info!=NULL);
    mstro_drc_get_oob_string(&target->oob_cookie, g_drc_info);
    DEBUG("Saved DRC cookie %s in ep-desc\n", target->oob_cookie);
    target->next = NULL;
    target->name = NULL;

    *result_p = target;
  }
  /* DEBUG("built ep description for type %d (%s) EP\n", */
  /*       (*result_p)->type, tntcl_ep_descriptor_names[(*result_p)->type]); */

  return MSTRO_OK;
}

/** Provide serialized description of EP.
 *
 * Result is a string of size at most MSTRO_EP_STRING_MAX.
 *
 * If EP is a linked list of endpoints, serialize them in order.
 *
 * If *result_p is NULL it will be allocated, otherwise
 * overwritten. */

#define MSTRO_EP_STRING_MAX (8*1024)

/* IPv4 sockaddr */
static inline
mstro_status
mstro_ep_desc_serialize__in4(char *dst,
                             const struct sockaddr_in *sin)
{
  mstro_status stat = MSTRO_UNIMPL;
  
  /* struct sockaddr_in is defined platform-independent and
   * already using network byte order as needed, so we can handle
   * it as a struct */
  tpl_node *tns;
  tpl_bin tb;
  
  tb.sz = sizeof(struct sockaddr_in);
  tb.addr = sin;
  tns = tpl_map("B", &tb);
  tpl_pack(tns,0);
  void *buf;
  size_t buflen;
  if(-1==tpl_dump(tns,TPL_MEM, &buf, &buflen)) {
    ERR("Not enough space to pack IPv4 endpoint\n");
    return MSTRO_FAIL;
  }
  /* DEBUG("tpl dump of in4 address has length %zu\n", buflen); */
  /* b64 encode */
  size_t needed;
  unsigned char *encoded = base64_encode(buf,buflen,&needed);
  if(needed>MSTRO_EP_STRING_MAX) {
    ERR("Cannot b64 encode IPv4 endpoint data, need %d\n", needed);
    stat=MSTRO_FAIL; goto BAILOUT;
  }
  if(encoded!=NULL) {
    strcpy(dst, (const char*)encoded);
    /* DEBUG("Encoded in4 address: %s\n", dst); */
  } else {
    ERR("Cannot b64 encode IPv4 endpoint data, no memory\n");
    stat=MSTRO_NOMEM;
    goto BAILOUT;
  }
  stat=MSTRO_OK;
BAILOUT:
  NFREE(encoded);
  /* clean up */
  tpl_free(tns);
  free(buf);

  return stat;
}

/** take base64-encoded string (0-terminated) and unpack it into *sin */
static inline
mstro_status
mstro_ep_desc_deserialize__in4(struct sockaddr_in *sin,
                               char *b64_strval)
{
  mstro_status stat=MSTRO_UNIMPL;
  
  tpl_node *tn;
  tpl_bin tb;
  tn = tpl_map( "B", &tb );

  /* undo base64 */
  size_t buflen;
  unsigned char *buf = base64_decode((unsigned char*)b64_strval,
                                     strlen(b64_strval),
                                     &buflen);
  if(buf==NULL) {
    stat=MSTRO_NOMEM;
    goto BAILOUT;
  }
  /* DEBUG("b64decode: buf of length %zu\n", buflen); */
  
  tpl_load( tn, TPL_MEM, buf, buflen);
  tpl_unpack( tn, 0 );
  tpl_free(tn);

  assert(tb.sz==sizeof(struct sockaddr_in));

  memcpy(sin, tb.addr, tb.sz);
  stat = MSTRO_OK;
BAILOUT:
  free(tb.addr);  /* our responsibility to free it */

  return stat;
}

static inline
mstro_status
mstro_ep_desc_serialize__in6(char *dst,
                             const struct sockaddr_in6 *sin)
{
  mstro_status stat = MSTRO_UNIMPL;
  /* struct sockaddr_in6 is defined platform-independent and
   * already using network byte order as needed, so we can handle
   * it as a struct */
  tpl_node *tns;
  tpl_bin tb;

  INFO("sockaddr_in6 with port %d\n", ntohs(sin->sin6_port));
  
  tb.sz = sizeof(struct sockaddr_in6);
  tb.addr = sin;
  tns = tpl_map("B", &tb);
  tpl_pack(tns,0);
  void *buf;
  size_t buflen;
  if(-1==tpl_dump(tns,TPL_MEM, &buf, &buflen)) {
    ERR("Not enough space to pack IPv6 endpoint\n");
    return MSTRO_FAIL;
  }
  /* b64 encode */
  size_t needed;
  unsigned char *encoded = base64_encode(buf, buflen, &needed);
  if(needed>MSTRO_EP_STRING_MAX) {
    ERR("Cannot b64 encode IPv6 endpoint data, need %d\n", needed);
    stat=MSTRO_FAIL;
    goto BAILOUT;
  }

  if(encoded!=NULL) {
    strcpy(dst, (char*)encoded);
  } else {
    ERR("Cannot b64 encode IPv6 endpoint data, no memory\n");
    stat=MSTRO_FAIL;
    goto BAILOUT;
  }
    
  stat=MSTRO_OK;
BAILOUT:
  /* clean up */
  NFREE(encoded);
  tpl_free(tns);
  free(buf);
  return stat;
}

static inline
mstro_status
mstro_ep_desc_deserialize__in6(struct sockaddr_in6 *sin,
                               char *b64_strval)
{
  mstro_status stat=MSTRO_UNIMPL;
  tpl_node *tn;
  tpl_bin tb;
  tn = tpl_map( "B", &tb );

  /* undo base64 */
  size_t buflen;
  unsigned char *buf = base64_decode((unsigned char*)b64_strval,
                                     strlen(b64_strval),
                                     &buflen);

  if(buf==NULL) {
    stat=MSTRO_NOMEM;
    goto BAILOUT;
  }

  tpl_load( tn, TPL_MEM, buf, buflen);
  tpl_unpack( tn, 0 );
  tpl_free(tn);
  
  assert(tb.sz==sizeof(struct sockaddr_in6));

  memcpy(sin, tb.addr, tb.sz);
  stat=MSTRO_OK;
BAILOUT:
  free(tb.addr);  /* our responsibility to free it */

  return stat;
}

mstro_status
mstro_ep_desc_serialize(char **result_p,
                        const struct mstro_endpoint *ep)
{
  unsigned char *encoded=NULL;
  mstro_status stat=MSTRO_UNIMPL;

  if(result_p==NULL)
    return MSTRO_INVOUT;
  if(ep==NULL)
    return MSTRO_INVARG;
  const struct mstro_endpoint_descriptor_ *epd = ep->descr;

  /* Read the documentation at
   * https://troydhanson.github.io/tpl/userguide.html#_linked_lists
   * for details */

  tpl_node *tn;
  const struct mstro_endpoint_descriptor_ *elt;
  struct serialized_endpoint_element serialized_element;

  assert(MSTRO_EP__MAX<=INT_MAX);
  
  tn = tpl_map("A(S(IssUUc#))", &serialized_element, MSTRO_OFI_KEY_LEN_MAX);
  char *strval = alloca(MSTRO_EP_STRING_MAX*sizeof(char));
  if(strval==NULL)
    return MSTRO_NOMEM;

  
  LL_FOREACH(epd,elt) {
    /* each one should serialize into STRVAL buffer, base64-encoded,
     * NULL-terminated */
    /* for code-reuse we don't use a 'switch-case' here */
    if(elt->type == MSTRO_EP_OFI_IN4) {
      stat = mstro_ep_desc_serialize__in4(strval, &elt->in4);
    } else if(elt->type == MSTRO_EP_OFI_IN6) {
      stat = mstro_ep_desc_serialize__in6(strval, &elt->in6);
    } else if(elt->type == MSTRO_EP_OFI_IB) {
      ERR("Infiniband endpoints unsupported\n");
      return MSTRO_FAIL;
    } else if(elt->type == MSTRO_EP_OFI_PSMX
              || elt->type == MSTRO_EP_OFI_PSMX2
              || elt->type == MSTRO_EP_OFI_GNI
              || elt->type == MSTRO_EP_OFI_BGQ
              || elt->type == MSTRO_EP_OFI_MLX) {
      /* uint64_t based endpoints */
      tpl_node *tns;
      //      uint64_t val;
      /* abusing C union properties we can refer to all entries by the
       * gni slots; we end up always transporting 6 uint64, but who cares */
      tns = tpl_map("UUUUUU",
                    & elt->gni[0], & elt->gni[1],
                    & elt->gni[2], & elt->gni[3],
                    & elt->gni[4], & elt->gni[5]);
      tpl_pack(tns, 0);
      void *buf=NULL;
      size_t buflen;
      if(-1==tpl_dump(tns,TPL_MEM, &buf, &buflen)) {
        ERR("Not enough space to pack fabric endpoint\n");
        return MSTRO_FAIL;
      }
      /* b64 encode */
      
      size_t needed;
      encoded = base64_encode(buf, buflen, &needed);
      if(needed>MSTRO_EP_STRING_MAX) {
        ERR("Cannot b64 encode fabric endpoint data, need %d\n", needed);
        stat=MSTRO_FAIL;
        return stat;
      }
      if(encoded!=NULL) {
        strcpy(strval, (char*)encoded);
      } else {
        ERR("Cannot b64 encode fabric endpoint data, no memory\n");
        stat=MSTRO_FAIL;
        goto BAILOUT_1;
      }
      stat=MSTRO_OK;
   BAILOUT_1:
      /* clean up */
      NFREE(encoded);      
      tpl_free(tns);
      free(buf);
    } else if(elt->type == MSTRO_EP_OFI_STR) {
      /* string-based endpoint descriptors */
      char *tmp;
      switch(elt->type) {
        /* string based methods */
        case MSTRO_EP_OFI_STR:
          tmp = elt->str;
          break;
        default:
          ERR("Confusion\n");
          return MSTRO_FAIL;
      }
      size_t len = strlen(tmp)+1;
      if(len>MSTRO_EP_STRING_MAX) {
        ERR("config endpoint string too long (%d) for serialization\n", len);
        return MSTRO_FAIL;
      }
      /* b64 encode */
      size_t needed;
      encoded = base64_encode((unsigned char*)tmp,
		              strlen(tmp)+1, &needed);
      if(needed>MSTRO_EP_STRING_MAX) {
        ERR("Cannot b64 encode fabric endpoint data, need %d\n", needed);
        stat=MSTRO_FAIL;
        goto BAILOUT_2;
      }
      if(encoded!=NULL) {
        strcpy(strval, (char*)encoded);
      } else {
        ERR("Cannot b64 encode fabric endpoint data, no memory\n");
        stat=MSTRO_NOMEM;
        goto BAILOUT_2;
      }
      
      stat=MSTRO_OK;
   BAILOUT_2:
      NFREE(encoded);
    } else {
      ERR("Unsupported MSTRO_EP type: %d (%s)\n",
          elt->type, mstro_ep_descriptor_names[elt->type]);
      return MSTRO_FAIL;
    }
    
    serialized_element.type = elt->type;
    serialized_element.strval=strdup(strval); /* shrink wrap size */
    if(serialized_element.strval==NULL)
      return MSTRO_NOMEM;

    mstro_drc_get_oob_string(&serialized_element.oob_cookie, g_drc_info);
    DEBUG("Added OOB info %s\n", serialized_element.oob_cookie);
    
    /* keys */
    serialized_element.info_addr = ep->component_info_addr;
    serialized_element.info_keysize = ep->component_info_keysize;
    memcpy(serialized_element.info_key,
           ep->component_info_raw_key,
           ep->component_info_keysize);
    DEBUG("Serialized MR addr 0x%" PRIx64 ", len %zu\n",
          serialized_element.info_addr, ep->component_info_keysize);

    /* DEBUG("Serializing %d:%s\n", serialized_element.type, */
    /*       serialized_element.strval); */
    tpl_pack(tn,1);
    free(serialized_element.strval);
    free(serialized_element.oob_cookie);
  }
  
  /* now write all to a string */
  size_t len;
  void *buf=NULL;
  tpl_dump(tn,TPL_MEM, &buf, &len);
  if(len>MSTRO_EP_STRING_MAX) {
    ERR("EP serialization took %zu bytes, but only %zu are supported\n",
        len, MSTRO_EP_STRING_MAX);
    return MSTRO_FAIL;
  }
  /* b64 encode */
  size_t needed;
  encoded = base64_encode(buf, len, &needed);
  if(needed>MSTRO_EP_STRING_MAX) {
    ERR("Cannot b64 encode endpoint descriptor list, need %d\n", needed);
    stat=MSTRO_FAIL;
    goto BAILOUT;
  }
  if(encoded==NULL) {
    ERR("Cannot b64 encode endpoint descriptor list, out of memory\n");
    stat=MSTRO_NOMEM;
    goto BAILOUT;
  } else {
    *result_p = (char*)encoded;
  }
#if 0
  {
    size_t count = 0;
    mstro_endpoint_descriptor_t tmp;
    LL_COUNT(ep,tmp,count);
    DEBUG("serialized %d EPs to |%s|\n",
          count, *result_p);
  }
#endif
BAILOUT:
  if(buf)
    free(buf);
        
  return stat;
}

mstro_status
mstro_ofi_pm_info(char **result_p)
{
  if(result_p==NULL)
    return MSTRO_INVOUT;
  mstro_status status = MSTRO_OK;
  
  unsigned char *encoded=NULL;
  if(g_endpoints==NULL) {
    ERR("No endpoints configured -- did you call mstro_pm_start()?\n");
    return MSTRO_FAIL;
  }

  /* Read the documentation at
   * https://troydhanson.github.io/tpl/userguide.html#_linked_lists
   * for details */

  tpl_node *tn;
  struct mstro_endpoint *elt;
  struct serialized_endpoint_element serialized_element;

  assert(MSTRO_EP__MAX<=INT_MAX);
  
  tn = tpl_map("A(S(IssUUc#))", &serialized_element, MSTRO_OFI_KEY_LEN_MAX);
  char *strval = alloca(MSTRO_EP_STRING_MAX*sizeof(char));
  if(strval==NULL)
    return MSTRO_NOMEM;

  mstro_status stat;
  
  LL_FOREACH(&(g_endpoints->eps[0]),elt) {
    mstro_endpoint_descriptor d = elt->descr;
    /* each one should serialize into STRVAL buffer, base64-encoded,
     * NULL-terminated */
    /* for code-reuse we don't use a 'switch-case' here */
    if(d->type == MSTRO_EP_OFI_IN4) {
      stat = mstro_ep_desc_serialize__in4(strval, &d->in4);
    } else if(d->type == MSTRO_EP_OFI_IN6) {
      stat = mstro_ep_desc_serialize__in6(strval, &d->in6);
    } else if(d->type == MSTRO_EP_OFI_IB) {
      ERR("Infiniband endpoints unsupported\n");
      return MSTRO_FAIL;
    } else if(d->type == MSTRO_EP_OFI_PSMX
              || d->type == MSTRO_EP_OFI_PSMX2
              || d->type == MSTRO_EP_OFI_GNI
              || d->type == MSTRO_EP_OFI_BGQ
              || d->type == MSTRO_EP_OFI_MLX) {
      /* uint64_t based endpoints */
      tpl_node *tns;
      /* abusing C union properties we can refer to all entries by the
       * gni slots; we end up always transporting 6 uint64, but who cares */
      tns = tpl_map("UUUUUU",
                    & d->gni[0], & d->gni[1],
                    & d->gni[2], & d->gni[3],
                    & d->gni[4], & d->gni[5]);
      tpl_pack(tns, 0);
      void *buf=NULL;
      size_t buflen;
      if(-1==tpl_dump(tns,TPL_MEM, &buf, &buflen)) {
        ERR("Not enough space to pack fabric endpoint\n");
        return MSTRO_FAIL;
      }
      /* b64 encode */
      
      size_t needed;
      encoded = base64_encode(buf, buflen, &needed);
      if(needed>MSTRO_EP_STRING_MAX) {
        ERR("Cannot b64 encode fabric endpoint data, need %d\n", needed);
        stat=MSTRO_FAIL;
        return stat;
      }
      if(encoded!=NULL) {
        strcpy(strval, (char*)encoded);
      } else {
        ERR("Cannot b64 encode fabric endpoint data, no memory\n");
        stat=MSTRO_FAIL;
        goto BAILOUT_1;
      }
      stat=MSTRO_OK;
   BAILOUT_1:
      /* clean up */
      NFREE(encoded);      
      tpl_free(tns);
      free(buf);
    } else if(d->type == MSTRO_EP_OFI_STR) {
      /* string-based endpoint descriptors */
      char *tmp;
      switch(d->type) {
        /* string based methods */
        case MSTRO_EP_OFI_STR:
          tmp = d->str;
          break;
        default:
          ERR("Confusion\n");
          return MSTRO_FAIL;
      }
      size_t len = strlen(tmp)+1;
      if(len>MSTRO_EP_STRING_MAX) {
        ERR("config endpoint string too long (%d) for serialization\n", len);
        return MSTRO_FAIL;
      }
      /* b64 encode */
      size_t needed;
      encoded = base64_encode((unsigned char*)tmp,
		              strlen(tmp)+1, &needed);
      if(needed>MSTRO_EP_STRING_MAX) {
        ERR("Cannot b64 encode fabric endpoint data, need %d\n", needed);
        stat=MSTRO_FAIL;
        goto BAILOUT_2;
      }
      if(encoded!=NULL) {
        strcpy(strval, (char*)encoded);
      } else {
        ERR("Cannot b64 encode fabric endpoint data, no memory\n");
        stat=MSTRO_NOMEM;
        goto BAILOUT_2;
      }
      
      stat=MSTRO_OK;
   BAILOUT_2:
      NFREE(encoded);
    } else {
      ERR("Unsupported MSTRO_EP type: %d\n",
          d->type, mstro_ep_descriptor_names[d->type]);
      return MSTRO_FAIL;
    }
    
    serialized_element.type = d->type;
    serialized_element.strval=strdup(strval); /* shrink wrap size */
    if(serialized_element.strval==NULL)
      return MSTRO_NOMEM;

    mstro_drc_get_oob_string(&serialized_element.oob_cookie, g_drc_info);
    DEBUG("Added OOB info %s\n", serialized_element.oob_cookie);

    serialized_element.info_addr = elt->component_info_addr;
    serialized_element.info_keysize = elt->component_info_keysize;
    memcpy(serialized_element.info_key,
           elt->component_info_raw_key,
           elt->component_info_keysize);
    DEBUG("Serialized for PM_INFO MR addr 0x%" PRIx64 ", len %zu\n",
              serialized_element.info_addr, elt->component_info_keysize);

    /* DEBUG("Serializing %d:%s\n", serialized_element.type, */
    /*       serialized_element.strval); */
    tpl_pack(tn,1);
    free(serialized_element.strval);
  }
  /* now write all to a string */
  size_t len;
  void *buf;
  tpl_dump(tn,TPL_MEM, &buf, &len);
  if(len>MSTRO_EP_STRING_MAX) {
    ERR("EP serialization took %zu bytes, but only %zu are supported\n",
        len, MSTRO_EP_STRING_MAX);
    return MSTRO_FAIL;
  }
  /* b64 encode */
  size_t needed;
  encoded = base64_encode(buf, len, &needed);
  if(needed>MSTRO_EP_STRING_MAX) {
    ERR("Cannot b64 encode endpoint descriptor list, need %d\n", needed);
    stat=MSTRO_FAIL;
    goto BAILOUT;
  }
  if(encoded==NULL) {
    ERR("Cannot b64 encode endpoint descriptor list, out of memory\n");
    stat=MSTRO_NOMEM;
    goto BAILOUT;
  } else {
    *result_p = (char*)encoded;
  }
#if 1
  {
    DEBUG("serialized %d EPs to |%s|\n",
          g_endpoints->size, *result_p);
  }
#endif
BAILOUT:
  free(buf);
  return status;
        
}


/** this one allocates the destination string */
static inline
mstro_status
mstro_ep_desc_deserialize__str(char **dst,
                               char *b64_strval)
{
  if(dst==NULL || *dst!=NULL)
    return MSTRO_INVOUT;
  if(b64_strval==NULL)
    return MSTRO_INVARG;

  mstro_status stat=MSTRO_UNIMPL;

  
  size_t buflen;
  unsigned char *buf = base64_decode((unsigned char*)b64_strval,
                                     strlen(b64_strval),
                                     &buflen);

  if(buf==NULL) {
    stat=MSTRO_NOMEM;
    goto BAILOUT;
  }
  *dst=(char*)buf;
BAILOUT:
  return stat;
}


/** unpack one or two uint64 values. DST must be six uint64 entries
 * wide.  This abuses structure union layout, just like at packing
 * time to handle the case where only a 1-element address is
 * transported: the junk we packed will be unpacked into the second
 * element */
static inline
mstro_status
mstro_ep_desc_deserialize__uint64(uint64_t (*dst)[6],
                                  char *b64_strval)
{
  tpl_node *tn=NULL;
  if(dst==NULL ||*dst==NULL) {
    ERR("dst or *dst is NULL\n");
    return MSTRO_INVOUT;
  }
  if(b64_strval==NULL) {
    ERR("b64strval is NULL\n");
    return MSTRO_INVARG;
  }
  mstro_status stat=MSTRO_UNIMPL;

  
  size_t buflen;
  unsigned char *buf = base64_decode((unsigned char*)b64_strval,
                                     strlen(b64_strval),
                                     &buflen);

  if(buf==NULL) {
    stat=MSTRO_NOMEM;
    goto BAILOUT;
  }

  uint64_t v[6];
  tn = tpl_map("UUUUUU", v, v+1, v+2, v+3, v+4, v+5);
  tpl_load(tn, TPL_MEM, buf, buflen);
  tpl_unpack( tn, 0 );
  tpl_free(tn);
  memcpy(*dst,v,sizeof(v));

  stat=MSTRO_OK;
BAILOUT:
  NFREE(buf);

  return stat;        
}


/** read the comments in mstro_ep_desc_serialize for the format */
mstro_status
mstro_ep_desc_deserialize(mstro_endpoint_descriptor *result_p,
                          const char *serialized_eps)
{
  size_t element_idx=0;
  tpl_node *tn=NULL;
  mstro_endpoint_descriptor *next = NULL;

  if(result_p==NULL) 
    return MSTRO_INVOUT;
  if(serialized_eps==NULL)
    return MSTRO_INVARG;
  mstro_status stat=MSTRO_UNIMPL;

  size_t buflen;
  unsigned char *buf = base64_decode((unsigned char*)serialized_eps,
                                     strlen(serialized_eps),
                                     &buflen);
  
  if(buf==NULL) {
    stat=MSTRO_NOMEM;
    goto BAILOUT;
  }

  struct serialized_endpoint_element serialized_element;
  //DEBUG("SE size: %zu, expecting %zu\n", sizeof(serialized_element),
  //      4+8+8+8+MSTRO_OFI_KEY_LEN_MAX);
  tn=tpl_map("A(S(IssUUc#))", &serialized_element, MSTRO_OFI_KEY_LEN_MAX);
  tpl_load( tn, TPL_MEM, buf, buflen);
  DEBUG("endpoint deserialization: got %d EPs\n", tpl_Alen(tn,1));

  /* target for list-append */
  next = result_p;

  while( tpl_unpack( tn, 1 ) > 0 ) {
    /* got another element */
    enum mstro_endpoint_type eptype
        = (enum mstro_endpoint_type) serialized_element.type;
    mstro_endpoint_descriptor target
        = malloc(sizeof(struct mstro_endpoint_descriptor_));
    if(target==NULL) {
      ERR("Failed to allocate EP descriptor\n");
      return MSTRO_NOMEM;
    }

    if(serialized_element.oob_cookie!=NULL) {
      target->oob_cookie=strdup(serialized_element.oob_cookie);
      DEBUG("oob cookie: %s\n", target->oob_cookie);
    } else {
      target->oob_cookie=NULL;
      DEBUG("no cookie\n");
    }
    
    target->name=NULL;
    
    if(eptype==MSTRO_EP_OFI_IN4) {
      stat = mstro_ep_desc_deserialize__in4(&(target->in4),
                                            serialized_element.strval);
    } else if(eptype==MSTRO_EP_OFI_IN6) {
      stat = mstro_ep_desc_deserialize__in6(&(target->in6),
                                            serialized_element.strval);
    } else if(eptype==MSTRO_EP_OFI_IB) {
      ERR("Endpoint typpe OFI_IB unsupported");
      stat = MSTRO_UNIMPL;
    } else if(eptype==MSTRO_EP_OFI_PSMX
              || eptype==MSTRO_EP_OFI_PSMX2
              || eptype==MSTRO_EP_OFI_GNI
              || eptype==MSTRO_EP_OFI_BGQ
              || eptype==MSTRO_EP_OFI_MLX) {
      /** transparently handles one .. six entries */
      stat = mstro_ep_desc_deserialize__uint64(&target->gni,
                                               serialized_element.strval);      
    } else if(eptype==MSTRO_EP_OFI_STR) {
      target->str=NULL;
      stat = mstro_ep_desc_deserialize__str(&(target->str),
                                            serialized_element.strval);      
    } else {
      ERR("Unsupported EP type: %d\n", eptype);
      stat = MSTRO_UNIMPL;
    }
    DEBUG("Parsed ep %d\n", element_idx);
    if(stat==MSTRO_OK) {
      /* DEBUG("%d. endpoint type %d (%s)\n", */
      /*       ++i, eptype, mstro_ep_descriptor_names[eptype]); */
      target->type = eptype;
      target->next = NULL;
      *next = target;
      next = &target->next;

      DEBUG("Found MR addr 0x%" PRIx64 ", keylen %" PRIu64 "\n",
            serialized_element.info_addr, serialized_element.info_keysize);
      target->info_addr = serialized_element.info_addr;
      target->info_keysize = serialized_element.info_keysize;
      assert(MSTRO_OFI_KEY_LEN_MAX>=serialized_element.info_keysize);
      memcpy(target->info_key,
             serialized_element.info_key,
             serialized_element.info_keysize);
    } else {
      INFO("dropped %zu. endpoint type %d (%s)\n",
           element_idx, eptype, mstro_ep_descriptor_names[eptype]);
      if(target->oob_cookie) free(target->oob_cookie);
      free(target);
    }
    free(serialized_element.strval);

    element_idx++;
  }
BAILOUT:
  free(buf);
  
  return stat;
}

#define MSTRO_EP_DESC_STRMAX 1024
const char *
mstro_ep_desc_describe(mstro_endpoint_descriptor desc)
{
  int s;
  /* FIXME: May need locking */
  if(desc==NULL)
    return "(null)";
  if(!desc->name) {
    char *buf = malloc(3* MSTRO_EP_DESC_STRMAX);
    if(buf!=NULL) {
      char *host = buf+MSTRO_EP_DESC_STRMAX;
      char *service = buf+2*MSTRO_EP_DESC_STRMAX;
      switch(desc->type) {
        case MSTRO_EP_OFI_IN4:
          s=getnameinfo((struct sockaddr*)&desc->in4, sizeof(struct sockaddr_in),
                      host, MSTRO_EP_DESC_STRMAX,
                      service, MSTRO_EP_DESC_STRMAX,
                      NI_NUMERICHOST | NI_NUMERICSERV);
	  if(s!=0) {
	    ERR("Failed to obtain IN4 address: %d (%s)\n",
	        errno, strerror(errno));
	    abort();
	  }
          break;
        case MSTRO_EP_OFI_IN6:
          s=getnameinfo((struct sockaddr*)&desc->in6, sizeof(struct sockaddr_in6),
                      host, MSTRO_EP_DESC_STRMAX,
                      service, MSTRO_EP_DESC_STRMAX,
                      NI_NUMERICHOST | NI_NUMERICSERV);
	  if(s!=0) {
	    ERR("Failed to obtain IN6 address: %d (%s)\n",
	        errno, strerror(errno));
	    abort();
	  }
          break;
#ifdef HAVE_IB
          /* case FI_SOCKADDR_IB: */
          /*   ept = MSTRO_EP_OFI_IB;         break; */
#endif
        case MSTRO_EP_OFI_PSMX:
        case MSTRO_EP_OFI_BGQ:
        case MSTRO_EP_OFI_MLX:
          s=snprintf(host, MSTRO_EP_DESC_STRMAX,
		     "[%" PRIu64 "]", desc->psmx);
          if(s<0 || s>MSTRO_EP_DESC_STRMAX) {
	    ERR("Failed to write u64 address: %d (errno %d = %s)\n",
	        s, errno, strerror(errno));
	    abort();
	  }
          service[0]='\0';
          break;
        case MSTRO_EP_OFI_PSMX2:
          s=snprintf(host, MSTRO_EP_DESC_STRMAX,
                   "[%"PRIu64":%"PRIu64"]", desc->psmx2[0], desc->psmx2[1]);
          if(s<0 || s>MSTRO_EP_DESC_STRMAX) {
	    ERR("Failed to write psmx2 address: %d (errno %d = %s)\n",
	        s, errno, strerror(errno));
	    abort();
	  }
          service[0]='\0';
          break;
        case MSTRO_EP_OFI_GNI:
          s=snprintf(host, MSTRO_EP_DESC_STRMAX,
                   "[%" PRIu64 ":%" PRIu64 ":%" PRIu64 ":%" PRIu64 ":%" PRIu64 ":%" PRIu64 "/%s]",
                   desc->gni[0], desc->gni[1],
                   desc->gni[2], desc->gni[3],
                   desc->gni[4], desc->gni[5],
                   desc->oob_cookie);
          if(s<0 || s>MSTRO_EP_DESC_STRMAX) {
	    ERR("Failed to write GNI address: %d (errno %d = %s)\n",
	        s, errno, strerror(errno));
	    abort();
	  }
          service[0]='\0';
	  break;
        case MSTRO_EP_OFI_STR:
          strcpy(host,desc->str);
          service[0]='\0';
          break;
        default:
          strcpy(host, "FIXME");
          service[0]='\0';
      }
      s=snprintf(buf, MSTRO_EP_DESC_STRMAX, "EPD type %d (%s) addr %s:%s",
	         desc->type, mstro_ep_descriptor_names[desc->type],
                 host, service);
      if(s<0 || s>MSTRO_EP_DESC_STRMAX) {
	      ERR("Failed to write endpoint description: %d (errno %d = %s)\n",
	      s, errno, strerror(errno));
	      abort();
      }
      
      desc->name = buf;
    }
  }
  return desc->name;
}

mstro_status
mstro_ep_desc_to_ofi_addr(uint32_t *addr_format,
                          void **addr,
                          size_t *addrlen,
                          const mstro_endpoint_descriptor epd)
{
  if(addr_format==NULL || addr==NULL || addrlen==NULL)
    return MSTRO_INVOUT;
  if(epd==NULL)
    return MSTRO_INVARG;
  
  switch(epd->type) {
    case MSTRO_EP_OFI_IN4:
      *addr_format = FI_SOCKADDR_IN;
      *addr = &epd->in4;
      *addrlen = sizeof(struct sockaddr_in);
      break;
    case MSTRO_EP_OFI_IN6:
      *addr_format = FI_SOCKADDR_IN6;
      *addr = &epd->in6;
      *addrlen = sizeof(struct sockaddr_in6);
      break;
#ifdef HAVE_IB
    case MSTRO_EP_OFI_IB:
      *addr_format = FI_SOCKADDR_IB;
      *addr = &epd->ib;
      *addrlen = sizeof(struct sockaddr_ib);
      break;
#endif
    case MSTRO_EP_OFI_PSMX:
      *addr_format = FI_ADDR_PSMX;
      *addr = &epd->psmx;
      *addrlen = sizeof(uint64_t);
      break;
#if FI_VERSION_GE(FI_VERSION(FI_MAJOR_VERSION,FI_MINOR_VERSION), FI_VERSION(1,5))
    case MSTRO_EP_OFI_PSMX2:
      *addr_format = FI_ADDR_PSMX2;
      *addr = &epd->psmx2;
      *addrlen = 2*sizeof(uint64_t);
      break;
#endif
    case MSTRO_EP_OFI_GNI:
      *addr_format = FI_ADDR_GNI;
      *addr = &epd->gni;
      *addrlen = 6*sizeof(uint64_t);
      break;
    default:
      ERR("Unimplemented mgr address format: %d\n", epd->type);
      return MSTRO_UNIMPL;
  }
  return MSTRO_OK;
}



/** the component descriptor read from the PM */
union mstro_component_descriptor g_pm_component_descriptor;


/* static inline */
mstro_status
mstro_ep_build_from_ofi(struct mstro_endpoint *dst,
                        struct fi_info *fi)
{
  int stat;
  mstro_status retstat=MSTRO_UNIMPL;
  if(dst==NULL)
    return MSTRO_INVOUT;
  if(fi==NULL)
    return MSTRO_INVARG;
  
  struct fid_fabric *fabric = NULL;
  struct fi_eq_attr eq_attr;
  struct fid_eq *eq = NULL;
  struct fid_domain *domain = NULL;
  struct fid_ep *ep = NULL;
  struct fi_av_attr av_attr;
  struct fid_av *av = NULL;
  struct fi_cq_attr cq_attr;
  struct fid_cq *cq = NULL;
  uint8_t *mr_key = NULL;
  
  /* create fabric object */
  stat = fi_fabric(fi->fabric_attr, &fabric, NULL);
  if(stat!=0) {
    ERR("fi_fabric failed: %d (%s)\n", stat, fi_strerror(-stat));
    retstat = MSTRO_FAIL; goto BAILOUT_FAIL;
  }

  if(fi->addr_format==FI_ADDR_GNI) {
    /* put cookie into
     *  fi->domain_attr->auth_key = (void *) &auth_key;
     *  fi->domain_attr->auth_key_size = sizeof(auth_key);
     */
    /* we only support one cookie per workflow currently. It is
     * allocated early on in ofi_init (on pool manager) or obtained
     * from PM_INFO (on clients) */
    assert(g_drc_info!=NULL);
    
    stat = mstro_drc_insert_ofi(fi, g_drc_info);
    if(stat!=MSTRO_OK) {
      ERR("Failed to insert DRC credential into fabric info\n");
      retstat=MSTRO_FAIL;
      goto BAILOUT_FAIL;
    }      
  }
  
  /* create domain */
  stat = fi_domain(fabric, fi, &domain, NULL);
  if(stat!=0) {
    ERR("fi_domain failed: %d (%s)\n", stat, fi_strerror(-stat));
    domain=NULL; /* we happened to see domain!=NULL on error on OSx */
    retstat=MSTRO_FAIL; goto BAILOUT_FAIL;
  }

  /* completion queue */
  memset(&cq_attr,0, sizeof(cq_attr));
  cq_attr.format = FI_CQ_FORMAT_TAGGED;
  cq_attr.size = 0; /* provider-specific default. Consider fi->rx_attr->size; */
  /** FIXME (investigate): on macOS I get a 'not supported' for some
      provider endpoints if set to FI_WAIT_UNSPEC; FD also does not
      work, and MUTEX_COND only works for some. But leaving it at 0
      means NONE, which means we can't do blocking waits...

      For now we don't do blocking waits, and we disable some
      providers like rxd/rxm, and things are fine */
 /*  cq_attr.wait_obj = FI_WAIT_MUTEX_COND; // works on macOS for most
  *  providers */
  /* cq_attr.wait_obj = FI_WAIT_UNSPEC; */
  cq_attr.wait_obj = FI_WAIT_NONE;
  stat = fi_cq_open(domain, &cq_attr, &cq, NULL);
  if(stat!=0) {
    ERR("fi_cq_open failed: %d (%s)\n", stat, fi_strerror(-stat));
    retstat=MSTRO_FAIL; goto BAILOUT_FAIL;
  }
  
  /* event queue */
  /* create event queue */
  memset(&eq_attr, 0, sizeof(eq_attr));
  eq_attr.size = 64; /* FIXME: check whether this makes sense */
  /* eq_attr.wait_obj = FI_WAIT_UNSPEC; /\* provider-optimal default *\/ */
  eq_attr.wait_obj = FI_WAIT_NONE;
  stat = fi_eq_open(fabric, &eq_attr, &eq, NULL);
  if(stat!=0) {
    ERR("fi_eq_open failed: %d (%s)\n", stat, fi_strerror(-stat));
    retstat=MSTRO_FAIL; goto BAILOUT_FAIL;
  }

  /* create endpoint */
  stat = fi_endpoint(domain, fi, &ep, NULL);
  if(stat!=0) {
    ERR("fi_endpoint failed: %d (%s)\n", stat, fi_strerror(-stat));
    retstat=MSTRO_FAIL; goto BAILOUT_FAIL;
  }

  /* INFO("This is a %s\n",  fi_tostr(fi,FI_TYPE_INFO)); */
  
  /* connectionless endpoint needs address vector */
  memset(&av_attr, 0, sizeof(av_attr));
  av_attr.type = FI_AV_MAP;
  stat = fi_av_open(domain, &av_attr, &av, NULL);
  if(stat!=0) {
    ERR("fi_av_open failed: %d (%s)\n", stat, fi_strerror(-stat));
    retstat=MSTRO_FAIL; goto BAILOUT_FAIL;
  }
  
  /* bind address vector to endpoint */
  stat = fi_ep_bind(ep, &av->fid, 0);
  if(stat!=0) {
    ERR("fi_ep_bind for av failed: %d (%s)\n", stat, fi_strerror(-stat));
    retstat=MSTRO_FAIL; goto BAILOUT_FAIL;
  }
  
  /* bind event queue to endpoint */
  stat = fi_ep_bind(ep, &eq->fid, 0);
  if(stat!=0) {
    ERR("fi_ep_bind for eq failed: %d (%s)\n", stat, fi_strerror(-stat));
    if(stat==-FI_ENOSYS) {
      ERR("Keep fingers crossed, we'll try without (psmx2 does not support EQ for instance)\n");
      fi_close((struct fid*)eq);
      eq=NULL;
    } else {
      retstat=MSTRO_FAIL; goto BAILOUT_FAIL;
    }      
  }
  /* bind cq to endpoint */
  stat = fi_ep_bind(ep, &cq->fid, FI_TRANSMIT|FI_RECV);
  if(stat!=0) {
    ERR("fi_ep_bind for cq failed: %d (%s)\n", stat, fi_strerror(-stat));
    retstat=MSTRO_FAIL; goto BAILOUT_FAIL;
  }
  
  /* enable! */
  stat = fi_enable(ep);
  if(stat!=0) {
    ERR("fi_enable failed: %d (%s)\n", stat, fi_strerror(-stat));
    retstat=MSTRO_FAIL; goto BAILOUT_FAIL;
  }

#if 0
  /* FIXME: this is a workaround until we understand
     https://github.com/ofiwg/libfabric/issues/5453
  */
  if(fi->addr_format==FI_SOCKADDR_IN) {
    struct sockaddr_in in4;
    size_t space = sizeof(in4);
    int s= fi_getname(&ep->fid, &in4, &space);
    if(s<0) {
      ERR("Failed to obtain IPv4 address: %d (%s)\n", s, fi_strerror(-s));
      retstat=MSTRO_FAIL; goto BAILOUT_FAIL;
    }
    if(in4.sin_port==0) {
      INFO("IPv4 endpoint has port==0, ignoring\n");
      retstat=MSTRO_FAIL; goto BAILOUT_FAIL;
    }
  } else if(fi->addr_format==FI_SOCKADDR_IN6) {
    struct sockaddr_in6 in6;
    size_t space = sizeof(in6);
    int s= fi_getname(&ep->fid, &in6, &space);
    if(s<0) {
      ERR("Failed to obtain IPv6 address: %d (%s)\n", s, fi_strerror(-s));
      retstat=MSTRO_FAIL; goto BAILOUT_FAIL;
    }
    if(in6.sin6_port==0) {
      INFO("IPv6 endpoint has port==0, ignoring\n");
      retstat=MSTRO_FAIL; goto BAILOUT_FAIL;
    }
  }
  /* end of workaround */
#endif

  /* register component block as RDMA-readable memory chunk */

  /* to be safe: mlock the descriptor. GNI needs it, others will like it */
  retstat = mstro_memlock(&g_component_descriptor, sizeof(g_component_descriptor));
  if(retstat!=MSTRO_OK) {
    ERR("Failed to mlock component descriptor (%d bytes at %p, pagesize %d): %d (%s)\n",
	sizeof(g_component_descriptor), &g_component_descriptor,sysconf(_SC_PAGESIZE),
        retstat, mstro_status_description(retstat));
    goto BAILOUT_FAIL;
  }

  struct fid_mr *mr;
  stat = fi_mr_reg(domain,
                   &g_component_descriptor, sizeof(g_component_descriptor),
                   FI_REMOTE_READ, 0, 43, 0, &mr, NULL);
  if(stat<0) {
    ERR("Failed to publish component descriptor on endpoint: %d (%s)\n",
        stat, fi_strerror(-stat));
    retstat=MSTRO_FAIL; goto BAILOUT_FAIL;
  }

  /* In an ideal world all providers would support fi_mr_raw_attr, as
   * the man page suggests (ofi issue #6095). Then we could play it
   * safe and always exchange raw attr/offset info. On SCALABLE mr
   * these conversions are no-ops. Instead we have to special-case things here ...
   */
  uint64_t addr=0;
  size_t keysize = fi->domain_attr->mr_key_size;
  if(keysize>MSTRO_OFI_KEY_LEN_MAX) {
    ERR("FIXME: provider requires MR key size exceeding hard-coded limit, consider increasing MSTRO_OFI_KEY_LEN_MAX (issue #81)\n");
    goto BAILOUT_FAIL;
  }

  mr_key=malloc(keysize);
  if(mr_key==NULL) {
    ERR("Failed to allocate raw key storage (%d bytes)\n", keysize);
    retstat=MSTRO_NOMEM; goto BAILOUT_FAIL;
  }

  if(fi->domain_attr->mr_mode & FI_MR_RAW) {
    stat= fi_mr_raw_attr(mr, &addr, mr_key, &keysize, 0);
    if(stat<0) {
      ERR("Failed to retrieve raw key for component info memory registration: %d (%s)\n",
          stat, fi_strerror(-stat));
      retstat=MSTRO_FAIL; goto BAILOUT_FAIL;
    }
  } else {
    uint64_t key = fi_mr_key(mr);
    if(key==FI_KEY_NOTAVAIL) {
      ERR("Failed to retrieve key for component info memory registration: %d (%s)\n");
      retstat=MSTRO_FAIL; goto BAILOUT_FAIL;
    }
    assert(sizeof(key)<=MSTRO_OFI_KEY_LEN_MAX); 
    /* keysize is 4 on verbs ... we try to let the compiler handle byte order issues here by casting */
    switch(keysize) {
        case 1: {
           uint8_t k = (uint8_t) key;
	   memcpy(mr_key, &k, sizeof(k));
	   break;
	}
        case 2: {
           uint16_t k = (uint16_t) key;
	   memcpy(mr_key, &k, sizeof(k));
	   break;
	}
        case 4: {
           uint32_t k = (uint32_t) key;
	   memcpy(mr_key, &k, sizeof(k));
	   break;
	}
        case 8: {
	   memcpy(mr_key, &key, sizeof(key));
	   break;
	}
        default: 
            ERR("Unsupported MR key size %d\n", keysize);
	    assert(0);
    }
  }
  if(fi->domain_attr->mr_mode & FI_MR_VIRT_ADDR) {
    if(fi->domain_attr->mr_mode & FI_MR_RAW) {
      DEBUG("Both FI_MR_RAW and FI_MR_VIRT_ADDR specified, addr after translation %p, local %p\n",
            addr, g_component_descriptor);
    } else {
      DEBUG("FI_MR_VIRT_ADDR set, passing local virtual address to remote side\n");
      assert(sizeof(void*)<=sizeof(uint64_t));
      addr=(uint64_t)&g_component_descriptor;
    }
  }
               
    

  {
    uint64_t k = 0;
    memcpy(&k, mr_key, keysize); /* just for printing */
    
    INFO("Component info RDMA block registered%s: local %p, MR %p, addr %" PRIx64 ", desc %p, keysize %zu (key uint64 start: 0x%" PRIx64 ")\n",
         fi->domain_attr->mr_mode & FI_MR_RAW ? " (raw keys)" : "",
         &g_component_descriptor, mr, addr, fi_mr_desc(mr), keysize, k);
  }

  dst->cq = cq;
  dst->av = av;
  dst->ep = ep;
  dst->domain = domain;
  dst->eq = eq;
  dst->fabric = fabric;
  dst->fi = fi;
  dst->component_info_mr = mr;
  dst->component_info_addr = addr;
  dst->component_info_keysize = keysize;
  dst->component_info_raw_key = mr_key;
                       

  dst->next = NULL;

  /* store serialized local address */
  retstat = mstro_ep_desc_create_ofi(&dst->descr, fi, ep);
  if(retstat!=MSTRO_OK) {
    ERR("Failed to construct worker mgmt ep description: %d\n", retstat);
    goto BAILOUT_FAIL;
  }
  
  retstat = mstro_ep_desc_serialize(&dst->addr_serialized,
                                    dst);
  if(retstat!=MSTRO_OK) {
    ERR("Failed to serialize worker mgmt ep description: %d\n", retstat);
    goto BAILOUT_FAIL;
  }
  goto BAILOUT;

BAILOUT_FAIL:
#define CLEANUP(thing,name) do {                        \
      if(thing!=NULL) {                                 \
          stat = fi_close((struct fid*)thing);          \
          if(stat!=0) {                                 \
            ERR("fi_close on %s failed: %d (%s)\n",       \
                name, stat, fi_strerror(-stat));        \
          }                                             \
      }                                                 \
  } while(0)

  /* recommended order as per OFI tutorials: */
  /* alias eps (none) */
  CLEANUP(mr_key, "MR key");
  CLEANUP(ep,"endpoint");
  /* poll sets (none) */
  CLEANUP(cq,"completion queue");
  /* counters (none) */
  CLEANUP(av,"address vector");
  CLEANUP(eq,"event queue");
  CLEANUP(domain,"domain");
  /* wait sets (none) */
  CLEANUP(fabric,"fabric");
#undef CLEANUP
  
BAILOUT:
  return retstat;
}

#define EP_DESC_BUF_MAX 1024

const char *
mstro_endpoint_describe(struct mstro_endpoint *ep)
{
  /* FIXME: this should lock the EP, save the string there, ... */
  /* for now this is only a debugging hack */
  static char ep_name_buf[EP_DESC_BUF_MAX];
  size_t size = EP_DESC_BUF_MAX;
  int ret;
  char strbuf[EP_DESC_BUF_MAX];
  char addrbuf[EP_DESC_BUF_MAX];

  ep_name_buf[0]='\0';
  
  size=0;
  ret = fi_getname(&ep->ep->fid,NULL,&size);
  if(ret!=-FI_ETOOSMALL) {
    ERR("fi_getname in EP describe failed: %d (%s)\n", ret, fi_strerror(-ret));
    goto DONE;
  }
  assert(size<EP_DESC_BUF_MAX);
  fi_getname(&ep->ep->fid,addrbuf,&size);
  
  size=0;
  fi_av_straddr(ep->av, addrbuf, NULL, &size);
  assert(size<EP_DESC_BUF_MAX);

  fi_av_straddr(ep->av, addrbuf, strbuf, &size);
  
  size=snprintf(ep_name_buf, EP_DESC_BUF_MAX,
                "OFI EP prov %s name %s straddr %s",
                ep->fi->fabric_attr->prov_name, 
                ep->fi->fabric_attr->name, strbuf);
  assert(size<1024);
DONE:
  return ep_name_buf;
}

/* order endpoint list.
 *
 * The goal is to have the high-speed ones first. Ideally OFI should
 * do that for us (and does to some degree), but we want to do more:
 * On a laptop used for debugging we want the loopback IP to be up
 * front so that firewalling rules can rely on that
 *
 * Performs a destructive in-place reodering of its argument.
 */

static
int mstro_ofi__fi_info_cmp(const struct fi_info *x1,
                           const struct fi_info *x2)
{
  assert(x1!=NULL); assert(x2!=NULL);

  uint32_t p1=FI_PROTO_UNSPEC, p2=FI_PROTO_UNSPEC;
  if(x1->ep_attr)
    p1=x1->ep_attr->protocol;

  if(x2->ep_attr)
    p2=x2->ep_attr->protocol;

  /* DEBUG("p1: %d, p2: %d\n", p1, p2); */

  /* break ties on anything IP-related to prefer localhost */
  {
    if(x1->addr_format==FI_SOCKADDR_IN
       && x2->addr_format==FI_SOCKADDR_IN
       && x1->src_addr && x2->src_addr) {
      /* DEBUG("both IPv4, preferring localhost\n"); */
      const struct sockaddr_in *a1 = (const struct sockaddr_in*)x1->src_addr;
      const struct sockaddr_in *a2 = (const struct sockaddr_in*)x2->src_addr;
      /* constant is 127.0.0.1 */
      if(a1->sin_addr.s_addr==htonl(0x7F000001)) {
        /* DEBUG("a1 ipv4 localhost\n"); */
        return -1;
      }
      if(a2->sin_addr.s_addr==htonl(0x7F000001)) {
        /* DEBUG("a2 ipv4 localhost\n"); */
        return +1;
      }
      return a1->sin_addr.s_addr < a2->sin_addr.s_addr ? -1 : +1;

    } else if (x1->addr_format==FI_SOCKADDR_IN6
               && x2->addr_format==FI_SOCKADDR_IN6
               && x1->src_addr && x2->src_addr) {
      /* DEBUG("both IPv6, preferring localhost\n"); */
      const struct sockaddr_in6 *a1 = (const struct sockaddr_in6*)x1->src_addr;
      const struct sockaddr_in6 *a2 = (const struct sockaddr_in6*)x2->src_addr;
      if(memcmp(&a1->sin6_addr.s6_addr,
                &in6addr_loopback,
                sizeof(struct in6_addr))==0) {
        /* DEBUG("a1 ipv6 localhost\n"); */
        return -1;
      }
      if(memcmp(&a2->sin6_addr.s6_addr,
                &in6addr_loopback,
                sizeof(struct in6_addr))==0){
        /* DEBUG("a2 ipv6 localhost\n"); */
        return +1;
      }
      size_t i;
      for(i=0; i<16; i++) {
        if(a1->sin6_addr.s6_addr[i] < a2->sin6_addr.s6_addr[i])
          return -1;
        else if (a1->sin6_addr.s6_addr[i] > a2->sin6_addr.s6_addr[i])
          return +1;
        else
          continue;
      }
      /* FIXME: also check for ::ffff:127.0.0.1, the ip4-in-ip6 loopback */
      /* FFFF7F000001 */
    }
  }

  /* otherwise prefer the most performant interconnects */

  /* shared mem */
  if(p1==FI_PROTO_SHM)
    return -1;
  if(p2==FI_PROTO_SHM)
    return +1;

  /* Aries */
  if(p1==FI_PROTO_GNI)
    return -1;
  if(p2==FI_PROTO_GNI)
    return +1;

  /* modern Mellanox */
  if(p1==FI_PROTO_RDMA_CM_IB_RC)
    return -1;
  if(p2==FI_PROTO_RDMA_CM_IB_RC)
    return +1;

  if(p1==FI_PROTO_IB_UD)
    return -1;
  if(p2==FI_PROTO_IB_UD)
    return +1;

  if(p1==FI_PROTO_RXM)
    return -1;
  if(p2==FI_PROTO_RXM)
    return +1;

  if(p1==FI_PROTO_UDP)
    return -1;
  if(p2==FI_PROTO_UDP)
    return +1;

  if(p1==FI_PROTO_SOCK_TCP)
    return -1;
  if(p2==FI_PROTO_SOCK_TCP)
    return +1;

  DEBUG("not ordering: %d/%d vs %d/%d\n",
        p1, x1->addr_format, p2, x2->addr_format);

  return 0;
}

static inline
mstro_status
mstro_ofi__order_fi_list(struct fi_info **fi)
{
  if(fi==NULL)
    return MSTRO_INVARG;
 
  struct fi_info *head = *fi;
  LL_SORT(head, mstro_ofi__fi_info_cmp);
  *fi = head;

  return MSTRO_OK;
}



/** Populate @ref g_endpoints with an enabled endpoint for each useful
 * OFI endpoint discovered */
mstro_status
mstro_ofi_init(void)
{
  int stat;
  size_t i=0;
  size_t num_fis = 0;
  mstro_status retstat = MSTRO_UNIMPL;
  struct fi_info *fi = NULL,
                 *hints = NULL;

  /* prepare for DRC */
  /* (only needed on GNI/Cray, but dummy emulation is provided) */
  if(g_drc_info==NULL) {
    DEBUG("No DRC cookie present, allocating one\n");
    retstat = mstro_drc_init(&g_drc_info);
    if(retstat!=MSTRO_OK) {
      ERR("Failed to create DRC credential\n");
      retstat=MSTRO_FAIL;
      goto BAILOUT_FAIL;
    }
  } else {
    DEBUG("DRC cookie already set, reusing\n");
  }

  /* obtain info */
  hints = fi_allocinfo();
  if(!hints) {
    ERR("fi_allocinfo failed\n");
    retstat=MSTRO_NOMEM; goto BAILOUT_FAIL;
  }

  hints->caps = MSTRO_OFI_CAPS;
  hints->mode = MSTRO_OFI_MODE;
  hints->ep_attr->type = MSTRO_OFI_EP_TYPE;
  hints->domain_attr->mr_mode = MSTRO_OFI_MRMODE;
  
  /* we really want 1.8 or above */
  stat = fi_getinfo(MSTRO_OFI_VERSION, NULL, NULL, 0, hints, &fi);
  fi_freeinfo(hints);
  
  if(stat!=0) {
    ERR("fi_getinfo failed: %d (%s)\n", stat, fi_strerror(-stat));
    retstat=MSTRO_FAIL; goto BAILOUT_FAIL;
  }

  /* order endpoint list.
   *
   * The goal is to have the high-speed ones first. Ideally OFI should
   * do that for us (and does to some degree), but we want to do more:
   * On a laptop used for debugging we want the loopback IP to be up
   * front so that firewalling rules can rely on that */
  stat = mstro_ofi__order_fi_list(&fi);
  if(stat!=MSTRO_OK) {
    ERR("Failed to order FI-list\n");
    retstat = MSTRO_FAIL; goto BAILOUT_FAIL;
  }

  /* allocate endpoints to accomodate all */
  struct fi_info *tmp;
  LL_COUNT(fi,tmp,num_fis);
  g_endpoints = calloc(num_fis, sizeof(struct mstro_endpoint_set));
  if(g_endpoints==NULL) {
    ERR("Failed to allocate endpoints structure\n");
    retstat =  MSTRO_NOMEM; goto BAILOUT_FAIL;
  }
  g_endpoints->num_free = 1+num_fis;
  g_endpoints->size = 0;
  g_endpoints->serialized_eps = NULL;

  INFO("%zu fis found\n", num_fis);

  /* to be safe: mlock the descriptor. GNI needs it, others will like it */
  assert(sizeof(g_component_descriptor)%sysconf(_SC_PAGESIZE)==0);
  retstat = mstro_memlock(&g_component_descriptor, sizeof(g_component_descriptor));
  if(retstat!=MSTRO_OK) {
    ERR("Failed to mlock our component descriptor (%d bytes at %p, pagesize %ul): %d (%s)\n",
	sizeof(g_component_descriptor), &g_component_descriptor, sysconf(_SC_PAGESIZE),
        retstat, mstro_status_description(retstat));
    goto BAILOUT_FAIL;
  }

  /* also do it for the PM descriptor block we'll RDMA read into */ 
  assert(sizeof(g_pm_component_descriptor)%sysconf(_SC_PAGESIZE)==0);
  retstat = mstro_memlock(&g_pm_component_descriptor, sizeof(g_pm_component_descriptor));
  if(retstat!=MSTRO_OK) {
    ERR("Failed to mlock PM component descriptor (%d bytes at %p, pagesize %ul): %d (%s)\n",
	sizeof(g_pm_component_descriptor), &g_pm_component_descriptor, sysconf(_SC_PAGESIZE),
        retstat, mstro_status_description(retstat));
    goto BAILOUT_FAIL;
  }

  /* create one entry in our global EP list per entry */
  LL_FOREACH(fi,tmp) {
    NOISE("index %zu\n", i);
    i++;
    DEBUG("potential endpoint: %s %s\n", tmp->fabric_attr->prov_name, 
		    tmp->fabric_attr->name);
    retstat = mstro_ep_build_from_ofi(&(g_endpoints->eps[g_endpoints->size]),
                                      tmp);
    
    if(retstat!=MSTRO_OK) {
      WARN("Failed to build EP %zu, trying others\n", i);
    } else {
      INFO("%zu. usable endpoint: %s\n",
            g_endpoints->size,
            mstro_endpoint_describe(&g_endpoints->eps[g_endpoints->size]));

      /* also register a region for incoming data */
      struct fid_mr *pm_info_mr = NULL;

      if((g_endpoints->eps[g_endpoints->size].fi->domain_attr->mr_mode & FI_MR_ENDPOINT)) {
        ERR("FIXME: missing support for FI_MR_ENDPOINT domains -- FIXME\n");
      }

      int ret = fi_mr_reg(g_endpoints->eps[g_endpoints->size].domain,
                          &g_pm_component_descriptor, sizeof(g_pm_component_descriptor),
                          FI_READ, 0, 42, 0, &pm_info_mr, NULL);
      if(ret<0) {
        ERR("Failed to register component descriptor read buffer on domain of ep %zu, skipping : %d (%s)\n",
            g_endpoints->size, ret, fi_strerror(-ret));
      } else {
        DEBUG("registered peer buffer for RDMA, addr %p, mr 0x%" PRIx64 ", desc %p\n",
              &g_pm_component_descriptor, pm_info_mr, fi_mr_desc(pm_info_mr));
        g_endpoints->eps[g_endpoints->size].peer_info_mr = pm_info_mr;
        g_endpoints->eps[g_endpoints->size].next
            = &g_endpoints->eps[g_endpoints->size+1];
        g_endpoints->size++;
      }
    }
  }
  /* fix pointer into invalid last entry */
  g_endpoints->eps[g_endpoints->size-1].next = NULL;
  
  if(g_endpoints!=NULL) {
    INFO("%zu usable endpoints\n", g_endpoints->size);
    retstat = MSTRO_OK;
  } else {
    ERR("No usable endpoints found\n");
  }
  
  goto BAILOUT;
BAILOUT_FAIL:
  ; /* FIXME: free resources */
  
BAILOUT:
  return retstat;
}

mstro_status
mstro_ofi_finalize(void)
{
  DEBUG("OFI finalization initiated\n");
  if(g_endpoints) {
    for(size_t i=0; i<g_endpoints->size; i++) {
      int s;
      struct mstro_endpoint *e = &g_endpoints->eps[i];

      if(e->peer_info_mr) {
        DEBUG("closing RDMA peer_info MR for ep %zu\n", i);
        s = fi_close((struct fid*)e->peer_info_mr);
        if(s<0) {
          ERR("Failed to close RDMA MR for ep %zu: %d (%s)\n",
              i, s, fi_strerror(-s));
        }
      }
      if(e->component_info_mr) {
        DEBUG("closing RDMA component_info MR for ep %zu\n", i);
        s = fi_close((struct fid*)e->component_info_mr);
        if(s<0) {
          ERR("Failed to close RDMA MR for ep %zu: %d (%s)\n",
              i, s, fi_strerror(-s));
        }
      }
      /* every EP has one lock on the component descriptor */
      mstro_memunlock(&g_component_descriptor, sizeof(g_component_descriptor));
      
#define CLOSE_FID(member, descr) do {                                   \
        if(e->member) {                                                 \
          DEBUG("Closing down %s for ep %zu\n", descr, i);              \
          int s = fi_close((struct fid*)e->member);                     \
          if(s<0) { ERR("Failed to close %s for ep %zu: %d (%s)\n",     \
                        descr, i, s, fi_strerror(-s)); }                \
        }                                                               \
      } while(0)
      
      CLOSE_FID(ep,"EP");
      CLOSE_FID(cq,"CQ");
      CLOSE_FID(av,"AV");
      CLOSE_FID(eq,"EQ");
      CLOSE_FID(domain,"DOMAIN");
      //      CLOSE_FID(fi,"FABRIC");
#undef CLOSE_FID
      
    }
  }
  DEBUG("OFI endpoints closed\n");
  
  /* It's safe to call this even if we've not yet locked the page; we
   * ignore the error that we'd get in this case */
  mstro_memunlock(&g_component_descriptor, sizeof(g_component_descriptor));
  mstro_memunlock(&g_pm_component_descriptor, sizeof(g_pm_component_descriptor));
  
  return MSTRO_OK;
}


/** messaging context: To make the communication fully asynchronous we
    need to store some information about the send/receive message for
    the completion queue. Since we sometimes also have to wait for
    completions (in the worker in particulary) we provide an
    (optional) semaphore mechanism. And finally, some open fabric
    providers need us to provide scratch space. While we're at it we
    do that.

    (No timings done whether this helps for providers that don't
    require it, but we really have no choice, since we want to support verbs) */
struct mstro_ofi_msg_context_ {
  union {
    struct fi_context  ofi_ctx;     /**< scratch space for OFI providers, don't touch */
    struct fi_context2 ofi_ctx2;    /**< scratch space for OFI providers, don't touch */
  };
  bool has_completion;              /**< indicates if pthread waiting
                                     * infrastructure is initialized in this
                                     * context and should be used*/       
  pthread_cond_t completion;        /**< Condition variable for originator of
                                     * message to */
  pthread_mutex_t lock;             /**< a lock, mostly because cond_wait
                                     * requires one */
  struct mstro_msg_envelope *msg;   /**< the message */
};
typedef struct mstro_ofi_msg_context_ * mstro_ofi_msg_context;


/** Create a message completion context for @arg msg.
    
    If @arg want_completion is true a mutex and a condition variable
    will be initialized and the completion handler can signal
    completion on that condition variable to the originator.

    Typical usage: Submission of the ctx to a libfabric call occurs
    while holding the lock, and then the caller does a cond_wait
    immediately to be woken up when the operation has completed.
    */

static
mstro_status
mstro_ofi__msg_context_create(mstro_ofi_msg_context *result_p,
                              struct mstro_msg_envelope *msg,
                              bool want_completion)
{
  mstro_status stat = MSTRO_UNIMPL;
  if(result_p==NULL)
    return MSTRO_INVOUT;
  /* RDMA ops don't have non-NULL msg */
  /* if(msg==NULL) */
  /*   return MSTRO_INVARG; */
  
  mstro_ofi_msg_context ctx
      = malloc(sizeof(struct mstro_ofi_msg_context_));
  if(ctx==NULL) {
    ERR("Failed to allocate ofi msg context\n");
    return MSTRO_NOMEM;
  }

  if(want_completion) {
    int res = pthread_mutex_init(&ctx->lock, NULL);
    if(res!=0) {
      ERR("Failed to init ofi msg ctx lock: %d\n", res);
      stat=MSTRO_FAIL;
      free(ctx);
      ctx=NULL;
      goto BAILOUT;
    } 
    res = pthread_cond_init(&ctx->completion, NULL);
    if(res!=0) {
      ERR("Failed to init ofi msg cond var: %d\n", res);
      pthread_mutex_destroy(&ctx->lock);
      free(ctx);
      ctx=NULL;
      stat=MSTRO_FAIL;
      goto BAILOUT;
    }
    ctx->has_completion = true;
  } else {
    ctx->has_completion = false;
  }
  ctx->msg = msg;
  stat = MSTRO_OK;
BAILOUT:
  *result_p = ctx;
  return stat;  
}

/** destroy an OFI message context */
static
mstro_status
mstro_ofi__msg_context_destroy(mstro_ofi_msg_context ctx)
{
  if(ctx==NULL)
    return MSTRO_INVARG;
  if(ctx->has_completion) {
    pthread_mutex_destroy(&ctx->lock);
    pthread_cond_destroy(&ctx->completion);
  }
  free(ctx);
  return MSTRO_OK;
}



/** select the best match between remote and local endpoints */
mstro_status
mstro_ofi__select_endpoint(struct mstro_endpoint_descriptor_ *remote,
                           struct mstro_endpoint **local_p,
                           fi_addr_t *remote_addr_p)
{
  mstro_status stat = MSTRO_UNIMPL;

  mstro_endpoint_descriptor tmp = NULL;
  struct mstro_endpoint *my_ep = NULL;

  mstro_ofi_msg_context ctx=NULL;
  stat = mstro_ofi__msg_context_create(&ctx, NULL, true);
  if(stat!=MSTRO_OK) {
    ERR("Failed to init msg context for RDMA op: %d (%s)\n",
        stat, mstro_status_description(stat));
    goto BAILOUT;
  };

  LL_FOREACH(remote, tmp) {
    DEBUG("Trying remote EP %s\n", mstro_ep_desc_describe(tmp));
    LL_FOREACH(g_endpoints->eps, my_ep) {
      DEBUG("Trying local EP %s\n", mstro_endpoint_describe(my_ep));
      bool suitable=false;
      stat = mstro_ofi__check_compatibility(&suitable, tmp, my_ep);
      if(stat!=MSTRO_OK) {
        ERR("Transport coupling check failed: %d\n", stat);
      } else if(suitable) {
        /* try transfer */
        /* FIXME: generalize to non-OFI */
        uint32_t addr_format;
        void *remote_addr;
        size_t addrlen;
        mstro_status stat
            = mstro_ep_desc_to_ofi_addr(&addr_format,
                                        &remote_addr, &addrlen,
                                        tmp);
        if(stat!=MSTRO_OK) {
          ERR("Cannot convert EPD to address: %d\n", stat);
          continue;
        }
        /* try that address in the EP */
        fi_addr_t translated_addr=FI_ADDR_UNSPEC;
        int ret = fi_av_insert(my_ep->av,
                               remote_addr, 1,
                               &translated_addr,
                               /* no flags, no context: we want
                                * synchronous insert */
                               0, NULL);
        if(ret!=1) {
          /* we asked to insert 1, so we should get 1 insert confirmed */
          ERR("failed to insert source address into endpoint av, skipping: %d\n",
              ret);
          continue; /* inner FOREACH */
        }
        DEBUG("AV of ep %s accepted address\n",
              mstro_endpoint_describe(my_ep));


        /* FIXME: Once we move to separate handler threads for message processing this work-around can go away */
        /* If we are handling a PM protocol message which causes a new
         * OFI endpoint selection (like transport does) we are running
         * on the PC thread. In that case we'd need to do the
         * fi_cq_read here, instead of relying the normal handler to
         * do it for us. But that is extremely hard without a full
         * refactor of the handler code. So we just skip it for
         * now. ... */
        if(pthread_self()==g_pc_thread || pthread_self()==g_pm_thread) {
          DEBUG("fi_read from PC or PM thread, skipping config block verification\n");
        } else {
          ret = pthread_mutex_lock(&ctx->lock);
          if(ret!=0) {
            ERR("Failed to acquire msg context lock: %d\n", ret);
            stat = MSTRO_FAIL;
            goto BAILOUT;
          }
          
          uint64_t mr_addr=0;
          /* since the EP accepted the address we interpolate that
           * MR_RAW will be set identically (while we really should be
           * checking a flag sent in the OOB info about the MR) */
          assert(!(my_ep->fi->domain_attr->mr_mode & FI_MR_RAW)); /* FIXME */
          mr_addr = tmp->info_addr;         /** FIXME: with RAW we'd need to translate */

          uint64_t mr_key=0;
          switch(tmp->info_keysize) {
            case 1: { uint8_t  x; memcpy(&x,tmp->info_key, sizeof(x)); mr_key = x; break; }
            case 2: { uint16_t x; memcpy(&x,tmp->info_key, sizeof(x)); mr_key = x; break; }
            case 4: { uint32_t x; memcpy(&x,tmp->info_key, sizeof(x)); mr_key = x; break; }
            case 8: {             memcpy(&mr_key,tmp->info_key, sizeof(mr_key));   break; }
            default: {
              ERR("Unhandled MR key length -- likely needs OFI based translation\n");
              stat=MSTRO_UNIMPL;
              goto BAILOUT;
            }
          }
          DEBUG("Checking for PM config block MR at (remote addr) 0x%" PRIx64 ", key %" PRIx64 "\n",
                mr_addr, mr_key);
          
          assert(ctx->msg==NULL);
          assert(my_ep->peer_info_mr!=NULL); /* incoming buffer has been registered at local endpoint set creation */
          void * local_buf_mr_desc = fi_mr_desc(my_ep->peer_info_mr);
          DEBUG("FI_READ with ctx %p\n", ctx);

          int num_retries = 3;
       RETRY_RDMA_CONFIG_BLOCK_READ:
          DEBUG("Doing fi_read into addr %p (%zu bytes), mr 0x%" PRIx64 ", desc %p\n",
                &g_pm_component_descriptor, sizeof(g_pm_component_descriptor),
                my_ep->peer_info_mr, local_buf_mr_desc);
          ret = fi_read(my_ep->ep,
                        &g_pm_component_descriptor, sizeof(g_pm_component_descriptor),
                        local_buf_mr_desc,
                        translated_addr, mr_addr, mr_key,
                        ctx);
          if(ret==-FI_EAGAIN) {
            INFO("RDMA read for PM config block needs to be retried\n");
            sleep(1);
            if(num_retries-->0)
              goto RETRY_RDMA_CONFIG_BLOCK_READ;
          }
          if(ret<0) {
            ERR("Failed to do RDMA read for PM config block: %d (%s)\n",
                ret, fi_strerror(-ret));
            pthread_mutex_unlock(&ctx->lock);
            continue;
          }
          DEBUG("going to sleep waiting for RDMA read to return\n");
          ret = pthread_cond_wait(&ctx->completion, &ctx->lock);
          if(ret!=0) {
            ERR("Failed to wait on msg ctx cond var for RDMA read: %d\n", ret);
            stat = MSTRO_FAIL;
            goto BAILOUT;
          }
          DEBUG("Found remote config block (%s). type %" PRIu64 ", protocol version: %d.%d.%d, workflow: %s, component: %s\n",
                g_pm_component_descriptor.version,
                g_pm_component_descriptor.type,
                g_pm_component_descriptor.protocol_version.major,
                g_pm_component_descriptor.protocol_version.minor,
                g_pm_component_descriptor.protocol_version.patch,
                g_pm_component_descriptor.workflow_name,
                g_pm_component_descriptor.component_name);
          if(g_pm_component_descriptor.type==MSTRO_COMPONENT_TYPE_UNDEF
             || 0!=strcmp(g_pm_component_descriptor.workflow_name,g_component_descriptor.workflow_name)) {
            INFO("Component descriptor mismatch, illegal type or other workflow, skipping\n");
            continue;
          }
        } /* end of possible config block verification */
        
        *remote_addr_p = translated_addr;
        *local_p = my_ep;
        stat=MSTRO_OK;
        INFO("Using local endpoint %s to communicate to remote at %s\n",
             mstro_endpoint_describe(*local_p),
             mstro_ep_desc_describe(tmp));

        goto BAILOUT;

      } else {
        DEBUG("(deemed unsuitable)\n");
      }
    }
  }

  ERR("No matching transport combination found\n");
  stat = MSTRO_FAIL;

BAILOUT:
  stat |= mstro_ofi__msg_context_destroy(ctx);

  return stat;
}


static
mstro_status
mstro_ofi__submit_message(struct mstro_endpoint *ep,
                          fi_addr_t dst,
                          const struct mstro_msg_envelope *env,
                          void *descriptor,
                          void* context)
{
  assert(env!=NULL);
  DEBUG("Sending message %s, size %zu, %p using ep %p\n",
        env->descriptor->name, env->payload_size, env, ep);
  
  int res;
  do {
    assert(env!=context); /* catch common mistake: context needs to be a
                           * mstro_ofi_msg_context */
    if(MSTRO_MSG_HAS_TAG(env)) {
      ERR("tagged messages not supported currently\n");
      return MSTRO_FAIL;
      /* res = fi_tsend(ep, env->data, env->size, */
      /*                descriptor, dst, msg->type, context); */
      /* DEBUG("tsend with tag %"PRIu64"\n", (uint64_t)msg->type); */

    } else {
      res = fi_send(ep->ep, env->data, env->payload_size, 
                    descriptor, dst, context);
    }
    if(res!=0) {
      if(res==-FI_EAGAIN) {
        NOISE("fi_send asked us to retry\n");
      } else {
        ERR("fi_send failed: %d (%s)\n", res, fi_strerror(-res));
        return MSTRO_OFI_FAIL;
      }
    }
  } while(res==-FI_EAGAIN);
  DEBUG("succeeded\n");
  return MSTRO_OK;
}

mstro_status
mstro_ofi__submit_message_wait(struct mstro_endpoint *ep, fi_addr_t dst,
                               struct mstro_msg_envelope *msg)
{
  mstro_ofi_msg_context ctx=NULL;
  mstro_status stat
      = mstro_ofi__msg_context_create(&ctx, msg, true);
  if(stat!=MSTRO_OK)
    return stat;

  int ret = pthread_mutex_lock(&ctx->lock);
  if(ret!=0) {
    ERR("Failed to acquire msg context lock: %d\n", ret);
    return MSTRO_FAIL;
  }
  
  stat = mstro_ofi__submit_message(ep, dst, msg, NULL, ctx);
  if(stat!=MSTRO_OK) 
    return stat;

  ret = pthread_cond_wait(&ctx->completion, &ctx->lock);
  if(ret!=0) {
    ERR("Failed to wait on msg ctx cond var: %d\n", ret);
    return MSTRO_FAIL;
  }

  DEBUG("completion of msg %p\n", msg);
  ret = pthread_mutex_unlock(&ctx->lock);
  if(ret!=0) {
    ERR("Failed to unlock msg ctx lock: %d\n", ret);
    return MSTRO_FAIL;
  }

  return mstro_ofi__msg_context_destroy(ctx);
}

mstro_status
mstro_ofi__submit_message_nowait(struct mstro_endpoint *ep, fi_addr_t dst,
                                 struct mstro_msg_envelope *msg)
{
  mstro_ofi_msg_context ctx=NULL;
  mstro_status stat
      = mstro_ofi__msg_context_create(&ctx, msg, false);
  if(stat!=MSTRO_OK) {
    ERR("Failed to create completion context: %d (%s)\n",
        stat, mstro_status_description(stat));
    return stat;
  } else {
    ;
  }
        
    
  /* for non-completion contexts the message handler frees the
   * context */
  return mstro_ofi__submit_message(ep, dst, msg, NULL, ctx);
}


mstro_status
mstro_ofi__expect_message_wait(struct fid_ep *ep, fi_addr_t src,
                               struct mstro_msg_envelope *msg)
{
  mstro_ofi_msg_context ctx=NULL;
  mstro_status stat
      = mstro_ofi__msg_context_create(&ctx, msg, true);
  if(stat!=MSTRO_OK)
    return stat;
  
  int ret = pthread_mutex_lock(&ctx->lock);
  if(ret!=0) {
    ERR("Failed to acquire msg context lock: %d\n", ret);
    return MSTRO_FAIL;
  }

  /* FIXME: split out do loop like for submit so that we don't need
   * fi_recv in arbitrary places anymore */
  
  int res;
  do {
    if(MSTRO_MSG_HAS_TAG(msg)) {
      /* for recv we provide specify the availabe size in the msg envelope */
      res = fi_trecv(ep, msg->data, msg->size, NULL,
                     src, msg->tag, 0, ctx);
      DEBUG("trecv with tag %"PRIu64"\n", msg->tag);
    } else {
      /* for recv we provide specify the availabe size in the msg envelope */
      res = fi_recv(ep, msg->data, msg->size, NULL, src, ctx);
    }
    if(res!=0) {
      switch(res) {
        case -FI_EAGAIN:
          INFO("fi_[t]recv wants us to try again\n");
          break;
        default:
          ERR("fi_[t]recv failed: %d\n", res, fi_strerror(-res));
          return MSTRO_OFI_FAIL;
      }
    }
  } while(res!=0);

  ret = pthread_cond_wait(&ctx->completion, &ctx->lock);
  if(ret!=0) {
    ERR("Failed to wait on msg ctx cond var: %d\n", ret);
    return MSTRO_FAIL;
  }

  DEBUG("completion of msg %p\n", msg);
  ret = pthread_mutex_unlock(&ctx->lock);
  if(ret!=0) {
    ERR("Failed to unlock msg ctx lock: %d\n", ret);
    return MSTRO_FAIL;
  }

  return mstro_ofi__msg_context_destroy(ctx);
  return MSTRO_OK;
}















/* FIXME: this needs to move to pool manager protocol code regions */



/* /\* next app id counter *\/ */
/* static _Atomic mstro_msg_source g_next_msg_source_id = 2; */

/* register a new app using its JOIN message. Allocates a new entry in
 * @ref g_mstro_pm_app_table and returns the key to identify it in
 * *entry_p
 *
 * The transport_methods slot will be consumed by the registration and
 * NULLed.
 */
mstro_status
mstro_pm__register_app(Mstro__Pool__Join *join_msg,
                       struct mstro_endpoint *ep,
                       struct mstro_pm_app_registry_entry **entry_p)
{
  mstro_status s=MSTRO_OK;

  assert(entry_p!=NULL);
  assert(join_msg!=NULL);

  /* unpack serialized address */
  struct mstro_endpoint_descriptor_ *epd=NULL;
  s = mstro_ep_desc_deserialize(&epd, join_msg->serialized_endpoint);
  if(s!=MSTRO_OK) {
    ERR("Failed to deserialize endpoint in JOIN message\n");
    goto BAILOUT_FREE;
  }
  INFO("EPD serial %s\n", join_msg->serialized_endpoint);
  INFO("EPD parsed as %s\n", mstro_ep_desc_describe(epd));
  
  uint32_t addr_format;
  void *remote_addr;
  size_t addrlen;
  mstro_status stat
      = mstro_ep_desc_to_ofi_addr(&addr_format,
                                  &remote_addr, &addrlen,
                                  epd);
  if(stat!=MSTRO_OK) {
    ERR("Cannot convert EPD to address: %d\n", stat);
    goto BAILOUT_FREE;
  }

  /* insert in endpoint */
  void *context=NULL;
  uint64_t flags=0;
  fi_addr_t translated_addr;
  int ret = fi_av_insert(ep->av,
                         remote_addr, 1,
                         &translated_addr,
                         flags, context);
  if(ret!=1) {
    /* we asked to insert 1, so we should see ret==1 */
    ERR("Failed to insert source address into ep: %d\n", ret);
    s = MSTRO_FAIL;
    goto BAILOUT_FREE;
  }

  DEBUG("App %s advertises %zu transport methods\n",
        join_msg->component_name, join_msg->transport_methods->n_supported);
  
  /* insert into registry table */
  s = mstro_pm_app_register(ep, translated_addr,
                            strdup(join_msg->serialized_endpoint),
                            join_msg->transport_methods,
                            NULL, join_msg->component_name,
                            entry_p);
  if(s!=MSTRO_OK) {
    ERR("Failed to register application: %s\n", s);
    goto BAILOUT_FREE;
  } else {
    join_msg->transport_methods = NULL; /* we kept a reference to it, caller will free the message */
    DEBUG("Assigned app ID %"PRIu64"\n", (*entry_p)->appid);
  }
  
BAILOUT:
  return s;
BAILOUT_FREE:
  /* leaking epd */
  goto BAILOUT;
}

















typedef mstro_status (*mstro_msg_handler)(const struct mstro_msg_envelope *envelope,
                                          struct mstro_endpoint *ep);

/** allocate a fresh 'short' message envelope, context, and post a receive for it on EP.
    Returns the message envelope in *envelope_p and the context in *ctxt_p.
*/
static inline
mstro_status
mstro_ofi__loop_post_recv(struct fid_ep *ep,
                          struct mstro_msg_envelope **envelope_p,
                          mstro_ofi_msg_context *ctxt_p)
{
  mstro_status status = mstro_msg_envelope_allocate(envelope_p);
  if(status!=MSTRO_OK) {
    goto BAILOUT;
  }
  status = mstro_ofi__msg_context_create(ctxt_p, *envelope_p, false);
  if(status!=MSTRO_OK) {
    goto BAILOUT;
  }

  int res;
  do {
    res = fi_recv(ep,
                  (*envelope_p)->data, (*envelope_p)->size,
                  NULL, FI_ADDR_UNSPEC, *ctxt_p);
    if(res!=0) {
      switch(res) {
        case -FI_EAGAIN:
          INFO("fi_recv wants us to try again\n");
          break;
        default:
          ERR("fi_recv failed: %d\n", res, fi_strerror(-res));
          status = MSTRO_OFI_FAIL;
          goto BAILOUT;
      }
    }
  } while(res!=0);
BAILOUT:
  return status;
}

static inline
mstro_status
mstro_ofi__maybe_notify_completion(mstro_ofi_msg_context ctx)
{
  int ret = pthread_mutex_lock(&ctx->lock);
  if(ret!=0) {
    ERR("msg handle could not lock completion context mutex: %d\n", ret);
    goto COMPLETION_NOTIFICATION_FAILED;
  }
  ret = pthread_cond_signal(&ctx->completion);
  if(ret!=0) {
    ERR("msg handler completion signalling failed: %d\n", ret);
    goto COMPLETION_NOTIFICATION_FAILED;
  }
  ret = pthread_mutex_unlock(&ctx->lock);
  if(ret!=0) {
    ERR("msg handler completion signalling failed: %d\n", ret);
    goto COMPLETION_NOTIFICATION_FAILED;
  }
COMPLETION_NOTIFICATION_FAILED:
  if(ret!=0) {
    return MSTRO_FAIL;
  } else {
    return MSTRO_OK;
  }
}

  /** Poll EP for one completion and handle it, if there is one
 *
 * If expected_slot_p is non-NULL it indicates that if the
 * completion is for a message in this known slot a new RECV should be
 * posted immediately.
 */
static inline
mstro_status
mstro_ofi__check_and_handle_cq(struct mstro_endpoint *ep,
                               struct mstro_msg_envelope** expected_slot_p,
                               mstro_ofi_msg_context *expected_ctx_p,
                               mstro_msg_handler incoming_msg_handler)
{
  assert(ep!=NULL);
  assert(incoming_msg_handler!=NULL);
  mstro_status status = MSTRO_OK;
  struct fi_cq_tagged_entry entry;
  int ret;
  ret = fi_cq_read(ep->cq, &entry, 1);

  /* no-data poll? */
  if(ret== -FI_EAGAIN) {
    return MSTRO_OK; /* FI_EAGAIN .. we'll be back */
  }

  /* actual errors */
  if(ret<=0) {
    struct fi_cq_err_entry err_entry;
    fi_cq_readerr(ep->cq, &err_entry, 0);
    ERR("Failure in fi_cq_read: %s %s\n",
        fi_strerror(err_entry.err),
        fi_cq_strerror(ep->cq,
                       err_entry.prov_errno,
                       err_entry.err_data, NULL, 0));
    return MSTRO_FAIL;
  }

  if(ret>1) {
    ERR("Unexpectedly large number of CQ entries: %d, expected 1\n", ret);
    return MSTRO_FAIL;
  }

  assert(ret==1);

  /* completions can be from our 'slot' or from elsewhere. The
   * context tells us. Sometimes we get spurious ones -- we
   * just drop them. */
  mstro_ofi_msg_context ctx = (mstro_ofi_msg_context)entry.op_context;
  assert(ctx!=NULL);

  bool re_post = (expected_slot_p!=NULL && ctx->msg==*expected_slot_p);
  /* if it was a known slot we re-post a fresh RECV immediately -- this allocates a fresh ctx */
  INFO("Completion of op with ctx %p message %p%s of size %zu on endpoint %p\n",
       ctx, ctx->msg, re_post ? " (slot i)" : "",
       entry.len, ep);
  
  if(re_post) {
    DEBUG("Posting fresh recv on slot %p", expected_slot_p);
    status = mstro_ofi__loop_post_recv(ep->ep, expected_slot_p, expected_ctx_p);
  }
  
  
  if(entry.flags&FI_RECV) {
    assert(ctx->msg!=NULL);
    ctx->msg->payload_size = entry.len; /* this permits us to not pass ofi 'entry' to pc/pm handler */
    const struct mstro_msg_envelope *msg = ctx->msg;
    status = incoming_msg_handler(msg, ep);
    
    if(status!=MSTRO_OK) {
      ERR("Error handling incoming message, dropping it: %d (%s)\n",
          status, mstro_status_description(status));
    }
    
    if(re_post) {
      ; /* fresh receive will be posted later, and slot messages have no completion watchers */
    } else {
      /* other recv */
      DEBUG("non-slot message, cleaning up\n");
      if(ctx->has_completion) {
        DEBUG("ctx asks for completion notification\n");
        status = mstro_ofi__maybe_notify_completion(ctx);          
      } else {
        DEBUG("ctx wants no completion notification, cleaning up ctx\n");
        mstro_msg_envelope_free(ctx->msg);
        status = mstro_ofi__msg_context_destroy(ctx);
      }
    }
    
  } else if(entry.flags&FI_SEND) {
    assert(ctx->msg!=NULL);
    struct mstro_msg_envelope *msg = ctx->msg;
    
    /* completion of our sent messages */
    DEBUG("Send of message %p type %s completed\n",
          msg, msg->descriptor->name);
    /* otherwise clean it up */
    if(ctx->has_completion) {
      DEBUG("ctx asks for completion notification\n");
      status = mstro_ofi__maybe_notify_completion(ctx);          
    } else {
      DEBUG("ctx wants no completion notification, cleaning up ctx\n");
      mstro_msg_envelope_free(msg);
      status = mstro_ofi__msg_context_destroy(ctx);
    }
    
  } else if(entry.flags&FI_RMA) {
    assert(ctx->msg==NULL);
    DEBUG("RDMA op completed, ctx %p\n", ctx);
    assert(ctx->has_completion); /* it does not make sense to
                                  * not have this set */
    status = mstro_ofi__maybe_notify_completion(ctx);          
  } else {
    DEBUG("Unexpected completion, flags %ull\n", entry.flags);
  }
  
  return status;      
}

static inline
mstro_status
mstro_ofi__loop(_Atomic bool *terminate,
                mstro_msg_handler incoming_msg_handler)
{
  mstro_status status = MSTRO_OK;
  size_t i;
  struct mstro_msg_envelope** slots
      =calloc(g_endpoints->size,
              sizeof(struct mstro_msg_envelope*));
  mstro_ofi_msg_context* ctxts
      =calloc(g_endpoints->size,
              sizeof(mstro_ofi_msg_context));
  if(slots==NULL || ctxts==NULL) {
    status=MSTRO_NOMEM;
    goto BAILOUT;
  }

  INFO("OFI loop starting\n");
  /* post receives on all endpoints */
  for(i=0; i<g_endpoints->size; i++) {
    status = mstro_ofi__loop_post_recv(g_endpoints->eps[i].ep,
                                       &(slots[i]), &(ctxts[i]));
    if(status!=MSTRO_OK) {
      goto BAILOUT;
    }
  }
  DEBUG("Posted a RECV on each of %d endpoints, iterating cq_read\n",
        g_endpoints->size);

  /* wait for completion on all of them. If a message arrives, handle
   * it and re-post a receive */
  mstro_status s=MSTRO_OK;
  do {
    for(i=0; false==atomic_load(terminate) && i<g_endpoints->size; i++) {
      mstro_ofi__check_and_handle_cq(&g_endpoints->eps[i],
                                     &(slots[i]), &(ctxts[i]),
                                     incoming_msg_handler);
    }
  } while(false==atomic_load(terminate));
  DEBUG("PM comm thread quits: loop status %d, global status %d\n",
        s, status);
  
BAILOUT:
   if(slots) {
     for(i=0; i<g_endpoints->size; i++) 
       mstro_msg_envelope_free(slots[i]);
     free(slots);
   }
   if(ctxts) {
     for(i=0; i<g_endpoints->size; i++) 
       mstro_ofi__msg_context_destroy(ctxts[i]);
     free(ctxts);
   }
   return status;
}

mstro_status
mstro_ofi_pm_loop(_Atomic bool *terminate)
{
  g_pool_app_id = MSTRO_APP_ID_MANAGER;

  return mstro_ofi__loop(terminate, mstro_pm_handle_msg);
}  

mstro_status
mstro_ofi_pc_loop(_Atomic bool *terminate)
{
  return mstro_ofi__loop(terminate, mstro_pc_handle_msg);
}


mstro_status
mstro_ofi__check_compatibility(bool *suitable_p,
                               struct mstro_endpoint_descriptor_ *remote,
                               struct mstro_endpoint *local)
{
  if(suitable_p==NULL)
    return MSTRO_INVOUT;
  if(remote==NULL ||local==NULL)
    return MSTRO_INVARG;
  switch(remote->type) {
    case MSTRO_EP_OFI_IN4:
    case MSTRO_EP_OFI_IN6:
    case MSTRO_EP_OFI_IB:
    case MSTRO_EP_OFI_PSMX:
    case MSTRO_EP_OFI_PSMX2:
    case MSTRO_EP_OFI_GNI:
    case MSTRO_EP_OFI_BGQ:
    case MSTRO_EP_OFI_MLX:
    case MSTRO_EP_OFI_STR: {
      /* these have a chance against current OFI-only endpoints */
      uint32_t addr_format;
      void *addr;
      size_t addrlen;
      mstro_status stat
          = mstro_ep_desc_to_ofi_addr(&addr_format, &addr, &addrlen, remote);
      if(stat!=MSTRO_OK) {
        ERR("Cannot convert EPD to address: %d\n", stat);
        *suitable_p =false;
/* #if !defined(OFI_ISSUE_5453_RESOLVED) */
      /* disable IPv6 for now */
      } else if(addr_format == FI_SOCKADDR_IN6) {
        *suitable_p = false;
/* #endif */
      } else {
        if(addr_format==local->fi->addr_format) {
          *suitable_p = true;
        } else
          *suitable_p = false;
      }
      break;
    }
    default:
      DEBUG("Non-OFI transport endpoint encountered\n");
      *suitable_p = false;
      break;
  }
  return MSTRO_OK;
}









/* FIXME: should go to pool.c */

/* internal entry point: started in a pthread, runs Pool client loop until it
 * is told to terminate (when g_pc_stop is set) */

static _Atomic bool g_pc_stop; /* static init 0 means: no */
/*
static void*
mstro_pc_thread(void *closure)
{
  assert(closure==NULL);
  DEBUG("pool client thread starting\n");
  mstro_status s = mstro_ofi_pc_loop(&g_pc_stop);
  if(s!=MSTRO_OK) {
    ERR("pc loop terminated with error %d (%s)\n",
        s, mstro_status_description(s));
  }
  DEBUG("pool client thread stopping\n");
  
  return NULL;
}
*/

static void*
mstro_pc_thread(void *closure)
{
  assert(closure==NULL);
  DEBUG("pool client thread starting\n");
#ifdef HAVE_MIO
  struct mio_thread mthread;
  mio_thread_init(&mthread);
#endif
  mstro_status s = mstro_ofi_pc_loop(&g_pc_stop);
  if(s!=MSTRO_OK) {
    ERR("pc loop terminated with error %d (%s)\n",
        s, mstro_status_description(s));
  }
  DEBUG("pool client thread stopping\n");
#ifdef HAVE_MIO
  mio_thread_fini(&mthread);
#endif
  return NULL;
}

mstro_status
mstro_pc_terminate(void)
{
  mstro_status status;

  /* FIXME: we need to send a LEAVE request, wait for that to succeed,
   * then kick the thread */
  INFO("pool client termination initiated\n");

  Mstro__Pool__Leave leave = MSTRO__POOL__LEAVE__INIT;

  Mstro__Pool__MstroMsg msg = MSTRO__POOL__MSTRO_MSG__INIT;
  status = mstro_pmp_package(&msg, (ProtobufCMessage*)&leave);
  if(status !=MSTRO_OK) {
    ERR("Failed to package %s into a pool manager message\n",
        leave.base.descriptor->name);
    goto BAILOUT;
  }
    
  status = mstro_pmp_send_nowait(MSTRO_APP_ID_MANAGER, &msg);
  
  /* The BYE message will be handled by the normal incoming message handler
   * We can tell it's done by checking g_mstro_pm_attached */
  switch(status) {
    case MSTRO_OK: {
      /* spin-wait */
      while(atomic_load(&g_mstro_pm_attached)) {
#if defined(_POSIX_PRIORITY_SCHEDULING)
        sched_yield();
#endif
        usleep(10); /*usec*/
      }
      INFO("Detached from pool manager\n");
      break;
    }
    case MSTRO_NO_PM: {
      /* simulate detach (local only) */
      atomic_store(&g_mstro_pm_attached, false);
      g_pool_app_id = MSTRO_APP_ID_INVALID;
      break;
    }
    default:
      ERR("Failed to send LEAVE message to pool manager: %d (%s)\n",
          status, mstro_status_description(status));
      return status;
  }

  /* terminate PC thread */
  atomic_store(&g_pc_stop, true);
  void *res;
  int s = pthread_join(g_pc_thread, &res);
  INFO("pool client thread terminated\n");
  if(s!=0) {
    ERR("pool client thread termination with error: %d (%s)\n",
        s, strerror(s));
    status = MSTRO_FAIL;
  } else {
    status = MSTRO_OK;
  }

  mstro_status status2 = mstro_ofi_finalize();
  if(status2!=MSTRO_OK) {
    ERR("Failed to finalize OFI layer: %d\n", status2);
  }
  if(status==MSTRO_OK) {
    status = status2;
  }

BAILOUT:
  return status;
}

/* end of FIXME (code for pool.c) */

/** Attach current application to a running pool manager */
mstro_status
mstro_pm_attach(const char *remote_pm_info)
{
  /* GNI cookies are special. We need to have the matching cookie before we can enable the endpoint. */
  mstro_endpoint_descriptor pm_epd=NULL;

  mstro_status s=mstro_ep_desc_deserialize(&pm_epd, remote_pm_info);
  
  if(s!=MSTRO_OK) {
    ERR("Failed to parse pool manager info: %d (%s%s)\n",
        s, mstro_status_description(s),
        /* parser errrors also if a NULL result */
        s==MSTRO_NOMEM ? " or invalid pool-manager info data" : "");
    goto BAILOUT;
  }

  mstro_endpoint_descriptor tmp;
  size_t num_entries=0;
  LL_COUNT(pm_epd, tmp, num_entries);
  INFO("Parsed %d pool manager endpoints\n", num_entries);

  {
    mstro_endpoint_descriptor tmp;
    size_t num_cookies = 0;
    
    LL_FOREACH(pm_epd, tmp) {
      INFO("Pool manager endpoint: %s\n", mstro_ep_desc_describe(tmp));
      if(tmp->type==MSTRO_EP_OFI_GNI) {
        INFO("GNI descriptor found, oob cookie %s\n", tmp->oob_cookie);
        if(num_cookies<1) {
          s = mstro_drc_init_from_oob_string(&g_drc_info, tmp->oob_cookie);
          if(s!=MSTRO_OK) {
            ERR("Failed to use provided DRC cookie\n");
            goto BAILOUT;
          } else {
            num_cookies++;
          }
        } else {
          WARN("Multiple GNI cookies found, first one will win\n");
        }
      }
    }
  }
  
  s=mstro_ofi_init();
  if(s!=MSTRO_OK) {
    ERR("Failed to initialize OFI layer\n");
    goto BAILOUT;
  }
  
  /* start handler thread: run pool client loop. It will stop when mstro_pc_terminate is called */
  /* we need to start it before endpoint selection since we do RDMA
   * read for the config block, and that needs to report completion
   * through our normal ofi_loop mechanism */
  int res = pthread_create(&g_pc_thread, NULL, mstro_pc_thread, NULL);
  if(res!=0) {
    ERR("Failed to start pool connection thread: %d (%s)\n",
        res, strerror(res));
    s=MSTRO_FAIL;
    goto BAILOUT;
  }
  g_component_descriptor.type = MSTRO_COMPONENT_TYPE_APP;
  
  
  s=mstro_ofi__select_endpoint(pm_epd, &g_pm_endpoint, &g_pm_addr);
  if(s!=MSTRO_OK) {
    ERR("Failed to select a suitable endpoint to talk to pool manager\n");
    goto BAILOUT;
  }
  assert(g_pm_endpoint!=NULL);
  
  /* register pool manager as a pmp app entry */
  s = mstro_pm_app_register_manager(g_pm_endpoint, g_pm_addr);
  if(s!=MSTRO_OK) {
    ERR("Failed to register Pool Manager in registry\n");
    goto BAILOUT;
  }
  
  
  /* send JOIN */
  
  
  /* FIXME make function set_transport_default() beware protobuf stack init
   */
  /* if we have OFI we'll add another entry */
  
  Mstro__Pool__TransportKind transport_kinds[] =
      { MSTRO__POOL__TRANSPORT_KIND__GFS
#ifdef HAVE_MIO
        ,MSTRO__POOL__TRANSPORT_KIND__MIO
#endif
      };
  size_t num_available_transports = sizeof(transport_kinds)/sizeof(transport_kinds[0]);
  
  Mstro__Pool__TransportMethods transport_methods
      = MSTRO__POOL__TRANSPORT_METHODS__INIT;
  transport_methods.n_supported = num_available_transports;
  transport_methods.supported
      = (Mstro__Pool__TransportKind*)&transport_kinds;
  
  const char* env_transport_default = getenv(MSTRO_ENV_TRANSPORT_DEFAULT);
  if (env_transport_default != NULL) {
    int i;
    int found = 0;
    for (	i = 0;	i < MSTRO__POOL__TRANSPORT_KIND__NUMBER_OF_KINDS; i++ ) { 
      if (!strcmp(env_transport_default, g_transport_registry[i].name)) {
        transport_methods.supported[0] = (Mstro__Pool__TransportKind)i;
        found = 1;
        INFO("Setting default transport method to %s (env %s)\n",
             g_transport_registry[i].name,
             MSTRO_ENV_TRANSPORT_DEFAULT);
      }
    }
    if (! found)
      WARN("default transport method given (env %s)  not found, using maestro defaults\n",  env_transport_default);
  }
  INFO("Preferred transport list: \n");
  size_t i;
  for (i = 0; i < transport_methods.n_supported; i++ )
    INFO("#%d %s \n", i, g_transport_registry[transport_methods.supported[i]].name);
  /**/
  
  Mstro__Pool__Join join = MSTRO__POOL__JOIN__INIT;
  join.protocol_version = MSTRO_POOL_PROTOCOL_VERSION;
  join.serialized_endpoint = g_pm_endpoint->addr_serialized;
  join.transport_methods = &transport_methods;
  join.component_name = g_initdata->component_name;
  
  Mstro__Pool__MstroMsg msg = MSTRO__POOL__MSTRO_MSG__INIT;
  s = mstro_pmp_package(&msg, (ProtobufCMessage*)&join);
  if(s!=MSTRO_OK) {
    ERR("Failed to package %s into a pool manager message\n",
        join.base.descriptor->name);
    goto BAILOUT;
  }
  
  s = mstro_pmp_send_nowait(MSTRO_APP_ID_MANAGER, &msg);

  /* The WELCOME message will be handled by the normal incoming
   * message handler.  We can tell it's done by checking
   * g_pool_app_id */
  
  switch(s) {
    case MSTRO_OK: {
      /* spin-wait */
      while(!atomic_load(&g_mstro_pm_attached)) {
#if defined(_POSIX_PRIORITY_SCHEDULING)
        sched_yield();
#endif
        usleep(10); /*usec*/
      }
      break;
    }
    case MSTRO_NO_PM: 
      ERR("Confusion: JOIN message failed with 'no PM info' error\n");
      /* fall-thru */
    default:
      ERR("Failed to send JOIN message: %d (%s)\n",
          s, mstro_status_description(s));
      return s;
  }
  
  
  /* Welcome received */
  INFO("Welcome message received, our app id is %"PRIu64"\n", 
       g_pool_app_id);
  

BAILOUT:
  return s;
}


