#include "i_pool_client.h"

#include "maestro/i_globals.h"
#include "maestro/logging.h"
#include "maestro/i_pool.h"
#include "maestro/i_pool_manager_registry.h"
#include "../transport/transport.h"
#include "maestro/i_ofi.h"
#include "i_subscription_registry.h"

#include <unistd.h>
#include <inttypes.h>

#ifdef HAVE_MIO
#include <mio.h> /* the MIO one */
#include "transport/transport_mio.h"
#endif


/* simplify logging */
#define DEBUG(...) LOG_DEBUG(MSTRO_LOG_MODULE_PC,__VA_ARGS__)
#define INFO(...)  LOG_INFO(MSTRO_LOG_MODULE_PC,__VA_ARGS__)
#define WARN(...)  LOG_WARN(MSTRO_LOG_MODULE_PC,__VA_ARGS__)
#define ERR(...)   LOG_ERR(MSTRO_LOG_MODULE_PC,__VA_ARGS__)


static inline
mstro_status
mstro_pc__handle_declare_ack(const Mstro__Pool__DeclareAck *declare_ack)
{
  assert(declare_ack!=NULL);
  DEBUG("CDO DeclareAck for serial %zu\n", declare_ack->serial);
  
  if(declare_ack->cdoid==NULL) 
    return MSTRO_INVMSG;

  struct mstro_cdo_id cdoid;
  cdoid.qw[0] = declare_ack->cdoid->qw0;
  cdoid.qw[1] = declare_ack->cdoid->qw1;

  /* feed into local CDO layer */
  return mstro_cdo_declare_completion(declare_ack->serial, &cdoid, declare_ack->channel);
}

static inline
mstro_status
mstro_pc__handle_subscribe_ack(const Mstro__Pool__SubscribeAck *sack)
{
  assert(sack!=NULL);
  assert(sack->handle!=NULL);
  DEBUG("SubscribeAck received, local id %" PRIu64", handle %" PRIu64 "\n",
        sack->serial, sack->handle->id);

  mstro_status s = mstro_subscription_register_bh(sack);
  if(s!=MSTRO_OK) {
    ERR("Failed to handle subscription ack for local id %" PRIu64 "\n",
        sack->serial);
  }
  return s;
}

static inline
mstro_status
mstro_pc__handle_resolve_reply(const Mstro__Pool__ResolveReply *reply)
{
  assert(reply!=NULL);
  assert(reply->query!=NULL);
  switch(reply->query->query_case) {
    case MSTRO__POOL__RESOLVE__QUERY_CDOID:
      DEBUG("Incoming RESOLVE-REPLY for a CDO-ID (result: |%s|)\n",
            reply->name);
      return mstro_pool_resolve_reply_bh(reply);
    default:
      ERR("FIXME: unhandled resolve-reply for a query of type %d\n",
          reply->query->query_case);
      return MSTRO_FAIL;
  }
}


/* store @arg new_attributes in the serialized attributes slot of @arg
 * cdo, and replace the attribute dict values accordingly. Consumes
 * @arg new_attributes. */
static inline
mstro_status
mstro_cdo__replace_attributes(mstro_cdo cdo,
                              Mstro__Pool__Attributes *new_attributes)
{
  /* We can't lock a CDO on it's own. But this function should only be
   * called before SEAL state change completes, or during WITHDRAW
   * etc. and so users can't be accessing the CDO legally, and it's
   * way before or after the CDO could be in the pool, so transport
   * etc can't touch it either. */
  assert(mstro_cdo_state_check(cdo, MSTRO_CDO_STATE_DECLARED));

  mstro_status s = MSTRO_UNIMPL;

  Mstro__Pool__Attributes *tmp = cdo->attributes_msg;
  cdo->attributes_msg = new_attributes;
  mstro__pool__attributes__free_unpacked(tmp, NULL);

  s = mstro_attribute_dict_clear(cdo->attributes);
  if(s!=MSTRO_OK) {
    ERR("Failed to clear existing attribute dictionary on CDO\n");
    goto BAILOUT;
  }

  /* insert the special ones */
  if(cdo->raw_ptr) {
    s = mstro_attribute_dict_set(cdo->attributes,
                                 MSTRO_ATTR_CORE_CDO_RAW_PTR,
                                 MSTRO_CDO_ATTR_VALUE_pointer,
                                 cdo->raw_ptr, true);
    if(s!=MSTRO_OK) {
      ERR("Failed to re-set raw-ptr on CDO\n");
      goto BAILOUT;
    }
  }

  if(cdo->mamba_array) {
    s = mstro_attribute_dict_set(cdo->attributes,
                                 MSTRO_ATTR_CORE_CDO_MAMBA_ARRAY,
                                 MSTRO_CDO_ATTR_VALUE_pointer,
                                 cdo->mamba_array, true);
    if(s!=MSTRO_OK) {
      ERR("Failed to re-set mamba_array on CDO\n");
      goto BAILOUT;
    }
  }

  /* insert all from message */
  s = mstro_cdo_attributes_update_incoming(cdo, cdo->attributes_msg, 0, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed to insert new attributes into CDO\n");
    goto BAILOUT;
  }
  WITH_CDO_ID_STR(idstr, &cdo->gid,
                  DEBUG("Replaced attributes on CDO |%s| (%s, @%p)\n",
                        cdo->name, idstr, cdo););

BAILOUT:
  return s;
}

static inline
mstro_status
mstro_pc__handle_poolop_ack(Mstro__Pool__PoolOpAck *poack)
{
  assert(poack!=NULL);

  DEBUG("PoolOpAck for a %s operation\n",
        (protobuf_c_enum_descriptor_get_value(
            &mstro__pool__pool_op_ack__pool_op__descriptor, poack->op))
        ->name);

  if(!poack->op)
    return MSTRO_INVMSG;

  if(poack->status!=MSTRO__POOL__POOL_OP_ACK__POOL_OP_STATUS__OK) {
    WARN("PoolOpAck contains error status\n");
  }

  mstro_cdo_state new_state;
  Mstro__Pool__Attributes *new_attributes = NULL;
  
  switch(poack->op) {
    case MSTRO__POOL__POOL_OP_ACK__POOL_OP__SEAL: {
      if(poack->payload_case == MSTRO__POOL__POOL_OP_ACK__PAYLOAD_UPDATED_ATTRIBUTES) {
        DEBUG("SEAL-ACK has updated attributes payload\n");
        assert(poack->updated_attributes!=NULL);
        /* blindly replace attributes, relying on the fact that the PM knows what's good for us */
        new_attributes = poack->updated_attributes;
        poack->updated_attributes = NULL; /* we consume this */
      }
      new_state = MSTRO_CDO_STATE_SEALED;
      goto shared_cdo_state_update;
    }
      
    case MSTRO__POOL__POOL_OP_ACK__POOL_OP__OFFER:
      new_state = MSTRO_CDO_STATE_OFFERED;
      goto shared_cdo_state_update;
      
    case MSTRO__POOL__POOL_OP_ACK__POOL_OP__REQUIRE:
      new_state = MSTRO_CDO_STATE_REQUIRED;
      goto shared_cdo_state_update;
      
    case MSTRO__POOL__POOL_OP_ACK__POOL_OP__RETRACT:
      new_state = MSTRO_CDO_STATE_RETRACTED_GLOBALLY;
      goto shared_cdo_state_update;
      
    case MSTRO__POOL__POOL_OP_ACK__POOL_OP__WITHDRAW:
      new_state = MSTRO_CDO_STATE_WITHDRAWN_GLOBALLY;
    shared_cdo_state_update: {
        assert(poack->operand_case==MSTRO__POOL__POOL_OP_ACK__OPERAND_CDOID);
        assert(poack->cdoid!=NULL);
        struct mstro_cdo_id id = {.qw[0] = poack->cdoid->qw0,
                                  .qw[1] = poack->cdoid->qw1};
        mstro_cdo cdo;
        mstro_cdo__find_cdo(&id, &cdo);
        if(cdo==NULL) {
          WITH_CDO_ID_STR(idstr, &id, {
              ERR("PoolOpAck for unknown CDO %s\n", idstr);
              return MSTRO_INVARG;
            });
        }
        if(new_attributes!=NULL) {
          mstro_cdo__replace_attributes(cdo, new_attributes);
        }
        mstro_cdo_state state_flags = (mstro_cdo_state_get(cdo)
                                       & MSTRO_CDO_STATE_FLAGS);
        
        mstro_cdo_state_set(cdo, new_state | state_flags);
        WITH_CDO_ID_STR(idstr, &id, {
            DEBUG("CDO %s now in state %d (%s)\n",
                  idstr, new_state, mstro_cdo_state_describe(new_state));
          });
        
        break;
      }
      
    case MSTRO__POOL__POOL_OP_ACK__POOL_OP__UNSUBSCRIBE: {
        mstro_status status = mstro_subscription_unregister_bh(poack->handle);
        if(status!=MSTRO_OK) {
          ERR("Failure unregistering subscription handle %" PRIu64 "\n",
              poack->handle->id);
          return status;
        }
        break;
      }
    default:
      ERR("PoolOpAck for invalid pool operation %d\n", poack->operand_case);
      return MSTRO_INVMSG;
  }
  return MSTRO_OK;
}


/** select the best way to transfer CDO, based on what the dst
 *   supports.  Fills in the (caller-provided) TransferTicket
 *   regarding the method and the ticket union.
 */
static inline
mstro_status
mstro_pc__select_transfer_method(mstro_cdo cdo,
                                 const Mstro__Pool__TransportMethods *dst_methods,
                                 Mstro__Pool__TransferTicket *selected)
{
  if(dst_methods==NULL)
    return MSTRO_INVARG;
  if(selected==NULL)
    return MSTRO_INVOUT;

  if(dst_methods->n_supported==0) {
    ERR("No transport method advertised by dst\n");
    return MSTRO_FAIL;
  }
  cdo=cdo; /* avoid 'unused param' warning */
  
  // Arbitrarily picking up the preferred transport method
  switch (dst_methods->supported[0]) {
    case MSTRO__POOL__TRANSPORT_KIND__GFS:
      selected->method = MSTRO__POOL__TRANSPORT_KIND__GFS;
      selected->ticket_case = MSTRO__POOL__TRANSFER_TICKET__TICKET_GFS; 	
      break;
    case MSTRO__POOL__TRANSPORT_KIND__MIO:
      selected->method = MSTRO__POOL__TRANSPORT_KIND__MIO;
      selected->ticket_case = MSTRO__POOL__TRANSFER_TICKET__TICKET_MIO;
      break;
    default: 
      ERR("Add a case in the TransportKind methods vs TicketCase switch\n");
      return MSTRO_UNIMPL;
  }
  return MSTRO_OK;
}

static inline
mstro_status
mstro_pc__construct_gfs_path_for_cdo(const mstro_cdo src_cdo,
                                     char **path)
{
  mstro_status s=MSTRO_UNIMPL;
  assert(path!=NULL);
  assert(src_cdo!=NULL);

  size_t l = g_mstro_transport_gfs_dir_len + strlen(src_cdo->name) + 1;
  *path = malloc(l);
  if(*path==NULL) {
    return MSTRO_NOMEM;
  }
  strcpy(*path, g_mstro_transport_gfs_dir);
  
  /* FIXME: possibly add a (configurable?) number of subdirectories
   * based on first segment of CDO name */
  
  strcpy((*path)+g_mstro_transport_gfs_dir_len, src_cdo->name);
  DEBUG("Constructed GFS transport path %s\n", *path);
  return MSTRO_OK;
}


static inline
mstro_status
mstro_pc__handle_initiate_transfer(const Mstro__Pool__InitiateTransfer* init)
{
  assert(init != NULL);
  mstro_cdo src_cdo = NULL;
  if (init->dst_serialized_endpoint == NULL) {
    ERR("Cannot initiate transfer without dst app endpoint\n");
    return MSTRO_INVMSG;
  }
  
  DEBUG("Initiating a transfer to endpoint %s\n",
        init->dst_serialized_endpoint);

  struct mstro_cdo_id cdoid = { .qw[0] = init->cdoid->qw0,
                                .qw[1] = init->cdoid->qw1 };
  WITH_CDO_ID_STR(idstr, &cdoid,
                  INFO("ticket for CDO gid %s\n",
                       idstr););
  if (!(MSTRO_OK == mstro_pool__find_source_cdo(&cdoid, init->dst_attributes,
                                                &src_cdo))) {
    WITH_CDO_ID_STR(idstr, &cdoid,
                    ERR("Cannot transfer CDO gid %s because not in local pool\n",
                        idstr););
    return MSTRO_FAIL;
  }

  if(init->dst_attributes==NULL) {
    WARN("No attributes on CDO\n");
  } else {
    if(init->dst_attributes->val_case==MSTRO__POOL__ATTRIBUTES__VAL_KV_MAP) {
      DEBUG("%zu attributes in kv-map\n", init->dst_attributes->kv_map->n_map);
    } else {
      WARN("non-kv attributes\n");
    }
  }

  DEBUG("Initiating transfer from src app %zu (me) to dst app %zu of CDO %s\n",
        g_pool_app_id, init->dst_appid->id, src_cdo->name);

  if(g_pool_app_id==init->dst_appid->id) {
    WARN("FIXME: We will be talking to ourselves via transport, should use a shortcut\n");
  }
  
  /* ensure we know how to talk to the recipient */
  struct mstro_pm_app_registry_entry *e;
  mstro_status status = mstro_pm_app_lookup(init->dst_appid->id, &e);
  if(e) {
    DEBUG("Found app %zu in local registry, good\n");
  } else {
    mstro_endpoint_descriptor dst_epd;
    status=mstro_ep_desc_deserialize(&dst_epd, init->dst_serialized_endpoint);
  
    if(status!=MSTRO_OK) {
      ERR("Failed to parse destination endpoint descriptor\n");
      return MSTRO_INVMSG;
    }

    struct mstro_endpoint *dst_ep=NULL;
    fi_addr_t dst_addr;
    status=mstro_ofi__select_endpoint(dst_epd, &dst_ep, &dst_addr);
    if(status!=MSTRO_OK) {
      ERR("DST endpoint not usable\n");
      return MSTRO_FAIL;
    }
    assert(dst_ep!=NULL);

    /* pass into registry */
    status = mstro_pc_app_register(dst_ep, dst_addr,
                                   strdup(init->dst_serialized_endpoint),
                                   init->methods,
                                   init->dst_appid->id,
                                   NULL);
    if(status!=MSTRO_OK) {
      ERR("Failed to register peer app %zu: %d\n", init->dst_appid->id);
      return MSTRO_FAIL;
    } else {
      DEBUG("Registered peer app %zu\n", init->dst_appid->id);
    }
  }
  
  /* Prepare the ticket */
  Mstro__Pool__UUID id = MSTRO__POOL__UUID__INIT;
  id.qw0 = cdoid.qw[0];
  id.qw1 = cdoid.qw[1];
  Mstro__Pool__TransferTicket ticket = MSTRO__POOL__TRANSFER_TICKET__INIT;
  ticket.cdoid = &id;

  const int64_t *size;
  enum mstro_cdo_attr_value_type type;
  if (! (MSTRO_OK == mstro_cdo_attribute_get(src_cdo,
                                             MSTRO_ATTR_CORE_CDO_SCOPE_LOCAL_SIZE,
                                             &type, (const void**)&size))) {
    ERR("Couldn't retrieve CDO %s local_size needed for transport\n", src_cdo->name);
    return MSTRO_FAIL;
  }
  int64_t realsize = *size;
  if(realsize==-1) {
    DEBUG("Source CDO empty, doing NULL transfer\n");
    realsize = 0;
  }
  if (!g_mio_available
      || (( realsize % getpagesize()) != 0 
          && init->methods->supported[0] == MSTRO__POOL__TRANSPORT_KIND__MIO)
      ){
    WARN("Not issuing a ticket with MIO. Either not available or CDO size (%zu)"
         " is not a multiple of the page size (%d)."
         " Will use maestro-core default transport (GFS)\n",
         realsize, getpagesize());
    init->methods->supported[0] = MSTRO__POOL__TRANSPORT_KIND__GFS;
  }

  status = mstro_pc__select_transfer_method(src_cdo, init->methods,
                                            &ticket);
  if(status!=MSTRO_OK) {
    ERR("Failed to select transport method\n");
    return MSTRO_FAIL;
  }
  
  Mstro__Pool__TransferTicketGFS gfs = MSTRO__POOL__TRANSFER_TICKET_GFS__INIT; 
  Mstro__Pool__TransferTicketMIO mio = MSTRO__POOL__TRANSFER_TICKET_MIO__INIT;

  ticket.data_size = realsize;
  ticket.gfs = &gfs;

  /* even conditionals are problematic, had to take out protobuf INITs ----^ */
  switch(ticket.ticket_case) {
    case MSTRO__POOL__TRANSFER_TICKET__TICKET_GFS:
      INFO("TICKET CASE GFS\n");
      status= mstro_pc__construct_gfs_path_for_cdo(src_cdo, &gfs.path);
      if(status!=MSTRO_OK) {
        ERR("Failed to construct GFS path for SRC-CDO: %d (%s)\n",
            status, mstro_status_description(status));
        return status;
      }
      gfs.keep_file = 0; // Arbitrarily rm the transport file on dst
      break;
    case MSTRO__POOL__TRANSFER_TICKET__TICKET_MIO:
      INFO("TICKET CASE MIO\n");
#ifdef HAVE_MIO
      struct mstro_cdo_id* semid; 
      semid = malloc(sizeof(struct mio_obj_id));
      if (semid == NULL) {
        ERR("No more memory.\n");
        return MSTRO_NOMEM;
      }
      struct mstro_cdo_id* objid; 
      objid = malloc(sizeof(struct mio_obj_id));
      if (objid == NULL) {
        ERR("No more memory.\n");
        return MSTRO_NOMEM;
      }
      char* semname = NULL;
#define MSTRO_MIO_SEM_STR_MAXLEN 	128
#warning semname should not me constructed here with MIO-specific length
      mstro_str_random(&semname, MSTRO_MIO_SEM_STR_MAXLEN);
      if (semname == NULL) {
        ERR("Couldn't prepare an id for semaphore obj\n");
        return MSTRO_FAIL;
      }
      status = mstro_cdo_id_from_name(semname, semid); /* So we do collision
                                                          detection in only
                                                          one place */
      if (status != MSTRO_OK) {
        ERR("Couldn't make an id from name for semaphore obj\n");
        return MSTRO_FAIL;
      }
      
      WITH_MIO_OBJ_STR(idstr, (struct mstro_cdo_id*)semid,
                       INFO("Semaphore has ID: %s\n",
                            idstr););
      WITH_CDO_ID_STR(idstr, &(src_cdo->gid),
                      INFO("(CDO associated has ID: %s)\n",
                           idstr););
      
      assert(sizeof(struct mstro_cdo_id) == 2*sizeof(uint64_t));
      assert(sizeof(struct mstro_cdo_id) == sizeof(struct mio_obj_id));
      
      mio.semid.len = sizeof(struct mstro_cdo_id);
      mio.semid.data = (uint8_t*)semid;
      mio.objid.len = sizeof(struct mstro_cdo_id);
      objid->qw[0] = src_cdo->gid.qw[0];
      objid->qw[1] = src_cdo->gid.qw[1];
      mio.objid.data = (uint8_t*)objid;
      
      mio.keep_obj = 0;
      
      ticket.mio = &mio;
#else
      ERR("Request to issue an MIO ticket, but built without MIO support\n");
      return MSTRO_FAIL;
#endif
      break;
    default:
      ERR("Unsupported ticket kind %d\n", ticket.ticket_case);
      return MSTRO_UNIMPL;
  }
  
 /**/

  if(src_cdo->attributes_msg==NULL) {
    ERR("source CDO has no attributes message data -- should not happen\n");
    return MSTRO_FAIL;
  }
  ticket.attributes = src_cdo->attributes_msg;
  
  INFO("Issued ticket for CDO %s, and starting execute process\n", src_cdo->name);
  
  //INFO("TransferTicket using path %s\n", ticket.gfs->path);
  INFO("TransferTicket cdo size %zu\n", ticket.data_size);
  
  /* Execute transport (non-blocking) */
  status = mstro_transport_execute(src_cdo, &ticket);
  if(MSTRO_OK != status) {
    ERR("Failure in transport execute for CDO %s\n",
        src_cdo->name);
    return MSTRO_FAIL;
  }
  

  /* do some clean-up if applicable */
  if (init->cp == 0) { // That means mv
    // TODO Detach raw pointer from CDO handle for Maestro resource
    // management while the CDO stays OFFERED by the new location? Or
    // is this part of a resource management protocol that originates
    // from pool manager and utilizes existing infrastructure for
    // movement?
    DEBUG("Unimplemented ticket case cp==0\n");    
  }
  
  /* Send ticket via pmp_send_nowait() */
  Mstro__Pool__MstroMsg msg = MSTRO__POOL__MSTRO_MSG__INIT;
  
  if (! (MSTRO_OK == mstro_pmp_package(&msg, (ProtobufCMessage*)&ticket))) {
    ERR("Failed to package %s into a pool manager message\n",
        ticket.base.descriptor->name);
    return MSTRO_FAIL;
  }
  
  status = mstro_pmp_send_nowait(init->dst_appid->id, &msg);
  if(status!=MSTRO_OK) {
    ERR("Cannot send ticket to %zu: %d (%s)\n",
        init->dst_appid->id, status, mstro_status_description(status));
    return MSTRO_FAIL;
  }
  
  return MSTRO_OK;
}

static inline
mstro_status
mstro_pc__handle_bye(const Mstro__Pool__Bye *bye)
{
  assert(bye!=NULL);
  DEBUG("Bye message on app %zu incoming\n", g_pool_app_id);


  /* reconfigure our app to be non-pool */
  atomic_store(&g_mstro_pm_attached, false);
  g_pool_app_id = MSTRO_APP_ID_INVALID;
  g_pool_appid.id = MSTRO_APP_ID_INVALID;
  g_pool_apptoken.appid = NULL;
  return MSTRO_OK;
}


static inline
mstro_status
mstro_pc__handle_transfer_ticket(Mstro__Pool__TransferTicket* ticket)
{
  if(ticket==NULL || ticket->cdoid==NULL) {
    ERR("Invalid transfer ticket: %p/%p\n", ticket, ticket ? ticket->cdoid : (void*)0xdeadbeef);
    return MSTRO_INVMSG;
  }
  mstro_status status;

  struct mstro_cdo_id cdoid = { .qw[0] = ticket->cdoid->qw0,
                                .qw[1] = ticket->cdoid->qw1 };

  WITH_CDO_ID_STR(idstr, &cdoid,
                  INFO("ticket for CDO gid %s, ticket kind %s\n",
                       idstr,
                       (protobuf_c_enum_descriptor_get_value(
                           &mstro__pool__transport_kind__descriptor,
                           ticket->method))
                       ->name););
  mstro_cdo dst_cdo;
  status = mstro_pool__find_sink_cdo(&cdoid, ticket->attributes,
                                     &dst_cdo);
  if(MSTRO_OK!=status && MSTRO_NOENT!=status) {
    WITH_CDO_ID_STR(idstr, &cdoid,
                    ERR("Cannot transfer CDO gid %s because not in local pool\n",
                        idstr););
    return MSTRO_FAIL;
  }

  if(MSTRO_NOENT==status) {
    WITH_CDO_ID_STR(idstr, &cdoid,
                    WARN("Incoming ticket for CDO %s, but CDO not present in local pool, skipping ticket\n",
                         idstr););
    if(ticket->force_offer) {
      WARN("Ticket has force-offer set, but this is unimplemented\n");
      /* FIXME: create a fresh CDO, use that CDO for incoming transport, then perform OFFER */
    }
  } else {
    DEBUG("Initiating incoming transfer to %zu (me) for CDO %s\n",
          g_pool_app_id, dst_cdo->name);
    
    /* Execute transport (non-blocking) */
    status = mstro_transport_execute(dst_cdo, ticket);
    if(MSTRO_OK != status) {
      ERR("Failure in transport execute for CDO %s\n",
          dst_cdo->name);
      return MSTRO_FAIL;
    }
  }

  /* FIXME: where do we wait for the transport to complete before sending completion? */
  
  /* update state, then tell manager */
  Mstro__Pool__TransferCompleted completion = MSTRO__POOL__TRANSFER_COMPLETED__INIT;
  completion.src_appid = ticket->srcid;
  completion.cdoid = ticket->cdoid;

  Mstro__Pool__MstroMsg msg = MSTRO__POOL__MSTRO_MSG__INIT;
  status = mstro_pmp_package(&msg, (ProtobufCMessage*)&completion);
  if(MSTRO_OK != status) {
    ERR("Failed to package %s into a pool manager message\n",
        completion.base.descriptor->name);
    return MSTRO_FAIL;
  }
  
  status = mstro_pmp_send_nowait(MSTRO_APP_ID_MANAGER, &msg);
  if(status!=MSTRO_OK) {
    ERR("Cannot send completion to pool manager: %d (%s)\n",
        status, mstro_status_description(status));
    return MSTRO_FAIL;
  }

  /* FIXME: query transport layer whether we should signal sender */
  WARN("Not asking transport layer whether sender needs notification\n");
#if 0
  /* FIXME: register sender as peer app */
  status = mstro_pmp_send_nowait(ticket->src_appid->id, &msg);
  if(status!=MSTRO_OK) {
    ERR("Cannot send completion to source %d: %d (%s)\n",
        ticket->src_appid->id, mstro_status_description(status));
    return MSTRO_FAIL;
  }
#endif
  
  
  return MSTRO_OK;
}


mstro_status
mstro_pc__handle_event(const Mstro__Pool__Event *ev)
{
  if(ev==NULL) {
    ERR("Invalid event\n");
    return MSTRO_INVMSG;
  }
  if(ev->subscription_handle==NULL) {
    ERR("Missing subscription handle\n");
    return MSTRO_INVMSG;
  }

  DEBUG("Pool event for sid %" PRIu64 ", kind %d (payload case %d)\n",
        ev->subscription_handle->id, ev->kind, ev->payload_case);

  uint64_t sid = ev->subscription_handle->id;

  return mstro_pool_event_consume(ev);
}



mstro_status
mstro_pc_handle_msg(const struct mstro_msg_envelope *envelope,
                    struct mstro_endpoint *ep)
{
  mstro_status status=MSTRO_OK;
  
  INFO("PC Handling incoming message size %" PRIu64 "\n",
       envelope->payload_size);

  Mstro__Pool__MstroMsg *msg
      = mstro__pool__mstro_msg__unpack(NULL,
                                       envelope->payload_size,
                                       envelope->data);

  if(msg==NULL) {
    ERR("Failed to unpack the message\n");
    status = MSTRO_INVMSG;
    goto BAILOUT;
  } else {
    INFO("... unpacked msg: type %s (%s), vsm %s\n",
         msg->base.descriptor->name,
         (msg->base.descriptor==&mstro__pool__mstro_msg__descriptor
          ? msg->declare->base.descriptor->name : "no subtype"),
         msg->opts && msg->opts->vsm_data ? "yes" : "no");
  }
  
  if(msg->base.descriptor != &mstro__pool__mstro_msg__descriptor) {
    ERR("Cannot handle messages other than MstroMsg messages\n");
    return MSTRO_INVMSG;
  }
  
  if(msg->opts && msg->opts->vsm_data!=NULL) {
    ERR("VSM, Unimplemented\n");
    status=MSTRO_UNIMPL;
    goto BAILOUT;
  }

  if(!msg->token) {
    ERR("Incoming message without token\n");
    status=MSTRO_INVMSG;
    goto BAILOUT;
  }
  if(!msg->token->appid) {
    ERR("Incoming message without App ID\n");
    status=MSTRO_INVMSG;
    goto BAILOUT;
  }
  
  if(msg->token->appid->id==MSTRO_APP_ID_INVALID) {
      ERR("Invalid app id %d\n");
      status=MSTRO_INVMSG;
      goto BAILOUT;
  }
  
  if(msg->token->appid->id!=MSTRO_APP_ID_MANAGER) {
    INFO("Message from peer %d\n", msg->token->appid->id);
  }

  status = MSTRO_OK;
  switch(msg->msg_case) {
    /* 'good' messages: */
    case MSTRO__POOL__MSTRO_MSG__MSG_WELCOME: {
      INFO("WELCOME message received, app id %" PRIu64 "\n",
           msg->welcome->token->appid->id);
      g_pool_app_id = msg->welcome->token->appid->id;
      g_pool_appid.id = g_pool_app_id;
      g_pool_apptoken.appid = &g_pool_appid;
      /* FIXME: possibly add security token at this time */
      
      atomic_store(&g_mstro_pm_attached, true);
      break;
    }

    case MSTRO__POOL__MSTRO_MSG__MSG_DECLARE_ACK:
      status = mstro_pc__handle_declare_ack(msg->declare_ack);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_ACK:
      status = mstro_pc__handle_poolop_ack(msg->ack);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_SUBSCRIBE_ACK:
      status = mstro_pc__handle_subscribe_ack(msg->subscribe_ack);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_BYE:
      status = mstro_pc__handle_bye(msg->bye);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_INITIATE_TRANSFER:
      INFO("PC handling InitiateTransfer msg\n");
      status = mstro_pc__handle_initiate_transfer(msg->initiate_transfer);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_TRANSFER_TICKET:
      INFO("PC handling TransferTicket msg\n");
      status = mstro_pc__handle_transfer_ticket(msg->transfer_ticket);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_EVENT:
      status = mstro_pc__handle_event(msg->event);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_RESOLVE_REPLY:
      status = mstro_pc__handle_resolve_reply(msg->resolve_reply);
      break;


      /* unhandled currently: */
    case MSTRO__POOL__MSTRO_MSG__MSG_TRANSFER_COMPLETED:
    case MSTRO__POOL__MSTRO_MSG__MSG_DEMAND_ATTR_RES:
    case MSTRO__POOL__MSTRO_MSG__MSG_QUERY_RES:
      WARN("Unhandled message kind: %s\n",
           msg->declare->base.descriptor->name);
      return MSTRO_UNIMPL;

      
      /* illegal on pool client: */
    case MSTRO__POOL__MSTRO_MSG__MSG_DECLARE:
    case MSTRO__POOL__MSTRO_MSG__MSG_SEAL:
    case MSTRO__POOL__MSTRO_MSG__MSG_SEAL_GROUP:
    case MSTRO__POOL__MSTRO_MSG__MSG_OFFER:
    case MSTRO__POOL__MSTRO_MSG__MSG_REQUIRE:
    case MSTRO__POOL__MSTRO_MSG__MSG_RETRACT:
    case MSTRO__POOL__MSTRO_MSG__MSG_DEMAND:
    case MSTRO__POOL__MSTRO_MSG__MSG_WITHDRAW:
    case MSTRO__POOL__MSTRO_MSG__MSG_DISPOSE:
    case MSTRO__POOL__MSTRO_MSG__MSG_JOIN: 
    case MSTRO__POOL__MSTRO_MSG__MSG_LEAVE:
    case MSTRO__POOL__MSTRO_MSG__MSG_SUBSCRIBE:
    case MSTRO__POOL__MSTRO_MSG__MSG_UNSUBSCRIBE:
    case MSTRO__POOL__MSTRO_MSG__MSG_LOG:
    case MSTRO__POOL__MSTRO_MSG__MSG_DEMAND_ATTR:
    case MSTRO__POOL__MSTRO_MSG__MSG_QUERY:
    case MSTRO__POOL__MSTRO_MSG__MSG_RESOLVE:
    case MSTRO__POOL__MSTRO_MSG__MSG__NOT_SET: {
    default:
      ERR("Invalid message kind (incoming on PC side): %s\n",
          msg->declare->base.descriptor->name);
      status=MSTRO_INVMSG;
      goto BAILOUT;
    }
  }

BAILOUT:
  if(msg!=NULL) {
    assert(msg->token!=&g_pool_apptoken);
    mstro__pool__mstro_msg__free_unpacked(msg, NULL);
  }
  return status;
}
