/* -*- mode:c -*- */
/** @file
 ** @brief Pool Manager implementation 
 **
 **/
/*
 * Copyright (C) 2019-2020 Cray Computer GmbH
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "i_pool_manager.h"

#include "maestro/i_globals.h"
#include "maestro/logging.h"
#include "i_subscription_registry.h"

#include "i_event.h"

#include <assert.h>
#include <inttypes.h>
#include "maestro/i_hashing.h"


/* simplify logging */
#define DEBUG(...) LOG_DEBUG(MSTRO_LOG_MODULE_PM,__VA_ARGS__)
#define INFO(...)  LOG_INFO(MSTRO_LOG_MODULE_PM,__VA_ARGS__)
#define WARN(...)  LOG_WARN(MSTRO_LOG_MODULE_PM,__VA_ARGS__)
#define ERR(...)   LOG_ERR(MSTRO_LOG_MODULE_PM,__VA_ARGS__)


/** Group registry */

/** A group record, containing the list of members for a group.

    This is created at SEAL time when the group is mentioned by some
    app for the first time.
*/
struct mstro_pm__group_record {
  struct mstro_cdo_id id;          /**< the proxy CDO ID */
  _Atomic(size_t) num_refs;        /**< number of live group handles for this record */
  size_t num_cdos;                 /**< the number of members */
  struct mstro_cdo_id *member_ids; /**< array of ::NUM_CDOS many CDO
                                    * IDs. FIXME: add hash for this */
  char **member_names;             /**< array of ::NUM_CDOS many CDO
                                      IDs.  FIXME: replace by long
                                      array of NUL-terminated strings
                                      and a hash to do look-ups */
};

/** The group registry is a hash table from proxy CDO-ID to group_record */
KHASH_INIT(group_reg, const struct mstro_cdo_id*,
           struct mstro_pm__group_record *, 1,
           kh_cdoidptr_hash_func, kh_cdoidptr_hash_equal)

    /** the global group registry. Protected by @ref g_pm_group_registry_mtx */
    khash_t(group_reg) g_pm_group_registry; /* we rely on kh_init just
                                             * doing the equivalent of
                                             * calloc, so this is
                                             * fine */
/** mutex protecting @ref g_pm_group_registry */
pthread_mutex_t g_pm_group_registry_mtx = PTHREAD_MUTEX_INITIALIZER;

#define WITH_LOCKED_GROUP_REGISTRY(body) \
  do {                                                                  \
    int wlgr_stat=pthread_mutex_lock(&g_pm_group_registry_mtx);         \
    if(wlgr_stat!=0) {                                                  \
      ERR("Failed to lock group registry: %d (%s)\n",                   \
          wlgr_stat, strerror(wlgr_stat));                              \
      abort();                                                          \
    }                                                                   \
                                                                        \
    do {                                                                \
      body;                                                             \
    } while(0);                                                         \
                                                                        \
    wlgr_stat=pthread_mutex_unlock(&g_pm_group_registry_mtx);           \
    if(wlgr_stat!=0) {                                                  \
      ERR("Failed to unlock group registry: %d (%s)\n",                 \
          wlgr_stat, strerror(wlgr_stat));                              \
      abort();                                                          \
    }                                                                   \
  } while(0)


/** fill in the PoolOpAck structure according to OP. */
static inline
mstro_status
mstro_pmp_fill_pool_op_ack(Mstro__Pool__PoolOpAck *poack,
                           Mstro__Pool__PoolOpAck__PoolOp op,
                           Mstro__Pool__UUID *uuid,
                           Mstro__Pool__SubscriptionHandle *h,
                           mstro_status status)
{
  assert(poack!=NULL
         && poack->base.descriptor == &mstro__pool__pool_op_ack__descriptor);
  assert(uuid!=NULL || h!=NULL);

  poack->op = op;
  switch(op) {
    case MSTRO__POOL__POOL_OP_ACK__POOL_OP__SEAL:
    case MSTRO__POOL__POOL_OP_ACK__POOL_OP__OFFER:
    case MSTRO__POOL__POOL_OP_ACK__POOL_OP__REQUIRE:
    case MSTRO__POOL__POOL_OP_ACK__POOL_OP__RETRACT:
    case MSTRO__POOL__POOL_OP_ACK__POOL_OP__WITHDRAW:
      poack->operand_case=MSTRO__POOL__POOL_OP_ACK__OPERAND_CDOID;
      poack->cdoid = uuid;
      break;
      
    case MSTRO__POOL__POOL_OP_ACK__POOL_OP__UNSUBSCRIBE:
      poack->operand_case=MSTRO__POOL__POOL_OP_ACK__OPERAND_HANDLE;
      poack->handle = h;
      break;

    default:
      ERR("unsupported PoolOp\n");
      return MSTRO_FAIL;
  }
  if(status==MSTRO_OK)
    poack->status = MSTRO__POOL__POOL_OP_ACK__POOL_OP_STATUS__OK;
  else
    poack->status = MSTRO__POOL__POOL_OP_ACK__POOL_OP_STATUS__FAIL;

  return MSTRO_OK;
}

static inline
mstro_status
mstro_pm__send_ack(mstro_app_id app_id,
                   Mstro__Pool__PoolOpAck__PoolOp op,
                   Mstro__Pool__UUID *uuid,
                   Mstro__Pool__SubscriptionHandle *h,
                   Mstro__Pool__Attributes *a,
                   mstro_status s)
{
  Mstro__Pool__PoolOpAck poack = MSTRO__POOL__POOL_OP_ACK__INIT;
  mstro_status status = mstro_pmp_fill_pool_op_ack(&poack, op, uuid, h, s);
  if(a!=NULL) {
    DEBUG("Including attribute update in PoolOpAck (likely a GROUP SEAL)\n");
    poack.payload_case = MSTRO__POOL__POOL_OP_ACK__PAYLOAD_UPDATED_ATTRIBUTES;
    poack.updated_attributes = a;
  }

  if(status!=MSTRO_OK) {
    const char *op_kind
        = (protobuf_c_enum_descriptor_get_value(
            &mstro__pool__pool_op_ack__pool_op__descriptor, op))
        ->name;
    ERR("Failed to construct ACK for %s\n", op_kind);
    return status;
  }
  
  Mstro__Pool__MstroMsg msg = MSTRO__POOL__MSTRO_MSG__INIT;
  status = mstro_pmp_package(&msg, (ProtobufCMessage*)&poack);
  if(status!=MSTRO_OK) {
    ERR("Failed to package %s-ACK into a pool manager message\n",
        poack.base.descriptor->name);
    return status;
  }
  /* send it off: this function automatically packs the message and sends it. The   */
  status = mstro_pmp_send_nowait(app_id, &msg);
  if(status!=MSTRO_OK) {
    ERR("Failed to send %s-ACK to %zu\n", app_id);
  }
  
  return status; 
}

/** forward declaration of continuation function context */
struct mstro_pm__continuation_ctx;

/** function that can be called as continuation of a message handler */
typedef mstro_status(*mstro_pm__continuation)(
    mstro_event event, struct mstro_pm__continuation_ctx *cont);

/** structure keeping context for bottom-half handlers of message
 * handling functions that are preempted (waiting) for subscription
 * acknowledgement.
 *
 * The continuation handler is expected to either schedule another
 * continuation, re-using the structure passed to it (like a private
 * stack frame), or free it using mstro_pm__continuation_free() before
 * returning.
 */
struct mstro_pm__continuation_ctx {
  /** the continuation function to call. It will be passed the event
   * that triggered it and the cont structure passed in the
   * mstro_pm__schedule_continuation() call.  */
  mstro_pm__continuation bh_handler;
  /* The data that is needed to express the 'current continuation'
   * context, mainly message-dependent info that the original message
   * handler had in arguments or derived from it, but that have gone
   * out of scope with it, and would have been recycled by the main
   * message dispatcher by now.
   */
  /** The original message that was handled when things were
   * interrupted */
  Mstro__Pool__MstroMsg *msg;

  /* extra information needed for only some specific handlers */
  union {
    struct {
      struct mstro_endpoint *ep;
    } ctx_join;
    struct {
      struct mstro_pm_app_registry_entry *regentry;
    } ctx_welcome;
  };
};

/** Deallocate a continuation structure
 *
 * This automatically deallocated the mstro_pm__continuation.msg, but
 * all other slots will not be touched (caller has to clear them
 * before if desired).
 */

static inline
void
mstro_pm__continuation_free(struct mstro_pm__continuation_ctx *c)
{
  assert(c!=NULL);
  assert(c->msg!=NULL);
  assert(c->msg->token!=&g_pool_apptoken);
  mstro__pool__mstro_msg__free_unpacked(c->msg, NULL);
  free(c);
}
  
/** the event handler dispatch function
 *
 * All subscriptions that require an ACK are dispatched through the i_event.h
 * infrastructure, with this function as the event handler
 */
static
void
mstro_pm__event_bh(mstro_event event,
                   void *closure)
{
  if(event==NULL) {
    ERR("NULL event\n"); abort();
  }
  if(closure==NULL) {
    ERR("NULL closure\n"); abort();
  }
  struct mstro_pm__continuation_ctx *cont
      = (struct mstro_pm__continuation_ctx*)closure;

  DEBUG("continuation %p, handler %p\n", cont, cont->bh_handler);
  mstro_status s = cont->bh_handler(event, cont);
  if(s!=MSTRO_OK) {
    ERR("Executing handler %p failed: %d (%s)\n",
        cont->bh_handler, s, mstro_status_description(s));
  }  
}

/** Event domain for handler continuations */
mstro_event_domain g_mstro_pm_continuations = NULL;

/** Create a new event and schedule a continuation handler for it.
 *
 * Returns the event that can trigger the event.
 */
static inline
mstro_status
mstro_pm__schedule_continuation(
    struct mstro_pm__continuation_ctx *cont,
    mstro_event *result)
{
  assert(g_mstro_pm_continuations!=NULL);
  mstro_event event;
  mstro_status s = mstro_event_create(g_mstro_pm_continuations,
                                      mstro_pm__event_bh,
                                      cont, NULL, true, &event);
  if(s!=MSTRO_OK) {
    ERR("Failed to schedule continuation event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT;
  }

  mstro_event_id id;
  s = mstro_event_id_get(event, &id);
  if(s!=MSTRO_OK) {
    ERR("Failed to obtain event id\n");
    goto BAILOUT;
  }
  
  DEBUG("Scheduled continuation, event %p, id %" PRIu64 "\n",
        event, id);
  *result = event;

BAILOUT:
  return s;  
}

/** create a new continuation object
 *
 * content beyond handler and msg must be filled in by caller as
 * needed.
 */
static inline
struct mstro_pm__continuation_ctx *
mstro_pm__continuation_create(
    mstro_pm__continuation bh_handler,
    Mstro__Pool__MstroMsg *msg)
{
  assert(bh_handler!=NULL);
  assert(msg!=NULL);
  struct mstro_pm__continuation_ctx *res
      = malloc(sizeof(struct mstro_pm__continuation_ctx));
  if(res) {
    res->bh_handler=bh_handler;
    res->msg = msg;
  }

  return res;
}
    

/* send event notification to subscribers.
 *
 * If the subscription requires an acknowledgement, schedule the
 * continuation, otherwise execute it immediately.
 *
 * If @arg ctx is non-null it will be reused as context, otherwise a
 * new one will be allocated. If reuse occurs, the other arguments
 * will still be set in the context (caller does not need to do it).
 *
 * The @arg pool_even_msg will not be referenced after return from
 * this function, so it's safe to allocate it on the stack of the
 * caller and have components in it that may be short-lived.
 */
static inline
mstro_status
mstro_pm__event_notify_and_continue(Mstro__Pool__Event *pool_event_msg,
                                    bool beforep,
                                    struct mstro_pm__continuation_ctx *ctx,
                                    mstro_pm__continuation cont_fun,
                                    Mstro__Pool__MstroMsg *msg,
                                    struct mstro_endpoint *ep,
                                    struct mstro_pm_app_registry_entry *regentry
                                    )
{
  if(pool_event_msg==NULL) {
    ERR("NULL pool event data\n");
    return MSTRO_INVARG;
  }
  if(cont_fun==NULL) {
    ERR("No continuation function given\n");
    return MSTRO_INVARG;
  } else {
    DEBUG("Cont-fun is %p\n", cont_fun);
  }
    
  if(msg==NULL) {
    ERR("No message available\n");
    return MSTRO_INVARG;
  }
  
  /* ep==NULL is permitted; only JOIN sets it */

  if(ctx==NULL) {
    ctx = mstro_pm__continuation_create(cont_fun, msg);
    if(ctx==NULL) {
      ERR("Failed to allocate continuation context\n");
      return MSTRO_NOMEM;
    }
  } else {
    DEBUG("Reusing caller-provided context %p\n", ctx);
    ctx->bh_handler = cont_fun;
    ctx->msg = msg;
  }
  /* special cases */
  if(ep) {
    assert(regentry==NULL);
    ctx->ctx_join.ep = ep;
  }
  if(regentry) {
    assert(ep==NULL);
    ctx->ctx_welcome.regentry = regentry;
  }

  mstro_event cont_event=NULL;
  mstro_status s = mstro_pm__schedule_continuation(ctx, &cont_event);
  if(s!=MSTRO_OK) {
    ERR("Failed to schedule event handler: %d (%s)\n",
        s, mstro_status_description(s));
    return s;
  }

  /* Send subscription info. We pass our continuation event so that
   *  the subscription engine can trigger our continuation when all
   *  the event acks have arrived. If none are needed the event will
   *  be triggered immediately.
   */
  DEBUG("Advertising event kind %d (%s:%s)\n",
        pool_event_msg->kind,
        protobuf_c_enum_descriptor_get_value(&mstro__pool__event_kind__descriptor,
                                             pool_event_msg->kind)->name,
        beforep ? "before" : "after");
  s = mstro_pool_event_advertise(pool_event_msg, beforep, cont_event);
                                                                 
  if(s!=MSTRO_OK) {
    ERR("Failed to advertise event: %d (%s)\n",
        s, mstro_status_description(s));
    return s;
  }
  return MSTRO_OK;
}


static inline
void
mstro_pm__msg_free(Mstro__Pool__MstroMsg *msg)
{
  if(msg!=NULL) {
    assert(msg->token!=&g_pool_apptoken);
    mstro__pool__mstro_msg__free_unpacked(msg, NULL);
  }
}


/** if the application receiving EV needs to be told the name, send
 * it, otherwise set it to NULL in the EV structure */
static inline
mstro_status
mstro_pm__possibly_fill_event_cdoname(
    const struct mstro_cdo_id *cdoid,
    mstro_app_id recipient,
    Mstro__Pool__Event *ev)
{
  recipient = recipient; /* avoid unused warning */
  
  /** FIXME: for the moment we just always send the name, could be improved */
  mstro_status s=
      mstro_pm_cdo_registry_cdo_name_lookup(cdoid,
                                            (const char **)&ev->cdo_name);
  if(s!=MSTRO_OK) {
    WITH_CDO_ID_STR(idstr, cdoid, {
        ERR("Cannot find CDO registry entry for |%s|\n",
            idstr);});
    return MSTRO_NOENT;
  }
  return MSTRO_OK;  
}


/*** DECLARE ***/
static inline
mstro_status
mstro_pm__handle_declare_phase3(mstro_event event,
                                struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("DECLARE phase 3/3, event %p\n", event);
  mstro_pm__msg_free(cont->msg);
  return MSTRO_OK;
}

static inline
mstro_status
mstro_pm__handle_declare_phase2(mstro_event event,
                                struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("DECLARE phase 2/3, event %p\n", event);

  mstro_app_id app_id = cont->msg->token->appid->id;
  struct mstro_cdo_id cdoid;
  mstro_status status
      = mstro_pm_cdo_registry_declare(app_id,
                                      cont->msg->declare->cdo_name,
                                      &cdoid);
  if(status!=MSTRO_OK) {
    ERR("Failed to add CDO declaration to registry\n");
    goto BAILOUT_FREE;
  }
  
  /* build DECLARE_ACK */
  Mstro__Pool__UUID uuid = MSTRO__POOL__UUID__INIT;
  uuid.qw0 = cdoid.qw[0];
  uuid.qw1 = cdoid.qw[1];
  
  Mstro__Pool__DeclareAck declare_ack = MSTRO__POOL__DECLARE_ACK__INIT;
  declare_ack.serial = cont->msg->declare->serial;
  declare_ack.cdoid = &uuid;
  declare_ack.channel = app_id; /* FIXME: sufficient for now */

  Mstro__Pool__MstroMsg msg_ack = MSTRO__POOL__MSTRO_MSG__INIT;
  status = mstro_pmp_package(&msg_ack, (ProtobufCMessage*)&declare_ack);
  if(status!=MSTRO_OK) {
    ERR("Failed to package %s into a pool manager message\n",
        declare_ack.base.descriptor->name);
    goto BAILOUT_FREE;
  }
  status = mstro_pmp_send_nowait(app_id, &msg_ack);
  if(status!=MSTRO_OK) {
    ERR("Failed to send DECLARE_ACK: %d (%s)\n",
        status, mstro_status_description(status));
  }
  
  /* produce declare:after event */
  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__DECLARE;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_DECLARE;
  ev.declare = cont->msg->declare;
  /* it's safe to refer to the Appid object since the Event object is
   * on stack and will only be alive until notify-and-continue is
   * done */
  ev.origin_id = cont->msg->token->appid;
  ev.cdo_name = cont->msg->declare->cdo_name;

  status = mstro_pm__event_notify_and_continue(
      &ev,
      false,
      NULL,
      mstro_pm__handle_declare_phase3,
      cont->msg, NULL, NULL);
  if(status!=MSTRO_OK) {
    ERR("Failed notifying subscribers of DECLARE:after event: %d (%s)\n",
        status, mstro_status_description(status));
    goto BAILOUT_FREE;
  }
  goto DONE;

BAILOUT_FREE:
  mstro_pm__msg_free(cont->msg);

DONE:
  return status;
}

static inline
mstro_status
mstro_pm__handle_declare(Mstro__Pool__MstroMsg *msg)
{
  Mstro__Pool__Declare *declare = msg->declare;
  mstro_app_id app_id = msg->token->appid->id;
  assert(declare!=NULL); assert(app_id!=MSTRO_APP_ID_INVALID);
  
  DEBUG("CDO declaration from %zu for %s (serial %zu)\n",
        app_id, declare->cdo_name, declare->serial);

  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__DECLARE;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_DECLARE;
  ev.declare = msg->declare;
  ev.origin_id = msg->token->appid;
  ev.cdo_name = msg->declare->cdo_name;

  mstro_status status
      = mstro_pm__event_notify_and_continue(
          &ev,
          true,
          NULL,
          mstro_pm__handle_declare_phase2,
          msg, NULL, NULL);
  if(status!=MSTRO_OK) {
    ERR("Failed notifying subscribers of DECLARE:before event: %d (%s)\n",
        status, mstro_status_description(status));
    goto BAILOUT;
  }
BAILOUT:
  return status;
}


/*** GROUP SEAL ***/

/* Create a new group record for CDOID with MEMBERS on APP_ID.
   Undeclared CDO members are declared on the PM CDO registry and
   moved from the undeclared to the declared list of MEMBERS
   destructively.
*/
static inline
mstro_status
mstro_pm__group_record_create(const struct mstro_cdo_id *cdoid,
                              mstro_app_id app_id,
                              Mstro__Pool__GroupMembers *members,
                              struct mstro_pm__group_record **result)
{
  if(cdoid==NULL || members==NULL)
    return MSTRO_INVARG;
  if(result==NULL)
    return MSTRO_NOMEM;
  mstro_status s=MSTRO_UNIMPL;
  struct mstro_pm__group_record *entry =NULL;

  if(members->sel!=NULL) {
    ERR("FIXME: No support for CDO selectors in groups\n");
    s=MSTRO_UNIMPL;
    goto BAILOUT;
  }
  
  /* build group-registry entry */
  entry = calloc(1,sizeof(struct mstro_pm__group_record));
  if(!entry) {
    s=MSTRO_NOMEM; goto BAILOUT;
  }

  memcpy(&entry->id, cdoid, sizeof(struct mstro_cdo_id));
  entry->num_refs = 1;
  entry->num_cdos = members->n_declared_members + members->n_undeclared_members;

  entry->member_ids   = malloc(sizeof(struct mstro_cdo_id) * entry->num_cdos);
  entry->member_names = calloc(entry->num_cdos, sizeof(char*));
  if(entry->member_ids==NULL || entry->member_names==NULL) {
    s=MSTRO_NOMEM;
    goto BAILOUT;
  }

  for(size_t i=0; i<members->n_declared_members; i++) {
    struct mstro_cdo_id c = { .qw[0] = members->declared_members[i]->qw0,
                              .qw[1] = members->declared_members[i]->qw1 };
    memcpy(&entry->member_ids[i], &c, sizeof(struct mstro_cdo_id));
    /* lookup and store name also */
  }

  /* add handles to per-group CDO handle table */
  s=MSTRO_OK;
  size_t offset = members->n_declared_members;
  for(size_t i=0; i<members->n_undeclared_members; i++) {
    char *name = members->undeclared_members[i];
    struct mstro_cdo_id c;
    s = mstro_pm_cdo_registry_declare(app_id, name, &c);
    if(s!=MSTRO_OK) {
      ERR("Failed to batch-declare CDO |%s|: %d (%s)\n", name, s, mstro_status_description(s));
      break;
    }
    WITH_CDO_ID_STR(cidstr, &c, {
        WITH_CDO_ID_STR(gidstr, cdoid, {
            DEBUG("Batch-declared CDO |%s| (%s) for group %s from app %zu\n",
                  name, cidstr, gidstr, app_id);});};);
    
    memcpy(&entry->member_ids[i+offset], &c, sizeof(struct mstro_cdo_id));
    entry->member_names[i+offset]=strdup(members->undeclared_members[i]);
    if(entry->member_names[i+offset]==NULL) {
      s=MSTRO_NOMEM;
      goto BAILOUT;
    }
  }
  /* now that all CDOs are declared we need to fix up the members
   * structure so that DEMANDind clients will not see undeclared CDOs
   * in the member attribute */
  /* NOTE: This relies on the protobuf allocator being system malloc/realloc/free */
  Mstro__Pool__UUID **tmp = realloc(members->declared_members,
                                    sizeof(Mstro__Pool__UUID*)*entry->num_cdos);
  if(tmp==NULL) {
    ERR("Failed to reallocate group member attribute\n");
    s=MSTRO_NOMEM;
  } else {
    members->declared_members = tmp;
    for(size_t i=0; i<members->n_undeclared_members; i++) {
      members->declared_members[i+offset] = malloc(sizeof(Mstro__Pool__UUID));
      if(members->declared_members[i+offset]==NULL) {
        ERR("Failed to allocate member UUID slot\n");
        /* FIXME: leaking here */
        s=MSTRO_NOMEM;
        goto BAILOUT;
        break;
      }
      mstro__pool__uuid__init(members->declared_members[i+offset]);
      members->declared_members[i+offset]->qw0 = entry->member_ids[i+offset].qw[0]; 
      members->declared_members[i+offset]->qw1 = entry->member_ids[i+offset].qw[1];
      free(members->undeclared_members[i]);
    }
    free(members->undeclared_members);
    members->undeclared_members=NULL;
    members->n_undeclared_members = 0;
    members->n_declared_members = entry->num_cdos;
  }                                 
  
BAILOUT:
  if(s!=MSTRO_OK) {
    if(entry) {
      if(entry->member_ids)
        free(entry->member_ids);
      if(entry->member_names) {
        for(size_t i=0; i<entry->num_cdos; i++) {
          if(entry->member_names[i]!=NULL) {
            free(entry->member_names[i]);
          } else {
            break;
          }
        }
        free(entry->member_names);
      }
    }
    free(entry);
  } else {
    *result = entry;
  }
  return s;
}

static inline
mstro_status
mstro_pm__group_registry_entry_dispose(struct mstro_pm__group_record *r)
{
  if(r==NULL)
    return MSTRO_INVARG;
  if(--(r->num_refs)>0) {
    /* no locking needed: num_refs is atomic, and when we reach 0 no
     * one else will step on our feet, or see a 0 */
    free(r->member_ids);
    for(size_t i=0; i<r->num_cdos; i++) {
      free(r->member_names[i]);
    }
    free(r->member_names);
    free(r);
  }
  return MSTRO_OK;
}

static inline
mstro_status
mstro_pm__group_registry_entry_check_consistency(struct mstro_pm__group_record *record,
                                                 const Mstro__Pool__GroupMembers *members)
{
  if(record==NULL)
    return MSTRO_INVARG;
  if(members==NULL)
    return MSTRO_OK;
  if(members->sel!=NULL) {
    ERR("FIXME: unimplemented case of group with selector and an exisiting record\n");
    return MSTRO_UNIMPL;
  }
  for(size_t i=0;i<members->n_declared_members;i++) {
    struct mstro_cdo_id c = { .qw[0] = members->declared_members[i]->qw0,
                              .qw[1] = members->declared_members[i]->qw1 };
    bool found = false;
    /* FIXME: avoid linear search here */
    for(size_t j=0; j<record->num_cdos; j++) {
      if(mstro_cdo_id__equal(&record->member_ids[j], &c)) {
        found = true;
        break;
      }
    }
    if(!found) {
      WITH_CDO_ID_STR(idstr, &c, {
          ERR("Group member %s not present in existing group record\n", idstr);});
      return MSTRO_NOMATCH;
    }
  }
  
  for(size_t i=0;i<members->n_undeclared_members;i++) {
    const char *c = members->undeclared_members[i];
    bool found = false;
    /* FIXME: avoid linear search here */
    for(size_t j=0; j<record->num_cdos; j++) {
      if(strcmp(record->member_names[j],c)==0) {
        found = true;
        break;
      }
    }
    if(!found) {
      ERR("Group member %s not present in existing group record\n", c);
    }
  }
  
  return MSTRO_OK;
}



static inline
mstro_status 
mstro_pm__group_seal(const struct mstro_cdo_id *cdoid,
                     mstro_app_id app_id,
                     Mstro__Pool__Attributes *attributes)
{
  mstro_status s = MSTRO_UNIMPL;
  size_t i;
  const Mstro__Pool__KvEntry *e=NULL;

  if(cdoid==NULL || attributes==NULL)
    return MSTRO_INVARG;
  if(attributes->val_case!=MSTRO__POOL__ATTRIBUTES__VAL_KV_MAP)
    return MSTRO_UNIMPL;
  
  for(i=0; i<attributes->kv_map->n_map; i++) {
    e = attributes->kv_map->map[i];
    if(strcmp(e->key,MSTRO_ATTR_CORE_CDO_GROUP_MEMBERS)==0) {
      break;
    }
  }
  if(i==attributes->kv_map->n_map) {
    ERR("Group has no members attribute\n");
    return MSTRO_INVARG;
  }

  assert(e->val->val_case == MSTRO__POOL__AVAL__VAL_BYTES);
  
  Mstro__Pool__GroupMembers *members 
      = mstro__pool__group_members__unpack(NULL, e->val->bytes.len, e->val->bytes.data);
  DEBUG("Group has %zu undeclared and %zu declared members. Selector is %p\n",
        members->n_undeclared_members, members->n_declared_members, members->sel);
  if(members->sel!=NULL) {
    ERR("GROUPS with CDO selectors unimplemented.\n");
    s= MSTRO_UNIMPL;
    goto BAILOUT;
  }


  /* register in group table */
  WITH_LOCKED_GROUP_REGISTRY({
      int ret;
      khiter_t slot = kh_put(group_reg, &g_pm_group_registry, cdoid, &ret);
      switch(ret) {
        case -1:
          ERR("Failed to register group\n");
          s=MSTRO_FAIL;
          break;
        case 0: {
          /* group already present, fetch handle and verify */
          WITH_CDO_ID_STR(idstr, cdoid, {
              DEBUG("Group %s record key already present, checking consistency\n", idstr);});
          slot = kh_get(group_reg, &g_pm_group_registry, cdoid);
          struct mstro_pm__group_record *rec = kh_val(&g_pm_group_registry, slot);
          s=mstro_pm__group_registry_entry_check_consistency(rec, members);
          break;
        }
        default: {
          /* inserted key, also insert entry */
          struct mstro_pm__group_record *entry = NULL;
          s = mstro_pm__group_record_create(cdoid, app_id, members, &entry);
          if(s!=MSTRO_OK) {
            ERR("Failed to construct new group registry entry\n");
          } else {          
            kh_val(&g_pm_group_registry,slot) = entry;
            WITH_CDO_ID_STR(idstr, cdoid, {
                DEBUG("Group %s record created, %zu members\n", idstr, entry->num_cdos);});
            s=MSTRO_OK;
          }
          break;
        }
      }
    });

BAILOUT:
  if(members!=NULL) {
    /* need to re-serialize and replace member attribute on CDO, then deallocate it */
    size_t size = mstro__pool__group_members__get_packed_size(members);
    uint8_t *buf = malloc(size*sizeof(uint8_t));
    if(!buf) {
      ERR("Failed to allocate replacement group_members buffer\n");
      s=MSTRO_NOMEM;
    } else {
      size_t written_size = mstro__pool__group_members__pack(members, buf);
      assert(written_size == size);

      /* Stuff into existing protobuf BYTES entry where we found the
       * serialized form of the original members data */
      e->val->bytes.len = size;
      free(e->val->bytes.data);
      e->val->bytes.data = buf;
      WITH_CDO_ID_STR(idstr, cdoid, {
          DEBUG("Updated group %s member attribute: Now %zu + %zu (decl/undecl) members: %zu bytes\n",
                idstr, members->n_declared_members, members->n_undeclared_members, written_size);});
    }
    mstro__pool__group_members__free_unpacked(members, NULL);
  }
  
  return s;
}

/*** SEAL ***/

static inline
mstro_status
mstro_pm__handle_seal_phase3(mstro_event event,
                             struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("SEAL phase 3/3, event %p\n", event);
  mstro_pm__msg_free(cont->msg);
  return MSTRO_OK;
}

static inline
mstro_status
mstro_pm__handle_seal_phase2(mstro_event event,
                             struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("SEAL phase 2/3, event %p\n", event);

  mstro_app_id app_id = cont->msg->token->appid->id;

  Mstro__Pool__Seal *seal = cont->msg->seal;

  struct mstro_cdo_id cdoid = { .qw[0] = seal->cdoid->qw0,
                                .qw[1] = seal->cdoid->qw1 };

  mstro_status status = MSTRO_OK;

  bool send_attribute_update = false;

  if(seal->attributes!=NULL) {
    DEBUG("Seal attributes: default-namespace |%s|, kind %s\n",
          seal->attributes->default_namespace == NULL
          ? seal->attributes->default_namespace : "(empty)",
          seal->attributes->val_case==MSTRO__POOL__ATTRIBUTES__VAL_KV_MAP
          ? "kv-map" : (
              seal->attributes->val_case
              == MSTRO__POOL__ATTRIBUTES__VAL_YAML_STRING
              ? "yaml" : "(other)"));
    if(seal->attributes->val_case==MSTRO__POOL__ATTRIBUTES__VAL_KV_MAP) {
      DEBUG("Seal attributes: %zu entries\n",
            seal->attributes->kv_map->n_map);
    }

    /* if this is a GROUP CDO perform extra work: Perform PM-side
     * group registration, declare undeclared CDOs in one go, etc. */
    if(seal->attributes->val_case!=MSTRO__POOL__ATTRIBUTES__VAL_KV_MAP) {
      WARN("If this SEAL is for a group CDO it will fail (FIXME)\n");
    } else if(seal->attributes->kv_map!=NULL) {
      size_t i;
      for(i=0; i<seal->attributes->kv_map->n_map; i++) {
        const Mstro__Pool__KvEntry *e = seal->attributes->kv_map->map[i];
        if(strcmp(e->key,MSTRO_ATTR_CORE_CDO_ISGROUP)==0) {
          DEBUG("Recognized as GROUP-CDO\n");
          status = mstro_pm__group_seal(&cdoid, app_id, seal->attributes);
          if(status!=MSTRO_OK) {
            WITH_CDO_ID_STR(idstr, &cdoid,{
                ERR("Failed to perform GROUP SEAL operation on %s: %d (%s)\n",
                    idstr, status, mstro_status_description(status));});
            goto BAILOUT_FREE;
          } else {
            send_attribute_update = true;
            break; /* no need to look further */
          }
        }
      }
    }

    /* FIXME: would be nice to be able to print out the actual attributes added here */
    status = mstro_pm_cdo_registry_store_attributes(&cdoid, app_id,
                                                    seal->attributes);
    if(status==MSTRO_OK) {
      WITH_CDO_ID_STR(idstr, &cdoid,{
          DEBUG("Set attributes on %s for app %d\n", idstr, app_id);
        });
    } else {
      WITH_CDO_ID_STR(idstr, &cdoid,{
          ERR("Failed to set attributes on %s for app %d\n", idstr, app_id);
        });
      goto BAILOUT_FREE;
    }
  }
  
  status = mstro_pm__send_ack(app_id,
                              MSTRO__POOL__POOL_OP_ACK__POOL_OP__SEAL,
                              seal->cdoid, NULL,
                              send_attribute_update ? seal->attributes : NULL,
                              status);
  if(status!=MSTRO_OK) {
    ERR("Failed to send ACK for SEAL: %d (%s)\n",
        status, mstro_status_description(status));
    goto BAILOUT_FREE;
  }
  /* we consumed the attributes allocation (in the registry), so must
   * ensure freeing the message will not kill it. As long as we can
   * rely on protobuf unpacking into separate allocation we can simply
   * NULL it and things will work. */
  seal->attributes = NULL;

  /* produce seal:after event */
  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__SEAL;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_SEAL;
  ev.seal = cont->msg->seal;
  ev.origin_id = cont->msg->token->appid;
  mstro_status s = mstro_pm__possibly_fill_event_cdoname(&cdoid, app_id, &ev);
  if(s!=MSTRO_OK) {
    goto BAILOUT_FREE;
  }

  status = mstro_pm__event_notify_and_continue(
      &ev,
      false,
      NULL,
      mstro_pm__handle_seal_phase3,
      cont->msg, NULL, NULL);
  if(status!=MSTRO_OK) {
    ERR("Failed notifying subscribers of SEAL:after event: %d (%s)\n",
        status, mstro_status_description(status));
    goto BAILOUT_FREE;
  }
  goto DONE;
  
BAILOUT_FREE:
  mstro_pm__msg_free(cont->msg);
DONE:
  return status;
}

static inline
mstro_status
mstro_pm__handle_seal(Mstro__Pool__MstroMsg *msg)
{
  Mstro__Pool__Seal *seal = msg->seal;
  mstro_app_id app_id = msg->token->appid->id;
  assert(seal!=NULL); assert(app_id!=MSTRO_APP_ID_INVALID);
  
  DEBUG("CDO SEAL from %zu\n", app_id);
  mstro_status s;
  
  if(seal->cdoid==NULL) {
    s=MSTRO_INVMSG;
    goto BAILOUT;
  }

  struct mstro_cdo_id cdoid;
  cdoid.qw[0] = seal->cdoid->qw0;
  cdoid.qw[1] = seal->cdoid->qw1;

  /* sanity check that it's been declared already */
  s= mstro_pm_cdo_registry_find(app_id, &cdoid);
  if(s!=MSTRO_OK) {
    ERR("CDO id not known here for app\n");
    s=MSTRO_INVMSG;
    goto BAILOUT;
  }

  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__SEAL;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_SEAL;
  ev.seal = msg->seal;
  ev.origin_id = msg->token->appid;
  s = mstro_pm__possibly_fill_event_cdoname(&cdoid, app_id, &ev);
  if(s!=MSTRO_OK) {
    goto BAILOUT_FREE;
  }

  s = mstro_pm__event_notify_and_continue(
      &ev,
      true,
      NULL,
      mstro_pm__handle_seal_phase2,
      msg, NULL, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed notifying subscribers of SEAL:before event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT_FREE;
  }
  
BAILOUT:
  return s;
  
BAILOUT_FREE:
  mstro_pm__msg_free(msg);
  return s;
}



/*** OFFER ***/

static inline
mstro_status
mstro_pm__handle_offer_phase3(mstro_event event,
                              struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("OFFER phase 3/3, event %p\n", event);
  mstro_pm__msg_free(cont->msg);
  return MSTRO_OK;
}

static inline
mstro_status
mstro_pm__handle_offer_phase2(mstro_event event,
                              struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("OFFER phase 2/3, event %p\n", event);

  mstro_app_id app_id = cont->msg->token->appid->id;

  struct mstro_cdo_id cdoid = { .qw[0] = cont->msg->offer->cdoid->qw0,
                                .qw[1] = cont->msg->offer->cdoid->qw1 };

  mstro_status s=mstro_pm_cdo_registry_update_state(&cdoid,
                                                    app_id,
                                                    MSTRO_CDO_STATE_OFFERED);
  if(s!=MSTRO_OK) {
    ERR("CDO id not known here for app\n");
  }

  s = mstro_pm__send_ack(app_id,
                         MSTRO__POOL__POOL_OP_ACK__POOL_OP__OFFER,
                         cont->msg->offer->cdoid, NULL, NULL, s);
  if(s!=MSTRO_OK) {
    ERR("Failed to send ACK for OFFER: %d (%s)\n", s, mstro_status_description(s));
  }

  /* produce offer:after event */
  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__OFFER;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_OFFER;
  ev.offer = cont->msg->offer;
  ev.origin_id = cont->msg->token->appid;
  s = mstro_pm__possibly_fill_event_cdoname(&cdoid, app_id, &ev);
  if(s!=MSTRO_OK) {
    goto BAILOUT_FREE;
  }

  s = mstro_pm__event_notify_and_continue(
      &ev,
      false,
      NULL,
      mstro_pm__handle_offer_phase3,
      cont->msg, NULL, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed notifying subscribers of OFFER:after event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT_FREE;
  }
  goto DONE;

BAILOUT_FREE:
  mstro_pm__msg_free(cont->msg);
DONE:
  return s;
}

static inline
mstro_status
mstro_pm__handle_offer(Mstro__Pool__MstroMsg *msg)
{
  Mstro__Pool__Offer *offer = msg->offer;
  mstro_app_id app_id = msg->token->appid->id;
  assert(offer!=NULL); assert(app_id!=MSTRO_APP_ID_INVALID);
  struct mstro_cdo_id cdoid = { .qw[0] = offer->cdoid->qw0,
                                .qw[1] = offer->cdoid->qw1 };
  
  DEBUG("CDO OFFER from %zu\n", app_id);
  mstro_status s=MSTRO_FAIL;
  
  if(offer->cdoid==NULL) {
    s=MSTRO_INVMSG;
    goto BAILOUT_FREE;
  }

  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__OFFER;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_OFFER;
  ev.offer = msg->offer;
  ev.origin_id = msg->token->appid;
  s = mstro_pm__possibly_fill_event_cdoname(&cdoid, app_id, &ev);
  if(s!=MSTRO_OK) {
    goto BAILOUT_FREE;
  }

  s = mstro_pm__event_notify_and_continue(
      &ev,
      true,
      NULL,
      mstro_pm__handle_offer_phase2,
      msg, NULL, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed notifying subscribers of OFFER:before event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT_FREE;
  }
BAILOUT:
  return s;
  
BAILOUT_FREE:
  mstro_pm__msg_free(msg);
  return s;
  
}


/*** REQUIRE ***/

static inline
mstro_status
mstro_pm__handle_require_phase3(mstro_event event,
                                struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("REQUIRE phase 3/3, event %p\n", event);
  mstro_pm__msg_free(cont->msg);
  return MSTRO_OK;
}

static inline
mstro_status
mstro_pm__handle_require_phase2(mstro_event event,
                                struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("REQUIRE phase 2/3, event %p\n", event);

  struct mstro_cdo_id cdoid = { .qw[0] = cont->msg->require->cdoid->qw0,
                                .qw[1] = cont->msg->require->cdoid->qw1 };
  mstro_app_id app_id = cont->msg->token->appid->id;

  mstro_status s=mstro_pm_cdo_registry_update_state(&cdoid,
                                                    app_id,
                                                    MSTRO_CDO_STATE_REQUIRED);

  if(s!=MSTRO_OK) {
    ERR("CDO id not known here for app\n");
    goto BAILOUT_FREE;
  }
  
  WARN("Not scheduling any action at REQUIRE time\n");

  s = mstro_pm__send_ack(app_id,
                         MSTRO__POOL__POOL_OP_ACK__POOL_OP__REQUIRE,
                         cont->msg->require->cdoid, NULL, NULL, s);
  if(s!=MSTRO_OK) {
    ERR("Failed to send REQUIRE-ACK: %d (%s)\n",
        s, mstro_status_description(s));
  }

  /* produce require:after event */
  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__REQUIRE;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_REQUIRE;
  ev.require = cont->msg->require;
  ev.origin_id = cont->msg->token->appid;
  s = mstro_pm__possibly_fill_event_cdoname(&cdoid, app_id, &ev);
  if(s!=MSTRO_OK) {
    goto BAILOUT_FREE;
  }
 
  s = mstro_pm__event_notify_and_continue(
      &ev,
      false,
      NULL,
      mstro_pm__handle_require_phase3,
      cont->msg, NULL, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed notifying subscribers of REQUIRE:after event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT_FREE;
  }
  goto DONE;

BAILOUT_FREE:
  mstro_pm__msg_free(cont->msg);
  
DONE:
  return s;
}

static inline
mstro_status
mstro_pm__handle_require(Mstro__Pool__MstroMsg *msg)
{
  Mstro__Pool__Require *require = msg->require;
  mstro_app_id app_id = msg->token->appid->id;
  assert(require!=NULL); assert(app_id!=MSTRO_APP_ID_INVALID);
  struct mstro_cdo_id cdoid = { .qw[0] = require->cdoid->qw0,
                                .qw[1] = require->cdoid->qw1 };

  DEBUG("CDO REQUIRE from %zu\n", app_id);
  mstro_status s;
  
  if(require->cdoid==NULL) {
    s=MSTRO_INVMSG;
    goto BAILOUT;
  }

  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__REQUIRE;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_REQUIRE;
  ev.require = msg->require;
  ev.origin_id = msg->token->appid;
  s = mstro_pm__possibly_fill_event_cdoname(&cdoid, app_id, &ev);
  if(s!=MSTRO_OK) {
    goto BAILOUT_FREE;
  }

  s = mstro_pm__event_notify_and_continue(
      &ev,
      true,
      NULL,
      mstro_pm__handle_require_phase2,
      msg, NULL, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed notifying subscribers of REQUIRE:before event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT_FREE;
  }
BAILOUT:
  return s;

BAILOUT_FREE:
  mstro_pm__msg_free(msg);
  return s;
}


/*** RETRACT ***/
static inline
mstro_status
mstro_pm__handle_retract_phase3(mstro_event event,
                                struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("RETRACT phase 3/3, event %p\n", event);
  mstro_pm__msg_free(cont->msg);
  return MSTRO_UNIMPL; /* FIXME: until retract actually works */
}

static inline
mstro_status
mstro_pm__handle_retract_phase2(mstro_event event,
                                struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("RETRACT phase 2/3, event %p\n", event);

  struct mstro_cdo_id cdoid = { .qw[0] = cont->msg->retract->cdoid->qw0,
                                .qw[1] = cont->msg->retract->cdoid->qw1 };
  mstro_app_id app_id = cont->msg->token->appid->id;

  ERR("RETRACT unimplemented\n"); /* FIXME; phase_3 sends the UNIMPL */

  /* produce require:after event */
  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__RETRACT;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_RETRACT;
  ev.retract = cont->msg->retract;
  ev.origin_id = cont->msg->token->appid;
  mstro_status s = mstro_pm__possibly_fill_event_cdoname(&cdoid, app_id, &ev);
  if(s!=MSTRO_OK) {
    goto BAILOUT_FREE;
  }

  s = mstro_pm__event_notify_and_continue(
      &ev,
      false,
      NULL,
      mstro_pm__handle_retract_phase3,
      cont->msg, NULL, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed notifying subscribers of RETRACT:after event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT_FREE;
  }
  goto DONE;

BAILOUT_FREE:
  mstro_pm__msg_free(cont->msg);
  
DONE:
  return s;
}

static inline
mstro_status
mstro_pm__handle_retract(Mstro__Pool__MstroMsg *msg)
{
  Mstro__Pool__Retract *retract = msg->retract;
  mstro_app_id app_id = msg->token->appid->id;
  assert(retract!=NULL);

  app_id=app_id; /* avoid unused arg warning */
  struct mstro_cdo_id cdoid = { .qw[0] = retract->cdoid->qw0,
                                .qw[1] = retract->cdoid->qw1 };
  


  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__RETRACT;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_RETRACT;
  ev.retract = msg->retract;
  ev.origin_id = msg->token->appid;
  mstro_status s =
      mstro_pm__possibly_fill_event_cdoname(&cdoid, app_id, &ev);
  if(s!=MSTRO_OK) {
    goto BAILOUT_FREE;
  }

  s = mstro_pm__event_notify_and_continue(
      &ev,
      true,
      NULL,
      mstro_pm__handle_retract_phase2,
      msg, NULL, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed notifying subscribers of RETRACT:before event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT_FREE;
  }
BAILOUT:
  return s;
  
BAILOUT_FREE:
  mstro_pm__msg_free(msg);
  return s;
}



/*** DEMAND ***/
static inline
mstro_status
mstro_pm__handle_demand_phase3(mstro_event event,
                               struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("DEMAND phase 3/3, event %p\n", event);
  mstro_pm__msg_free(cont->msg);
  return MSTRO_OK;
}

static inline
mstro_status
mstro_pm__handle_demand_phase2(mstro_event event,
                               struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("DEMAND phase 2/3, event %p\n", event);

  struct mstro_cdo_id cdoid = { .qw[0] = cont->msg->demand->cdoid->qw0,
                                .qw[1] = cont->msg->demand->cdoid->qw1 };
  mstro_app_id app_id = cont->msg->token->appid->id;

  mstro_status s=mstro_pm_cdo_registry_update_state(
      &cdoid,
      app_id,
      MSTRO_CDO_STATE_REQUIRED|MSTRO_CDO_STATE_IN_TRANSPORT);
  if(s!=MSTRO_OK) {
    ERR("CDO id not known here for app\n");
    goto BAILOUT_FREE;
  }

  /* trigger transport process */
  s=mstro_pm_demand_queue_insert(&cdoid, app_id);
  if(s!=MSTRO_OK) {
    WITH_CDO_ID_STR(idstr, &cdoid,{
        ERR("Failed to trigger transport process for %s to %d: %d\n",
            idstr, app_id, s);
      });
  };
  /* DEMAND has no ACK. The DEMAND is implicitly acknowledged when the
   * requestor is ready to send out the TRANSFER_COMPLETED */

  /* produce demand:after event */
  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__DEMAND;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_DEMAND;
  ev.demand = cont->msg->demand;
  ev.origin_id = cont->msg->token->appid;
  s = mstro_pm__possibly_fill_event_cdoname(&cdoid, app_id, &ev);
  if(s!=MSTRO_OK) {
    goto BAILOUT_FREE;
  }

  s = mstro_pm__event_notify_and_continue(
      &ev,
      false,
      NULL,
      mstro_pm__handle_demand_phase3,
      cont->msg, NULL, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed notifying subscribers of DEMAND:after event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT_FREE;
  }
  goto DONE;

BAILOUT_FREE:
  mstro_pm__msg_free(cont->msg);
  
DONE:
  return s;
}

static inline
mstro_status
mstro_pm__handle_demand(Mstro__Pool__MstroMsg *msg)
{
  Mstro__Pool__Demand *demand = msg->demand;
  mstro_app_id app_id = msg->token->appid->id;
  assert(demand!=NULL); assert(app_id!=MSTRO_APP_ID_INVALID);

  struct mstro_cdo_id cdoid = { .qw[0] = demand->cdoid->qw0,
                                .qw[1] = demand->cdoid->qw1 };
  

  DEBUG("CDO DEMAND from %zu\n", app_id);
  mstro_status s;
  
  if(demand->cdoid==NULL) {
    s=MSTRO_INVMSG;
    goto BAILOUT;
  }

    Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__DEMAND;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_DEMAND;
  ev.demand = msg->demand;
  ev.origin_id = msg->token->appid;
  s = mstro_pm__possibly_fill_event_cdoname(&cdoid, app_id, &ev);
  if(s!=MSTRO_OK) {
    goto BAILOUT_FREE;
  }

  s = mstro_pm__event_notify_and_continue(
      &ev,
      true,
      NULL,
      mstro_pm__handle_demand_phase2,
      msg, NULL, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed notifying subscribers of DEMAND:before event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT_FREE;
  }
BAILOUT:
  return s;

BAILOUT_FREE:
  mstro_pm__msg_free(msg);
  return s;
}



/*** WITHDRAW ***/
static inline
mstro_status
mstro_pm__handle_withdraw_phase3(mstro_event event,
                                 struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("WITHDRAW phase 3/3, event %p\n", event);
  mstro_pm__msg_free(cont->msg);
  return MSTRO_OK;
}

static inline
mstro_status
mstro_pm__handle_withdraw_phase2(mstro_event event,
                                 struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("WITHDRAW phase 2/3, event %p\n", event);
  mstro_app_id app_id = cont->msg->token->appid->id;
  Mstro__Pool__Withdraw *withdraw = cont->msg->withdraw;
  struct mstro_cdo_id cdoid = { .qw[0] = withdraw->cdoid->qw0,
                                .qw[1] = withdraw->cdoid->qw1 };
  mstro_status s = MSTRO_OK;

  s=mstro_pm_cdo_registry_update_state(&cdoid,
                                       app_id,
                                       MSTRO_CDO_STATE_WITHDRAWN);
  if(s!=MSTRO_OK) {
    ERR("CDO id not known here for app\n");
  }

  s = mstro_pm__send_ack(app_id,
                         MSTRO__POOL__POOL_OP_ACK__POOL_OP__WITHDRAW,
                         withdraw->cdoid, NULL, NULL, s);

  if(s!=MSTRO_OK) {
    ERR("Failed to send WITHDRAW-ACK: %d (%s)\n",
        s, mstro_status_description(s));
  }

  /* produce require:after event */
  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__WITHDRAW;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_WITHDRAW;
  ev.withdraw = cont->msg->withdraw;
  ev.origin_id = cont->msg->token->appid;
  s = mstro_pm__possibly_fill_event_cdoname(&cdoid, app_id, &ev);
  if(s!=MSTRO_OK) {
    goto BAILOUT_FREE;
  }

  s = mstro_pm__event_notify_and_continue(
      &ev,
      false,
      NULL,
      mstro_pm__handle_withdraw_phase3,
      cont->msg, NULL, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed notifying subscribers of WITHDRAW:after event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT_FREE;
  }
  goto DONE;
  
BAILOUT_FREE:
  mstro_pm__msg_free(cont->msg);
DONE:
  return s;
}

static inline
mstro_status
mstro_pm__handle_withdraw(Mstro__Pool__MstroMsg *msg)
{
  Mstro__Pool__Withdraw *withdraw = msg->withdraw;
  mstro_app_id app_id = msg->token->appid->id;
  assert(withdraw!=NULL); assert(app_id!=MSTRO_APP_ID_INVALID);
  DEBUG("CDO WITHDRAW from %zu\n", app_id);
  mstro_status s;
  
  if(withdraw->cdoid==NULL) {
    s=MSTRO_INVMSG;
    goto BAILOUT;
  }

  struct mstro_cdo_id cdoid = { .qw[0] = withdraw->cdoid->qw0,
                                .qw[1] = withdraw->cdoid->qw1 };

  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__WITHDRAW;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_WITHDRAW;
  ev.withdraw = msg->withdraw;
  ev.origin_id = msg->token->appid;
  s = mstro_pm__possibly_fill_event_cdoname(&cdoid, app_id, &ev);
  if(s!=MSTRO_OK) {
    goto BAILOUT_FREE;
  }

  s = mstro_pm__event_notify_and_continue(
      &ev,
      true,
      NULL,
      mstro_pm__handle_withdraw_phase2,
      msg, NULL, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed notifying subscribers of WITHDRAW:before event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT_FREE;
  }

BAILOUT:
  return s;
BAILOUT_FREE:
  mstro_pm__msg_free(msg);
  return s;
}


/*** DISPOSE ***/

static inline
mstro_status
mstro_pm__handle_dispose_phase3(mstro_event event,
                                struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("DISPOSE phase 3/3, event %p\n", event);
  mstro_pm__msg_free(cont->msg);
  return MSTRO_UNIMPL;
}

static inline
mstro_status
mstro_pm__handle_dispose_phase2(mstro_event event,
                                struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("DISPOSE phase 2/3, event %p\n", event);
  mstro_app_id app_id = cont->msg->token->appid->id;
  Mstro__Pool__Dispose *dispose = cont->msg->dispose;
  struct mstro_cdo_id cdoid = { .qw[0] = dispose->cdoid->qw0,
                                .qw[1] = dispose->cdoid->qw1 };
  mstro_status s = MSTRO_OK;

  ERR("DISPOSE unimplemented on PM\n"); /* FIXME: UNIMPL is sent by phase3 */
  /* remove it from the registry */

  /* produce require:after event */
  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__DISPOSE;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_DISPOSE;
  ev.dispose = cont->msg->dispose;
  ev.origin_id = cont->msg->token->appid;
  s = mstro_pm__possibly_fill_event_cdoname(&cdoid, app_id, &ev);
  if(s!=MSTRO_OK) {
    goto BAILOUT_FREE;
  }

  s = mstro_pm__event_notify_and_continue(
      &ev,
      false,
      NULL,
      mstro_pm__handle_dispose_phase3,
      cont->msg, NULL, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed notifying subscribers of DISPOSE:after event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT_FREE;
  }
  goto DONE;
  
BAILOUT_FREE:
  mstro_pm__msg_free(cont->msg);
DONE:
  return s;
}

static inline
mstro_status
mstro_pm__handle_dispose(Mstro__Pool__MstroMsg *msg)
{
  Mstro__Pool__Dispose *dispose = msg->dispose;
  mstro_app_id app_id = msg->token->appid->id;
  app_id=app_id; /* avoid unused arg warning */
  assert(dispose!=NULL);

  struct mstro_cdo_id cdoid = { .qw[0] = dispose->cdoid->qw0,
                                .qw[1] = dispose->cdoid->qw1 };


  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__DISPOSE;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_DISPOSE;
  ev.dispose = msg->dispose;
  ev.origin_id = msg->token->appid;
  mstro_status s =
      mstro_pm__possibly_fill_event_cdoname(&cdoid, app_id, &ev);
  if(s!=MSTRO_OK) {
    goto BAILOUT_FREE;
  }

  s = mstro_pm__event_notify_and_continue(
      &ev,
      true,
      NULL,
      mstro_pm__handle_dispose_phase2,
      msg, NULL, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed notifying subscribers of DISPOSE:before event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT_FREE;
  }
  goto DONE;
  
BAILOUT_FREE:
  mstro_pm__msg_free(msg);
DONE:
  return s;
}



static inline
mstro_status
mstro_pm__handle_transport_completed_phase2(mstro_event event,
                                            struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("TRANSPORT-COMPLETED phase 2/2, event %p\n", event);
  mstro_pm__msg_free(cont->msg);
  return MSTRO_OK;
}

static inline
mstro_status
mstro_pm__handle_transfer_completed(Mstro__Pool__MstroMsg *msg)
{
  Mstro__Pool__TransferCompleted *completion = msg->transfer_completed;
  mstro_app_id app_id = msg->token->appid->id;
  assert(completion!=NULL); assert(app_id!=MSTRO_APP_ID_INVALID);

  struct mstro_cdo_id cdoid = { .qw[0] = completion->cdoid->qw0,
                                .qw[1] = completion->cdoid->qw1 };
  
  DEBUG("TRANSFER COMPLETION from %zu\n", app_id);

  /* produce completed:after event */
  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__TRANSFER_COMPLETED;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_TRANSFER_COMPLETED;
  ev.transfer_completed = msg->transfer_completed;
  ev.origin_id = msg->token->appid;
  mstro_status s =
      mstro_pm__possibly_fill_event_cdoname(&cdoid, app_id, &ev);
  if(s!=MSTRO_OK) {
    goto BAILOUT_FREE;
  }

  s = mstro_pm__event_notify_and_continue(
      &ev,
      false,
      NULL,
      mstro_pm__handle_transport_completed_phase2,
      msg, NULL, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed notifying subscribers of TRANSPORT_COMPLETED:after event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT_FREE;
  }
  goto DONE;
  
BAILOUT_FREE:
  mstro_pm__msg_free(msg);
DONE:
  return s;
}


/*** LEAVE ***/
static inline
mstro_status
mstro_pm__handle_leave_phase3(mstro_event event,
                              struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("LEAVE phase 3/3, event %p\n", event);
  mstro_pm__msg_free(cont->msg);
  return MSTRO_UNIMPL;
}

static inline
mstro_status
mstro_pm__handle_leave_phase2(mstro_event event,
                              struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("LEAVE phase 2/3, event %p\n", event);
  mstro_app_id app_id = cont->msg->token->appid->id;

  /* this is a very much blocking op: It checks that all outstanding
   * requests depending on this app are done, pulls off persisten CDOs,
   * then removes the app itself from the registry */
  
  
  /* FIXME It's one of the main reasons we should consider using ULT
   * and preemption on the message handlers at some point */
  mstro_status s = mstro_pm_app_deregister(app_id);
  if(s!=MSTRO_OK) {
    ERR("Failed to deregister app %zu\n", app_id);
  }

  INFO("Granting LEAVE request from app %zu\n", app_id);
  
  Mstro__Pool__Bye bye = MSTRO__POOL__BYE__INIT;

  Mstro__Pool__MstroMsg msg_bye = MSTRO__POOL__MSTRO_MSG__INIT;
  s = mstro_pmp_package(&msg_bye, (ProtobufCMessage*)&bye);
  if(s!=MSTRO_OK) {
    ERR("Failed to package %s into a pool manager message\n",
        bye.base.descriptor->name);
    goto BAILOUT;
  }
  s = mstro_pmp_send_nowait(app_id, &msg_bye);
  if(s!=MSTRO_OK) {
    ERR("Cannot send BYE reply to %zu: %d (%s)\n",
        app_id, s, mstro_status_description(s));
    goto BAILOUT;
  }

  /* produce leave:after event */
  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__APP_LEAVE;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_LEAVE;
  ev.leave = cont->msg->leave;
  ev.origin_id = cont->msg->token->appid;

  s = mstro_pm__event_notify_and_continue(
      &ev,
      false,
      NULL,
      mstro_pm__handle_leave_phase3,
      cont->msg, NULL, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed notifying subscribers of LEAVE:after event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT_FREE;
  }
  goto DONE;
  
BAILOUT:
BAILOUT_FREE:
  mstro_pm__msg_free(cont->msg);
DONE:
  return s;
}

static inline
mstro_status
mstro_pm__handle_leave(Mstro__Pool__MstroMsg *msg)
{
  Mstro__Pool__Leave *leave = msg->leave;
  mstro_app_id app_id = msg->token->appid->id;
  assert(leave!=NULL); assert(app_id!=MSTRO_APP_ID_INVALID);

  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__APP_LEAVE;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_LEAVE;
  ev.leave = msg->leave;
  ev.origin_id = msg->token->appid;

  mstro_status s
      = mstro_pm__event_notify_and_continue(
          &ev,
          true,
          NULL,
          mstro_pm__handle_leave_phase2,
          msg, NULL, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed notifying subscribers of LEAVE:before event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT_FREE;
  }
BAILOUT:
  return s;

BAILOUT_FREE:
  mstro_pm__msg_free(msg);
  return s;
}



/* FIXME: this function needs to be refactored and moved to pool
 * manager proto; it currently is in ofi.c */
mstro_status
mstro_pm__register_app(Mstro__Pool__Join *join_msg,
                       struct mstro_endpoint *ep,
                       struct mstro_pm_app_registry_entry **entry_p);

static inline
const char *
mstro_pm__transport_method_name(Mstro__Pool__TransportKind tk)
{
  return protobuf_c_enum_descriptor_get_value
      (&mstro__pool__transport_kind__descriptor, tk)
      ->name;
}



/*** JOIN ***/
static
mstro_status
mstro_pm__handle_join_phase5(mstro_event event,
                             struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("JOIN phase 5/5, event %p\n", event);
  mstro_pm__msg_free(cont->msg);
  return MSTRO_OK;
}

static
mstro_status
mstro_pm__handle_join_phase4(mstro_event event,
                             struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("JOIN phase 4/5, event %p\n", event);

  Mstro__Pool__Appid id = MSTRO__POOL__APPID__INIT;
  id.id = cont->ctx_welcome.regentry->appid;
  
  Mstro__Pool__Apptoken token = MSTRO__POOL__APPTOKEN__INIT;
  token.appid = &id;
  
  Mstro__Pool__Welcome welcome = MSTRO__POOL__WELCOME__INIT;
  welcome.token = &token;
  
  Mstro__Pool__MstroMsg msg_welcome = MSTRO__POOL__MSTRO_MSG__INIT;
  mstro_status status = mstro_pmp_package(&msg_welcome,
                                          (ProtobufCMessage*)&welcome);
  if(status!=MSTRO_OK) {
    ERR("Failed to package %s into a pool manager message\n",
        welcome.base.descriptor->name);
    goto BAILOUT_FREE;
  }
  
  status = mstro_pmp_send_nowait(cont->ctx_welcome.regentry->appid,
                                 &msg_welcome);
  if(status!=MSTRO_OK) {
    ERR("Cannot send WELCOME reply to %zu: %d (%s)\n",
        cont->ctx_welcome.regentry->appid,
        status, mstro_status_description(status));
    goto BAILOUT_FREE;
  }
  
  /* broadcast WELCOME:after event */
  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__APP_WELCOME;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_WELCOME;
  ev.welcome = &welcome;
  ev.origin_id = cont->msg->token->appid;

  status = mstro_pm__event_notify_and_continue(
      &ev,
      false,
      cont,
      mstro_pm__handle_join_phase5,
      cont->msg, NULL, NULL);
  if(status!=MSTRO_OK) {
    ERR("Failed notifying subscribers of WELCOME:after event\n");
    goto BAILOUT_FREE;
  }
  goto DONE;
  
BAILOUT_FREE:
  mstro_pm__msg_free(cont->msg);
DONE:
  return status;
}

static
mstro_status
mstro_pm__handle_join_phase3(mstro_event event,
                             struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("JOIN phase 3/5, event %p\n", event);

  /* broadcast WELCOME:before event */
  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__APP_WELCOME;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_WELCOME;
  /* FIXME: instead of a full welcome message just the app id would be enough */
  Mstro__Pool__Appid id = MSTRO__POOL__APPID__INIT;
  id.id =  cont->ctx_welcome.regentry->appid;
  Mstro__Pool__Apptoken token = MSTRO__POOL__APPTOKEN__INIT;
  token.appid = &id;
  Mstro__Pool__Welcome welcome = MSTRO__POOL__WELCOME__INIT;
  welcome.token = &token;
  ev.welcome = &welcome;
  
  mstro_status status = mstro_pm__event_notify_and_continue(
      &ev,
      true,
      cont,
      mstro_pm__handle_join_phase4,
      cont->msg, NULL, cont->ctx_welcome.regentry);
  if(status!=MSTRO_OK) {
    ERR("Failed notifying subscribers of WELCOME:before event\n");
    goto BAILOUT_FREE;
  }
  goto DONE;
BAILOUT_FREE:
  mstro_pm__msg_free(cont->msg);

DONE:
  return status;
}

static
mstro_status
mstro_pm__handle_join_phase2(mstro_event event,
                             struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("JOIN phase 2/5, event %p\n", event);

  /* yes, this is an exceedingly ugly way of getting a printout of
   * the supported transports */
  const char* transports[5]= {"","","","",""};
  Mstro__Pool__TransportMethods *tm = cont->msg->join->transport_methods;
  
  size_t n = tm->n_supported;
  if(n>4) {
    transports[4] = "and more";
    n=4;
  }
  switch(n) {
    case 4:
      transports[3]
          = mstro_pm__transport_method_name(
              tm->supported[3]);
      /* fall-thru */
    case 3:
      transports[2]
          = mstro_pm__transport_method_name(
              tm->supported[2]);
      /* fall-thru */
    case 2:
      transports[1]
          = mstro_pm__transport_method_name(
              tm->supported[1]);
      /* fall-thru */
    case 1:
      transports[0]
          = mstro_pm__transport_method_name(
              tm->supported[0]);
      break;
    default:
      ERR("Not handling all transport methods advertised (have %zu, handling 4)\n", n);
  };
  
  struct mstro_pm_app_registry_entry *regentry = NULL;
  mstro_status status = mstro_pm__register_app(cont->msg->join,
                                               cont->ctx_join.ep,
                                               &regentry);
  if(status!=MSTRO_OK) {
    ERR("Failed to register caller as new app\n");
    goto BAILOUT_FREE;
  }
  
  INFO("Registered app at id %zu, Transport: %s %s %s %s %s\n",
       regentry->appid,
       transports[0], transports[1], transports[2],
       transports[3], transports[4]);

  /* produce join:after event */

  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__APP_JOIN;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_JOIN;
  ev.join = cont->msg->join;
  /* it's safe to refer to the Appid object since the Event object is
   * on stack and will only be alive until notify-and-continue is
   * done */
  Mstro__Pool__Appid aid = MSTRO__POOL__APPID__INIT;
  aid.id = regentry->appid;
  ev.origin_id = &aid;


  status = mstro_pm__event_notify_and_continue(
      &ev,
      false,
      NULL,
      mstro_pm__handle_join_phase3,
      cont->msg, NULL, regentry);
  if(status!=MSTRO_OK) {
    ERR("Failed notifying subscribers of JOIN:after event: %d (%s)\n",
        status, mstro_status_description(status));
    goto BAILOUT_FREE;
  }
  goto DONE;
BAILOUT_FREE:
  mstro_pm__msg_free(cont->msg);
  /* FIXME: free regentry and context */

DONE:
  return status;
}

/** handle JOIN message that came in on EP */
static inline
mstro_status
mstro_pm__handle_join(Mstro__Pool__MstroMsg *msg,
                      struct mstro_endpoint *ep)
{
  Mstro__Pool__Join *join = msg->join;
  mstro_status status = MSTRO_OK;
  
  INFO("JOIN message received. Caller %s (PM proto version %d) advertises endpoint %s\n",
       join->component_name, join->protocol_version, join->serialized_endpoint);
  
  if(join->transport_methods==NULL
     || (join->transport_methods->n_supported==0)) {
    ERR("No transport methods advertised\n");
    status=MSTRO_INVMSG;
    goto BAILOUT_FREE;
  }

  /* broadcast JOIN:before event */
  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__APP_JOIN;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_JOIN;
  ev.join = msg->join;
  /* origin-id not assigned  yet. */
  ev.origin_id = NULL;


  status = mstro_pm__event_notify_and_continue(
      &ev,
      true,
      NULL,
      mstro_pm__handle_join_phase2,
      msg, ep, NULL);
  if(status!=MSTRO_OK) {
    ERR("Failed notifying subscribers of JOIN:before event: %d (%s)\n",
        status, mstro_status_description(status));
    goto BAILOUT_FREE;
  }
  goto DONE;
BAILOUT_FREE:
  mstro_pm__msg_free(msg);

DONE:
  return status;
}


/*** SUBSCRIBE ***/

static inline
mstro_status
mstro_pm__handle_subscribe(Mstro__Pool__MstroMsg *msg)
{
  Mstro__Pool__Subscribe *subscribe = msg->subscribe;
  mstro_app_id app_id = msg->token->appid->id;
  assert(subscribe!=NULL); assert(app_id!=MSTRO_APP_ID_INVALID);
  DEBUG("SUBSCRIBE from %zu\n", app_id);
  mstro_status s=MSTRO_UNIMPL;

  Mstro__Pool__SubscriptionHandle *handle = NULL;
  s = mstro_subscription_message_register(subscribe, app_id, &handle);
  if(s!=MSTRO_OK) {
    ERR("Failed to register subscription for app %" PRIu64 " (local id % " PRIu64 ")\n",
        app_id, subscribe->local_id);
    goto BAILOUT;
  }

  Mstro__Pool__SubscribeAck sack = MSTRO__POOL__SUBSCRIBE_ACK__INIT;
  sack.serial = subscribe->local_id;
  sack.handle = handle;

  Mstro__Pool__MstroMsg msg_sack = MSTRO__POOL__MSTRO_MSG__INIT;
  s = mstro_pmp_package(&msg_sack, (ProtobufCMessage*)&sack);
  if(s!=MSTRO_OK) {
    ERR("Failed to package %s-ACK into a pool manager message\n",
        sack.base.descriptor->name);
    goto BAILOUT;
  }
  
  s = mstro_pmp_send_nowait(app_id, &msg_sack);
  if(s!=MSTRO_OK) {
    ERR("Failed to send %s-ACK to %zu\n", app_id);
    goto BAILOUT;
  }
  
  s=MSTRO_OK;
BAILOUT:
  mstro_pm__msg_free(msg);
  return s;
}


/*** UNSUBSCRIBE ***/

static inline
mstro_status
mstro_pm__handle_unsubscribe_phase2(mstro_event event,
                                    struct mstro_pm__continuation_ctx *cont)
{
  DEBUG("UNSUBSCRIBE phase 2/3, event %p\n", event);
  mstro_app_id app_id = cont->msg->token->appid->id;

  /* unregister kills the subscription handle, so we need to copy it */
  Mstro__Pool__SubscriptionHandle *handle
      = cont->msg->unsubscribe->subscription_handle;
  
  mstro_status s = mstro_subscription_message_unregister(
      cont->msg->unsubscribe, app_id);
  if(s!=MSTRO_OK) {
    ERR("Failed to drop subscription %" PRIu64 "\n", handle->id);
  }  
  
  
BAILOUT:
  s = mstro_pm__send_ack(app_id,
                         MSTRO__POOL__POOL_OP_ACK__POOL_OP__UNSUBSCRIBE,
                         NULL, handle, NULL, s);
  mstro_pm__msg_free(cont->msg);
  return s;
}

static inline
mstro_status
mstro_pm__handle_unsubscribe(Mstro__Pool__MstroMsg *msg)
{
  Mstro__Pool__Unsubscribe *unsubscribe = msg->unsubscribe;
  mstro_app_id app_id = msg->token->appid->id;
  assert(unsubscribe!=NULL); assert(app_id!=MSTRO_APP_ID_INVALID);
  DEBUG("UNSUBSCRIBE from %zu\n", app_id);
  mstro_status s=MSTRO_UNIMPL;

  /* broadcast JOIN event */
  Mstro__Pool__Event ev = MSTRO__POOL__EVENT__INIT;
  /* serial and handle set later */
  ev.kind = MSTRO__POOL__EVENT_KIND__UNSUBSCRIBE;
  ev.payload_case = MSTRO__POOL__EVENT__PAYLOAD_UNSUBSCRIBE;
  ev.unsubscribe = msg->unsubscribe;
  /* it's safe to refer to the Appid object since the Event object is
   * on stack and will only be alive until notify-and-continue is
   * done */
  ev.origin_id = msg->token->appid;

  s = mstro_pm__event_notify_and_continue(
      &ev,
      true,
      NULL,
      mstro_pm__handle_unsubscribe_phase2,
      msg, NULL, NULL);
  if(s!=MSTRO_OK) {
    ERR("Failed notifying subscribers of UNSUBSCRIBE:before event: %d (%s)\n",
        s, mstro_status_description(s));
    goto BAILOUT_FREE;
  }
  goto DONE;
BAILOUT_FREE:
  mstro_pm__msg_free(msg);
DONE:
  return s;
}


/*** EVENT_ACK ***/
mstro_status
mstro_pm__handle_event_ack(Mstro__Pool__MstroMsg *msg)
{
  Mstro__Pool__EventAck *ack = msg->event_ack;
  
  if(ack==NULL)
    return MSTRO_INVARG;

  /* pass in to subscription engine. It will do teh accounting, and in
   * the end trigger continuations for those handlers that sent the
   * notification event. */
  mstro_status s = mstro_subscription_message_event_ack(ack);
  if(s!=MSTRO_OK) {
    ERR("Failed to process event ack %" PRIu64 ": %d (%s)\n",
        ack->serial, s, mstro_status_description(s));
    goto BAILOUT;
  }
  
BAILOUT:
  mstro_pm__msg_free(msg);
  return s;
}

/* FIXME: missing event handling here */
mstro_status
mstro_pm__handle_resolve(Mstro__Pool__MstroMsg *msg)
{
  Mstro__Pool__Resolve *resolve = msg->resolve;
  mstro_app_id app_id = msg->token->appid->id;
  assert(resolve!=NULL); assert(app_id!=MSTRO_APP_ID_INVALID);
  DEBUG("RESOLVE from %zu\n", app_id);
  mstro_status s=MSTRO_UNIMPL;

  Mstro__Pool__ResolveReply reply = MSTRO__POOL__RESOLVE_REPLY__INIT;
  reply.query = resolve;

  switch(resolve->query_case) {
    case MSTRO__POOL__RESOLVE__QUERY_CDOID: {
      reply.result_case = MSTRO__POOL__RESOLVE_REPLY__RESULT_NAME;
      struct mstro_cdo_id id = {.qw[0] = resolve->cdoid->qw0,
                                .qw[1] = resolve->cdoid->qw1};
      s= mstro_pm_cdo_registry_cdo_name_lookup(&id, (const char **)&reply.name);
      if(s!=MSTRO_OK) {
        WITH_CDO_ID_STR(idstr, &id, {
            ERR("Cannot find CDO registry entry for |%s|\n",
                idstr);});
        reply.name = NULL;
      } else {
        ; /* set in lookup call above */
      }
      WITH_CDO_ID_STR(idstr, &id, {
          DEBUG("Resolved CDO ID %s to |%s| for app %" PRIi64 "\n",
                idstr, reply.name, app_id);});
      s=MSTRO_OK;
      break;
    }
    case MSTRO__POOL__RESOLVE__QUERY_APPID: {
      reply.result_case = MSTRO__POOL__RESOLVE_REPLY__RESULT_NAME;
      mstro_app_id appid = resolve->appid;
      struct mstro_pm_app_registry_entry *entry;
      s = mstro_pm_app_lookup(appid, &entry);
      if(s!=MSTRO_OK) {
        ERR("Cannot find app for ID " PRIi64 "\n", appid);
        reply.name = NULL;
      } else {
        reply.name = entry->component_name;
      }
      DEBUG("Resolved APP ID %" PRIi64 " to |%s| for app %" PRIi64 "\n",
            appid, reply.name, app_id);
      s=MSTRO_OK;
      break;
    }

    case MSTRO__POOL__RESOLVE__QUERY__NOT_SET: {
      ERR("Invalid resolve query received\n");
      s=MSTRO_INVARG; goto BAILOUT;
      break;
    }
    default: {
      ERR("Unsupported resolve query type: %d\n",
          resolve->query_case);
      s=MSTRO_UNIMPL; goto BAILOUT;
    }
  }
  /* send reply */
  Mstro__Pool__MstroMsg rmsg = MSTRO__POOL__MSTRO_MSG__INIT;
  s = mstro_pmp_package(&rmsg, (ProtobufCMessage*)&reply);
  if(s!=MSTRO_OK) {
    ERR("Failed to package ResolveReply into a pool manager message\n",
        reply.base.descriptor->name);
    goto BAILOUT;
  }
  s = mstro_pmp_send_nowait(app_id, &rmsg);
  if(s!=MSTRO_OK) {
    ERR("Failed to send ResolveReply to %zu\n", app_id);
    goto BAILOUT;
  }

BAILOUT:
  mstro_pm__msg_free(msg);
  return s;
}


/* Pool manager "Handle incoming Message" dispatcher function */
mstro_status
mstro_pm_handle_msg(const struct mstro_msg_envelope *envelope,
                    struct mstro_endpoint *ep)
{
  /* for now we serialize things in here. Instead we could push to a
   * queue and have a queue runner in a different thread pick things
   * up */
  mstro_status status=MSTRO_OK;
  INFO("PM Handling incoming message size %d\n", envelope->payload_size);

  Mstro__Pool__MstroMsg *msg
      = mstro__pool__mstro_msg__unpack(NULL, envelope->payload_size,
                                       envelope->data);
  
  if(msg==NULL) {
    ERR("Failed to unpack the message\n");
    status = MSTRO_INVMSG;
    goto BAILOUT;
  } else {
    INFO("... unpacked msg: type %s (%s), vsm %s\n",
         msg->base.descriptor->name,
         (msg->base.descriptor==&mstro__pool__mstro_msg__descriptor
          ? msg->declare->base.descriptor->name : "no subtype"),
         msg->opts && msg->opts->vsm_data ? "yes" : "no");
  }
  
  if(msg->base.descriptor != &mstro__pool__mstro_msg__descriptor) {
    ERR("Cannot handle messages other than MstroPool messages\n");
    mstro_pm__msg_free(msg);
    status = MSTRO_INVMSG;
    goto BAILOUT;
  }
  
  if(msg->opts && msg->opts->vsm_data!=NULL) {
    ERR("VSM, Unimplemented\n");
    mstro_pm__msg_free(msg);
    status=MSTRO_UNIMPL;
    goto BAILOUT;
  }

  /* FIXME: even JOIN should have security token */
  if(msg->msg_case!=MSTRO__POOL__MSTRO_MSG__MSG_JOIN) {
    if(!msg->token) {
      ERR("Incoming message without token\n");
      mstro_pm__msg_free(msg);
      status=MSTRO_INVMSG;
      goto BAILOUT;
    }
    if(!msg->token->appid) {
      ERR("Incoming message without App ID\n");
      mstro_pm__msg_free(msg);
      status=MSTRO_INVMSG;
      goto BAILOUT;
    }

    /* FIXME: should check whether the app is JOINED */
    if(msg->token->appid->id==MSTRO_APP_ID_INVALID) {
      ERR("Invalid app id -- are you JOINed?\n");
      mstro_pm__msg_free(msg);
      status=MSTRO_INVMSG;
      goto BAILOUT;
    }
  }
  
  switch(msg->msg_case) {
    /* 'good' messages: */
    case MSTRO__POOL__MSTRO_MSG__MSG_DECLARE:
      status = mstro_pm__handle_declare(msg);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_SEAL:
      status = mstro_pm__handle_seal(msg);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_OFFER:
      status = mstro_pm__handle_offer(msg);
      break;
      
    case MSTRO__POOL__MSTRO_MSG__MSG_REQUIRE:
      status = mstro_pm__handle_require(msg);
      break;
        
    case MSTRO__POOL__MSTRO_MSG__MSG_RETRACT:
      status = mstro_pm__handle_retract(msg);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_DEMAND:
      status = mstro_pm__handle_demand(msg);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_WITHDRAW:
      status = mstro_pm__handle_withdraw(msg);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_DISPOSE:
      status = mstro_pm__handle_dispose(msg);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_LEAVE:
      status = mstro_pm__handle_leave(msg);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_JOIN: 
      status = mstro_pm__handle_join(msg, ep);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_TRANSFER_COMPLETED:
      status = mstro_pm__handle_transfer_completed(msg);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_SUBSCRIBE:
      status = mstro_pm__handle_subscribe(msg);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_UNSUBSCRIBE:
      status = mstro_pm__handle_unsubscribe(msg);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_EVENT_ACK:
      status = mstro_pm__handle_event_ack(msg);
      break;

    case MSTRO__POOL__MSTRO_MSG__MSG_RESOLVE:
      status = mstro_pm__handle_resolve(msg);
      break;


      /* currently unimplemented: */
    case MSTRO__POOL__MSTRO_MSG__MSG_SEAL_GROUP:


    case MSTRO__POOL__MSTRO_MSG__MSG_DEMAND_ATTR:

    case MSTRO__POOL__MSTRO_MSG__MSG_QUERY:

    case MSTRO__POOL__MSTRO_MSG__MSG_LOG:

    case MSTRO__POOL__MSTRO_MSG__MSG_INITIATE_TRANSFER:

    case MSTRO__POOL__MSTRO_MSG__MSG_TRANSFER_TICKET: {
      WARN("%s message received, dropping it, not even sending ACK\n",
           msg->base.descriptor->name);
      mstro_pm__msg_free(msg);
      status = MSTRO_UNIMPL;
      break;
    }


      /* 'illegal' messages: */
    case MSTRO__POOL__MSTRO_MSG__MSG_WELCOME:
    case MSTRO__POOL__MSTRO_MSG__MSG_BYE:
    case MSTRO__POOL__MSTRO_MSG__MSG_DECLARE_ACK:
    case MSTRO__POOL__MSTRO_MSG__MSG_ACK:
    case MSTRO__POOL__MSTRO_MSG__MSG_EVENT:
    case MSTRO__POOL__MSTRO_MSG__MSG_DEMAND_ATTR_RES:
    case MSTRO__POOL__MSTRO_MSG__MSG_QUERY_RES:
    case MSTRO__POOL__MSTRO_MSG__MSG__NOT_SET: {
      ERR("Invalid message kind (incoming on PM side): %s\n",
          msg->base.descriptor->name);
      mstro_pm__msg_free(msg);
      status=MSTRO_INVARG;
      goto BAILOUT;
    }

    default:
      WARN("Unknown message kind: %s\n",
           msg->base.descriptor->name);
      mstro_pm__msg_free(msg);
      status = MSTRO_INVARG;
      break;
  }

BAILOUT:
  /* messages get deallocated by sub-handlers, not here */
  
  return status;
}
