/* -*- mode:c -*- */
/** @file
 ** @brief Maestro Pool Manager component -- OFI based implementation
 **/
/*
 * Copyright (C) 2019 Cray Computer GmbH
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "maestro/status.h"
#include "maestro/pool_manager.h"
#include "maestro/i_state.h"
#include "maestro/logging.h"
#include "maestro/i_ofi.h"
#include "maestro/i_globals.h"
#include "maestro/logging.h"

#include "i_event.h"

#include <inttypes.h>
#include <unistd.h>
#include <stdatomic.h>
#include <stdbool.h>
#include <pthread.h>
#include <assert.h>


/* simplify logging */
#define DEBUG(...) LOG_DEBUG(MSTRO_LOG_MODULE_PM,__VA_ARGS__)
#define INFO(...)  LOG_INFO(MSTRO_LOG_MODULE_PM,__VA_ARGS__)
#define WARN(...)  LOG_WARN(MSTRO_LOG_MODULE_PM,__VA_ARGS__)
#define ERR(...)   LOG_ERR(MSTRO_LOG_MODULE_PM,__VA_ARGS__)


/* internal entry point: started in a pthread, runs PM loop until it
 * is told to terminate (when g_pm_stop is set) */

static _Atomic bool g_pm_stop; /* static init 0 means: no */

static _Atomic bool g_pm_transport_init_thread_started = false;

static void*
mstro_pm_thread(void *closure)
{
  assert(closure==NULL);
  DEBUG("PM thread starting\n");
  mstro_status s = mstro_ofi_pm_loop(&g_pm_stop);
  if(s!=MSTRO_OK) {
    ERR("pm loop terminated with error %d (%s)\n",
        s, mstro_status_description(s));
  }
  DEBUG("PM thread stopping\n");
  
  return NULL;
}

static void*
mstro_pm_transport_init_thread_fun(void *closure)
{
  assert(closure==NULL);
  g_pm_transport_init_thread_started = true;

  DEBUG("PM transport initiator thread starting\n");
  while(!atomic_load(&g_pm_stop)) {
    mstro_status s = mstro_pm_handle_demand_queue();
    if(s!=MSTRO_OK) {
      ERR("Error in demand queue runner: %d (%s)\n",
          s, mstro_status_description(s));
    }
    pthread_testcancel();
  }

  DEBUG("PM transport initiator thread stopping.\n");

  return NULL;
}


/* User-side entry point: dedicate current thread to pool manager work */
mstro_status
mstro_pm_start(void)
{
  mstro_status stat = MSTRO_UNIMPL;

  atomic_store(&g_pm_stop,false);

  /* if(!(g_pool_manager_size>=1)) { */
  /*   return MSTRO_INVARG; */
  /* } */

  /* if(g_pool_manager_size!=1) { */
  /*   WARN("DEFICIENCY: Cannot run distributed pool manager yet\n"); */
  /*   return MSTRO_UNIMPL; */
  /* } */

  const struct mstro_core_initdata *d;
  stat = mstro_core_state_get(&d);

  if(stat!=MSTRO_OK) {
    ERR("Failed to get initdata: %d\n", stat);
    goto BAILOUT;
  }
  
  if(d==NULL) {
    ERR("Trying to start pool manager before mstro_init()\n");
    stat=MSTRO_FAIL;
    goto BAILOUT;
  }
  
  /* INFO("Starting Pool Manager for %s," */
  /*      "expecting %" PRIi64 " partners\n", */
  /*      d->workflow_name, g_pool_manager_size); */
  INFO("Starting Pool Manager for workflow %s\n",
       d->workflow_name);

  stat = mstro_event_domain_create("Pool Manager message handler continuations",
                                   &g_mstro_pm_continuations);
  if(stat!=MSTRO_OK) {
    ERR("Failed to create message handler continuation domain\n");
    goto BAILOUT;
  }
  
  /* organize quorum for pool_manager_size */
  
  /* create connection */

  /* create openfabric endpoint on all interfaces, listen on them,
   * create advertisable destination string */

  stat = mstro_ofi_init();  

  /* all endpoints are now enabled. */
  if(stat!=MSTRO_OK) {
     ERR("mstro_ofi_init() failed\n");
  } else {
    /* configure global app ID for manager role */
    g_pool_app_id = MSTRO_APP_ID_MANAGER;
    g_pool_appid.id = g_pool_app_id;
    g_pool_apptoken.appid = &g_pool_appid;
    g_component_descriptor.type = MSTRO_COMPONENT_TYPE_PM;
    
    /* run pool manager loop. It will return after receiving a TERM message */
    int s = pthread_create(&g_pm_thread, NULL, mstro_pm_thread, NULL);
    if(s!=0) {
      ERR("Failed to start pool manager thread: %d\n", s);
      stat=MSTRO_FAIL;
    }
    s = pthread_create(&g_pm_transport_init_thread, NULL,
                       mstro_pm_transport_init_thread_fun, NULL);
    if(s!=0) {
      ERR("Failed to start pool manager transport initator thread: %d\n", s);
      stat=MSTRO_FAIL;
    }
  }

BAILOUT:
  return stat;
}

mstro_status
mstro_pm_terminate(void)
{
  mstro_status status;
  
  INFO("PM termination initiated\n");
  if(g_pm_stop==true) {
    DEBUG("Duplicate PM termination request, ignoring it\n");
    return MSTRO_OK;
  }
  
  atomic_store(&g_pm_stop, true);

  if(g_pm_transport_init_thread_started) {
    int stat=pthread_cancel(g_pm_transport_init_thread);
    if(stat!=0) {
      ERR("Failed to cancel transport initiator thread: %d (%s)\n",
          stat, strerror(stat));
    } else {
      void *res;
      stat=pthread_join(g_pm_transport_init_thread, &res);
      if(stat!=0) {
        ERR("Failed to join transport initiator thread: %d\n");
      }
      g_pm_transport_init_thread_started = false;
    }
  };

  void *res;
  int s = pthread_join(g_pm_thread, &res);
  INFO("PM thread terminated\n");

  status = mstro_ofi_finalize();
  if(status==MSTRO_OK) {
    if(s!=0)
      status = MSTRO_FAIL;
  }
  return status;
}

mstro_status
mstro_pm_getinfo(char **result_p)
{
  if(result_p==NULL)
    return MSTRO_INVOUT;
  return mstro_ofi_pm_info(result_p);
}

/**@brief Deatch a client app from a running pool manager */
mstro_status
mstro_pm_detach(void)
{
  mstro_status s=MSTRO_UNIMPL;
  WARN("Not detaching from pool manager yet\n");

  goto BAILOUT;
BAILOUT:
  return s;
}

