/* -*- mode:c -*- */
/** @file
 ** @brief Pool manager registry
 **/

/*
 * Copyright (C) 2019 Cray Computer GmbH
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the copyright holder nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "maestro/i_pool_manager_registry.h"
#include "maestro/logging.h"

#include "maestro/i_uthash.h"
#include "maestro/i_utlist.h"
#include "maestro/i_attributes.h"

#include "protocols/mstro_pool.pb-c.h"

#include "maestro/i_pool_manager_protocol.h"

#include "i_subscription_registry.h"

#include <assert.h>
#include <stdint.h>
#include <inttypes.h>
#include <pthread.h>
#include <stdatomic.h>


/* simplify logging */
#define NOISE(...) LOG_NOISE(MSTRO_LOG_MODULE_PM,__VA_ARGS__)
#define DEBUG(...) LOG_DEBUG(MSTRO_LOG_MODULE_PM,__VA_ARGS__)
#define INFO(...)  LOG_INFO(MSTRO_LOG_MODULE_PM,__VA_ARGS__)
#define WARN(...)  LOG_WARN(MSTRO_LOG_MODULE_PM,__VA_ARGS__)
#define ERR(...)   LOG_ERR(MSTRO_LOG_MODULE_PM,__VA_ARGS__)


/** the app registry table */
static struct mstro_pm_app_registry_entry *g_mstro_pm_app_registry = NULL;

/** lock for the registry */
pthread_mutex_t g_mstro_pm_app_registry_lock = PTHREAD_MUTEX_INITIALIZER;

/** macro to make locked uthash accesses to the registry. Not super
 * safe, beware. It's meant to only be used around HASH_* functions */
#define WITH_LOCKED_APP_REGISTRY(body) \
  do {                                                                  \
  int wlr_stat=pthread_mutex_lock(&g_mstro_pm_app_registry_lock);       \
  if(wlr_stat!=0) {                                                     \
    ERR("Failed to lock app registry: %d (%s)\n",                       \
        wlr_stat, strerror(wlr_stat));                                  \
    abort();                                                            \
  }                                                                     \
                                                                        \
  do {                                                                  \
    body;                                                               \
  } while(0);                                                           \
                                                                        \
  wlr_stat=pthread_mutex_unlock(&g_mstro_pm_app_registry_lock);         \
  if(wlr_stat!=0) {                                                     \
    ERR("Failed to unlock app registry: %d (%s)\n",                     \
        wlr_stat, strerror(wlr_stat));                                  \
    abort();                                                            \
  }                                                                     \
  } while(0)

/** the global next-available-app-id counter */
static _Atomic mstro_app_id g_next_msg_source_id = MSTRO_APP_ID_MANAGER+1;

static
mstro_app_id
mstro_pm_app__next_id(void)
{
  mstro_app_id next = atomic_fetch_add(&g_next_msg_source_id, 1);
  if(next<UINT64_MAX) {
    return next;
  } else {
    /* slow path: table is quite full apparently */
    ERR("Unimplemented slow-path for >2^64-1 maestro pool participants, FIXME\n");
    abort();
  }
}

static
mstro_status
mstro_pm_app_reg__new_entry(struct mstro_pm_app_registry_entry **entry_p, mstro_app_id id)
{
  assert(entry_p!=NULL);
  
  struct mstro_pm_app_registry_entry *e
      =malloc(sizeof(struct mstro_pm_app_registry_entry));
  if(e==NULL) {
    return MSTRO_NOMEM;
  }
  
  e->appid = id;
  e->ep = NULL;
  e->serialized_desc = NULL;
  e->transport_methods = NULL;
  e->component_name = NULL;

  *entry_p = e;
  return MSTRO_OK;
}


static const Mstro__Pool__TransportKind default_transport_kinds[2] =   { MSTRO__POOL__TRANSPORT_KIND__GFS, MSTRO__POOL__TRANSPORT_KIND__MIO};

static const Mstro__Pool__TransportMethods default_transport_methods = {
  PROTOBUF_C_MESSAGE_INIT (&mstro__pool__transport_methods__descriptor), \
  2, (Mstro__Pool__TransportKind*)&default_transport_kinds};

mstro_status
mstro_pm_app_register(struct mstro_endpoint *ep,
                      fi_addr_t addr,
                      char* serialized_desc,
                      Mstro__Pool__TransportMethods *transport_methods,
                      mstro_app_id *id_p,
                      const char *component_name,
                      struct mstro_pm_app_registry_entry **entry_p)
{
  if(ep==NULL)
    return MSTRO_INVARG;
  if(id_p==NULL && entry_p==NULL)
    return MSTRO_INVOUT;

  
  struct mstro_pm_app_registry_entry *e=NULL;
  mstro_status status
      = mstro_pm_app_reg__new_entry(&e, mstro_pm_app__next_id());
  if(status==MSTRO_OK) {
    e->ep = ep;
    e->addr = addr;
	e->serialized_desc = serialized_desc;
    e->transport_methods = transport_methods;
    e->component_name = strdup(component_name);

    INFO("Registered app with transport methods %p\n",
         e->transport_methods);

    WITH_LOCKED_APP_REGISTRY({
        HASH_ADD(hh, g_mstro_pm_app_registry,
                 appid, sizeof(mstro_app_id), e);
      });

    /* return data to caller if interesed */
    if(id_p!=NULL) {
      *id_p = e->appid;
    }
    if(entry_p!=NULL) {
      *entry_p = e;
    }

  }
  return status;
}

mstro_status
mstro_pc_app_register(struct mstro_endpoint *ep,
                      fi_addr_t addr,
                      char* serialized_desc,
                      Mstro__Pool__TransportMethods *transport_methods,
                      mstro_app_id id,
                      struct mstro_pm_app_registry_entry **entry_p)
{
  if(ep==NULL)
    return MSTRO_INVARG;
  if(id==MSTRO_APP_ID_INVALID)
    return MSTRO_INVARG;

  struct mstro_pm_app_registry_entry *e=NULL;
  mstro_status status = mstro_pm_app_reg__new_entry(&e, id);
  if(status==MSTRO_OK) {
    e->ep = ep;
    e->addr = addr;
    e->serialized_desc = serialized_desc;
    e->transport_methods = transport_methods;

    /* FIXME: we should check that there is no previous entry and
     * document whether that's an error or a legal way of overriding
     * things */
    WITH_LOCKED_APP_REGISTRY({
        HASH_ADD(hh, g_mstro_pm_app_registry,
                 appid, sizeof(mstro_app_id), e);
      });

    /* return data to caller if interesed */
    if(entry_p!=NULL) {
      *entry_p = e;
    }

  }
  return status;
}

mstro_status
mstro_pm_app_register_manager(struct mstro_endpoint *ep,
                              fi_addr_t addr)
{
  if(ep==NULL)
    return MSTRO_INVARG;

  struct mstro_pm_app_registry_entry *e=NULL;
  mstro_pm_app_lookup(MSTRO_APP_ID_MANAGER, &e);
  if(e!=NULL) {
    WARN("Overwriting existing pool manager registry entry!\n");
    e=NULL;
  }

  mstro_status status = mstro_pm_app_reg__new_entry(&e, MSTRO_APP_ID_MANAGER);
  if(status==MSTRO_OK) {
    e->ep = ep;
    e->addr = addr;

    WITH_LOCKED_APP_REGISTRY({
        HASH_ADD(hh, g_mstro_pm_app_registry,
                 appid, sizeof(mstro_app_id), e);
      });
  }
  return status;
}






mstro_status
mstro_pm_app_lookup(mstro_app_id appid,
                    struct mstro_pm_app_registry_entry **app_entry_p)
{
  assert(app_entry_p!=NULL);
  mstro_status status=MSTRO_OK;
  
  WITH_LOCKED_APP_REGISTRY({
      struct mstro_pm_app_registry_entry *e=NULL;
      HASH_FIND(hh, g_mstro_pm_app_registry, &appid, sizeof(mstro_app_id), e);
      if(e!=NULL) {
        *app_entry_p = e;
      } else {
        *app_entry_p = NULL;
        status=MSTRO_FAIL;
      }
    });
  return status;
}




/** pair of CDO name and chosen CDO ID */
struct cdo_name_id_pair {
  struct cdo_name_id_pair *next;  /**< this is a linked list */
  char *cdo_name;                 /**< the CDO name */ 
  struct mstro_cdo_id cdo_id;     /**< the CDO ID */
};
  
/** Table of colliding names for the same CDO ID, and their resolution */
struct cdo_id_collision_entry {
  UT_hash_handle hh;                   /**< hash on cdo_id */
  struct mstro_cdo_id cdo_id;          /**< the CDO ID that had a collision */
  struct cdo_name_id_pair *collisions; /**< list of name/ID pairs to resolve the collision */
};

/** Table of collisions encountered */
/* FIXME: currenly not implemented */
static struct cdo_id_collision_entry *g_cdo_id_collision_table = NULL;


/** Per-Application CDO knowledge structure. */
struct per_app_cdo_entry {
  UT_hash_handle hh;                /**< hash on app */
  mstro_app_id app;                 /**< application id [currently JOINED] */
  mstro_cdo_state cdo_state;            /**< state on that app */
  Mstro__Pool__Attributes *attributes;  /**< attributes reported from this app */
};
  
/** Entry used to keep all CDOs known to the pool manager in sync */
struct mstro_pm_cdo_registry_entry {
  UT_hash_handle hh;                           /**< hash on CDO id */
  struct mstro_cdo_id cdo_id;                  /**< the global cdo id */
  char *cdo_name;                              /**< the CDO printable name */
  struct per_app_cdo_entry *app_to_attributes; /**< app-to-attributes table for the cdo id */
};


/** the CDO registry */
static struct mstro_pm_cdo_registry_entry *g_mstro_pm_cdo_registry = NULL;

/** lock for the CDO registry */
pthread_mutex_t g_mstro_pm_cdo_registry_lock = PTHREAD_MUTEX_INITIALIZER;

/** macro to make locked uthash accesses to the cdo registry. Not super
 * safe, beware. It's meant to only be used around HASH_* functions */
#define WITH_LOCKED_CDO_REGISTRY(body) \
  do {                                                                  \
  int wlr_stat=pthread_mutex_lock(&g_mstro_pm_cdo_registry_lock);       \
  if(wlr_stat!=0) {                                                     \
    ERR("Failed to lock cdo registry: %d (%s)\n",                       \
        wlr_stat, strerror(wlr_stat));                                  \
    abort();                                                            \
  }                                                                     \
                                                                        \
  do {                                                                  \
    body;                                                               \
  } while(0);                                                           \
                                                                        \
  wlr_stat=pthread_mutex_unlock(&g_mstro_pm_cdo_registry_lock);         \
  if(wlr_stat!=0) {                                                     \
    ERR("Failed to unlock cdo registry: %d (%s)\n",                     \
        wlr_stat, strerror(wlr_stat));                                  \
    abort();                                                            \
  }                                                                     \
  } while(0)


/** Outstanding require and demand queue */

struct mstro_pm_demand_queue_entry {
  struct mstro_pm_demand_queue_entry *next; /**< singly linked list */
  struct mstro_cdo_id cdo_id; /**< the CDO ID */
  mstro_app_id requestor; /**< the requesting app */
  Mstro__Pool__Attributes *req_attributes; /**< the requesting side's CDO attributes */
  Mstro__Pool__TransportMethods *req_methods; /**< requesting side's transfer method wishes */
  mstro_app_id provider; /**< source satisfying this request */
};



static inline
mstro_status
mstro_pm_demand_queue_entry__destroy(struct mstro_pm_demand_queue_entry *res)
{
  /* FIXME: could pool for reuse */
  /* the attributes are read-only and belong the the app's entry, we must not free them */
  free(res);
  return MSTRO_OK;
}


/** list of outstanding demand queue entries */
static struct mstro_pm_demand_queue_entry *g_mstro_pm_demand_queue = NULL;
/** Lock for mstro_pm_demand_queue */
static pthread_mutex_t g_mstro_pm_demand_queue_mtx = PTHREAD_MUTEX_INITIALIZER;
/** wait object for mstro_pm_demand_queue */
pthread_cond_t  g_mstro_pm_demand_queue_cvar = PTHREAD_COND_INITIALIZER;

/** notify whoever needs to be informed about changes on the CDO registry.
 *
 * pokes @ref g_mstro_pm_demand_queue_cvar
 */
static inline
mstro_status
mstro_pm__notify_cdo_registry_change(void)
{
  int s = pthread_mutex_lock(&g_mstro_pm_demand_queue_mtx);
  if(s!=0) {
    ERR("Failed to lock demand queue: %d\n", s);
    return MSTRO_FAIL;
  }
  s=pthread_cond_broadcast(&g_mstro_pm_demand_queue_cvar);
  if(s!=0) {
    ERR("Failed to broadcast demand queue update: %d\n", s);
    pthread_mutex_unlock(&g_mstro_pm_demand_queue_mtx);
    return MSTRO_FAIL;
  }
  s=pthread_mutex_unlock(&g_mstro_pm_demand_queue_mtx);
  if(s!=0) {
    ERR("Failed to lock demand queue: %d\n", s);
    return MSTRO_FAIL;
  }

  return MSTRO_OK;
}

/** find best candidate to serve CDO to app_id
 *
 * Returns best match in *best, or MSTRO_APP_ID_INVALID if none exists.
 */
static inline
mstro_status
mstro_pm__find_best_provider(
    struct per_app_cdo_entry *app_to_attributes_table,
    mstro_app_id app_id,
    struct per_app_cdo_entry **best)
{
  struct per_app_cdo_entry *entry, *tmp;
  *best = NULL;
  app_id=app_id; /* silence unused arg warning */
  /* FIXME: this is just greedy */
  HASH_ITER(hh, app_to_attributes_table, entry, tmp) {
    if(entry->cdo_state==MSTRO_CDO_STATE_OFFERED) {
      *best = entry;
      break;
    }
  }
  return MSTRO_OK;
}
  
  

static inline
mstro_status
mstro_pm__handle_demands(struct mstro_pm_demand_queue_entry *q)
{
  mstro_status status = MSTRO_OK;
  struct mstro_pm_demand_queue_entry *unhandled=NULL;
  size_t len=0;
  struct mstro_pm_demand_queue_entry *e;
  LL_COUNT(q,e,len);
  
  NOISE("Handling %d demand queue entries\n", len);

  while(q!=NULL) {
    e = q;
    q = q->next;
    /* try to handle e */
    WITH_LOCKED_CDO_REGISTRY(
        struct mstro_pm_cdo_registry_entry *regentry=NULL;
        HASH_FIND(hh, g_mstro_pm_cdo_registry,
                  &e->cdo_id, sizeof(struct mstro_cdo_id), regentry);
        if(!regentry) {
          /* should not happen */
          WITH_CDO_ID_STR(idstr, &e->cdo_id,
                          ERR("No CDO registry entry for DEMANDed CDO %s\n",
                              idstr););
        } else {
          struct per_app_cdo_entry *candidate=NULL;
          mstro_pm__find_best_provider(regentry->app_to_attributes,
                                       e->requestor,
                                       &candidate);
          if(candidate==NULL) {
            /* no offer for this demand */
            LL_PREPEND(unhandled, e);
          } else {
            Mstro__Pool__UUID cdoid = MSTRO__POOL__UUID__INIT;
            cdoid.qw0 = e->cdo_id.qw[0];
            cdoid.qw1 = e->cdo_id.qw[1];

            Mstro__Pool__Appid dst_appid = MSTRO__POOL__APPID__INIT;
            dst_appid.id = e->requestor;
            INFO("Initiating a transfer for requestor %zu\n",
                 e->requestor);
            
            Mstro__Pool__InitiateTransfer it
                = MSTRO__POOL__INITIATE_TRANSFER__INIT;
            it.cdoid = &cdoid;
            it.dst_appid = &dst_appid;
            it.dst_attributes = e->req_attributes;

            struct mstro_pm_app_registry_entry *app_entry = NULL;
            mstro_status s;
            s = mstro_pm_app_lookup(dst_appid.id, &app_entry);
            if(s != MSTRO_OK || app_entry==NULL) {
              ERR("Target %zu not in app table\n", e->requestor);
              return MSTRO_FAIL;
            }
            INFO("InitiateTransfer msg packed with serialized ep desc %s\n",
                 app_entry->serialized_desc);
            it.dst_serialized_endpoint = app_entry->serialized_desc;
            INFO("entry has req methods %p\n", e->req_methods);
            it.methods = e->req_methods;
            it.cp = 0; // FIXME When PM starts doing smart stuff. I am
                       // here arbitrarily choosing not to keep a
                       // local copy on the src after
                       // transfer. Perhaps PM wants to communicate to
                       // keep the transport file/obj on the dst, in
                       // which TODO case add a "keep_file" field in
                       // the InitiateTransfer msg

            Mstro__Pool__MstroMsg msg = MSTRO__POOL__MSTRO_MSG__INIT;

            status = mstro_pmp_package(&msg, (ProtobufCMessage*)&it);
            if(status!=MSTRO_OK) {
              ERR("Failed to package %s into a pool manager message\n",
                  it.base.descriptor->name);
              goto BAILOUT_UNLOCK;
            }

            status = mstro_pmp_send_nowait(candidate->app, &msg);
            
            switch(status) {
              case MSTRO_OK:
                WITH_CDO_ID_STR(
                    idstr, &e->cdo_id,
                    INFO("Sent INITIATE-TRANSFER for %s to app %zu (for app %zu)\n",
                         idstr,
                         candidate->app, it.dst_appid->id);
                                );
                break;
              default:
                ERR("Failed to send INITIATE_TRANSFER message to app %zu: %d (%s)\n",
                    candidate->app, status, mstro_status_description(status));
                goto BAILOUT_UNLOCK;
            }
            DEBUG("Disposing demand queue entry that has been handled\n");
            status = mstro_pm_demand_queue_entry__destroy(e);
            if(status!=MSTRO_OK) {
              ERR("Failed to de-allocate demand queue entry\n");
              goto BAILOUT_UNLOCK;
            }
          }
        }
   BAILOUT_UNLOCK:
        ;
                             );
  }
  
  if(unhandled!=NULL) {
    /* append the leftovers to queue for another round later */
    int s=pthread_mutex_lock(&g_mstro_pm_demand_queue_mtx);
    if(s!=0) {
      ERR("Failed to lock demand queue mutex: %d\n", s);
      return MSTRO_FAIL;
    }
    LL_CONCAT(g_mstro_pm_demand_queue, unhandled);
    s=pthread_mutex_unlock(&g_mstro_pm_demand_queue_mtx);
    if(s!=0) {
      ERR("Failed to unlock demand queue mutex: %d\n", s);
      return MSTRO_FAIL;
    }
  }    

  return status;
}

static inline
void unlock_demand_queue(void *arg)
{
  arg=arg; /* silence unused arg warning */
  pthread_mutex_unlock(&g_mstro_pm_demand_queue_mtx);
}

/* This function is called by the transport initiator thread to process the demand queue once:
 *  wait on cvar, drain queue, unlock, process drained queue entries, quit */
mstro_status
mstro_pm_handle_demand_queue(void)
{
  mstro_status status;
  int s;
  struct mstro_pm_demand_queue_entry *q=NULL;

  /* this function will be canceled by the pool manager thread at the
   * cond_wait */
  pthread_cleanup_push(unlock_demand_queue, NULL);

  /* disable cancelation */
  s = pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
  if(s!=0) {
    ERR("failed to disable pthread cancel: %d\n", s);
    status = MSTRO_FAIL;
    goto BAILOUT;
  }

  int s = pthread_mutex_lock(&g_mstro_pm_demand_queue_mtx);
  if(s!=0) {
    ERR("Failed to lock demand queue mutex: %d (%s)\n", s, strerror(s));
    status = MSTRO_FAIL;
    goto BAILOUT;
  }
  
  while(g_mstro_pm_demand_queue!=NULL) {
    /* drain */
    q = g_mstro_pm_demand_queue;
    g_mstro_pm_demand_queue = NULL;
    s = pthread_mutex_unlock(&g_mstro_pm_demand_queue_mtx);
    if(s!=0) {
      ERR("Failed to unlock demand queue mutex: %d (%s)\n", s, strerror(s));
      status = MSTRO_FAIL;
      goto BAILOUT;
    }
    
    status = mstro_pm__handle_demands(q);
    if(status!=MSTRO_OK) {
      ERR("Failure processing demand queue entries\n");
      goto BAILOUT;
    }

    /* re-obtain lock */
    s = pthread_mutex_lock(&g_mstro_pm_demand_queue_mtx);
    if(s!=0) {
      ERR("Failed to lock demand queue mutex: %d (%s)\n", s, strerror(s));
      status = MSTRO_FAIL;
      goto BAILOUT;
    }

    s=pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
    if(s!=0) {
      ERR("failed to enable thread cancelation: %d\n", s);
      status=MSTRO_FAIL;
      goto BAILOUT;
    }
    
    /* this will jump into cleanup/non-local exit if we were canceled */
    pthread_testcancel();
    s=pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
    if(s!=0) {
      ERR("failed to disable thread cancelation: %d\n", s);
      status=MSTRO_FAIL;
      goto BAILOUT;
    }
    /* if someone added new entries we'll loop */
  }
  /* no entries during last processing step; we hold the lock: go to sleep */
  
  s=pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  if(s!=0) {
    ERR("failed to enable thread cancelation: %d\n", s);
    status=MSTRO_FAIL;
    goto BAILOUT;
  }

  /* go to sleep until something interesting happens */
  s = pthread_cond_wait(&g_mstro_pm_demand_queue_cvar,
                        &g_mstro_pm_demand_queue_mtx);
  if(s!=0) {
    ERR("Failed to wait on demand queue: %d (%s)\n", s, strerror(s));
    status = MSTRO_FAIL;
    goto BAILOUT;
  }
  
  /* once we wake up we hold the lock again. Pull off one round of
   * entries, then return. We'll be called again by the thread running
   * it */
  q=g_mstro_pm_demand_queue;
  g_mstro_pm_demand_queue = NULL;

  /* unlocks the mutex (just like when getting canceled) */
  pthread_cleanup_pop(1);
  
  status = mstro_pm__handle_demands(q);
  if(status!=MSTRO_OK) {
    ERR("Failure processing demand queue entries\n");
    goto BAILOUT;
  }
  
BAILOUT:
  return status;
}
    

mstro_status
mstro_pm_cdo_registry_declare(mstro_app_id app,
                              const char *cdo_name,
                              struct mstro_cdo_id *cdo_id_p)
{
  struct mstro_pm_app_registry_entry *e;
  assert(app!=MSTRO_APP_ID_INVALID);
      
  if(cdo_name==NULL)
    return MSTRO_INVARG;
  if(cdo_id_p==NULL)
    return MSTRO_INVOUT;

  /* compute default ID */
  mstro_status status = mstro_cdo_id_from_name(cdo_name, cdo_id_p);
  if(status!=MSTRO_OK) {
    ERR("Failed to compute CDO ID\n");
    goto BAILOUT;
  }

  /* check whether ID has been seen before */
  struct mstro_pm_cdo_registry_entry *regentry=NULL;
  WITH_LOCKED_CDO_REGISTRY(
      HASH_FIND(hh, g_mstro_pm_cdo_registry,
                cdo_id_p, sizeof(struct mstro_cdo_id), regentry);
      if(!regentry) {
        WITH_CDO_ID_STR(idstr,cdo_id_p, {
            DEBUG("New CDO %s -> %s\n", cdo_name, idstr);
          });
        regentry = malloc(sizeof(struct mstro_pm_cdo_registry_entry));
        if(regentry==NULL) {
          ERR("Cannot allocate new CDO registry entry\n");
          status=MSTRO_NOMEM;
          goto BAILOUT_UNLOCK;
        }
        regentry->cdo_name = strdup(cdo_name);
        if(regentry->cdo_name==NULL) {
          ERR("Cannot allocate name of new CDO regentry\n");
          free(regentry);
          status=MSTRO_NOMEM;
          goto BAILOUT_UNLOCK;
        }
        memcpy(&(regentry->cdo_id),cdo_id_p,sizeof(struct mstro_cdo_id));
        regentry->app_to_attributes = NULL;
        HASH_ADD(hh,g_mstro_pm_cdo_registry,
                 cdo_id, sizeof(struct mstro_cdo_id), regentry);
      }
      
      /* We now have a regentry, possibly without app registration */
      if(strcmp(regentry->cdo_name,cdo_name)!=0) {
        WITH_CDO_ID_STR(idstr,cdo_id_p, {
            ERR("Jackpot! You hit a hash collision on the CDO ID: |%s| and |%s| both map to %s\n",
                regentry->cdo_name, cdo_name, idstr);
            abort();
          });
      } else {
        /* have previous registration for this CDO */
        struct per_app_cdo_entry *per_app_entry = malloc(sizeof(struct per_app_cdo_entry));
        if(per_app_entry==NULL) {
          ERR("Cannot allocate new per-app CDO entry\n");
          status=MSTRO_NOMEM;
          goto BAILOUT;
        }
        per_app_entry->app = app;
        per_app_entry->attributes = NULL;
        per_app_entry->cdo_state = MSTRO_CDO_STATE_DECLARED;
        HASH_ADD(hh, regentry->app_to_attributes,
                 app, sizeof(mstro_app_id), per_app_entry);
      }
      DEBUG("CDO %s now has registered by %d applications\n",
            cdo_name, HASH_COUNT(regentry->app_to_attributes));
      status = MSTRO_OK;
 BAILOUT_UNLOCK:
      ;
                           );
  
BAILOUT:
  return status;
}

mstro_status
mstro_pm_cdo_registry_find(mstro_app_id app, const struct mstro_cdo_id *id)
{
  bool found=false;
  
  struct mstro_pm_cdo_registry_entry *regentry=NULL;
  WITH_LOCKED_CDO_REGISTRY(
      HASH_FIND(hh, g_mstro_pm_cdo_registry,
                id, sizeof(struct mstro_cdo_id), regentry);
      if(regentry) {
        struct per_app_cdo_entry *per_app_entry;
        HASH_FIND(hh, regentry->app_to_attributes,
                  &app, sizeof(app), per_app_entry);
        if(per_app_entry) 
          found=true;
      }
                           );
  if(found)
    return MSTRO_OK;
  else
    return MSTRO_FAIL;
}

mstro_status
mstro_pm_cdo_registry_store_attributes(const struct mstro_cdo_id *cdoid,
                                       mstro_app_id app_id,
                                       Mstro__Pool__Attributes *attributes)
{
  struct mstro_pm_cdo_registry_entry *regentry=NULL;
  bool found=false;
  
  WITH_LOCKED_CDO_REGISTRY(
      HASH_FIND(hh, g_mstro_pm_cdo_registry,
                cdoid, sizeof(struct mstro_cdo_id), regentry);
      if(regentry) {
        struct per_app_cdo_entry *per_app_entry;
        HASH_FIND(hh, regentry->app_to_attributes,
                  &app_id, sizeof(app_id), per_app_entry);
        if(per_app_entry) {
          if(per_app_entry->attributes==NULL) {
            per_app_entry->attributes = attributes;
            found=true;
          } else {
            WITH_CDO_ID_STR(idstr, cdoid, {
                ERR("Trying to overwrite attributes for %s from app %d\n",
                    idstr, app_id);
              });
          }
        }
      });
 
  return found? MSTRO_OK : MSTRO_FAIL;
}

mstro_status
mstro_pm_cdo_registry_update_state(
    const struct mstro_cdo_id *cdoid,
    const mstro_app_id app_id,
    mstro_cdo_state new_state)
{
  struct mstro_pm_cdo_registry_entry *regentry=NULL;
  bool found=false;
  
  WITH_LOCKED_CDO_REGISTRY(
      HASH_FIND(hh, g_mstro_pm_cdo_registry,
                cdoid, sizeof(struct mstro_cdo_id), regentry);
      if(regentry) {
        struct per_app_cdo_entry *per_app_entry;
        HASH_FIND(hh, regentry->app_to_attributes,
                  &app_id, sizeof(app_id), per_app_entry);
        if(per_app_entry) {
          /* FIXME: could check state machine stepping here */
          WITH_CDO_ID_STR(idstr, cdoid, {
              DEBUG("Stepping CDO state for %s for app %zu from %d (%s) to %d (%s)\n",
                    idstr, app_id,
                    per_app_entry->cdo_state,
                    mstro_cdo_state_describe(per_app_entry->cdo_state),
                    new_state,
                    mstro_cdo_state_describe(new_state)
                    );
            });
          per_app_entry->cdo_state = new_state;
          found=true;
        } 
      });
  
  return found? MSTRO_OK : MSTRO_FAIL;
}

/** De-Register the application @ref id */
mstro_status
mstro_pm_app_deregister(mstro_app_id key)
{
  mstro_status status=MSTRO_OK;

  WARN("Not checking whether app %zu holds permanent CDOs that need to be preserved\n", key);
  WARN("Not checking whether app %zu offers any CDOs that will be needed by others\n", key);
      
  if(key==MSTRO_APP_ID_MANAGER) {
    DEBUG("De-registering pool manager connection\n");
  }
  
  WITH_LOCKED_CDO_REGISTRY({
      struct mstro_pm_cdo_registry_entry *e;
      struct mstro_pm_cdo_registry_entry *tmp;
      HASH_ITER(hh, g_mstro_pm_cdo_registry,e,tmp) {
        struct per_app_cdo_entry *app_record;
        HASH_FIND(hh, e->app_to_attributes,
                  &key, sizeof(mstro_app_id),
                  app_record);
        if(app_record==NULL) {
          ;/* this app not on list for this CDO */
        } else {
          switch(app_record->cdo_state) {
            case MSTRO_CDO_STATE_DECLARED:
              break; /* ok, declared and never sealed */
            case MSTRO_CDO_STATE_SEALED:
              break; /* ok, sealed and never offered */
            case MSTRO_CDO_STATE_OFFERED:
              WITH_CDO_ID_STR(idstr, &e->cdo_id,
                              ERR("LEAVE request from %zu when CDO %s is still OFFERED\n",
                                  key, idstr);
                              );
              break;
            case MSTRO_CDO_STATE_WITHDRAWN_GLOBALLY:
            case MSTRO_CDO_STATE_WITHDRAWN:
            case MSTRO_CDO_STATE_RETRACTED_GLOBALLY:
            case MSTRO_CDO_STATE_RETRACTED:
            case MSTRO_CDO_STATE_DEAD:
              break; /* ok, no one could need it */
            default:
              WITH_CDO_ID_STR(idstr, &e->cdo_id,
                              ERR("Unexpected PM cdo registry state for CDO %s: %d (%s)\n",
                                  idstr, app_record->cdo_state,
                                  mstro_cdo_state_describe(app_record->cdo_state));
                              );
              break;
          }
          HASH_DEL(e->app_to_attributes, app_record);
          free(app_record);
        }
      }
    });
  DEBUG("Dropped all CDO registry entries for app %zu\n", key);

  /* We can not drop the reference to the app itself because we could
   * not send the BYE otherwise. This may actually be a good thing */
  /* WITH_LOCKED_APP_REGISTRY({ */
  /*     struct mstro_pm_app_registry_entry *e=NULL; */
  /*     HASH_FIND(hh, g_mstro_pm_app_registry, &key, sizeof(key), e); */
  /*     if(e!=NULL) { */
  /*       HASH_DEL(g_mstro_pm_app_registry, e); */
  /*     } else { */
  /*       ERR("Registry entry for app %zu vanished\n", key); */
  /*     }       */
  /*   }); */
  /* DEBUG("Dropped all registry entries for app %zu\n", key); */

  return MSTRO_OK;
}

/** look up per-app cdo entry. Must be called unter CDO_REGISTRY_LOCK. */
static inline
mstro_status
mstro_pm_cdo_app_lookup(const struct mstro_cdo_id* cdo_id,
                        mstro_app_id app_id,
                        struct per_app_cdo_entry **entry)
{
  struct mstro_pm_cdo_registry_entry *regentry=NULL;

  HASH_FIND(hh, g_mstro_pm_cdo_registry,
            cdo_id, sizeof(struct mstro_cdo_id), regentry);
  if(!regentry) {
    *entry = NULL;
    return MSTRO_FAIL;
  } else {
    struct per_app_cdo_entry *per_app_entry;
    HASH_FIND(hh, regentry->app_to_attributes,
              &app_id, sizeof(app_id), *entry);
    if(*entry==NULL)
      return MSTRO_FAIL;
    else
      return MSTRO_OK;
  }
}

mstro_status
mstro_pm_cdo_registry_cdo_name_lookup(const struct mstro_cdo_id *id,
                                      const char **cdo_name)
{
  mstro_status s;
  struct mstro_pm_cdo_registry_entry *regentry=NULL;
  WITH_LOCKED_CDO_REGISTRY(
      HASH_FIND(hh, g_mstro_pm_cdo_registry,
                id, sizeof(struct mstro_cdo_id), regentry);
      if(regentry) {
        *cdo_name = regentry->cdo_name;
        s=MSTRO_OK;
      } else {
        s=MSTRO_NOMATCH;
      });
  return s;
} 

static inline
mstro_status
mstro_pm_demand_queue_entry__create(const struct mstro_cdo_id *cdo_id,
                                    mstro_app_id app,
                                    struct mstro_pm_demand_queue_entry **res)
{
  if(res==NULL)
    return MSTRO_INVOUT;
  *res=malloc(sizeof(struct mstro_pm_demand_queue_entry));
  if(NULL==*res) {
    return MSTRO_NOMEM;
  } else {
    mstro_status status=MSTRO_OK;
    struct mstro_pm_app_registry_entry *app_entry;
    status = mstro_pm_app_lookup(app, &app_entry);

    (*res)->cdo_id.qw[0] = cdo_id->qw[0];
    (*res)->cdo_id.qw[1] = cdo_id->qw[1];
    (*res)->provider = MSTRO_APP_ID_INVALID;
    (*res)->req_methods = app_entry->transport_methods;
    (*res)->requestor = app;
    /* (*res)->req_attributes set below inside locked region */


    WITH_LOCKED_CDO_REGISTRY({
        struct per_app_cdo_entry *entry;
        mstro_status status = mstro_pm_cdo_app_lookup(cdo_id, app, &entry);
        if(status!=MSTRO_OK) {
          ERR("Failed to find CDO entry for app %zu\n", app);
        } else {
          if(entry->cdo_state & MSTRO_CDO_STATE_INJECTED) {
            WARN("operation on injected cdo, fasten seat belts (untested)\n");
          } else {
            DEBUG("demand queue entry for cdo in state %d (%s)\n",
                  entry->cdo_state,
                  mstro_cdo_state_describe(entry->cdo_state));
            assert(   entry->cdo_state & MSTRO_CDO_STATE_REQUIRED
                   || entry->cdo_state & MSTRO_CDO_STATE_DEMANDED );
          }
          if(entry->cdo_state & MSTRO_CDO_STATE_IN_TRANSPORT) {
            ; /* normal */
          } else {
            WARN("operation on CDO that is not marked 'in-transport'\n");
          }
          (*res)->req_attributes = entry->attributes;
        }
      });
    return status;
  }
}

mstro_status
mstro_pm_demand_queue_insert(const struct mstro_cdo_id *cdo_id,
                             mstro_app_id requestor)
{
  if(cdo_id==NULL)
    return MSTRO_INVARG;
  if(requestor==MSTRO_APP_ID_INVALID)
    return MSTRO_INVARG;

  struct mstro_pm_demand_queue_entry *e;
  mstro_status status = mstro_pm_demand_queue_entry__create(cdo_id, requestor, &e);
  if(status!=MSTRO_OK) {
    ERR("Cannot allocate demand queue entry\n");
    return status;
  }

  if(status==MSTRO_OK) {
    /* add to queue */
    int s=pthread_mutex_lock(&g_mstro_pm_demand_queue_mtx);
    if(s!=0) {
      ERR("Failed to lock demand queue mutex: %d (%s)\n",
          s, strerror(s));
      status=MSTRO_FAIL;
    } else {
      LL_PREPEND(g_mstro_pm_demand_queue, e);
      s=pthread_cond_broadcast(&g_mstro_pm_demand_queue_cvar);
      if(s!=0) {
        ERR("Failed to signal change in demand queue: %d (%s)\n",
            s, strerror(s));
        status=MSTRO_FAIL;
      }
      s=pthread_mutex_unlock(&g_mstro_pm_demand_queue_mtx);
      if(s!=0) {
        ERR("Failed to unlock demand queue mutex: %d (%s)\n",
            s, strerror(s));
        status=MSTRO_FAIL;
      }
    }
  }
BAILOUT:
  return status;
}


mstro_status
mstro_pm_cdo_registry_transfer_completed(const struct mstro_cdo_id *cdo_id,
                                         mstro_app_id app_id)
{
  /* There are 3 essentially different cases:
     IN_TRANSPORT set on
     - DEMANDED
     - REQUIRED
     - other 'base state'

     For DEMANDED we must consider the completion of transfer
     notification as the removal from the pool.

     For REQUIRED we set SATISFIED -- telling us that we've
     proactively done a transfer for a required CDO that has not been
     DEMANDED, and that is also considered OFFERED until we see the
     DEMAND.

     Other base states indicate that we have been very proactive, and
     should have INJECTED set. Again, we'll set SATISFIED (to indicate
     availability) and keep the INJECTED for bookkeeping.
  */

  mstro_status status = MSTRO_OK;

  WITH_LOCKED_CDO_REGISTRY({
      struct per_app_cdo_entry *entry;
      mstro_status status = mstro_pm_cdo_app_lookup(cdo_id, app_id, &entry);
      if(status!=MSTRO_OK) {
        ERR("Failed to find CDO entry for app %zu\n", app_id);
        status = MSTRO_INVARG;
      } else {
        if(! (entry->cdo_state & MSTRO_CDO_STATE_IN_TRANSPORT)) {
          WITH_CDO_ID_STR(idstr, cdo_id,
                          ERR("Unexpected PM cdo registry state for CDO %s: %d (%s)\n",
                              idstr);
                          );
          status = MSTRO_INVARG;
        } else {
          /* ok, it was in-transport */
          if(entry->cdo_state & MSTRO_CDO_STATE_DEMANDED) {
            WITH_CDO_ID_STR(idstr, cdo_id,
                            DEBUG("Clearing IN_TRANSPORT state from DEMANDED CDO %s\n",
                                  idstr););
            entry->cdo_state &= ~MSTRO_CDO_STATE_IN_TRANSPORT;
          } else if(entry->cdo_state & MSTRO_CDO_STATE_REQUIRED) {
            WITH_CDO_ID_STR(idstr, cdo_id,
                            DEBUG("Clearing IN_TRANSPORT state from REQUIRED CDO %s\n",
                                  idstr););
            entry->cdo_state &= ~MSTRO_CDO_STATE_IN_TRANSPORT;
            entry->cdo_state |= MSTRO_CDO_STATE_SATISFIED;
          } else if(entry->cdo_state & MSTRO_CDO_STATE_INJECTED) {
            WITH_CDO_ID_STR(idstr, cdo_id,
                            DEBUG("Clearing IN_TRANSPORT state from INJECTED CDO %s\n",
                                  idstr););
            entry->cdo_state &= ~MSTRO_CDO_STATE_IN_TRANSPORT;
            entry->cdo_state |= MSTRO_CDO_STATE_SATISFIED;
          } else {
            WITH_CDO_ID_STR(idstr, cdo_id,
                            ERR("Unhandled CDO state %s for CDO %s\n",
                                mstro_cdo_state_describe(entry->cdo_state), idstr););
            status = MSTRO_UNIMPL;
          }
        }
      }
    });
  return status;
}


mstro_status
mstro_pm_cdo_app_match(mstro_app_id origin, const struct mstro_cdo_id *id,
                       const struct mstro_cdo_selector_ *cdo_selector)
{
  if(origin==g_pool_app_id) {
    ERR("FIXME: unhandled case PM-is-source-of-PM-event\n");
    return MSTRO_UNIMPL;
  }
  if(id==NULL) {
    return MSTRO_INVARG;
  }
  if(cdo_selector==NULL) {
    DEBUG("NULL selector always matches -- hope that's what you meant\n");
    return MSTRO_OK;
  }

  /* WITH_CDO_ID_STR(str,id, { */
  /*     DEBUG("Trying to match %s origin %" PRIappid " for selector %p, query |%s|\n", */
  /*           str, origin, cdo_selector, cdo_selector->query); */
  /*   }); */
  mstro_status status = MSTRO_FAIL;
  /* now fetch attributes from per-app-CDO entry, call subscription-module checker and return result */
  WITH_LOCKED_CDO_REGISTRY({
      struct per_app_cdo_entry *entry;
      status = mstro_pm_cdo_app_lookup(id, origin, &entry);
      if(status!=MSTRO_OK) {
        ERR("Failed to find CDO entry for app %zu\n", origin);
      } else {
        status = mstro_subscription_selector_eval(id, cdo_selector, 
                                                  entry->attributes);
      }
    });
  return status;
}
